﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file contains a programming example to display information  
//       about installed ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using ALLDAQ.Boards;
using ADQInfo.Utility;
using System.Diagnostics;

namespace ADQInfo
{
    public partial class FormMain : Form
    {
        const double _picoSecondsPerSecond = 1e+12;

        private uint _adqOpenError = AdqCommon.NotOpen;

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            try
            {
                TreeNode treeNode = new TreeNode("ALLDAQ hardware");
                tvInfo.Nodes.Add(treeNode);
                
                _adqOpenError = OpenAdqDriver();
                if (_adqOpenError != AdqErrors.ADQ_ERROR_NONE) return;

                GetSystemInfo(treeNode);
                tvInfo.ExpandAll();
                tvInfo.Nodes[0].EnsureVisible();

                CloseAdqDriver();
            }
            catch (DllNotFoundException ex)
            {
                tvInfo.Nodes.Add(new TreeNode(string.Format("DllNotFoundException: {0}", ex.Message)));
            }
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                tvInfo.Nodes.Add(new TreeNode(FormatErrorText("adqOpen", adqError)));
            }
            return adqError;
        }

        private void GetSystemInfo(TreeNode parentTreeNode)
        {
            var systemInfo = new AdqDriver.ADQSystemInfo(AdqDriver.ADQSystemInfo_ID);

            uint adqError = AdqDriver.adqGetSystemInfo(ref systemInfo);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                TreeNode treeNode = new TreeNode("Library version: " + AdqCommon.FormatVersion(systemInfo.DLLVersion));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Number of boards: {0}", systemInfo.NumberOfBoards));
                parentTreeNode.Nodes.Add(treeNode);

                GetBoardInfo(systemInfo.NumberOfBoards, parentTreeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetSystemInfo", adqError);
            }
        }

        private void GetBoardInfo(uint numberOfBoards, TreeNode parentTreeNode)
        {
            TreeNode boardNode;
            TreeNode treeNode;
            var boardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);

            for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
            {
                uint adqError = AdqDriver.adqGetBoardInfo(boardIdx, ref boardInfo);
                if (adqError == AdqErrors.ADQ_ERROR_NONE)
                {
                    string model = AdqCommon.GetModelName(boardInfo.BoardName);

                    boardNode = new TreeNode(
                        string.Format("Board number: {0} - {1}", boardIdx, model));
                    parentTreeNode.Nodes.Add(boardNode);

                    treeNode = new TreeNode(
                        string.Format("Name: {0}", boardInfo.BoardName));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Vendor ID: 0x{0:X4}", boardInfo.VendorID));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Device ID: 0x{0:X4}", boardInfo.DeviceID));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Serial number: 0x{0:X8}", boardInfo.SerialNumber));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Firmware version: {0}", boardInfo.FirmwareVersion));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Description: {0}", boardInfo.BoardDescription));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode(
                        string.Format("Driver name: {0}", boardInfo.DriverName));
                    boardNode.Nodes.Add(treeNode);

                    treeNode = new TreeNode("Driver version: " + AdqCommon.FormatVersion(boardInfo.DriverVersion));
                    boardNode.Nodes.Add(treeNode);

                    if (boardInfo.TTLVoltageLevel > 0)
                    {
                        AddTTLVoltageLevels(boardNode, boardInfo.TTLVoltageLevel);
                        AddCurrentTTLVoltageLevel(boardIdx, boardNode);
                    }

                    treeNode = new TreeNode("Bus type: " + AdqCommon.GetBusTypeName(boardInfo.BusType));
                    boardNode.Nodes.Add(treeNode);

                    if (boardInfo.BusType == AdqConstants.ADQ_BUS_TYPE_PCI)
                    {
                        AddBoardPCIInfo(boardIdx, treeNode);
                    }

                    treeNode = new TreeNode("Installed location: " + (boardInfo.Remote ? "Remote" : "Local"));
                    boardNode.Nodes.Add(treeNode);

                    if (boardInfo.Remote)
                    {
                        treeNode = new TreeNode(
                            string.Format("Remote IP address: {0}", boardInfo.RemoteIPAddressOrHostName));
                        boardNode.Nodes.Add(treeNode);
                    }

                    treeNode = new TreeNode(
                        string.Format("Number of modules: {0}", boardInfo.NumberOfModules));
                    boardNode.Nodes.Add(treeNode);

                    GetModuleInfo(boardIdx, boardInfo.NumberOfModules, boardNode);
                }
                else
                {
                    treeNode = new TreeNode(string.Format("Board {0} - ", boardIdx) + FormatErrorText("adqGetBoardInfo", adqError));
                    parentTreeNode.Nodes.Add(treeNode);
                    continue;
                }
            }
        }

        private void AddTTLVoltageLevels(TreeNode parentTreeNode, uint voltageLevels)
        {
            var voltageLevelsTreeNode = new TreeNode("TTL voltage levels");
            parentTreeNode.Nodes.Add(voltageLevelsTreeNode);

            AddTTLVoltageLevelTreeNode(voltageLevelsTreeNode, voltageLevels, AdqConstants.ADQ_TTL_VOLTAGE_LEVEL_3_3_VOLTS);
            AddTTLVoltageLevelTreeNode(voltageLevelsTreeNode, voltageLevels, AdqConstants.ADQ_TTL_VOLTAGE_LEVEL_5_0_VOLTS);
        }

        private void AddTTLVoltageLevelTreeNode(TreeNode parentTreeNode, uint voltageLevels, uint voltageConstant)
        {
            if ((voltageLevels & voltageConstant) == voltageConstant)
            {
                var treeNode = new TreeNode(AdqCommon.GetTTLVoltageLevelText(voltageConstant));
                parentTreeNode.Nodes.Add(treeNode);
            }
        }

        private void AddCurrentTTLVoltageLevel(uint boardIndex, TreeNode parentTreeNode)
        {
            uint adqError = AdqDriver.adqGetBoardTTLVoltage(boardIndex, out uint voltage);
            if (adqError == AdqErrors.ADQ_ERROR_NONE) 
            {
                var treeNode = new TreeNode(
                    string.Format("Current TTL voltage level: {0}", AdqCommon.GetTTLVoltageLevelText(voltage)));
                parentTreeNode.Nodes.Add(treeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetBoardTTLVoltage", adqError);
            }
        }

        private void AddBoardPCIInfo(uint boardIndex, TreeNode parentTreeNode)
        {
            var boardPciInfo = new AdqDriver.ADQBoardPCIInfo(AdqDriver.ADQBoardPCIInfo_ID);
            uint adqError = AdqDriver.adqGetBoardPCIInfo(boardIndex, ref boardPciInfo);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                var treeNode = new TreeNode(string.Format("PCI-Bus number: {0}", boardPciInfo.BusNumber));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(string.Format("PCI Device number: {0}", boardPciInfo.DeviceNumber));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(string.Format("PCI Function number: {0}", boardPciInfo.FunctionNumber));
                parentTreeNode.Nodes.Add(treeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetBoardPCIInfo", adqError);
            }
        }

        private void GetModuleInfo(uint boardIndex, uint numberOfModules, TreeNode parentTreeNode)
        {
            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                var moduleNode = new TreeNode(
                    string.Format("Module number: {0}", moduleIdx));
                parentTreeNode.Nodes.Add(moduleNode);

                uint adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError == AdqErrors.ADQ_ERROR_NONE)
                {
                    string moduleTypeName = AdqCommon.GetModuleTypeName(moduleType);
                    moduleNode.Text += " - " + moduleTypeName;

                    var treeNode = new TreeNode(
                        string.Format("Module type: {0}", moduleTypeName));
                    moduleNode.Nodes.Add(treeNode);

                    if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
                    {
                        GetDIOModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DO)
                    {
                        GetDOModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DI)
                    {
                        GetDIModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_CNTR)
                    {
                        GetCounterModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_EXTINT)
                    {
                        GetExternalIrqInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_AI)
                    {
                        GetAIModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_AO)
                    {
                        GetAOModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_I2C)
                    {
                        GetI2CModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_PWM)
                    {
                        GetPWMModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_FIN)
                    {
                        GetFINModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_I2SO)
                    {
                        GetI2SOModuleInfo(boardIndex, moduleIdx, moduleNode);
                    }
                }
                else
                {
                    AddFunctionErrorTreeNode(moduleNode, "adqGetModuleType", adqError);
                    continue;
                }
            }
        }

        private void GetDIOModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDioModuleInfo = new AdqDriver.ADQDIOModuleInfo(AdqDriver.ADQDIOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDioModuleInfo.Size);
            Marshal.StructureToPtr(adqDioModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDioModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDioModuleInfo = (AdqDriver.ADQDIOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {

                Debug.Assert(adqDioModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqDioModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);
                AddDigitalModuleCommonAttributes(parentTreeNode, adqDioModuleInfo.PortWidthBits, adqDioModuleInfo.IsOptoIsolated, adqDioModuleInfo.IsTTL);

                AddCurrentDioDirection(boardIndex, moduleIndex, parentTreeNode);
                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);

                AddMaximumSampleFrequency(boardIndex, moduleIndex, adqGeneralModuleInfo.ClockType, parentTreeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private void AddCurrentDioDirection(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            uint adqError = AdqDriver.adqDIOGetDirection(boardIndex, moduleIndex, out uint dioDirection);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                var childNode = new TreeNode(string.Format("Current DIO direction: {0}", AdqCommon.GetDioDirectionText(dioDirection)));
                parentTreeNode.Nodes.Add(childNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqDIOGetDirection", adqError);
            }
        }

        private void GetDOModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDoModuleInfo = new AdqDriver.ADQDOModuleInfo(AdqDriver.ADQDIOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDoModuleInfo.Size);
            Marshal.StructureToPtr(adqDoModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDoModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDoModuleInfo = (AdqDriver.ADQDOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqDoModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqDoModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);
                AddDigitalModuleCommonAttributes(parentTreeNode, adqDoModuleInfo.PortWidthBits, adqDoModuleInfo.IsOptoIsolated, adqDoModuleInfo.IsTTL);
                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private void GetDIModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDiModuleInfo = new AdqDriver.ADQDIModuleInfo(AdqDriver.ADQDIModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDiModuleInfo.Size);
            Marshal.StructureToPtr(adqDiModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDiModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDiModuleInfo = (AdqDriver.ADQDIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqDiModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqDiModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);

                AddDigitalModuleCommonAttributes(parentTreeNode, adqDiModuleInfo.PortWidthBits, adqDiModuleInfo.IsOptoIsolated, adqDiModuleInfo.IsTTL);

                var treeNode = new TreeNode(
                    string.Format("Number of digital filters: {0}", adqDiModuleInfo.NumberOfFilters));
                parentTreeNode.Nodes.Add(treeNode);

                if (adqDiModuleInfo.NumberOfFilters > 0)
                {
                    AddDigitalFilters(parentTreeNode, ref adqDiModuleInfo);
                    AddCurrentDigitalFilter(boardIndex, moduleIndex, parentTreeNode);
                }

                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                AddMaximumSampleFrequency(boardIndex, moduleIndex, adqGeneralModuleInfo.ClockType, parentTreeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private void AddDigitalModuleCommonAttributes(TreeNode parentTreeNode, uint portWidthBits, bool isOptoIsolated, bool isTTL)
        {
            TreeNode treeNode = new TreeNode(
                string.Format("Port width: {0} Bit", portWidthBits));
            parentTreeNode.Nodes.Add(treeNode);

            treeNode = new TreeNode(
                string.Format("Is optoisolated: {0}", isOptoIsolated));
            parentTreeNode.Nodes.Add(treeNode);

            treeNode = new TreeNode(
                string.Format("Is TTL: {0}", isTTL));
            parentTreeNode.Nodes.Add(treeNode);
        }

        private void AddDigitalFilters(TreeNode parentTreeNode, ref AdqDriver.ADQDIModuleInfo diModuleInfo)
        {
            var filtersTreeNode = new TreeNode("Digital filters");
            parentTreeNode.Nodes.Add(filtersTreeNode);

            TreeNode treeNode;
            for (uint i = 0; i < diModuleInfo.NumberOfFilters; i++)
            {
                treeNode = new TreeNode(
                    string.Format("Digital filter {0}: {1}", i, 
                    AdqCommon.FormatDoubleValue(diModuleInfo.FilterTimePicoseconds[i] / _picoSecondsPerSecond, "s")));
                filtersTreeNode.Nodes.Add(treeNode);
            }
        }

        private void AddCurrentDigitalFilter(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            TreeNode treeNode;
            uint adqError = AdqDriver.adqDIOGetFilterPicoseconds(boardIndex, moduleIndex, out ulong filterValue);
            if (adqError == AdqErrors.ADQ_ERROR_NONE) 
            {
                treeNode = new TreeNode(
                string.Format("Current filter value: {0}", AdqCommon.FormatDoubleValue((filterValue / _picoSecondsPerSecond), "s")));
                parentTreeNode.Nodes.Add(treeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqDIOGetFilterPicoseconds", adqError);
            }
        }

        private void AddGeneralModuleInfo(TreeNode parentTreeNode, ref AdqDriver.ADQGeneralModuleInfo adqGeneralModuleInfo)
        {
            TreeNode treeNode;
            treeNode = new TreeNode(
                string.Format("Description: {0}", adqGeneralModuleInfo.ModuleDescription));
            parentTreeNode.Nodes.Add(treeNode);

            treeNode = new TreeNode(
                string.Format("Clock type: {0}", AdqCommon.GetClockType(adqGeneralModuleInfo.ClockType)));
            parentTreeNode.Nodes.Add(treeNode);

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, "The ClockMinimumDivider of the module is zero");

                treeNode = new TreeNode(
                    string.Format("Clock base frequency: {0}", AdqCommon.FormatDoubleValue(adqGeneralModuleInfo.ClockBaseFrequencyHz, "Hz")));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Clock minimum divider: {0}", adqGeneralModuleInfo.ClockMinimumDivider));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Maximum frequency: {0}", 
                    AdqCommon.FormatDoubleValue(adqGeneralModuleInfo.ClockBaseFrequencyHz / adqGeneralModuleInfo.ClockMinimumDivider, "Hz")));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Internal data buffer size: {0} values", adqGeneralModuleInfo.InternalDataBufferSizeValues));
                parentTreeNode.Nodes.Add(treeNode);
            }

            treeNode = new TreeNode(
                string.Format("Is interrupt source: {0}", adqGeneralModuleInfo.IsIRQSource));
            parentTreeNode.Nodes.Add(treeNode);
        }

        private void AddInterruptInfo(uint boardIndex, uint moduleIndex, TreeNode parentNode)
        {
            TreeNode treeNode;
            var adqIrqInfo = new AdqDriver.ADQIRQInfo(AdqDriver.ADQIRQInfo_ID);
            uint adqError = AdqDriver.adqGetIRQInfo(boardIndex, moduleIndex, ref adqIrqInfo);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                TreeNode irqInfoTreeNode = new TreeNode("Interrupt info");
                parentNode.Nodes.Add(irqInfoTreeNode);

                treeNode = new TreeNode(
                    string.Format("Description: {0}", adqIrqInfo.IRQDescription));
                irqInfoTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Supports bit change: {0}", adqIrqInfo.SupportsBitChange));
                irqInfoTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Supports bit pattern compare: {0}", adqIrqInfo.SupportsBitPatternCompare));
                irqInfoTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Port status error: 0x{0:X8}", adqIrqInfo.SupportsPortStatusError));
                irqInfoTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Fault recovery: 0x{0:X8}", adqIrqInfo.SupportsFaultRecovery));
                irqInfoTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("External interrupt: 0x{0:X8}", adqIrqInfo.SupportsExternal));
                irqInfoTreeNode.Nodes.Add(treeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentNode, "adqGetIRQInfo", adqError);
            }
        }

        private void GetCounterModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            TreeNode treeNode;
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqCntrModuleInfo = new AdqDriver.ADQCNTRModuleInfo(AdqDriver.ADQCNTRModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqCntrModuleInfo.Size);
            Marshal.StructureToPtr(adqCntrModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqCntrModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqCntrModuleInfo = (AdqDriver.ADQCNTRModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQCNTRModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqCntrModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqCntrModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);

                treeNode = new TreeNode(
                    string.Format("Counter type: {0}", AdqCommon.GetCounterTypeName(adqCntrModuleInfo.CounterType)));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Counter width: {0} Bit, max. start value: {1}, hex 0x{1:X}", 
                        adqCntrModuleInfo.WidthBits, GetCounterMaxStartValue(adqCntrModuleInfo.WidthBits)));
                parentTreeNode.Nodes.Add(treeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }

                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                AddMaximumSampleFrequency(boardIndex, moduleIndex, adqGeneralModuleInfo.ClockType, parentTreeNode);
                AddCounterClockSources(parentTreeNode, adqCntrModuleInfo.ClockSources);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private uint GetCounterMaxStartValue(uint widthBits)
        {
            return (uint)((1UL << (int)widthBits) - 1UL);
        }

        private void AddMaximumSampleFrequency(uint boardIndex, uint moduleIndex, uint clockType, TreeNode parentTreeNode)
        {
            if (!IsClockControlledModule(clockType)) return;

            uint adqError = AdqDriver.adqGetMaximumSampleFrequency(boardIndex, moduleIndex, 1, out double maxSampleFrequency);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                var childNode = new TreeNode(string.Format("Maximum sample frequency: {0}", AdqCommon.FormatDoubleValue(maxSampleFrequency, "Hz")));
                parentTreeNode.Nodes.Add(childNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetMaximumSampleFrequency", adqError);
            }
        }

        private bool IsClockControlledModule(uint clockType)
        {
            return clockType == AdqConstants.ADQ_CLOCK_TYPE_HARDWARE || clockType == AdqConstants.ADQ_CLOCK_TYPE_SOFTWARE;
        }

        private void AddCounterClockSources(TreeNode parentTreeNode, uint clockSources)
        {
            var clockSourcesTreeNode = new TreeNode("Counter clock sources");
            parentTreeNode.Nodes.Add(clockSourcesTreeNode);

            AddCounterClockSourceTreeNode(clockSourcesTreeNode, clockSources, AdqConstants.ADQ_CNTR_CLOCK_SOURCE_EXTERNAL);
            AddCounterClockSourceTreeNode(clockSourcesTreeNode, clockSources, AdqConstants.ADQ_CNTR_CLOCK_SOURCE_INTERNAL_10_MHZ);
            AddCounterClockSourceTreeNode(clockSourcesTreeNode, clockSources, AdqConstants.ADQ_CNTR_CLOCK_SOURCE_INTERNAL_66_MHZ);
        }

        private void AddCounterClockSourceTreeNode(TreeNode parentTreeNode, uint clockSources, uint clockSourceConstant)
        {
            if ((clockSources & clockSourceConstant) == clockSourceConstant)
            {
                var treeNode = new TreeNode(AdqCommon.GetCounterClockSourceName(clockSourceConstant));
                parentTreeNode.Nodes.Add(treeNode);
            }
        }

        private void GetExternalIrqInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);

            const uint moduleInfoSize = 0;
            uint adqError = AdqDriver.adqGetModuleInfo(boardIndex, moduleIndex, ref adqGeneralModuleInfo, moduleInfoSize, IntPtr.Zero, out uint infoSize);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(moduleInfoSize == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, moduleInfoSize, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);
                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private void GetAIModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqAiModuleInfo = new AdqDriver.ADQAIModuleInfo(AdqDriver.ADQAIModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqAiModuleInfo.Size);
            Marshal.StructureToPtr(adqAiModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqAiModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqAiModuleInfo = (AdqDriver.ADQAIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQAIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqAiModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqAiModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }

                TreeNode treeNode = new TreeNode(
                    string.Format("Shared data rate: {0}", adqAiModuleInfo.SharedDataRate ? "Yes" : "No"));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Number of channels: {0}", adqAiModuleInfo.NumberOfChannels));
                parentTreeNode.Nodes.Add(treeNode);

                AddCurrentHardwareRange(boardIndex, moduleIndex, adqAiModuleInfo.NumberOfChannels, treeNode);
                AddSelectedCalibration(boardIndex, moduleIndex, parentTreeNode);
                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                AddAiMaximumSampleFrequencyList(boardIndex, moduleIndex, adqAiModuleInfo.NumberOfChannels, 
                    adqGeneralModuleInfo.ClockType, parentTreeNode);

                if (adqAiModuleInfo.RangeInfoPerChannel)
                {
                    treeNode = new TreeNode("Variable ranges per channel encountered; Please use a newer version of this program.");
                    parentTreeNode.Nodes.Add(treeNode);
                    return;
                }

                AddAiRangesInfo(parentTreeNode, ref adqAiModuleInfo);
            }
        }

        private void AddSelectedCalibration(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            uint adqError = AdqDriver.adqAnalogGetSelectedCalibration(boardIndex, moduleIndex, out uint selectedCalibration);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                var childNode = new TreeNode(string.Format("Selected calibration: {0}", AdqCommon.GetCalibrationText(selectedCalibration)));
                parentTreeNode.Nodes.Add(childNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqAnalogGetSelectedCalibration", adqError);
            }
        }

        private void AddHardwareTriggerInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var hwTriggerInfo = new AdqDriver.ADQHWTriggerInfo(AdqDriver.ADQHWTriggerInfo_ID);
            uint adqError = AdqDriver.adqGetHWTriggerInfo(boardIndex, moduleIndex, ref hwTriggerInfo);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                var hwTriggerTreeNode = new TreeNode("Hardware trigger info");
                parentTreeNode.Nodes.Add(hwTriggerTreeNode);

                if (hwTriggerInfo.HWTriggerStartSource == 0 
                    && hwTriggerInfo.HWTriggerStopSource == 0 
                    && !hwTriggerInfo.SupportsChannelListPerTrigger)
                {
                    hwTriggerTreeNode.Nodes.Add(new TreeNode("No hardware trigger supported"));
                    return;
                }

                var treeNode = new TreeNode(
                    string.Format("Hardware trigger start source: 0x{0:X8}", hwTriggerInfo.HWTriggerStartSource));
                hwTriggerTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Hardware trigger stop source: 0x{0:X8}", hwTriggerInfo.HWTriggerStopSource));
                hwTriggerTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Supports channel list per trigger: {0}", hwTriggerInfo.SupportsChannelListPerTrigger ? "Yes" : "No"));
                hwTriggerTreeNode.Nodes.Add(treeNode);
            }
        }

        private void AddAiMaximumSampleFrequencyList(uint boardIndex, uint moduleIndex, uint numberOfChannels, uint clockType, TreeNode parentTreeNode)
        {
            if (!IsClockControlledModule(clockType)) return;

            var treeNode = new TreeNode("Maximum sample frequencies");
            parentTreeNode.Nodes.Add(treeNode);

            for (uint numChannels = 1; numChannels <= numberOfChannels; numChannels++)
            {
                uint adqError = AdqDriver.adqGetMaximumSampleFrequency(boardIndex, moduleIndex, numChannels, out double maxSampleFrequency);
                if (adqError == AdqErrors.ADQ_ERROR_NONE)
                {
                    string channels = numChannels == 1 ? "channel" : "channels";
                    var childNode = new TreeNode(string.Format("{0} {1}: {2}",
                        numChannels, channels, AdqCommon.FormatDoubleValue(maxSampleFrequency, "Hz")));
                    treeNode.Nodes.Add(childNode);
                }
                else
                {
                    AddFunctionErrorTreeNode(treeNode, "adqGetMaximumSampleFrequency", adqError);
                }
            }
        }

        private void AddAiRangesInfo(TreeNode parentTreeNode, ref AdqDriver.ADQAIModuleInfo adqAiModuleInfo)
        {
            var treeNode = new TreeNode(
                string.Format("Number of Ranges: {0}", adqAiModuleInfo.NumberOfRanges));
            parentTreeNode.Nodes.Add(treeNode);

            for (int rangeIdx = 0; rangeIdx < (int)adqAiModuleInfo.NumberOfRanges; rangeIdx++)
            {
                AdqDriver.ADQAIRange aiRange = adqAiModuleInfo.Ranges[rangeIdx];
                AddRangeInfo(parentTreeNode,
                    rangeIdx,
                    aiRange.DigitalMin,
                    aiRange.DigitalMax,
                    aiRange.PhysicalUnit,
                    aiRange.PhysicalMin,
                    aiRange.PhysicalMax);
            }
        }

        private void AddCurrentHardwareRange(uint boardIndex, uint moduleIndex, uint numberOfChannels, TreeNode parentTreeNode)
        {
            for (uint channelIdx = 0; channelIdx < numberOfChannels; channelIdx++)
            {
                uint adqError = AdqDriver.adqAIGetChannelHardwareRange(boardIndex, moduleIndex, channelIdx, out uint rangeIndex);
                if (adqError == AdqErrors.ADQ_ERROR_NONE)
                {
                    var treeNode = new TreeNode(string.Format("Channel {0} current range: {1}", channelIdx, rangeIndex));
                    parentTreeNode.Nodes.Add(treeNode);
                }
                else
                {
                    // The API-function adqAIGetChannelHardwareRange is only supported for AI modules with safety switching on excessive range overflow, e.g. ADQ-412
                    if (!IsErrorUnsuppored(adqError))
                    {
                        AddFunctionErrorTreeNode(parentTreeNode, "adqAIGetChannelHardwareRange", adqError);
                    }
                    break;
                }
            }
        }

        private bool IsErrorUnsuppored(uint adqError)
        {
            return adqError == (AdqErrors.ADQ_ERROR_TYPE_ERROR | AdqErrors.ADQ_ERROR_PARAMETER_2 | AdqErrors.ADQ_ERROR_UNSUPPORTED);
        }

        private void GetAOModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqAoModuleInfo = new AdqDriver.ADQAOModuleInfo(AdqDriver.ADQAOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqAoModuleInfo.Size);
            Marshal.StructureToPtr(adqAoModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqAoModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqAoModuleInfo = (AdqDriver.ADQAOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQAOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqAoModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqAoModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);
                AddSelectedCalibration(boardIndex, moduleIndex, parentTreeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }

                AddHardwareTriggerInfo(boardIndex, moduleIndex, parentTreeNode);
                AddAoRangesInfo(parentTreeNode, ref adqAoModuleInfo);
            }
        }

        private void AddAoRangesInfo(TreeNode parentTreeNode, ref AdqDriver.ADQAOModuleInfo adqAoModuleInfo)
        {
            TreeNode treeNode = new TreeNode(
                string.Format("Number of ranges: {0}", adqAoModuleInfo.NumberOfRanges));
            parentTreeNode.Nodes.Add(treeNode);

            for (int rangeIdx = 0; rangeIdx < (int)adqAoModuleInfo.NumberOfRanges; rangeIdx++)
            {
                AdqDriver.ADQAORange aoRange = adqAoModuleInfo.Ranges[rangeIdx];

                AddRangeInfo(parentTreeNode, 
                    rangeIdx, 
                    aoRange.DigitalMin, 
                    aoRange.DigitalMax, 
                    aoRange.PhysicalUnit, 
                    aoRange.PhysicalMin, 
                    aoRange.PhysicalMax);
            }
        }

        private static void AddRangeInfo(
            TreeNode parentTreeNode, 
            int rangeIndex, 
            uint digitalMin, 
            uint digitalMax, 
            uint physicalUnit, 
            double physicalMin, 
            double physicalMax)
        {
            TreeNode rangeTreeNode;
            rangeTreeNode = new TreeNode(string.Format("Range {0}", rangeIndex));
            parentTreeNode.Nodes.Add(rangeTreeNode);

            string unitName = AdqCommon.GetPhysicalUnitName(physicalUnit);
            TreeNode treeNode = new TreeNode(
                string.Format("Physical unit: {0}", unitName));
            rangeTreeNode.Nodes.Add(treeNode);

            treeNode = new TreeNode(
                string.Format("Physical range: {0:0.######} ... {1:0.#######} {2}",
                    physicalMin, physicalMax, unitName));
            rangeTreeNode.Nodes.Add(treeNode);

            treeNode = new TreeNode(
                string.Format("Digital range: {0} ... {1}, hex 0x{0:X} ... 0x{1:X}", digitalMin, digitalMax));
            rangeTreeNode.Nodes.Add(treeNode);

            double physResolution = (physicalMax - physicalMin) / (digitalMax - digitalMin);
            treeNode = new TreeNode("Physical resolution: " 
                + AdqCommon.FormatDoubleValue(physResolution, AdqCommon.GetPhysicalUnitSymbol(physicalUnit)));
            rangeTreeNode.Nodes.Add(treeNode);
        }

        private void GetI2CModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqI2cModuleInfo = new AdqDriver.ADQI2CModuleInfo(AdqDriver.ADQI2CModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqI2cModuleInfo.Size);
            Marshal.StructureToPtr(adqI2cModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqI2cModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqI2cModuleInfo = (AdqDriver.ADQI2CModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQI2CModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqI2cModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqI2cModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);

                TreeNode treeNode = new TreeNode(
                    string.Format("Data transfer rate: {0}", AdqCommon.FormatBitsPerSecond(adqI2cModuleInfo.SpeedBitsPerSecond)));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Supports 7-bit addresses: {0}", adqI2cModuleInfo.Supports7BitAddresses ? "Yes" : "No"));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Supports 10-bit addresses: {0}", adqI2cModuleInfo.Supports10BitAddresses ? "Yes" : "No"));
                parentTreeNode.Nodes.Add(treeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }
            }
        }

        private void GetPWMModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqPwmModuleInfo = new AdqDriver.ADQPWMModuleInfo(AdqDriver.ADQPWMModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqPwmModuleInfo.Size);
            Marshal.StructureToPtr(adqPwmModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqPwmModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqPwmModuleInfo = (AdqDriver.ADQPWMModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQPWMModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(adqPwmModuleInfo.Size == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, adqPwmModuleInfo.Size, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);

                TreeNode treeNode = new TreeNode(
                    string.Format("Base frequency: {0}", AdqCommon.FormatDoubleValue(adqPwmModuleInfo.BaseFrequencyHz, "Hz")));
                parentTreeNode.Nodes.Add(treeNode);

                treeNode = new TreeNode(
                    string.Format("Minimum divider: {0}", adqPwmModuleInfo.MinimumDivider));
                parentTreeNode.Nodes.Add(treeNode);

                AddPWMMaximumFrequency(boardIndex, moduleIndex, parentTreeNode);

                if (adqGeneralModuleInfo.IsIRQSource)
                {
                    AddInterruptInfo(boardIndex, moduleIndex, parentTreeNode);
                }
            }
        }

        private void AddPWMMaximumFrequency(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            uint adqError = AdqDriver.adqPWMGetMaximumFrequency(boardIndex, moduleIndex, out double maxFrequency);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                TreeNode treeNode = new TreeNode(
                    string.Format("Max frequency: {0}", AdqCommon.FormatDoubleValue(maxFrequency, "Hz")));
                parentTreeNode.Nodes.Add(treeNode);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqPWMGetMaximumFrequency", adqError);
            }
        }

        private void GetFINModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            GetGeneralModuleInfo(boardIndex, moduleIndex, parentTreeNode);
        }

        private void GetI2SOModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            GetGeneralModuleInfo(boardIndex, moduleIndex, parentTreeNode);
        }

        private void GetGeneralModuleInfo(uint boardIndex, uint moduleIndex, TreeNode parentTreeNode)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);

            const uint moduleInfoSize = 0;
            uint adqError = AdqDriver.adqGetModuleInfo(boardIndex, moduleIndex, ref adqGeneralModuleInfo, moduleInfoSize, IntPtr.Zero, out uint infoSize);

            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                Debug.Assert(moduleInfoSize == infoSize,
                    string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                    boardIndex, moduleIndex, moduleInfoSize, infoSize));

                AddGeneralModuleInfo(parentTreeNode, ref adqGeneralModuleInfo);
            }
            else
            {
                AddFunctionErrorTreeNode(parentTreeNode, "adqGetModuleInfo", adqError);
            }
        }

        private void AddFunctionErrorTreeNode(TreeNode parentTreeNode, string functionName, uint adqError)
        {
            parentTreeNode.Nodes.Add(new TreeNode(FormatErrorText(functionName, adqError)));
        }

        private string FormatErrorText(string functionName, uint adqError)
        {
            return string.Format("API-function {0} returned error number 0x{1:X}", functionName, adqError);
        }

        private void CloseAdqDriver()
        {
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                _adqOpenError = AdqCommon.NotOpen;
            }
        }

        private void BtnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseAdqDriver();
        }

        private void BtnInfo_Click(object sender, EventArgs e)
        {
            var aboutBox = new AboutBox();
            aboutBox.ShowDialog();
        }

        private void CopyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (tvInfo.SelectedNode == null) return;

            Clipboard.SetText(tvInfo.SelectedNode.Text);
        }

        private void TvInfo_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Right) return;

            TreeNode clickedNode = tvInfo.GetNodeAt(e.X, e.Y);
            if (clickedNode == null) return;

            tvInfo.SelectedNode = clickedNode;
            contextMenuStrip1.Show(tvInfo, new System.Drawing.Point(e.X, e.Y));
        }
    }
}
