﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the AIReadContinuous project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using AIReadContinuous.Utility;
using ALLDAQ.Boards;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace AIReadContinuous.ALLDAQ
{
    public class AnalogInputModule
    {
        private const int _notificationsPerSecond = 10;

        private readonly double _maxFrequencyHz;
        private AdqDriver.ADQAIChannel[] _channelList;
        private BackgroundWorker _worker;
        private uint _notifyAfterFrames;
        private uint _adqError;
        private string _errorFunctionName;

        private uint[] _frames;
        private double[] _physicalFrames;
        private double[] _physicalFrames1;
        private double[] _physicalFrames2;
        private uint _physicalIndex;

        public event EventHandler<InfoTextChangedEventArgs> ModuleInfoTextChanged;
        public event EventHandler<ModulePropertyChangedEventArgs> ModulePropertyChanged;

        public AnalogInputModule(string boardName, uint boardIndex, uint moduleIndex, 
            ref AdqDriver.ADQGeneralModuleInfo generalInfo, ref AdqDriver.ADQAIModuleInfo aiModuleInfo)
        {
            ModelName = AdqCommon.GetModelName(boardName);
            BoardIndex = boardIndex;
            ModuleIndex = moduleIndex;
            NumberOfChannels = aiModuleInfo.NumberOfChannels;
            NumberOfRanges = aiModuleInfo.NumberOfRanges;
            _maxFrequencyHz = generalInfo.ClockBaseFrequencyHz / generalInfo.ClockMinimumDivider;
            DesiredSampleFrequency = Math.Round(_maxFrequencyHz * AdqCommon.DefaultDesiredSampleFrequencyFactor);
            BuildModuleName();
            DisplaySeconds = AdqCommon.DefaultDisplayNumberOfFrames / DesiredSampleFrequency; 

            AddRanges(ref aiModuleInfo);
            DefaultRange = Ranges[0];
            GetTriggerInfo(boardIndex, moduleIndex);
        }

        private void AddRanges(ref AdqDriver.ADQAIModuleInfo adqAiModuleInfo)
        {
            for (uint rangeIdx = 0; rangeIdx < adqAiModuleInfo.NumberOfRanges; rangeIdx++)
            {
                AdqDriver.ADQAIRange range = adqAiModuleInfo.Ranges[rangeIdx];

                var analogRange = new AnalogRange(
                    rangeIdx,
                    range.PhysicalUnit,
                    range.DigitalMin,
                    range.DigitalMax,
                    range.PhysicalMin,
                    range.PhysicalMax);
                Ranges.Add(analogRange);
            }
        }

        public string ModelName { get; }
        public uint BoardIndex { get; }
        public uint ModuleIndex { get; }
        public string ModuleName { get; private set; }
        public uint NumberOfChannels { get; }
        public uint NumberOfRanges { get; }
        public List<ChannelSetting> ChannelSettings { get; } = new List<ChannelSetting>();
        public HardwareTrigger HardwareTrigger { get; private set; }

        public string TriggerDescription
        {
            get { return HardwareTrigger != null ? HardwareTrigger.TriggerDescription : "No hardware trigger supported"; }
        }


        private void BuildModuleName()
        {
            ModuleName = string.Format("{0}, Board {1}, Module {2}, {3} {4}, {5} {6}, max. {7}",
                ModelName,
                BoardIndex,
                ModuleIndex,
                NumberOfChannels,
                NumberOfChannels == 1 ? "Channel" : "Channels",
                NumberOfRanges,
                NumberOfRanges == 1 ? "Range" : "Ranges",
                AdqCommon.FormatFrequency(_maxFrequencyHz));
        }

        private uint GetTriggerInfo(uint boardIndex, uint moduleIndex)
        {
            var hwTriggerInfo = new AdqDriver.ADQHWTriggerInfo(AdqDriver.ADQHWTriggerInfo_ID);
            uint adqError = AdqDriver.adqGetHWTriggerInfo(boardIndex, moduleIndex, ref hwTriggerInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorInfo("adqGetHWTriggerInfo", adqError);
            }
            else
            {
                if (hwTriggerInfo.HWTriggerStartSource != 0 || hwTriggerInfo.HWTriggerStopSource != 0)
                {
                    HardwareTrigger = new HardwareTrigger(BoardIndex, ModuleIndex, ref hwTriggerInfo);
                }
            }

            return adqError;
        }

        public void AddChannelSetting(uint channelIndex, AnalogRange range)
        {
            ChannelSettings.Add(new ChannelSetting(channelIndex, range));
        }

        public void UpdateChannelList()
        {
            if (_channelList == null)
            {
                _channelList = new AdqDriver.ADQAIChannel[NumberOfChannels];
            }

            for (int i = 0; i < _channelList.Length; i++)
            {
                _channelList[i].AcquireChannel = false;
                _channelList[i].RangeIndex = 0;
            }

            MinResolution = double.MaxValue;
            foreach (ChannelSetting chSetting in ChannelSettings)
            {
                _channelList[chSetting.ChannelIndex].AcquireChannel = true;
                _channelList[chSetting.ChannelIndex].RangeIndex = chSetting.Range.RangeIndex;
                if (MinResolution > chSetting.Range.PhysicalResolution)
                {
                    MinResolution = chSetting.Range.PhysicalResolution;
                }
            }
            FrameLength = (uint)ChannelSettings.Count;
        }

        public void InitAcquisition()
        {
            if (ModuleIsBusy) return;
            Reset();
            ConfigureModule();
            ModuleIsBusy = true;
            if (_worker == null)
            {
                _worker = new BackgroundWorker
                {
                    WorkerReportsProgress = true,
                    WorkerSupportsCancellation = true
                };
                _worker.DoWork += Worker_DoWork;
                _worker.ProgressChanged += Worker_ProgressChanged;
                _worker.RunWorkerCompleted += Worker_RunWorkerCompleted;
            }
        }

        private void ConfigureModule()
        {
            CheckCommonSettings();
            SetChannelListAndFrequencySampleTime();

            if (DisplaySeconds < SampleTime)
            {
                throw new ApplicationException(string.Format("The display time ({0} sec) should be larger than the sample time {1} sec.", DisplaySeconds, SampleTime));
            }

            uint displayNumberOfFrames = (uint)(DisplaySeconds * ActualSampleFrequency);
            _notifyAfterFrames = Math.Min(displayNumberOfFrames, (uint)(ActualSampleFrequency / _notificationsPerSecond));

            if (StopAfterFrames > 0 && StopAfterFrames < _notifyAfterFrames)
            {
                _notifyAfterFrames = (uint)StopAfterFrames;
            }

            uint numberOfValues = _notifyAfterFrames * FrameLength;

            if (_frames == null || _frames.Length != numberOfValues)
            {
                _frames = new uint[numberOfValues];
            }

            if (StopAfterFrames > 0 && StopAfterFrames < displayNumberOfFrames)
            {
                displayNumberOfFrames = (uint)StopAfterFrames;
            }

            DisplayBufferNumberOfFrames = displayNumberOfFrames;

            uint numberOfPhysicalValues = displayNumberOfFrames * FrameLength; 

            if (_physicalFrames1 == null || _physicalFrames1.Length != numberOfPhysicalValues)
            {
                _physicalFrames1 = new double[numberOfPhysicalValues];
                _physicalFrames2 = new double[numberOfPhysicalValues];
            }
            DisplayBufferLength = numberOfPhysicalValues;
            RaisePropertyChanged(AdqCommon.PropertyType.DisplayBufferLength);
        }

        private void CheckCommonSettings()
        {
            if (DesiredSampleFrequency == 0)
            {
                throw new ApplicationException("Please enter a valid sample frequency.");
            }

            if (DisplaySeconds == 0)
            {
                throw new ApplicationException("Please enter a valid display time.");
            }

            if (FrameLength == 0)
            {
                throw new ApplicationException("Please select channels in the channel list.");
            }
        }

        private bool SetChannelListAndFrequencySampleTime()
        {
            uint adqError = AdqDriver.adqAISetChannelListAndFrequency(BoardIndex, ModuleIndex,
                 _channelList, NumberOfChannels, DesiredSampleFrequency, out double sampleFrequency);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                UpdateActualSampleFrequencyAndSampleTime(sampleFrequency);
            }
            else
            {
                throw new ApplicationException(string.Format("Function adqAISetChannelListAndFrequency returned Error number: 0x{0:X}", adqError));
            }
            return adqError == AdqErrors.ADQ_ERROR_NONE;
        }

        private bool IsUsingStartTrigger()
        {
            return HardwareTrigger != null && HardwareTrigger.StartTriggerEnabled;
        }

        private void EnableHardwareTrigger()
        {
            if (HardwareTrigger != null)
            {
                HardwareTrigger.EnableHardwareTrigger();
            }
        }

        public void StartAcquisition()
        {
            UpdateModuleInfo("Acquisition initialized");
            _worker.RunWorkerAsync();
        }

        private void Worker_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker worker = sender as BackgroundWorker;

            _errorFunctionName = string.Empty;
            TotalFramesAcquired = 0;
            TotalFramesDisplayed = 0;
            TotalFramesSkipped = 0;
            worker.ReportProgress((int)AdqCommon.PropertyType.AcquisitionStarting);

            _physicalFrames = _physicalFrames1;
            _physicalIndex = 0;
            GCHandle gchFrames = GCHandle.Alloc(_frames, GCHandleType.Pinned);

            if (IsUsingStartTrigger())
            {
                worker.ReportProgress((int)AdqCommon.PropertyType.WaitingForStartTrigger);
            }
            EnableHardwareTrigger();

            _adqError = AdqDriver.adqAIReadContinuousStart(BoardIndex,
                                     ModuleIndex,
                                     _notifyAfterFrames,
                                     StopAfterFrames,
                                     null,
                                     IntPtr.Zero);

            if (_adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                gchFrames.Free();
                _errorFunctionName = "adqAIReadContinuousStart";
                return;
            }

            if (IsUsingStartTrigger())
            {
                worker.ReportProgress((int)AdqCommon.PropertyType.StartTriggerEventOccured);
            }
            else
            {
                worker.ReportProgress((int)AdqCommon.PropertyType.AcquisitionStarted);
            }


            while (true)
            {
                if (worker.CancellationPending)
                {
                    e.Cancel = true;
                    break;
                }

                _adqError = AdqDriver.adqAIReadContinuous(BoardIndex, ModuleIndex, FrameLength, _notifyAfterFrames, _frames, true, out uint numberOfFramesReturned);
                TotalFramesAcquired += numberOfFramesReturned;
                if (_adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    if (StopAfterFrames > 0 && StopAfterFrames == TotalFramesAcquired)
                    {
                        _adqError = AdqErrors.ADQ_ERROR_NONE;
                        break;
                    }

                    if (numberOfFramesReturned > 0)
                    {
                        ProcessAcquiredFrames(worker, numberOfFramesReturned);
                    }

                    _errorFunctionName = "adqAIReadContinuous";
                    break;
                }

                if (numberOfFramesReturned > 0)
                {
                    ProcessAcquiredFrames(worker, numberOfFramesReturned);
                }
            }
            gchFrames.Free();
            AdqDriver.adqStop(BoardIndex, ModuleIndex);
        }

        private void ProcessAcquiredFrames(BackgroundWorker worker, uint numberOfFramesReturned)
        {
            CalculatePhysicalFrames(numberOfFramesReturned);

            if (RequestUpdateNumberOfFrames)
            {
                RequestUpdateNumberOfFrames = false;
                worker.ReportProgress((int)AdqCommon.PropertyType.UpdateNumberOfFrames);
            }
        }

        private void CalculatePhysicalFrames(uint numberOfFramesReturned)
        {
            uint numberOfValuesReturned = numberOfFramesReturned * FrameLength;
            uint digitalIndex = 0;

            while (digitalIndex < numberOfValuesReturned)
            {
                uint processed = ProcessFrames(numberOfValuesReturned, digitalIndex);
                digitalIndex += processed;
                _physicalIndex += processed;
                if (_physicalIndex >= DisplayBufferLength)
                {
                    SwitchDisplayFrameBuffer();
                    _physicalIndex = 0;
                }
            }
        }

        private uint ProcessFrames(uint numberOfValuesReturned, uint digitalIndex)
        {
            uint numProcessed = 0;
            uint valIdx;
            uint physValIndex;
            for (uint idx = 0; idx < FrameLength; idx++)
            {
                ChannelSetting channelSetting = ChannelSettings[(int)idx];
                AnalogRange range = channelSetting.Range;

                physValIndex = _physicalIndex + idx;
                valIdx = digitalIndex + idx;
                while (valIdx < numberOfValuesReturned && physValIndex < DisplayBufferLength)
                {
                    _physicalFrames[physValIndex] = range.DigitalToPhysical(_frames[valIdx]);
                    valIdx += FrameLength;
                    physValIndex += FrameLength;
                    numProcessed++;
                }
            }
            return numProcessed;
        }

        private void SwitchDisplayFrameBuffer()
        {
            if (RequestFramesToDisplay)
            {
                RequestFramesToDisplay = false;
                if (ReferenceEquals(_physicalFrames, _physicalFrames1))
                {
                    DisplayBuffer = _physicalFrames1;
                    _physicalFrames = _physicalFrames2;
                }
                else
                {
                    DisplayBuffer = _physicalFrames2;
                    _physicalFrames = _physicalFrames1;
                }
                TotalFramesDisplayed += DisplayBufferNumberOfFrames;
                RaisePropertyChanged(AdqCommon.PropertyType.DisplayBuffer);
            }
            else
            {
                TotalFramesSkipped += DisplayBufferNumberOfFrames;
            }
        }

        private void Worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            RaisePropertyChanged((AdqCommon.PropertyType)e.ProgressPercentage);
        }

        private void Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            ModuleIsBusy = false;

            RaisePropertyChanged(AdqCommon.PropertyType.AcquisitionEnded);
            RaisePropertyChanged(AdqCommon.PropertyType.UpdateNumberOfFrames);

            if (_adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorInfo(_errorFunctionName, _adqError);
                RaisePropertyChanged(AdqCommon.PropertyType.AcquisitionErrorOccurred);
            }
            UpdateModuleInfo("Acquisition ended.");
        }

        public void CancelAcquisition()
        {
            _worker.CancelAsync();
        }

        public uint StopAcquisition()
        {
            CancelAcquisition();

            uint adqError = AdqDriver.adqStop(BoardIndex, ModuleIndex);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorInfo("adqStop", adqError);
            }
            return adqError;
        }

        public uint Reset()
        {
            uint adqError = AdqDriver.adqResetModule(BoardIndex, ModuleIndex);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                UpdateModuleInfo(string.Format("Reset board {0} module {1}", BoardIndex, ModuleIndex));
            }
            else
            {
                SetErrorInfo("adqResetModule", adqError);
            }
            return adqError;
        }

        private void SetErrorInfo(string functionName, uint adqError)
        {
            UpdateModuleInfo(string.Format("Function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateModuleInfo(string info)
        {
            if (ModuleInfoTextChanged != null)
            {
                var args = new InfoTextChangedEventArgs
                {
                    InfoText = info
                };
                ModuleInfoTextChanged.Invoke(this, args);
            }
        }

        public bool ModuleIsBusy { get; private set; }
        public bool RequestFramesToDisplay { get; set; }
        public bool RequestUpdateNumberOfFrames { get; set; }
        public ulong TotalFramesAcquired { get; private set; }
        public ulong TotalFramesDisplayed { get; private set; }
        public ulong TotalFramesSkipped { get; private set; }
        public ulong StopAfterFrames { get; set; }
        public double DisplaySeconds { get; set; }
        public uint FrameLength { get; set; }
        public uint DisplayBufferNumberOfFrames { get; private set; }
        public uint DisplayBufferLength { get; private set; }
        public double[] DisplayBuffer { get; private set; }
        public double MinResolution { get; private set; }

        private void RaisePropertyChanged(AdqCommon.PropertyType propertyType)
        {
            if (ModulePropertyChanged != null)
            {
                var args = new ModulePropertyChangedEventArgs
                {
                    ChangedProperty = propertyType
                };
                ModulePropertyChanged.Invoke(this, args);
            }
        }

        private void UpdateActualSampleFrequencyAndSampleTime(double sampleFrequency)
        {
            ActualSampleFrequency = sampleFrequency;
            SampleTime = 1.0 / sampleFrequency;

            RaisePropertyChanged(AdqCommon.PropertyType.ActualSampleFrequency);
        }

        public double ActualSampleFrequency { get; private set; }
        public double SampleTime { get; private set; }
        public double DesiredSampleFrequency { get; set; }
        public AnalogRange DefaultRange { get; }

        public List<AnalogRange> Ranges { get; } = new List<AnalogRange>();

        public void ClearModuleResources()
        {
            _worker = null;
            _frames = null;
            _physicalFrames1 = null;
            _physicalFrames2 = null;
        }
    }
}
