﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the AIReadContinuous project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using AIReadContinuous.Utility;
using ALLDAQ.Boards;
using System;

namespace AIReadContinuous.ALLDAQ
{
    public class HardwareTrigger
    {
        private AdqDriver.ADQHWTriggerEnable _hardwareTriggerEnable = new AdqDriver.ADQHWTriggerEnable(AdqDriver.ADQHWTriggerEnable_ID);
        private readonly uint _boardIndex;
        private readonly uint _moduleIndex;

        public HardwareTrigger(uint boardIndex, uint moduleIndex, ref AdqDriver.ADQHWTriggerInfo hwTriggerInfo)
        {
            _boardIndex = boardIndex;
            _moduleIndex = moduleIndex;
            TriggerStartSource = hwTriggerInfo.HWTriggerStartSource;
            TriggerStopSource = hwTriggerInfo.HWTriggerStopSource;
            SupportsChannelListPerTrigger = hwTriggerInfo.SupportsChannelListPerTrigger;
            TriggerTimeoutMilliseconds = AdqCommon.DefaultTriggerTimeout;
            TriggerDescription = AdqCommon.GetTriggerDescription(ref hwTriggerInfo);
        }

        public string TriggerDescription { get; }
        public uint TriggerStartSource { get; }
        public uint TriggerStopSource { get; }
        public bool SupportsChannelListPerTrigger { get; }

        public uint TriggerStartRisingEdges
        {
            get { return _hardwareTriggerEnable.HWTriggerStartRisingEdge; }
            set { _hardwareTriggerEnable.HWTriggerStartRisingEdge = value; }
        }

        public uint TriggerStartFallingEdges
        {
            get { return _hardwareTriggerEnable.HWTriggerStartFallingEdge; }
            set { _hardwareTriggerEnable.HWTriggerStartFallingEdge = value; }
        }

        public uint TriggerStopRisingEdges
        {
            get { return _hardwareTriggerEnable.HWTriggerStopRisingEdge; }
            set { _hardwareTriggerEnable.HWTriggerStopRisingEdge = value; }
        }

        public uint TriggerStopFallingEdges
        {
            get { return _hardwareTriggerEnable.HWTriggerStopFallingEdge; }
            set { _hardwareTriggerEnable.HWTriggerStopFallingEdge = value; }
        }

        public bool ChannelListPerTrigger
        {
            get { return _hardwareTriggerEnable.ChannelListPerTrigger; }
            set { _hardwareTriggerEnable.ChannelListPerTrigger = value; }
        }

        public bool TriggerAutomaticRearm
        {
            get { return _hardwareTriggerEnable.AutomaticRearm; }
            set { _hardwareTriggerEnable.AutomaticRearm = value; }
        }

        public bool UseTriggerTimeout { get; set; }
        public uint TriggerTimeoutMilliseconds { get; set; }

        public bool StartTriggerEnabled
        {
            get { return _hardwareTriggerEnable.HWTriggerStartRisingEdge != 0 || _hardwareTriggerEnable.HWTriggerStartFallingEdge != 0; }
        }

        public bool EnableHardwareTrigger()
        {
            uint timeout = GetTimeoutValue();

            uint adqError = AdqDriver.adqEnableHWTrigger(_boardIndex, _moduleIndex, ref _hardwareTriggerEnable, true, timeout);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                throw new ApplicationException(string.Format("Function adqEnableHWTrigger returned Error number: 0x{0:X}", adqError));
            }

            return adqError == AdqErrors.ADQ_ERROR_NONE;
        }

        private uint GetTimeoutValue()
        {
            uint timeout = 0;
            if (_hardwareTriggerEnable.HWTriggerStartRisingEdge != 0 || _hardwareTriggerEnable.HWTriggerStartFallingEdge != 0)
            {
                timeout = UseTriggerTimeout ? TriggerTimeoutMilliseconds : AdqConstants.ADQ_HW_START_TRIGGER_TIMEOUT_INFINITE;
            }
            return timeout;
        }
    }
}
