﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the AIReadInterval project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using AIReadInterval.Utility;
using ALLDAQ_DLL.Utilities;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace AIReadInterval.ALLDAQ
{
    public enum SignalValuesType : uint
    {
        AverageValues,
        RMSValues,
        MinimumValues,
        MaximumValues,
        P2PValues,
        MainFrequencies,
        FrequProportions,
        ThdValues,
        ThdrPlusNValues,
        ThdfPlusNValues,
        RMSValuesSpectrum,
        MainFrequenciesCoherent,
        FrequProportionsCoherent,
        ThdValuesCoherent,
        ThdrPlusNValuesCoherent,
        ThdfPlusNValuesCoherent,
        RMSValuesSpectrumCoherent
    };

    public class SignalUtility
    {
        private const double _dBuRelativeSize = 0.7746;

        private AnalogInputModule _aiModule;
        private List<ChannelSetting> _channelSettings;
        private double _sineFrequencyHz;
        private double _sampleFrequency;
        private double _overlap;
        private double _frequencyIgnoreBelow;
        private double _frequencyIgnoreAbove;

        public event EventHandler<InfoTextChangedEventArgs> OnInfoTextChanged;

        public uint FrameLength
        {
            get; private set;
        }
        public uint SpectrumSize
        {
            get; private set;
        }
        public StringValueItem<uint> WindowFunctionItem
        {
            get; private set;
        }

        public uint GetChannelIndex(int idx)
        {
            return _channelSettings[idx].ChannelIndex;
        }

        private double _frequencyStep;
        public double FrequencyStep
        {
            get
            {
                return _frequencyStep;
            }
        }

        private double _frequencyStepCoherent;
        public double FrequencyStepCoherent
        {
            get
            {
                return _frequencyStepCoherent;
            }
        }

        public List<double[]> FrequencySpectrums { get; } = new List<double[]>();
        public List<double[]> FrequencySpectrumsCoherent { get; } = new List<double[]>();
        public List<double[]> FrequencySpectrumsCoherentDBU { get; } = new List<double[]>();
        public List<double[]> FrequencySpectrumsDBU { get; } = new List<double[]>();

        public double[] RmsValues
        {
            get; private set;
        }
        public double[] MinValues
        {
            get; private set;
        }
        public double[] MaxValues
        {
            get; private set;
        }
        public double[] AverageValues
        {
            get; private set;
        }
        public double[] P2PValues
        {
            get; private set;
        }
        public double[] MainFrequencies
        {
            get; private set;
        }
        public double[] FrequProportions
        {
            get; private set;
        }
        public double[] ThdValues
        {
            get; private set;
        }
        public double[] ThdrPlusNValues
        {
            get; private set;
        }
        public double[] ThdfPlusNValues
        {
            get; private set;
        }
        public double[] RmsValuesSpectrum
        {
            get; private set;
        }
        public double[] MainFrequenciesCoherent
        {
            get; private set;
        }
        public double[] FrequProportionsCoherent
        {
            get; private set;
        }
        public double[] ThdValuesCoherent
        {
            get; private set;
        }
        public double[] ThdrPlusNValuesCoherent
        {
            get; private set;
        }
        public double[] ThdfPlusNValuesCoherent
        {
            get; private set;
        }
        public double[] RmsValuesSpectrumCoherent
        {
            get; private set;
        }
        public double[] SelectedValuesType
        {
            get; private set;
        }

        private SignalValuesType _signalValuesType = SignalValuesType.AverageValues;
        public SignalValuesType SignalValuesType
        {
            get
            {
                return _signalValuesType;
            }
            set
            {
                if (_signalValuesType != value)
                {
                    _signalValuesType = value;
                    SetSelectedValueType();
                }
            }
        }
        private void SetSelectedValueType()
        {
            switch (_signalValuesType)
            {
                case SignalValuesType.AverageValues:
                    SelectedValuesType = AverageValues;
                    break;
                case SignalValuesType.RMSValues:
                    SelectedValuesType = RmsValues;
                    break;
                case SignalValuesType.MinimumValues:
                    SelectedValuesType = MinValues;
                    break;
                case SignalValuesType.MaximumValues:
                    SelectedValuesType = MaxValues;
                    break;
                case SignalValuesType.P2PValues:
                    SelectedValuesType = P2PValues;
                    break;
                case SignalValuesType.MainFrequencies:
                    SelectedValuesType = MainFrequencies;
                    break;
                case SignalValuesType.FrequProportions:
                    SelectedValuesType = FrequProportions;
                    break;
                case SignalValuesType.ThdValues:
                    SelectedValuesType = ThdValues;
                    break;
                case SignalValuesType.ThdrPlusNValues:
                    SelectedValuesType = ThdrPlusNValues;
                    break;
                case SignalValuesType.ThdfPlusNValues:
                    SelectedValuesType = ThdfPlusNValues;
                    break;
                case SignalValuesType.RMSValuesSpectrum:
                    SelectedValuesType = RmsValuesSpectrum;
                    break;
                case SignalValuesType.MainFrequenciesCoherent:
                    SelectedValuesType = MainFrequenciesCoherent;
                    break;
                case SignalValuesType.FrequProportionsCoherent:
                    SelectedValuesType = FrequProportionsCoherent;
                    break;
                case SignalValuesType.ThdValuesCoherent:
                    SelectedValuesType = ThdValuesCoherent;
                    break;
                case SignalValuesType.ThdrPlusNValuesCoherent:
                    SelectedValuesType = ThdrPlusNValuesCoherent;
                    break;
                case SignalValuesType.ThdfPlusNValuesCoherent:
                    SelectedValuesType = ThdfPlusNValuesCoherent;
                    break;
                case SignalValuesType.RMSValuesSpectrumCoherent:
                    SelectedValuesType = RmsValuesSpectrumCoherent;
                    break;
                default:
                    SelectedValuesType = AverageValues;
                    break;
            }
        }

        public void InitFftSettings(AnalogInputModule aiModule, uint spectrumSize, double sineFrequencyHz, StringValueItem<uint> windowFunctionItem,
            double overlap, double frequencyIgnoreBelow, double frequencyIgnoreAbove)
        {
            CheckFftValues(spectrumSize, sineFrequencyHz, overlap, frequencyIgnoreBelow, frequencyIgnoreAbove);
            _aiModule = aiModule;
            _channelSettings = aiModule.ChannelSettings;
            _sampleFrequency = aiModule.ActualSampleFrequency;
            _sineFrequencyHz = sineFrequencyHz;
            WindowFunctionItem = windowFunctionItem;
            _overlap = overlap;
            _frequencyIgnoreBelow = frequencyIgnoreBelow;
            _frequencyIgnoreAbove = frequencyIgnoreAbove;
            InitFrequencySpectrums(spectrumSize, aiModule.FrameLength);
        }

        private static void CheckFftValues(uint spectrumSize, double sineFrequencyHz, double overlap, double frequencyIgnoreBelow, double frequencyIgnoreAbove)
        {
            if (spectrumSize == 0)
            {
                throw new ApplicationException("The spectrum size must be greater than 0.");
            }
            if (sineFrequencyHz < 0)
            {
                throw new ApplicationException("The sine frequency must not be negative.");
            }
            if (overlap < 0 || overlap > 0.98)
            {
                throw new ApplicationException("The overlap value is not within the valid range from 0 to 0.98.");
            }
            if (frequencyIgnoreBelow < 0 || frequencyIgnoreAbove < 0)
            {
                throw new ApplicationException("The filter frequency values must not be negative.");
            }
        }

        private void InitFrequencySpectrums(uint spectrumSize, uint frameLength)
        {
            if (SpectrumSize != spectrumSize || FrameLength != frameLength)
            {
                FrequencySpectrums.Clear();
                FrequencySpectrumsCoherent.Clear();
                FrequencySpectrumsDBU.Clear();
                FrequencySpectrumsCoherentDBU.Clear();
                if (spectrumSize > 0)
                {
                    for (uint idx = 0; idx < frameLength; idx++)
                    {
                        FrequencySpectrums.Add(new double[spectrumSize]);
                        FrequencySpectrumsCoherent.Add(new double[spectrumSize]);
                        FrequencySpectrumsDBU.Add(new double[spectrumSize]);
                        FrequencySpectrumsCoherentDBU.Add(new double[spectrumSize]);
                    }
                }
                SpectrumSize = spectrumSize;

                if (FrameLength != frameLength)
                {
                    FrameLength = frameLength;
                    RmsValues = new double[frameLength];
                    MinValues = new double[frameLength];
                    MaxValues = new double[frameLength];
                    AverageValues = new double[frameLength];
                    P2PValues = new double[frameLength];
                    MainFrequencies = new double[frameLength];
                    FrequProportions = new double[frameLength];
                    ThdValues = new double[frameLength];
                    ThdrPlusNValues = new double[frameLength];
                    ThdfPlusNValues = new double[frameLength];
                    RmsValuesSpectrum = new double[frameLength];

                    MainFrequenciesCoherent = new double[frameLength];
                    FrequProportionsCoherent = new double[frameLength];
                    ThdValuesCoherent = new double[frameLength];
                    ThdrPlusNValuesCoherent = new double[frameLength];
                    ThdfPlusNValuesCoherent = new double[frameLength];
                    RmsValuesSpectrumCoherent = new double[frameLength];

                    SetSelectedValueType();
                }
            }
        }

        public void CalculateSignalValues()
        {
            CalculateFft();
            CalculateSignalSpectrumValues();
            CalculateSignalSpectrumValuesCoherent();
            CalculateSignalSpectrumValuesCoherentDBU();
            CalculateSignalSpectrumValuesDBU();
            CalculateSignalValues2();
        }

        private void CalculateFft()
        {
            AnalogRange range;
            uint windowFunction = WindowFunctionItem.Value;
            for (int idx = 0; idx < FrameLength; idx++)
            {
                double[] frequencySpectrum = FrequencySpectrums[idx];
                GCHandle gch = GCHandle.Alloc(frequencySpectrum, GCHandleType.Pinned);
                range = _channelSettings[idx].Range;

                uint adqError = AdqUtility.adqUtilityFrequencySpectrum(
                    _sampleFrequency, _aiModule.Frames, (uint)idx, FrameLength, _aiModule.NumberOfFramesRead,
                    range.DigitalMin, range.DigitalMax, range.PhysicalMin, range.PhysicalMax, frequencySpectrum, out _frequencyStep,
                    SpectrumSize, windowFunction, _overlap);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityFrequencySpectrum", adqError);
                }
                gch.Free();
            }

            for (int idx = 0; idx < FrameLength; idx++)
            {
                double[] frequencySpectrumCoherent = FrequencySpectrumsCoherent[idx];
                GCHandle gch = GCHandle.Alloc(frequencySpectrumCoherent, GCHandleType.Pinned);
                range = _channelSettings[idx].Range;

                uint adqError = AdqUtility.adqUtilityFrequencySpectrumCoherentSampling(
                    _sampleFrequency, _sineFrequencyHz, _aiModule.Frames, (uint)idx, FrameLength, _aiModule.NumberOfFramesRead,
                    range.DigitalMin, range.DigitalMax, range.PhysicalMin, range.PhysicalMax, frequencySpectrumCoherent, out _frequencyStepCoherent,
                    SpectrumSize, windowFunction, _overlap);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityFrequencySpectrumCoherentSampling", adqError);
                }
                gch.Free();
            }
        }

        private void CalculateSignalSpectrumValues()
        {
            int idx = 0;
            foreach (double[] spectrum in FrequencySpectrums)
            {
                GCHandle gch = GCHandle.Alloc(spectrum, GCHandleType.Pinned);
                uint adqError = AdqUtility.adqUtilityCalculateCharacteristicValuesFromFrequencySpectrum(
                    spectrum, FrequencyStep, SpectrumSize, _frequencyIgnoreBelow, _frequencyIgnoreAbove,
                    out double mainFrequency, out double proportion, out double thd, out double thdrPlusNoise,
                    out double thdfPlusNoise, out double rmsValue);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityCalculateCharacteristicValuesFromFrequencySpectrum", adqError);
                }
                gch.Free();
                MainFrequencies[idx] = mainFrequency;
                FrequProportions[idx] = proportion;
                ThdValues[idx] = thd;
                ThdrPlusNValues[idx] = thdrPlusNoise;
                ThdfPlusNValues[idx] = thdfPlusNoise;
                RmsValuesSpectrum[idx] = rmsValue;
                idx++;
            }
        }

        private void CalculateSignalSpectrumValuesCoherent()
        {
            int idx = 0;
            foreach (double[] spectrum in FrequencySpectrumsCoherent)
            {
                GCHandle gch = GCHandle.Alloc(spectrum, GCHandleType.Pinned);
                uint adqError = AdqUtility.adqUtilityCalculateCharacteristicValuesFromFrequencySpectrum(
                    spectrum, FrequencyStep, SpectrumSize, _frequencyIgnoreBelow, _frequencyIgnoreAbove,
                    out double mainFrequency, out double proportion, out double thd, out double thdrPlusNoise,
                    out double thdfPlusNoise, out double rmsValue);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityCalculateCharacteristicValuesFromFrequencySpectrum", adqError);
                }
                gch.Free();
                MainFrequenciesCoherent[idx] = mainFrequency;
                FrequProportionsCoherent[idx] = proportion;
                ThdValuesCoherent[idx] = thd;
                ThdrPlusNValuesCoherent[idx] = thdrPlusNoise;
                ThdfPlusNValuesCoherent[idx] = thdfPlusNoise;
                RmsValuesSpectrumCoherent[idx] = rmsValue;
                idx++;
            }
        }

        private void CalculateSignalSpectrumValuesCoherentDBU()
        {
            int idx = 0;
            double[] array;
            foreach (double[] spectrum in FrequencySpectrumsCoherent)
            {
                array = FrequencySpectrumsCoherentDBU[idx];
                for (int i = 1; i < spectrum.Length; i++)
                {
                    uint adqError = AdqUtility.adqUtilityConvertLinearTodBu(_dBuRelativeSize, spectrum[i], out array[i]);
                    if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqUtilityConvertLinearTodBu", adqError);
                    }
                }
                idx++;
            }
        }

        private void CalculateSignalSpectrumValuesDBU()
        {
            int idx = 0;
            double[] array;
            foreach (double[] spectrum in FrequencySpectrums)
            {
                array = FrequencySpectrumsDBU[idx];
                for (int i = 1; i < spectrum.Length; i++)
                {
                    uint adqError = AdqUtility.adqUtilityConvertLinearTodBu(_dBuRelativeSize, spectrum[i], out array[i]);
                    if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqUtilityConvertLinearTodBu", adqError);
                    }
                }
                idx++;
            }
        }

        private void CalculateSignalValues2()
        {
            AnalogRange range;
            GCHandle gch = GCHandle.Alloc(_aiModule.Frames, GCHandleType.Pinned);
            for (int idx = 0; idx < FrameLength; idx++)
            {
                range = _channelSettings[idx].Range;
                uint adqError = AdqUtility.adqUtilityCalculateRMSValue(
                    _aiModule.Frames, (uint)idx, FrameLength, _aiModule.NumberOfFramesRead,
                    range.DigitalMin, range.DigitalMax, range.PhysicalMin, range.PhysicalMax, out double rmsValue);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityCalculateRMSValue", adqError);
                }

                RmsValues[idx] = rmsValue;
                adqError = AdqUtility.adqUtilityCalculateMinMaxAverageAndP2PValue(
                    _aiModule.Frames, (uint)idx, FrameLength, _aiModule.NumberOfFramesRead,
                    range.DigitalMin, range.DigitalMax, range.PhysicalMin, range.PhysicalMax,
                    out double minValue, out double maxValue, out double averageValue, out double p2pValue);
                if (adqError != AdqUtilityErrors.ADQ_UTILITY_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqUtilityCalculateMinMaxAverageAndP2PValue", adqError);
                }
                MinValues[idx] = minValue;
                MaxValues[idx] = maxValue;
                AverageValues[idx] = averageValue;
                P2PValues[idx] = p2pValue;
            }
            gch.Free();
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            RaiseInfoTextChanged(string.Format("Utility function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        protected virtual void RaiseInfoTextChanged(string infoText)
        {
            if (OnInfoTextChanged != null)
            {
                var args = new InfoTextChangedEventArgs
                {
                    InfoText = infoText
                };
                OnInfoTextChanged(this, args);
            }
        }
    }
}
