﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the AIReadInterval project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ_DLL.Devices.Boards;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;

namespace AIReadInterval.Utility
{
    public static class AdqCommon
    {
        private const double _oneMicrosecond = 0.000001;
        private const double _oneMillisecond = 0.001;
        private const double _oneSecond = 1.0;
        private const double _oneMinute = 60;
        private const double _oneHour = 60 * 60;

        public static uint NotOpen = AdqErrors.ADQ_ERROR_TYPE_ERROR | AdqErrors.ADQ_ERROR_NOT_OPEN;
        public static uint DefaultNumberOfFrames = 32768;
        public static double DefaultWindowOverlapping = 0.5;

        public static Color[] ChannelColors = {
            Color.DeepSkyBlue,
            Color.LimeGreen,
            Color.MediumOrchid,
            Color.LightCoral,
            Color.SteelBlue,
            Color.YellowGreen,
            Color.Turquoise,
            Color.HotPink,
            Color.Khaki,
            Color.Tan,
            Color.DarkSeaGreen,
            Color.CornflowerBlue,
            Color.Plum,
            Color.CadetBlue,
            Color.PeachPuff,
            Color.LightSalmon};
        public static Color[] ChannelColorsDBUSimilar =
        {
            Color.LightSkyBlue,
            Color.MediumSeaGreen,
            Color.Orchid,
            Color.Coral,
            Color.CadetBlue,
            Color.Yellow,
            Color.PowderBlue,
            Color.LightPink,
            Color.DarkKhaki,
            Color.Wheat,
            Color.SeaGreen,
            Color.DarkCyan,
            Color.LightCoral,
            Color.LightSlateGray,
            Color.PaleGoldenrod,
            Color.Salmon
        };
        public static Color ChannelColorsRed = Color.Red;

        public static MarkerStyle[] MarkerStyles = {
            MarkerStyle.Circle,
            MarkerStyle.Cross,
            MarkerStyle.Diamond,
            MarkerStyle.Square,
            MarkerStyle.Star10,
            MarkerStyle.Star4,
            MarkerStyle.Star5,
            MarkerStyle.Star6,
            MarkerStyle.Triangle
        };

        public static string GetPhysicalUnitName(uint physicalUnit)
        {
            return GetDictionaryValue(_physicalUnitNames, physicalUnit);
        }

        public static string GetPhysicalUnitSymbol(uint physicalUnit)
        {
            return GetDictionaryValue(_physicalUnitSymbols, physicalUnit);
        }

        private static string GetDictionaryValue(IDictionary<uint, string> dictionary, uint key)
        {
            bool found = dictionary.TryGetValue(key, out string stringValue);
            if (!found)
            {
                stringValue = string.Format("0x{0:X8}", key);
            }
            return stringValue;
        }

        private static readonly Dictionary<uint, string> _physicalUnitNames = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_PHYSICAL_UNIT_VOLT, "Volt"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_AMPERE, "Ampere"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_OHM, "Ohm"}
        };

        private static readonly Dictionary<uint, string> _physicalUnitSymbols = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_PHYSICAL_UNIT_VOLT, "V"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_AMPERE, "A"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_OHM, "Ω"}
        };

        public static string GetModelName(string boardName)
        {
            int posADQ = boardName.IndexOf("ADQ");
            if (posADQ < 0)
            {
                posADQ = 0;
            }

            int posEnd = boardName.IndexOf(" ", posADQ);
            if (posEnd < 0)
            {
                posEnd = boardName.Length;
            }

            return boardName.Substring(posADQ, posEnd - posADQ);
        }

        public static string FormatFrequency(double frequency)
        {
            string s;
            if (frequency >= 1000000)
            {
                s = string.Format("{0:0.######}MHz", frequency / 1000000);
            }
            else if (frequency >= 1000)
            {
                s = string.Format("{0:0.######}kHz", frequency / 1000);
            }
            else
            {
                s = string.Format("{0:0.########}Hz", frequency);
            }
            return s;
        }

        public static string FormatTimeToWait(double seconds)
        {
            string s = string.Empty;
            try
            {
                TimeSpan ts = TimeSpan.FromSeconds(seconds);

                s = string.Concat(ts.Days >= 1 ? string.Concat(ts.Days, "d ") : "",
                                  ts.Hours >= 1 ? string.Concat(ts.Hours, "h ") : "",
                                  ts.Minutes >= 1 ? string.Concat(ts.Minutes, "min ") : "",
                                  ts.Seconds >= 1 ? string.Concat(ts.Seconds, "sec ") : "",
                                  ts.Milliseconds >= 1 ? string.Concat(ts.Milliseconds, "ms") : "");
                if (string.IsNullOrEmpty(s))
                {
                    s = (seconds > 0) ? "< 1 ms" : "0";
                }
            }
            catch (OverflowException)
            {
                ;
            }
            return s;
        }

        public static string FormatDuration(double seconds)
        {
            if (seconds == 0)
            {
                return "0";
            }

            var sb = new StringBuilder();
            try
            {
                TimeSpan ts = TimeSpan.FromSeconds(seconds);
                string s = "";

                if (ts.Days >= 1)
                {
                    sb.AppendFormat("{0}:", ts.Days);
                    s = "d";
                }
                if (sb.Length > 0 || ts.Hours >= 1)
                {
                    if (sb.Length == 0)
                    {
                        s = "h";
                    }
                    sb.AppendFormat("{0:00}:", ts.Hours);
                }
                if (sb.Length > 0 || ts.Minutes >= 1)
                {
                    if (sb.Length == 0)
                    {
                        s = "min";
                    }
                    sb.AppendFormat("{0:00}:", ts.Minutes);
                }
                if (sb.Length > 0 || ts.Seconds >= 1)
                {
                    if (sb.Length == 0)
                    {
                        s = "sec";
                    }
                    sb.AppendFormat("{0:00}.", ts.Seconds);
                }
                if (sb.Length > 0 || ts.Milliseconds >= 1)
                {
                    if (sb.Length == 0)
                    {
                        s = "msec";
                    }
                    sb.AppendFormat("{0:000}", ts.Milliseconds);
                    sb.Append(s);
                }
            }
            catch (OverflowException)
            {
                ;
            }
            return sb.ToString();
        }

        public static string FormatSampleTime(double seconds)
        {
            if (seconds == 0.0)
            {
                return "0";
            }

            string s;

            if (seconds < _oneMicrosecond)
            {
                s = string.Format("{0:0##}ns", seconds * 1000000000);
            }
            else if (seconds < _oneMillisecond)
            {
                s = string.Format(CultureInfo.InvariantCulture, "{0:0.#######}µs", seconds * 1000000);
            }
            else if (seconds < _oneSecond)
            {
                s = string.Format(CultureInfo.InvariantCulture, "{0:0.#######}ms", seconds * 1000);
            }
            else if (seconds < _oneMinute)
            {
                s = string.Format(CultureInfo.InvariantCulture, "{0:0.0000000}s", seconds);
            }
            else if (seconds < _oneHour)
            {
                int minutes = (int)(seconds / 60.0);
                double sec = seconds % 60.0;
                s = string.Format(CultureInfo.InvariantCulture, "{0:00}:{1:00.0000000}", minutes, sec);
            }
            else
            {
                int hours = (int)(seconds / _oneHour);
                int minutes = (int)(seconds % _oneHour);
                double sec = seconds % _oneMinute;
                s = string.Format(CultureInfo.InvariantCulture, "{0:00}:{1:00}:{2:00.0000000}", hours, minutes, sec);
            }
            return s;
        }

        public static string FormatPhysicalValue(double physicalValue)
        {
            return FormatPhysicalValue("{0:+0.##;-0.##}", physicalValue);
        }

        public static string FormatPhysicalValue(string formatString, double physicalValue)
        {
            string physicalTxt;
            double tmpVal = Math.Abs(physicalValue);
            if (tmpVal == 0)
            {
                physicalTxt = "0";
            }
            else if (tmpVal < 0.000001)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000000000.0) + "n";
            }
            else if (tmpVal < 0.001)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000000.0) + "µ";
            }
            else if (tmpVal < 1.0)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000.0) + "m";
            }
            else
            {
                physicalTxt = string.Format(formatString, physicalValue);
            }
            return physicalTxt;
        }

        public static void SetDoubleBuffered(Control control)
        {
            // http://www.csharp-examples.net/set-doublebuffered/
            // set instance non-public property with name "DoubleBuffered" to true
            typeof(Control).InvokeMember("DoubleBuffered",
                BindingFlags.SetProperty | BindingFlags.Instance | BindingFlags.NonPublic,
                null, control, new object[] { true });
        }
    }
}
