﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the CommonSettings project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using ALLDAQ.Boards;
using CommonSettings.Utility;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace CommonSettings.ALLDAQ
{
    public class ModuleLocator
    {
        private static uint _adqOpenError = AdqCommon.NotOpen;
        private readonly ICollection<AnalogModule> _analogModules;
        private readonly ICollection<Board> _dioBoards;
        private readonly ICollection<DigitalInputModule> _diModules;

        public event EventHandler<InfoTextChangedEventArgs> OnInfoTextChanged;   

        public ModuleLocator(ICollection<AnalogModule> analogModules, ICollection<DigitalInputModule> diModules, 
            ICollection<Board> dioBoards, EventHandler<InfoTextChangedEventArgs> InfoTextChangedHandler)
        {
            _analogModules = analogModules;
            _diModules = diModules;
            _dioBoards = dioBoards;
            OnInfoTextChanged += InfoTextChangedHandler;
        }

        public void InitModuleLocator()
        {
            _adqOpenError = OpenAdqDriver();
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                uint numberOfBoards = GetNumberOfBoards();
                for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
                {
                    var adqBoardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);
                    uint adqError = AdqDriver.adqGetBoardInfo(boardIdx, ref adqBoardInfo);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetBoardInfo", adqError);
                        continue;
                    }

                    if (SupportsTTLVoltageSelection(adqBoardInfo.TTLVoltageLevel))
                    {
                        var dioBoard = new Board(adqBoardInfo.BoardName, boardIdx, ref adqBoardInfo);
                        _dioBoards.Add(dioBoard);
                    }

                    adqError = AdqDriver.adqGetNumberOfModules(boardIdx, out uint numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetNumberOfModules", adqError);
                        continue;
                    }

                    adqError = GetModuleTypes(adqBoardInfo.BoardName, boardIdx, numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
        }

        private uint GetNumberOfBoards()
        {
            uint adqError = AdqDriver.adqGetNumberOfBoards(out uint numBoards);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetNumberOfBoards", adqError);
                return 0;
            }
            return numBoards;
        }

        private uint GetModuleTypes(string boardName, uint boardIndex, uint numberOfModules)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;
            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqGetModuleType", adqError);
                    continue;
                }

                if (moduleType == AdqConstants.ADQ_MODULE_TYPE_AI || moduleType == AdqConstants.ADQ_MODULE_TYPE_AO)
                {
                    AddAnalogModule(boardName, boardIndex, moduleIdx, moduleType);
                }
                else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DI)
                {
                    adqError = GetDiModuleInfo(boardName, boardIndex, moduleIdx);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
            return adqError;
        }

        private void AddAnalogModule(string boardName, uint boardIndex, uint moduleIdx, uint moduleType)
        {
            _analogModules.Add(new AnalogModule(boardName, boardIndex, moduleIdx, moduleType));
        }

        private uint GetDiModuleInfo(string boardName, uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDiModuleInfo = new AdqDriver.ADQDIModuleInfo(AdqDriver.ADQDIModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDiModuleInfo.Size);
            Marshal.StructureToPtr(adqDiModuleInfo, moduleInfo, false);
            uint adqError = AdqDriver.adqGetModuleInfo(boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDiModuleInfo.Size, moduleInfo, out uint infoSize);
            adqDiModuleInfo = (AdqDriver.ADQDIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDiModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDiModuleInfo.Size, infoSize));

            if (adqDiModuleInfo.NumberOfFilters > 1)
            {
                var diModule = new DigitalInputModule(boardName, boardIndex, moduleIndex, ref adqDiModuleInfo);
                _diModules.Add(diModule);
            }
            return adqError;
        }

        private static bool SupportsTTLVoltageSelection(uint voltageLevels)
        {
            if (voltageLevels == 0) return false;
            return (voltageLevels & (voltageLevels - 1)) != 0;
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqOpen", adqError);
            }
            return adqError;
        }

        public void CloseAdqDriver()
        {
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                _adqOpenError = AdqCommon.NotOpen;
            }
        }
        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            RaiseInfoTextChanged(AdqCommon.FormatFunctionErrorNumber(functionName, adqError));
        }

        private void RaiseInfoTextChanged(string text)
        {
            OnInfoTextChanged?.Invoke(this, new InfoTextChangedEventArgs(text));
        }
    }
}
