﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the CommonSettings project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using ALLDAQ.Boards;
using CommonSettings.ALLDAQ;
using CommonSettings.Utility;

namespace CommonSettings
{
    public partial class FormMain : Form
    {
        private ModuleLocator _moduleLocator;
        private readonly List<AnalogModule> _analogModules = new List<AnalogModule>();
        private readonly List<Board> _dioBoards = new List<Board>();
        private readonly List<DigitalInputModule> _diModules = new List<DigitalInputModule>();
        private AnalogModule _selectedAnalogModule;
        private Board _selectedDioBoard;
        private DigitalInputModule _selectedDiModule;

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Load(object sender, EventArgs e)
        {
            try
            {
                _moduleLocator = new ModuleLocator(_analogModules, _diModules, _dioBoards, OnInfoTextChanged);
                _moduleLocator.InitModuleLocator();

                FillAiModuleCombobox();
                FillDioBoardCombobox();
                FillDiModuleCombobox();

                AddInfoTextLine(string.Format("Number of analog input and output modules found: {0}", _analogModules.Count));
                AddInfoTextLine(string.Format("Number of digital boards with TTL-voltage switching found: {0}", _dioBoards.Count));
                AddInfoTextLine(string.Format("Number of digital input modules with digital filters found: {0}", _diModules.Count));

                SelectFirstComboboxItem(coBoxAiModules);
                SelectFirstComboboxItem(coBoxDioBoards);
                SelectFirstComboboxItem(coBoxDiModules);
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ALLDAQ device driver is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void FillAiModuleCombobox()
        {
            coBoxAiModules.Items.Clear();
            foreach (AnalogModule aiModule in _analogModules)
            {
                coBoxAiModules.Items.Add(aiModule.ModuleName);
            }
        }

        private void FillDioBoardCombobox()
        {
            coBoxDioBoards.Items.Clear();
            foreach (Board dioBoard in _dioBoards)
            {
                coBoxDioBoards.Items.Add(dioBoard.BoardName);
            }
        }

        private void FillDiModuleCombobox()
        {
            coBoxDiModules.Items.Clear();
            foreach(DigitalInputModule diModule in _diModules)
            {
                coBoxDiModules.Items.Add(diModule.ModuleName);
            }
        }

        private void SelectFirstComboboxItem(ComboBox coBox)
        {
            if (coBox.Items.Count > 0)
            {
                coBox.SelectedIndex = 0;
            }
        }

        private void CoBoxAiModule_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_selectedAnalogModule != null)
            {
                _selectedAnalogModule.OnInfoTextChanged -= OnInfoTextChanged;
            }

            if (coBoxAiModules.SelectedIndex < 0)
            {
                _selectedAnalogModule = null;
            }
            else
            {
                AnalogModule oldSelectedModule = _selectedAnalogModule;
                _selectedAnalogModule = _analogModules[coBoxAiModules.SelectedIndex];
                _selectedAnalogModule.OnInfoTextChanged += OnInfoTextChanged;
                if (_selectedAnalogModule != oldSelectedModule)
                {
                    GetAnalogCalibration();
                }
            }
        }

        private void CoBoxDioBoard_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_selectedDioBoard != null)
            {
                _selectedDioBoard.OnInfoTextChanged -= OnInfoTextChanged;
            }

            if (coBoxDioBoards.SelectedIndex < 0)
            {
                _selectedDioBoard = null;
            }
            else
            {
                Board oldSelectedBoard = _selectedDioBoard;
                _selectedDioBoard = _dioBoards[coBoxDioBoards.SelectedIndex];
                _selectedDioBoard.OnInfoTextChanged += OnInfoTextChanged;
                if(oldSelectedBoard != _selectedDioBoard)
                {
                    FillVoltageLevelsComboBox();
                    GetConfiguredTTLVoltage();
                }
            }
        }

        private void FillVoltageLevelsComboBox()
        {
            coBoxVoltageLevels.Items.Clear();
            foreach (KeyValuePair<uint, string> voltageLevel in _selectedDioBoard.VoltageLevels)
            {
                coBoxVoltageLevels.Items.Add(voltageLevel.Value);
            }
            SelectFirstComboboxItem(coBoxVoltageLevels);
        }

        private void CoBoxDiModule_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_selectedDiModule != null)
            {
                _selectedDiModule.OnInfoTextChanged -= OnInfoTextChanged;
            }

            if (coBoxDiModules.SelectedIndex < 0)
            {
                _selectedDiModule = null;
                return;
            }

            DigitalInputModule oldSelectedModule = _selectedDiModule;
            _selectedDiModule = _diModules[coBoxDiModules.SelectedIndex];
            _selectedDiModule.OnInfoTextChanged += OnInfoTextChanged;
            if(oldSelectedModule != _selectedDiModule)
            {
                FillFilterCombobox();
                GetDigitalFilter();
            }
        }

        private void FillFilterCombobox()
        {
            coBoxFilters.Items.Clear();
            if (_selectedDiModule == null) return;

            foreach (ulong filterTime in _selectedDiModule.FilterTimes)
            {
                coBoxFilters.Items.Add(FormatFilterTime(filterTime));
            }
            SelectFirstComboboxItem(coBoxFilters);
        }

        private void BtnSet_Calibration_Click(object sender, EventArgs e)
        {
            if (_selectedAnalogModule == null) return;
            if (RbtnFactory.Checked)
            {
                SetAnalogCalibration(AdqConstants.ADQ_CALIBRATION_FACTORY);
            }
            else if (RbtnUser.Checked)
            {
                SetAnalogCalibration(AdqConstants.ADQ_CALIBRATION_USER_1);
            }
        }

        private void BtnGet_Calibration_Click(object sender, EventArgs e)
        {
            if (_selectedAnalogModule == null) return; 
            GetAnalogCalibration();
        }

        private void GetAnalogCalibration()
        {
            bool succeeded = _selectedAnalogModule.GetSelectedCalibration(out uint calibrationID);
            if (succeeded)
            {
                string s = string.Format("Board {0}, Module {1}, Get Calibration: {2}",
                                        _selectedAnalogModule.BoardIndex, 
                                        _selectedAnalogModule.ModuleIndex, 
                                        AdqCommon.GetCalibrationText(calibrationID));
                AddInfoTextLine(s);
            }
        }

        private void SetAnalogCalibration(uint calibrationID)
        {
            bool succeeded = _selectedAnalogModule.SetSelectedCalibration(calibrationID);
            if (succeeded)
            {
                string s = string.Format("Board {0}, Module {1}, set Calibration: {2}",
                                        _selectedAnalogModule.BoardIndex, 
                                        _selectedAnalogModule.ModuleIndex, 
                                        AdqCommon.GetCalibrationText(calibrationID));
                AddInfoTextLine(s);
            }
        }

        private void BtnGetTTLVoltage_Click(object sender, EventArgs e)
        {
            if (_selectedDioBoard == null) return;

            GetConfiguredTTLVoltage();
        }

        private void BtnSetTTLVoltage_Click(object sender, EventArgs e)
        {
            if (_selectedDioBoard == null) return;

            SetDigitalVoltage(_selectedDioBoard.VoltageLevels[coBoxVoltageLevels.SelectedIndex]);
        }

        private void SetDigitalVoltage(KeyValuePair<uint, string> voltageLevel)
        {
            bool succeeded = _selectedDioBoard.SetTTLVoltage(voltageLevel.Key);
            if (succeeded)
            {
                AddInfoTextLine(string.Format("Board {0}, Set TTL-Voltage: {1}",
                                    _selectedDioBoard.BoardIndex, voltageLevel.Value));
            }
        }

        private void GetConfiguredTTLVoltage()
        {
            bool succeeded = _selectedDioBoard.GetTTLVoltage(out uint voltage);
            if (succeeded)
            {
                AddInfoTextLine(string.Format("Board {0}, Get TTL-Voltage: {1}",
                                    _selectedDioBoard.BoardIndex, AdqCommon.GetVoltageText(voltage)));
            }
        }

        private void BtnGetFilter_Click(object sender, EventArgs e)
        {
            if (_selectedDiModule == null) return;

            GetDigitalFilter();
        }

        private void BtnSetFilter_Click(object sender, EventArgs e)
        {
            if (_selectedDiModule == null) return;

            SetDigitalFilter(_selectedDiModule.FilterTimes[coBoxFilters.SelectedIndex]);
        }

        private void GetDigitalFilter()
        {
            bool succeeded = _selectedDiModule.GetDigitalFilter(out ulong filterPicoseconds);
            if (succeeded)
            {
                string s = string.Format("Board {0}, Module {1}, Get Filter: {2}",
                                        _selectedDiModule.BoardIndex, 
                                        _selectedDiModule.ModuleIndex, 
                                        FormatFilterTime(filterPicoseconds));
                AddInfoTextLine(s);
            }
        }

        private void SetDigitalFilter(ulong filterPicoseconds)
        {
            bool succeeded = _selectedDiModule.SetDigitalFilter(filterPicoseconds);
            if (succeeded)
            {
                string s = string.Format("Board {0}, Module {1}, Set Filter: {2}",
                                        _selectedDiModule.BoardIndex, 
                                        _selectedDiModule.ModuleIndex, 
                                        FormatFilterTime(filterPicoseconds));
                AddInfoTextLine(s);
            }
        }

        private string FormatFilterTime(ulong filterPicoseconds)
        {
            return AdqCommon.FormatDoubleValue(filterPicoseconds / 1e+12, "s");
        }

        private void BtnResetAnalogModule_Click(object sender, EventArgs e)
        {
            if (_selectedAnalogModule == null) return;

            bool succeeded = _selectedAnalogModule.Reset();
            if (succeeded)
            {
                AddInfoTextLine(string.Format("Reset board {0}, module {1}",
                    _selectedAnalogModule.BoardIndex, _selectedAnalogModule.ModuleIndex));
            }
        }

        private void BtnResetDioBoard_Click(object sender, EventArgs e)
        {
            if (_selectedDioBoard == null) return;

            bool succeeded = _selectedDioBoard.Reset();
            if (succeeded)
            {
                AddInfoTextLine(string.Format("Reset board {0}", _selectedDioBoard.BoardIndex));
            }
        }

        private void BtnResetDiModule_Click(object sender, EventArgs e)
        {
            if (_selectedDiModule == null) return;

            bool succeeded = _selectedDiModule.Reset();
            if (succeeded)
            {
                AddInfoTextLine(string.Format("Reset board {0}, module {1}",
                    _selectedDiModule.BoardIndex, _selectedDiModule.ModuleIndex));
            }
        }

        private void AddInfoTextLine(string text)
        {
            textBoxInfo.AppendText(string.Format("[{0}] {1}" + Environment.NewLine,
                    DateTime.Now.ToString("HH:mm:ss.fff"), text));
        }

        private void OnInfoTextChanged(object sender, InfoTextChangedEventArgs e)
        {
            AddInfoTextLine(e.InfoText);
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseAdqDriver();
        }

        private void BtnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void CloseAdqDriver()
        {
            if (_moduleLocator != null)
            {
                _moduleLocator.CloseAdqDriver();
            }
        }

        private void BtnInfo_Click(object sender, EventArgs e)
        {
            var aboutBox = new AboutBox();
            aboutBox.ShowDialog();
        }

        private void BtnClear_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
        }
    }
}