﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//      This file contains a programming example for Intel 82C54 ICs
//      used on ALLDAQ Data Acquisition Boards
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using Counter8254.Utility;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Counter8254.ALLDAQ
{
    public class ModuleLocator
    {
        private const int _maxCounters = 3;
        private uint adqOpenError = AdqCommon.NotOpen;
        private readonly ICollection<Counter8254Modules> _counter8254Modules;
        private readonly uint _moduleType;
        public event EventHandler<InfoTextChangedEventArgs> OnInfoTextChanged;

        public ModuleLocator(ICollection<Counter8254Modules> counter8254Modules, 
            uint moduleType, EventHandler<InfoTextChangedEventArgs> infoTextChangedHandler)
        {
            _counter8254Modules = counter8254Modules;
            _moduleType = moduleType;
            OnInfoTextChanged += infoTextChangedHandler;
        }

        public void InitModuleLocator()
        {
            adqOpenError = OpenAdqDriver();
            if (adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                uint numberOfBoards = GetNumberOfBoards();
                for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
                {
                    uint adqError = AdqDriver.adqGetNumberOfModules(boardIdx, out uint numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetNumberOfModules", adqError);
                        continue;
                    }

                    adqError = GetModuleTypes(boardIdx, numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqOpen", adqError);
            }
            return adqError;
        }

        public void CloseAdqDriver()
        {
            if (adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                adqOpenError = AdqCommon.NotOpen;
            }
        }

        private string GetBoardName(uint boardIndex)
        {
            var boardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);

            uint adqError = AdqDriver.adqGetBoardInfo(boardIndex, ref boardInfo);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetBoardInfo", adqError);
                return "";
            }
            return boardInfo.BoardName;
        }

        private uint GetNumberOfBoards()
        {
            uint adqError = AdqDriver.adqGetNumberOfBoards(out uint numBoards);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetNumberOfBoards", adqError);
                return 0;
            }
            return numBoards;
        }

        private uint GetModuleTypes(uint boardIndex, uint numberOfModules)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;
            int counterIndex = 0;
            Counter8254Modules counters = null;
            string boardName = string.Empty;
            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqGetModuleType", adqError);
                    break;
                }

                if (moduleType == _moduleType && moduleType == AdqConstants.ADQ_MODULE_TYPE_CNTR)
                {
                    if (string.IsNullOrEmpty(boardName))
                    {
                        boardName = GetBoardName(boardIndex);
                    }

                    CounterModule cntr = GetCounterModule(boardIndex, moduleIdx);
                    if (cntr == null) break;

                    if (counterIndex == 0)
                    {
                        counters = new Counter8254Modules(boardIndex, boardName);
                    }
                    counters.AddCounter(cntr);
                    counterIndex++;
                }

                if (IsLastCounterInGroup(moduleIdx, numberOfModules, moduleType, counterIndex))
                {
                    _counter8254Modules.Add(counters);
                    counterIndex = 0;
                }
            }
            return adqError;
        }

        private bool IsLastCounterInGroup(uint moduleIndex, uint numberOfModules, uint moduleType, int counterIndex)
        {
            if (counterIndex == _maxCounters) return true;

            if (counterIndex > 0)
            {
                if (moduleType != _moduleType || moduleIndex == (numberOfModules - 1U))
                {
                    return true;
                }
            }
            return false;
        }

        private CounterModule GetCounterModule(uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqCounterModuleInfo = new AdqDriver.ADQCNTRModuleInfo(AdqDriver.ADQCNTRModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqCounterModuleInfo.Size);
            Marshal.StructureToPtr(adqCounterModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqCounterModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqCounterModuleInfo = (AdqDriver.ADQCNTRModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQCNTRModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return null;
            }

            Debug.Assert(adqCounterModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqCounterModuleInfo.Size, infoSize));

            if (adqCounterModuleInfo.CounterType != AdqConstants.ADQ_CNTR_TYPE_8254)
            {
                UpdateInfoText(string.Format("Board {0}, Module {1}, counter type is not supported by this program.", boardIndex, moduleIndex));
                return null;
            }

            Debug.Assert(adqCounterModuleInfo.WidthBits == 16,
                string.Format("Board {0}, Module {1}, API-function adqGetModuleInfo counter 8254 width is {2}, but should be 16.", 
                boardIndex, moduleIndex, adqCounterModuleInfo.WidthBits));

            return new CounterModule(boardIndex, moduleIndex, adqCounterModuleInfo.WidthBits, adqCounterModuleInfo.ClockSources);
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            UpdateInfoText(string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateInfoText(string text)
        {
            InfoTextChangedEventArgs args = new InfoTextChangedEventArgs
            {
                InfoText = text
            };
            RaiseOnModuleInfoTextChanged(args);
        }

        protected virtual void RaiseOnModuleInfoTextChanged(InfoTextChangedEventArgs e)
        {
            OnInfoTextChanged?.Invoke(this, e);
        }
    }
}
