﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//      This file contains a programming example for Intel 82C54 ICs
//      used on ALLDAQ Data Acquisition Boards
//-----------------------------------------------------------------------------

using System;
using System.Windows.Forms;
using ALLDAQ.Boards;
using System.Collections.Generic;
using Counter8254.Utility;
using Counter8254.ALLDAQ;

namespace Counter8254
{
    public partial class FormMain : Form
    {
        private int _selectedCountersIndex;
        private ModuleLocator _moduleLocator;
        private List<Counter8254Modules> _counter8254Modules;
        private CounterModule _counter0;
        private CounterModule _counter1;
        private CounterModule _counter2;
        private CounterUI _counterUi0;
        private CounterUI _counterUi1;
        private CounterUI _counterUi2;

        public FormMain()
        {
            InitializeComponent();
            InitCounterUiElements();
        }

        private void InitCounterUiElements()
        {
            _counterUi0 = new CounterUI
            {
                GrpBoxModuleNumber = grpBoxCounter0,
                CoBoxCounterMode = coBoxMode0,
                CoBoxClockSource = coBoxClkSource0,
                CoBoxStartValue = coBoxCntrStartValue0,
                TxtBoxCounterValue = txtBoxCtrValue0,
                ButtonSetConfig = btnSetConfig0,
                ButtonRead = btnRead0,
                ButtonReset = btnReset0,
                ButtonReadInInterval = btnReadInInterval0,
                ButtonStop = btnStop0
            };

            _counterUi1 = new CounterUI
            {
                GrpBoxModuleNumber = grpBoxCounter1,
                CoBoxCounterMode = coBoxMode1,
                CoBoxClockSource = coBoxClkSource1,
                CoBoxStartValue = coBoxCntrStartValue1,
                TxtBoxCounterValue = txtBoxCtrValue1,
                ButtonSetConfig = btnSetConfig1,
                ButtonRead = btnRead1,
                ButtonReset = btnReset1,
                ButtonReadInInterval = btnReadInInterval1,
                ButtonStop = btnStop1
            };

            _counterUi2 = new CounterUI
            {
                GrpBoxModuleNumber = grpBoxCounter2,
                CoBoxCounterMode = coBoxMode2,
                CoBoxClockSource = coBoxClkSource2,
                CoBoxStartValue = coBoxCntrStartValue2,
                TxtBoxCounterValue = txtBoxCtrValue2,
                ButtonSetConfig = btnSetConfig2,
                ButtonRead = btnRead2,
                ButtonReset = btnReset2,
                ButtonReadInInterval = btnReadInInterval2,
                ButtonStop = btnStop2
            };
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            EnableUi(false);
            try
            {
                _counter8254Modules = new List<Counter8254Modules>();
                _moduleLocator = new ModuleLocator(_counter8254Modules, AdqConstants.ADQ_MODULE_TYPE_CNTR, OnInfoTextChanged);
                _moduleLocator.InitModuleLocator();
                FillCountersCombobox();
                _selectedCountersIndex = -1;
                if (coBoxCounters.Items.Count > 0)
                {
                    coBoxCounters.SelectedIndex = 0;
                }
            }
            catch (DllNotFoundException ex)
            {
                AppendInfoText("The device driver is not properly installed.");
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void FillCountersCombobox()
        {
            coBoxCounters.Items.Clear();
            foreach (Counter8254Modules counters8254 in _counter8254Modules)
            {
                coBoxCounters.Items.Add(counters8254.ToString());
            }
        }

        private void OnInfoTextChanged(object sender, InfoTextChangedEventArgs e)
        {
            AppendInfoText(e.InfoText);
        }

        private void CoBoxCounters_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_selectedCountersIndex == coBoxCounters.SelectedIndex) return;

            StopIntervalTimers();
            _selectedCountersIndex = coBoxCounters.SelectedIndex;
            Counter8254Modules cntrModules = null;
            if (coBoxCounters.SelectedIndex >= 0)
            {
                cntrModules = _counter8254Modules[coBoxCounters.SelectedIndex];
            }
            ChangeCountersSelection(cntrModules);
        }

        private void ChangeCountersSelection(Counter8254Modules counterModules)
        {
            if (counterModules != null)
            {
                _counter0 = counterModules.GetCounter(0);
                _counter1 = counterModules.GetCounter(1);
                _counter2 = counterModules.GetCounter(2);
            }
            else
            {
                _counter0 = null;
                _counter1 = null;
                _counter2 = null;
            }
            InitCounterUiElements(_counter0, _counterUi0);
            InitCounterUiElements(_counter1, _counterUi1);
            InitCounterUiElements(_counter2, _counterUi2);
        }

        private void InitCounterUiElements(CounterModule counter, CounterUI counterUI)
        {
            counterUI.TxtBoxCounterValue.Text = "0";
            counterUI.CoBoxClockSource.Items.Clear();
            if (counter != null)
            {
                counterUI.GrpBoxModuleNumber.Text = string.Format(" Module {0} ", counter.ModuleIndex);
                counterUI.CoBoxStartValue.Text = string.Format("{0}", counter.CounterMaxValue);
                FillCounterClockSources(counter, counterUI);
                counterUI.CoBoxCounterMode.SelectedIndex = 0;
                counterUI.EnableUI(true);
            }
            else
            {
                counterUI.GrpBoxModuleNumber.Text = " Not available ";
                counterUI.CoBoxStartValue.Text = string.Empty;
                counterUI.CoBoxCounterMode.SelectedIndex = -1;
                counterUI.EnableUI(false);
            }
        }

        private void FillCounterClockSources(CounterModule counter, CounterUI counterUI)
        {
            for (int i = 0; i < counter.ClockSourceValues.Count; i++)
            {
                var item = new ClockSourceComboBoxItem(counter.ClockSourceValues[i]);
                counterUI.CoBoxClockSource.Items.Add(item);
            }
            counterUI.CoBoxClockSource.SelectedIndex = 0;
        }

        private void BtnSetConfig0_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer0, _counter0, _counterUi0);
            try
            {
                ConfigureCounter(_counter0, _counterUi0);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void BtnSetConfig1_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer1, _counter1, _counterUi1);
            try
            {
                ConfigureCounter(_counter1, _counterUi1);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void BtnSetConfig2_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer2, _counter2, _counterUi2);
            try
            {
                ConfigureCounter(_counter2, _counterUi2);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private bool ConfigureCounter(CounterModule counter, CounterUI counterUI)
        {
            uint clockSource = ((ClockSourceComboBoxItem)(counterUI.CoBoxClockSource).SelectedItem).ClockSourceValue;
            uint counterMode = AdqCommon.Counter8254Modes[counterUI.CoBoxCounterMode.SelectedIndex];
            uint startValue = uint.Parse(counterUI.CoBoxStartValue.Text);
            return WriteCounterConfiguration(counter, clockSource, counterMode, startValue, counterUI.CoBoxCounterMode.Text);
        }

        private bool WriteCounterConfiguration(CounterModule counter, uint clockSource, uint counterMode, uint startValue, string ctrModeName)
        {
            uint adqError = counter.Config(clockSource, counterMode, startValue);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                DisplayError(counter.ModuleIndex, adqError);
                return false;
            }
            
            DisplayConfiguration(counter.ModuleIndex, ctrModeName, startValue);
            return true;
        }

        private void BtnRead0_Click(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter0, _counterUi0);
            DisplayReadSingle(_counter0.ModuleIndex, adqError);
        }

        private void BtnRead1_Click(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter1, _counterUi1);
            DisplayReadSingle(_counter1.ModuleIndex, adqError);
        }

        private void BtnRead2_Click(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter2, _counterUi2);
            DisplayReadSingle(_counter2.ModuleIndex, adqError);
        }

        private uint ReadCounter(CounterModule counter, CounterUI counterUI) 
        {
            uint adqError = counter.ReadSingle(out uint value);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                counterUI.TxtBoxCounterValue.Text = value.ToString();
            }
            else
            {
                counterUI.TxtBoxCounterValue.Text = "Error " + adqError.ToString();
            }
            return adqError;
        }

        private void BtnReset0_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer0, _counter0, _counterUi0);
            ResetCounterModule(_counter0, _counterUi0);
        }

        private void BtnReset1_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer1, _counter1, _counterUi1);
            ResetCounterModule(_counter1, _counterUi1);
        }

        private void BtnReset2_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer2, _counter2, _counterUi2);
            ResetCounterModule(_counter2, _counterUi2);
        }

        private void ResetCounterModule(CounterModule counter, CounterUI counterUI)
        {
            uint adqError = counter.Reset();
            counterUI.TxtBoxCounterValue.Text = "0";
            DisplayReset(counter.ModuleIndex, adqError);
        }

        private void BtnReadInInterval0_Click(object sender, EventArgs e)
        {
            try
            {
                bool canStart = ConfigureCounter(_counter0, _counterUi0);
                StartReadInInterval(canStart, timer0, _counter0, _counterUi0);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void BtnReadInInterval1_Click(object sender, EventArgs e)
        {
            try
            {
                bool canStart = ConfigureCounter(_counter1, _counterUi1);
                StartReadInInterval(canStart, timer1, _counter1, _counterUi1);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void BtnReadInInterval2_Click(object sender, EventArgs e)
        {
            try
            {
                bool canStart = ConfigureCounter(_counter2, _counterUi2);
                StartReadInInterval(canStart, timer2, _counter2, _counterUi2);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void StartReadInInterval(bool canStart, Timer timer, CounterModule counter, CounterUI counterUI)
        {
            counterUI.ButtonReadInInterval.Enabled = !canStart;
            if (canStart)
            {
                DisplayReadInInterval("Start", counter.ModuleIndex);
                timer.Enabled = true;
            }
        }

        private void Timer0_Tick(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter0, _counterUi0);
            if (IsCriticalError(adqError))
            {
                StopIntervalTimer(timer0, _counter0, _counterUi0);
                DisplayError(_counter0.ModuleIndex, adqError);
            }
        }

        private void Timer1_Tick(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter1, _counterUi1);
            if (IsCriticalError(adqError))
            {
                StopIntervalTimer(timer1, _counter1, _counterUi1);
                DisplayError(_counter1.ModuleIndex, adqError);
            }
        }

        private void Timer2_Tick(object sender, EventArgs e)
        {
            uint adqError = ReadCounter(_counter2, _counterUi2);
            if (IsCriticalError(adqError))
            {
                StopIntervalTimer(timer2, _counter2, _counterUi2);
                DisplayError(_counter2.ModuleIndex, adqError);
            }
        }

        private bool IsCriticalError(uint adqError)
        {
            return (adqError != AdqErrors.ADQ_ERROR_NONE && adqError != AdqErrors.ADQ_ERROR_INTERNAL);
        }

        private void BtnStop0_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer0, _counter0, _counterUi0);
        }

        private void BtnStop1_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer1, _counter1, _counterUi1);
        }

        private void BtnStop2_Click(object sender, EventArgs e)
        {
            StopIntervalTimer(timer2, _counter2, _counterUi2);
        }

        private void StopIntervalTimer(Timer timer, CounterModule counter, CounterUI counterUI)
        {
            if (timer.Enabled)
            {
                timer.Enabled = false;
                if (counter != null)
                {
                    DisplayReadInInterval("Stop", counter.ModuleIndex);
                }
            }
            counterUI.ButtonReadInInterval.Enabled = counter != null;
        }

        private void StopIntervalTimers()
        {
            StopIntervalTimer(timer0, _counter0, _counterUi0);
            StopIntervalTimer(timer1, _counter1, _counterUi1);
            StopIntervalTimer(timer2, _counter2, _counterUi2);
        }

        private void CloseAdqDriver()
        {
            if (_moduleLocator != null)
            {
                StopIntervalTimers();
                _moduleLocator.CloseAdqDriver();
            }
        }

        private void DisplayConfiguration(uint moduleIndex, string ctrMode, uint startValue)
        {
            AppendInfoText("Configure module " + moduleIndex.ToString()
                + ", Counter mode: " + ctrMode
                + ", Start value: " + startValue.ToString());
        }

        private void DisplayReadSingle(uint moduleIndex, uint adqError)
        {
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                AppendInfoText("Read single value from module " + moduleIndex.ToString());
                return;
            }
            DisplayError(moduleIndex, adqError);
        }

        private void DisplayReadInInterval(string startOrStop, uint moduleIndex)
        {
            AppendInfoText(startOrStop + " interval reading from module " + moduleIndex.ToString());
        }

        private void DisplayReset(uint moduleIndex, uint adqError)
        {
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                AppendInfoText("Reset module " + moduleIndex.ToString());
                return;
            }
            DisplayErrorText("adqReset", adqError);
        }

        private void DisplayError(uint moduleIndex, uint adqError)
        {
            AppendInfoText(string.Format("Error on counter module {0}, Error 0x{1:X8}", moduleIndex, adqError));
        }

        private void DisplayErrorText(string functionName, uint adqError)
        {
            AppendInfoText(string.Format("Function {0} returned Error number: 0x{1:X}",
                    functionName,
                    adqError));
        }
        
        private void AppendInfoText(string infoText)
        {
            textBoxInfo.AppendText("[" + DateTime.Now.ToString("HH:mm:ss.fff") + "] " + infoText + Environment.NewLine);
        }

        private void EnableUi(bool enable)
        {
            _counterUi0.EnableUI(_counter0 != null && enable);
            _counterUi1.EnableUI(_counter1 != null && enable);
            _counterUi2.EnableUI(_counter2 != null && enable);
        }

        private void BtnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseAdqDriver();
        }

        private void BtnInfo_Click(object sender, EventArgs e)
        {
            AboutBox aboutBox = new AboutBox();
            aboutBox.ShowDialog();
        }

        private void BtnClear_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
        }
    }
}
