﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the DIReadContinuous project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using CommonSettings.Utility;
using DIReadContinuous.Utility;
using System;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace DIReadContinuous.ALLDAQ
{
    class DigitalInputModule
    {
        private const int _notificationsPerSecond = 10;
        private string _errorFunctionName;
        private uint _adqError;
        private readonly uint _moduleType;
        private uint _notifyAfterValues;

        public event EventHandler<InfoTextChangedEventArgs> InfoTextChanged;
        public event EventHandler<ModulePropertyChangedEventArgs> ModulePropertyChanged;
        private BackgroundWorker _worker;

        private uint[] _bufferValues;
        private uint[] _bufferValues1;
        private uint[] _bufferValues2;

        public DigitalInputModule(string boardName, uint boardIndex, uint moduleIndex,
            ref AdqDriver.ADQGeneralModuleInfo generalModuleInfo, uint portWidthBits)
        {
            ModelName = AdqCommon.GetModelName(boardName);
            BoardIndex = boardIndex;
            ModuleIndex = moduleIndex;
            PortWidthBits = portWidthBits;
            _moduleType = generalModuleInfo.ModuleType;
            DisplayNumberOfValues = AdqCommon.DefaultDisplayNumberOfValues;
            MaxFrequency = generalModuleInfo.ClockBaseFrequencyHz / generalModuleInfo.ClockMinimumDivider;
            DesiredFrequency = (MaxFrequency < AdqCommon.DefaultDesiredSampleFrequency) ? MaxFrequency : AdqCommon.DefaultDesiredSampleFrequency;
            BuildModuleName();
        }

        public string ModelName { get; }
        public uint BoardIndex { get; }
        public uint ModuleIndex { get; }
        public string ModuleName { get; private set; }
        public uint PortWidthBits { get; }
        public ulong StopAfterValues { get; set; }
        public bool ModuleIsBusy { get; private set; }
        public double DesiredFrequency { get; set; }
        public double MaxFrequency { get; }
        public double ActualSampleFrequency { get; private set; }
        public uint DisplayNumberOfValues { get; set; }

        public bool RequestValuesToDisplay { get; set; }
        public bool RequestUpdateNumberOfValues { get; set; }
        public uint DisplayBufferNumberOfValues { get; private set; }
        public int BitsSelected { get; set; }
        public double ActualSampleTime { get; private set; }
        public ulong TotalValuesAcquired { get; private set; }
        public ulong TotalValuesDisplayed { get; private set; }
        public ulong TotalValuesSkipped { get; private set; }
        public uint[] NewValuesDisplayBuffer { get; private set; }
        public uint NumberOfValidValues { get; private set; }

        private void BuildModuleName()
        {
            ModuleName = string.Format("{0}, Board {1}, Module {2}, Type {3}, {4} Bit, max. {5}",
                ModelName,
                BoardIndex,
                ModuleIndex,
                AdqCommon.GetModuleTypeName(_moduleType),
                PortWidthBits,
                AdqCommon.FormatFrequency(MaxFrequency));
        }

        private void SetErrorInfo(string functionName, uint adqError)
        {
            UpdateModuleInfo(string.Format("Function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateModuleInfo(string info)
        {
            if (InfoTextChanged != null)
            {
                var args = new InfoTextChangedEventArgs
                {
                    InfoText = info
                };
                InfoTextChanged.Invoke(this, args);
            }
        }

        private void RaisePropertyChanged(PropertyType propertyType)
        {
            if (ModulePropertyChanged != null)
            {
                var args = new ModulePropertyChangedEventArgs
                {
                    ChangedProperty = propertyType
                };
                ModulePropertyChanged.Invoke(this, args);
            }
        }

        private void SetDirection()
        {
            if (_moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
            {
                uint adqError = AdqDriver.adqDIOSetDirection(BoardIndex, ModuleIndex, AdqConstants.ADQ_DIO_INPUT);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    throw new ApplicationException(string.Format("Function adqSetDirection returned Error number: 0x{0:X}", adqError));
                }
            }
        }

        private void ConfigureModule()
        {
            SetDirection();
            CheckCommonSettings();
            SetFrequency();
            _notifyAfterValues = Math.Min(DisplayNumberOfValues, (uint)(ActualSampleFrequency / _notificationsPerSecond));
            if(_notifyAfterValues < AdqCommon.DefaultMinNotifyAfterValues)
            {
                _notifyAfterValues = AdqCommon.DefaultMinNotifyAfterValues;
            }

            if (StopAfterValues > 0 && StopAfterValues < _notifyAfterValues)
            {
                _notifyAfterValues = (uint)StopAfterValues;
            }

            if (StopAfterValues > 0 && StopAfterValues < DisplayNumberOfValues)
            {
                DisplayNumberOfValues = (uint)StopAfterValues;
            }

            uint numberOfValues = _notifyAfterValues;
            if (_bufferValues1 == null || _bufferValues1.Length != numberOfValues)
            {
                _bufferValues1 = new uint[numberOfValues];
                _bufferValues2 = new uint[numberOfValues];
            }
            DisplayBufferNumberOfValues = DisplayNumberOfValues;
            RaisePropertyChanged(PropertyType.UpdateValuesToDisplay);
        }

        private void CheckCommonSettings()
        {
            if (DesiredFrequency <= 0)
            {
                throw new ApplicationException("Please enter a valid sample frequency.");
            }
            if (DisplayNumberOfValues == 0)
            {
                throw new ApplicationException("Please enter a valid number of values to display.");
            }
            if (BitsSelected == 0)
            {
                throw new ApplicationException("Please select bits in the bit list.");
            }
        }

        protected void SetFrequency()
        {
            uint adqError = AdqDriver.adqSetFrequency(BoardIndex, ModuleIndex, DesiredFrequency, out double actualFrequency);
            if(adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                throw new ApplicationException(string.Format("Function adqSetFrequency returned Error number: 0x{0:X}", adqError));
            }

            ActualSampleFrequency = actualFrequency;
            ActualSampleTime = 1.0 / actualFrequency;
            RaisePropertyChanged(PropertyType.ActualSampleFrequency);
        }

        public void InitAcquisition()
        {
            if (ModuleIsBusy) return;
            Reset();
            ConfigureModule();
            ModuleIsBusy = true;
            if (_worker == null)
            {
                _worker = new BackgroundWorker
                {
                    WorkerReportsProgress = true,
                    WorkerSupportsCancellation = true
                };
                _worker.DoWork += Worker_DoWork;
                _worker.ProgressChanged += Worker_ProgressChanged;
                _worker.RunWorkerCompleted += Worker_RunWorkerCompleted;
            }
        }

        public void StartAcquisition()
        {
            UpdateModuleInfo("Acquisition initialized");
            _worker.RunWorkerAsync();
        }

        public void CancelAcquisition()
        {
            _worker.CancelAsync();
        }

        public uint StopAcquisition()
        {
            CancelAcquisition();
            uint adqError = AdqDriver.adqStop(BoardIndex, ModuleIndex);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorInfo("adqStop", adqError);
            }
            return adqError;
        }

        private void Worker_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker worker = sender as BackgroundWorker;
            _errorFunctionName = string.Empty;
            TotalValuesAcquired = 0;
            TotalValuesDisplayed = 0;
            TotalValuesSkipped = 0;
            worker.ReportProgress((int)PropertyType.AcquisitionStarting);

            _bufferValues = _bufferValues1;
            GCHandle gchValues1 = GCHandle.Alloc(_bufferValues1, GCHandleType.Pinned);
            GCHandle gchValues2 = GCHandle.Alloc(_bufferValues2, GCHandleType.Pinned);

            _adqError = AdqDriver.adqReadContinuousStart(BoardIndex, ModuleIndex, _notifyAfterValues, StopAfterValues, null, IntPtr.Zero);
            if (_adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                _errorFunctionName = "adqReadContinuousStart";
                return;
            }

            worker.ReportProgress((int)PropertyType.AcquisitionStarted);
            while (true)
            {
                if (worker.CancellationPending)
                {
                    e.Cancel = true;
                    break;
                }
                _adqError = AdqDriver.adqReadContinuous(BoardIndex, ModuleIndex, _notifyAfterValues, _bufferValues, true, out uint numberOfValuesReturned);
                TotalValuesAcquired += numberOfValuesReturned;
                if (_adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    if(StopAfterValues > 0 && StopAfterValues == TotalValuesAcquired)
                    {
                        _adqError = AdqErrors.ADQ_ERROR_NONE;
                        break;
                    }

                    if(numberOfValuesReturned > 0)
                    {
                        ProcessAcquiredValues(worker, numberOfValuesReturned);
                    }
                    _errorFunctionName = "adqReadContinuous";
                    break;
                }

                if (numberOfValuesReturned > 0)
                {
                    ProcessAcquiredValues(worker, numberOfValuesReturned);
                }
            }
            gchValues1.Free();
            gchValues2.Free();
            AdqDriver.adqStop(BoardIndex, ModuleIndex);
        }

        private void Worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            RaisePropertyChanged((PropertyType)e.ProgressPercentage);
        }

        private void Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            ModuleIsBusy = false;
            RaisePropertyChanged(PropertyType.AcquisitionEnded);
            RaisePropertyChanged(PropertyType.UpdateNumberOfValues);
            if (_adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorInfo(_errorFunctionName, _adqError);
                RaisePropertyChanged(PropertyType.AcquisitionErrorOccurred);
            }
            UpdateModuleInfo("Acquisition ended.");
        }

        private void ProcessAcquiredValues(BackgroundWorker worker,uint numberOfValuesReturned)
        {
            SwitchDisplayValueBuffer(numberOfValuesReturned);
            if (RequestUpdateNumberOfValues)
            {
                RequestUpdateNumberOfValues = false;
                worker.ReportProgress((int)PropertyType.UpdateNumberOfValues);
            }
        }

        private void SwitchDisplayValueBuffer(uint numberOfValuesReturned)
        {
            NumberOfValidValues = numberOfValuesReturned;
            if (RequestValuesToDisplay)
            {
                RequestValuesToDisplay = false;
                if (ReferenceEquals(_bufferValues, _bufferValues1))
                {
                    NewValuesDisplayBuffer = _bufferValues1;
                    _bufferValues = _bufferValues2;
                }
                else if (ReferenceEquals(_bufferValues, _bufferValues2))
                {
                    NewValuesDisplayBuffer = _bufferValues2;
                    _bufferValues = _bufferValues1;
                }
                TotalValuesDisplayed += numberOfValuesReturned;
                RaisePropertyChanged(PropertyType.DisplayBuffer);
            }
            else
            {
                TotalValuesSkipped += numberOfValuesReturned;
            }
        }

        public void ClearModuleResources()
        {
            _worker = null;
            _bufferValues = null;
            _bufferValues1 = null;
            _bufferValues2 = null;
        }

        public uint Reset()
        {
            uint adqError = AdqDriver.adqResetModule(BoardIndex, ModuleIndex);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                UpdateModuleInfo(string.Format("Reset board {0} module {1}", BoardIndex, ModuleIndex));
            }
            else
            {
                SetErrorInfo("adqResetModule", adqError);
            }
            return adqError;
        }
    }
}