﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the DIReadContinuous project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using CommonSettings.Utility;
using DIReadContinuous.Utility;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace DIReadContinuous.ALLDAQ
{
    internal class ModuleLocator
    {
        private uint _adqOpenError = AdqCommon.NotOpen;
        private readonly ICollection<DigitalInputModule> _modules;
        public event EventHandler<InfoTextChangedEventArgs> ModuleInfoTextChanged;

        public ModuleLocator(ICollection<DigitalInputModule> modules,
            EventHandler<InfoTextChangedEventArgs> infoTextChangedHandler)
        {
            _modules = modules;
            ModuleInfoTextChanged += infoTextChangedHandler;
        }

        public void FindModules()
        {
            _adqOpenError = OpenAdqDriver();
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                uint numberOfBoards = GetNumberOfBoards();
                for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
                {
                    uint adqError = AdqDriver.adqGetNumberOfModules(boardIdx, out uint numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetNumberOfModules", adqError);
                        continue;
                    }
                    adqError = GetModuleTypes(boardIdx, numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
        }

        private uint GetModuleTypes(uint boardIndex, uint numberOfModules)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;
            string boardName = null;
            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqGetModuleType", adqError);
                    continue;
                }

                if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DI || moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
                {
                    if (boardName == null)
                    {
                        boardName = GetBoardName(boardIndex);
                    }

                    if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DI)
                    {
                        adqError = GetDIModuleInfo(boardName, boardIndex, moduleIdx);
                        if (adqError != AdqErrors.ADQ_ERROR_NONE)
                        {
                            continue;
                        }
                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
                    {
                        adqError = GetDIOModuleInfo(boardName, boardIndex, moduleIdx);
                        if (adqError != AdqErrors.ADQ_ERROR_NONE)
                        {
                            continue;
                        }
                    }
                }
            }
            return adqError;
        }

        private uint GetDIModuleInfo(string boardName, uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDiModuleInfo = new AdqDriver.ADQDIModuleInfo(AdqDriver.ADQDIModuleInfo_ID);
            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDiModuleInfo.Size);
            Marshal.StructureToPtr(adqDiModuleInfo, moduleInfo, false);
            uint adqError = AdqDriver.adqGetModuleInfo(boardIndex,
                                                  moduleIndex,
                                                  ref adqGeneralModuleInfo,
                                                  adqDiModuleInfo.Size,
                                                  moduleInfo,
                                                  out uint infoSize);
            adqDiModuleInfo = (AdqDriver.ADQDIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDiModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDiModuleInfo.Size, infoSize));

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));
                _modules.Add(new DigitalInputModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDiModuleInfo.PortWidthBits));
            }
            return adqError;
        }

        private uint GetDIOModuleInfo(string boardName, uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDioModuleInfo = new AdqDriver.ADQDIOModuleInfo(AdqDriver.ADQDIOModuleInfo_ID);
            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDioModuleInfo.Size);
            Marshal.StructureToPtr(adqDioModuleInfo, moduleInfo, false);
            uint adqError = AdqDriver.adqGetModuleInfo(boardIndex,
                                                  moduleIndex,
                                                  ref adqGeneralModuleInfo,
                                                  adqDioModuleInfo.Size,
                                                  moduleInfo,
                                                  out uint infoSize);
            adqDioModuleInfo = (AdqDriver.ADQDIOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDioModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDioModuleInfo.Size, infoSize));

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));
                _modules.Add(new DigitalInputModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDioModuleInfo.PortWidthBits));
            }

            return adqError;
        }

        private bool IsClockControlledModule(uint clockType)
        {
            return clockType == AdqConstants.ADQ_CLOCK_TYPE_HARDWARE || clockType == AdqConstants.ADQ_CLOCK_TYPE_SOFTWARE;
        }

        private string GetBoardName(uint boardIndex)
        {
            var boardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);
            uint adqError = AdqDriver.adqGetBoardInfo(boardIndex, ref boardInfo);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetBoardInfo", adqError);
                return "";
            }
            return boardInfo.BoardName;
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqOpen", adqError);
            }
            return adqError;
        }

        public void CloseAdqDriver()
        {
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                _adqOpenError = AdqCommon.NotOpen;
            }
        }

        private uint GetNumberOfBoards()
        {
            uint adqError = AdqDriver.adqGetNumberOfBoards(out uint numBoards);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetNumberOfBoards", adqError);
                return 0;
            }
            return numBoards;
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            UpdateInfoText(string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateInfoText(string text)
        {
            InfoTextChangedEventArgs args = new InfoTextChangedEventArgs
            {
                InfoText = text
            };
            RaiseOnModuleInfoTextChanged(args);
        }

        protected virtual void RaiseOnModuleInfoTextChanged(InfoTextChangedEventArgs e)
        {
            ModuleInfoTextChanged?.Invoke(this, e);
        }
    }
}