﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the DOSingle project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using DOSingle.Utility;
using System;

namespace DOSingle.ALLDAQ
{
    public class DigitalModule
    {
        public event EventHandler<InfoTextChangedEventArgs> InfoTextChanged;

        public DigitalModule(string boardName, uint boardIndex, uint moduleIndex, 
            ref AdqDriver.ADQGeneralModuleInfo generalInfo, uint portWidthBits)
        {
            BoardIndex = boardIndex;
            ModuleIndex = moduleIndex;
            ModuleType = generalInfo.ModuleType;
            Description = generalInfo.ModuleDescription;
            PortWidthBits = portWidthBits;
            MaxPortValue = GetMaxPortValue(portWidthBits);
            BuildModuleName(boardName);
            InitDefaultOutputValues(portWidthBits);
        }

        public uint BoardIndex { get; }
        public uint ModuleIndex { get; }
        public uint ModuleType { get; }
        public string Description { get; }
        public uint PortWidthBits { get; }
        public uint MaxPortValue { get; }
        public string ModuleName { get; private set; }
        public uint OutputValue1 { get; set; }
        public uint OutputValue2 { get; set; }
        public uint OutputValue3 { get; set; }

        private uint GetMaxPortValue(uint widthBits)
        {
            return (widthBits < 32) ? (1U << (int)widthBits) - 1 : uint.MaxValue;
        }

        public void BuildModuleName(string boardName)
        {
            ModuleName = string.Format("{0}, Board {1}, Module {2}, Type {3}, {4} Bit",
                AdqCommon.GetModelName(boardName),
                BoardIndex,
                ModuleIndex,
                AdqCommon.GetModuleTypeName(ModuleType),
                PortWidthBits);
        }

        private void InitDefaultOutputValues(uint widthBits)
        {
            int bitIdx = 0;
            while (bitIdx < widthBits)
            {
                OutputValue1 |= (uint)(1 << bitIdx);
                bitIdx += 2;
            }

            bitIdx = 1;
            while (bitIdx < widthBits)
            {
                OutputValue2 |= (uint)(1 << bitIdx);
                bitIdx += 2;
            }
        }

        public bool WriteSingle(uint value)
        {
            bool success = SetDioDirectionToOutput();
            if (!success) return false;

            uint adqError = AdqDriver.adqWriteSingle(BoardIndex, ModuleIndex, value);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqWriteSingle", adqError);
            }
            return adqError == AdqErrors.ADQ_ERROR_NONE;
        }

        private bool SetDioDirectionToOutput()
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;
            if (ModuleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
            {
                adqError = AdqDriver.adqDIOSetDirection(BoardIndex, ModuleIndex, AdqConstants.ADQ_DIO_OUTPUT);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqDIOSetDirection", adqError);
                }
            }
            return adqError == AdqErrors.ADQ_ERROR_NONE;
        }

        public bool ResetModule()
        {
            uint adqError = AdqDriver.adqResetModule(BoardIndex, ModuleIndex);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqResetModule", adqError);
            }

            return adqError == AdqErrors.ADQ_ERROR_NONE;
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            UpdateInfoText(string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateInfoText(string text)
        {
            InfoTextChangedEventArgs args = new InfoTextChangedEventArgs
            {
                InfoText = text
            };
            RaiseOnInfoTextChanged(args);
        }

        protected virtual void RaiseOnInfoTextChanged(InfoTextChangedEventArgs e)
        {
            InfoTextChanged?.Invoke(this, e);
        }
    }
}
