﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the DOSingle project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using DOSingle.Utility;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace DOSingle.ALLDAQ
{
    public class ModuleLocator
    {
        private uint _adqOpenError = AdqCommon.NotOpen;
        private readonly ICollection<DigitalModule> _doModules;
        private readonly ICollection<uint> _moduleTypes;

        public event EventHandler<InfoTextChangedEventArgs> InfoTextChanged;

        public ModuleLocator(ICollection<DigitalModule> doModules, ICollection<uint> moduleTypes,
            EventHandler<InfoTextChangedEventArgs> infoTextChangedHandler)
        {
            _doModules = doModules;
            _moduleTypes = moduleTypes;
            InfoTextChanged += infoTextChangedHandler;
        }

        public void FindModules()
        {
            _adqOpenError = OpenAdqDriver();
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                uint numberOfBoards = GetNumberOfBoards();
                for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
                {
                    uint adqError = AdqDriver.adqGetNumberOfModules(boardIdx, out uint numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetNumberOfModules", adqError);
                        continue;
                    }

                    adqError = GetModuleTypes(boardIdx, numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqOpen", adqError);
            }
            return adqError;
        }

        private uint GetModuleTypes(uint boardIndex, uint numberOfModules)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;

            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqGetModuleType", adqError);
                    continue;
                }

                if (_moduleTypes.Contains(moduleType))
                {
                    if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DO)
                    {
                        GetDoModuleInfo(boardIndex, moduleIdx);

                    }
                    else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
                    {
                        GetDioModuleInfo(boardIndex, moduleIdx);
                    }
                }
            }
            return adqError;
        }

        private uint GetDoModuleInfo(uint boardIndex, uint moduleIndex)
        {
            uint adqError;
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDoModuleInfo = new AdqDriver.ADQDOModuleInfo(AdqDriver.ADQDOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDoModuleInfo.Size);
            Marshal.StructureToPtr(adqDoModuleInfo, moduleInfo, false);

            adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDoModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDoModuleInfo = (AdqDriver.ADQDOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDoModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDoModuleInfo.Size, infoSize));

            string boardName = GetBoardName(boardIndex);
            var adqModule = new DigitalModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDoModuleInfo.PortWidthBits);
            _doModules.Add(adqModule);
            return adqError;
        }

        private uint GetDioModuleInfo(uint boardIndex, uint moduleIndex)
        {
            uint adqError;
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDioModuleInfo = new AdqDriver.ADQDIOModuleInfo(AdqDriver.ADQDIOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDioModuleInfo.Size);
            Marshal.StructureToPtr(adqDioModuleInfo, moduleInfo, false);

            adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDioModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDioModuleInfo = (AdqDriver.ADQDIOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDioModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDioModuleInfo.Size, infoSize));

            string boardName = GetBoardName(boardIndex);
            var adqModule = new DigitalModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDioModuleInfo.PortWidthBits);
            _doModules.Add(adqModule);
            return adqError;
        }

        private string GetBoardName(uint boardIndex)
        {
            var boardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);

            uint adqError = AdqDriver.adqGetBoardInfo(boardIndex, ref boardInfo);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetBoardInfo", adqError);
                return "";
            }
            return boardInfo.BoardName;
        }

        public void CloseAdqDriver()
        {
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                _adqOpenError = AdqCommon.NotOpen;
            }
        }

        private uint GetNumberOfBoards()
        {
            uint adqError = AdqDriver.adqGetNumberOfBoards(out uint numBoards);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetNumberOfBoards", adqError);
                return 0;
            }
            return numBoards;
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            UpdateInfoText(string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateInfoText(string text)
        {
            InfoTextChangedEventArgs args = new InfoTextChangedEventArgs
            {
                InfoText = text
            };
            RaiseOnInfoTextChanged(args);
        }

        protected virtual void RaiseOnInfoTextChanged(InfoTextChangedEventArgs e)
        {
            InfoTextChanged?.Invoke(this, e);
        }
    }
}
