﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the DOSingle project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using System;
using System.Windows.Forms;
using ALLDAQ.Boards;
using System.Collections.Generic;
using DOSingle.Utility;
using DOSingle.ALLDAQ;
using System.Globalization;

namespace DOSingle
{
    public partial class FormMain : Form
    {
        private ModuleLocator _moduleLocator;
        private readonly List<DigitalModule> _doModules = new List<DigitalModule>();
        private DigitalModule _selectedModule;

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            EnableUI(false);
            try
            {
                List<uint> moduleTypes = InitModuleTypesToFind();
                _moduleLocator = new ModuleLocator(_doModules, moduleTypes, InfoTextChanged);
                _moduleLocator.FindModules();

                FillModuleCombobox();
                SelectFirstModule();
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The device driver is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void SelectFirstModule()
        {
            if (coBoxModules.Items.Count > 0)
            {
                coBoxModules.SelectedIndex = 0;
            }
        }

        private List<uint> InitModuleTypesToFind()
        {
            List<uint> moduleTypes = new List<uint>
            {
                AdqConstants.ADQ_MODULE_TYPE_DO,
                AdqConstants.ADQ_MODULE_TYPE_DIO
            };
            return moduleTypes;
        }

        private void FillModuleCombobox()
        {
            coBoxModules.Items.Clear();
            foreach (DigitalModule doModule in _doModules)
            {
                coBoxModules.Items.Add(doModule.ModuleName);
            }
        }

        private void BtnWriteValue1_Click(object sender, EventArgs e)
        {
            WriteEnteredValue(textBoxValue1, valueNumber: 1);
        }

        private void BtnWriteValue2_Click(object sender, EventArgs e)
        {
            WriteEnteredValue(textBoxValue2, valueNumber: 2);
        }

        private void BtnWriteValue3_Click(object sender, EventArgs e)
        {
            WriteEnteredValue(textBoxValue3, valueNumber: 3);
        }

        private void WriteEnteredValue(TextBox txtBox, int valueNumber)
        {
            try
            {
                uint valueToWrite = ConvertStringValueToUint(txtBox.Text);
                bool succeeded = _selectedModule.WriteSingle(valueToWrite);
                if (succeeded)
                {
                    DisplayValue(valueToWrite);
                    UpdateModuleOutputValue(valueToWrite, valueNumber);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                txtBox.Focus();
            }
        }

        private void UpdateModuleOutputValue(uint valueToWrite, int valueNumber)
        {
            switch (valueNumber)
            {
                case 1:
                    _selectedModule.OutputValue1 = valueToWrite;
                    break;
                case 2:
                    _selectedModule.OutputValue2 = valueToWrite;
                    break;
                case 3:
                    _selectedModule.OutputValue3 = valueToWrite;
                    break;
            }
        }

        private uint ConvertStringValueToUint(string valueStr)
        {
            uint x = uint.Parse(valueStr, NumberStyles.AllowHexSpecifier);
            if (x > _selectedModule.MaxPortValue)
            {
                throw new ArgumentOutOfRangeException(string.Format("Value out of range. Maximum allowed value: 0x{0:X}", _selectedModule.MaxPortValue));
            }
            return x;
        }

        private void DisplayValue(uint value)
        {
            AddInfoTextLine(string.Format("Write to Board {0}, Module {1}",
                                   _selectedModule.BoardIndex, _selectedModule.ModuleIndex)
                + string.Format(", Value hex: 0x{0:X2}", value)
                + ", bin: " + AdqCommon.IntToBin(value, _selectedModule.PortWidthBits)
                + ", dec: " + value.ToString());
        }

        private void BtnReset_Click(object sender, EventArgs e)
        {
            bool success = _selectedModule.ResetModule();
            if (success)
            {
                AddInfoTextLine(string.Format("Reset board {0} module {1}",
                    _selectedModule.BoardIndex,
                    _selectedModule.ModuleIndex));
            }
        }

        private void CoBoxModules_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_selectedModule != null)
            {
                _selectedModule.InfoTextChanged -= InfoTextChanged;
            }

            EnableUI(coBoxModules.SelectedIndex >= 0);
            if (coBoxModules.SelectedIndex >= 0)
            {
                _selectedModule = _doModules[coBoxModules.SelectedIndex];
                _selectedModule.InfoTextChanged += InfoTextChanged;
                textBoxValue1.Text = string.Format("{0:X}", _selectedModule.OutputValue1);
                textBoxValue2.Text = string.Format("{0:X}", _selectedModule.OutputValue2);
                textBoxValue3.Text = string.Format("{0:X}", _selectedModule.OutputValue3);
            }
            else
            {
                _selectedModule = null;
            }
            ShowModuleDescription(_selectedModule);
        }

        private void EnableUI(bool enabled)
        {
            btnWriteValue1.Enabled = enabled;
            btnWriteValue2.Enabled = enabled;
            btnWriteValue3.Enabled = enabled;
            btnReset.Enabled = enabled;
        }

        private void ShowModuleDescription(DigitalModule doModule)
        {
            textBoxDescription.Text = (doModule == null) ? "" : doModule.Description;
        }

        private void InfoTextChanged(object sender, InfoTextChangedEventArgs e)
        {
            AddInfoTextLine(e.InfoText);
        }

        private void AddInfoTextLine(string text)
        {
            textBoxInfo.AppendText(string.Format("[{0}] {1}" + Environment.NewLine,
                    DateTime.Now.ToString("HH:mm:ss.fff"), text));
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseAdqDriver();
        }

        private void CloseAdqDriver()
        {
            if (_moduleLocator != null)
            {
                _moduleLocator.CloseAdqDriver();
            }
        }

        private void BtnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void BtnClear_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
        }

        private void BtnInfo_Click(object sender, EventArgs e)
        {
            AboutBox aboutBox = new AboutBox();
            aboutBox.ShowDialog();
        }
    }
}
