﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the FINRead project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using FINRead.Utility;
using System;
using System.ComponentModel;
using System.Threading.Tasks;

namespace FINRead.Models
{
    public class FinModule : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;

        public FinModule(string boardName, uint boardIndex, uint moduleIndex)
        {
            BoardIndex = boardIndex;
            ModuleIndex = moduleIndex;
            SetModuleName(boardName);
            TimeoutMilliseconds = AdqCommon.DefaultTimeoutMilliseconds;
        }

        public uint BoardIndex { get; }
        public uint ModuleIndex { get; }
        public string ModuleName { get; private set; }
        public uint TimeoutMilliseconds { get; set; }

        public async Task<FinReadData> FinReadAsync()
        {
            IsModuleBusy = true;
            FinReadData finData = null;
            await Task.Run(() =>
            {
                finData = FinRead(BoardIndex, ModuleIndex, TimeoutMilliseconds);
            });
            IsModuleBusy = false;
            return finData;
        }

        private static FinReadData FinRead(uint boardIndex, uint moduleIndex, uint timeoutMilliseconds)
        {
            var finData = new FinReadData();

            uint adqError = AdqDriver.adqFINRead(boardIndex, moduleIndex, out double frequency, out double onTime, timeoutMilliseconds);
            finData.HasError = adqError != AdqErrors.ADQ_ERROR_NONE;
            finData.Frequency = frequency;
            finData.HighTime = onTime;
            if (finData.HasError)
            {
                finData.ErrorText = GetErrorText(adqError, timeoutMilliseconds);
            }
            return finData;
        }

        private static string GetErrorText(uint adqError, uint timeoutMilliseconds)
        {
            string s = string.Format("API-function adqFINRead returned Error number 0x{0:X}", adqError);
            if (adqError == AdqErrors.ADQ_ERROR_TIMEOUT)
            {
                s += Environment.NewLine + string.Format("Time-out reached after {0} milliseconds", timeoutMilliseconds);
            }
            return s;
        }

        private bool _isModuleBusy;
        public bool IsModuleBusy
        {
            get { return _isModuleBusy; }
            private set
            {
                _isModuleBusy = value;
                NotifyPropertyChanged(nameof(IsModuleBusy));
            }
        }

        public bool ResetModule()
        {
            uint adqError = AdqDriver.adqResetModule(BoardIndex, ModuleIndex);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqResetModule", adqError);
                return false;
            }
            return true;
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            ModuleInfoText = string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError);
        }

        private string _moduleInfoText;
        public string ModuleInfoText
        {
            get { return _moduleInfoText; }
            private set
            {
                _moduleInfoText = value;
                NotifyPropertyChanged(nameof(ModuleInfoText));
            }
        }

        void NotifyPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        private void SetModuleName(string boardName)
        {
            ModuleName = AdqCommon.GetModelName(boardName)
                + string.Format(", Board {0}, Module {1}", BoardIndex, ModuleIndex);
        }
    }
}
