﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the Interrupt project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using Interrupt.Models;
using Interrupt.Utility;
using System.Collections.ObjectModel;
using System.ComponentModel;

namespace Interrupt.ViewModels
{
    public class BitChangeViewModel : ErrorInfoViewModelBase
    {
        private IrqModule _irqModule;

        public void UpdateIrqModule(IrqModule irqModule)
        {
            if (SupportsIrqType)
            {
                _irqModule.BitChangeStatusChanged -= BitChangeStatusChanged;
                _irqModule.PropertyChanged -= BitChangeViewModel_PropertyChanged;
                _irqModule.PortBitPatternCompareSettingsChanged -= BitPatternCompareSettingsChanged;
            }

            _irqModule = irqModule;
            SupportsIrqType = irqModule != null && irqModule.SupportsBitChange;

            if (SupportsIrqType)
            {
                UpdateBitMaskViewModels();
                UpdateRisingEdgesText(irqModule.PortBitChangeRisingEdge);
                UpdateRisingBitMaskEdges(irqModule.PortBitChangeRisingEdge);
                UpdateFallingEdgesText(irqModule.PortBitChangeFallingEdge);
                UpdateFallingBitMaskEdges(irqModule.PortBitChangeFallingEdge);
                _irqModule.BitChangeStatusChanged += BitChangeStatusChanged;
                _irqModule.PropertyChanged += BitChangeViewModel_PropertyChanged;
                _irqModule.PortBitPatternCompareSettingsChanged += BitPatternCompareSettingsChanged;
            }
            else
            {
                BitMaskEdgeItems.Clear();
                BitEdgeIndicatorItems.Clear();
                UpdateRisingEdgesText(0);
                UpdateFallingEdgesText(0);
            }
            ClearStatusIndicators();
            NotifyPropertyChanged(nameof(CanUseBitChange));
            NotifyPropertyChanged(nameof(SupportsIrqType));

        }

        public bool SupportsIrqType { get; private set; }

        public bool CanUseBitChange
        {
            get { return SupportsIrqType && _irqModule != null && _irqModule.IsModuleIdle && !_irqModule.BitPatternCompareIsEnabled; }
        }

        private void BitPatternCompareSettingsChanged(object sender, PortBitPatternCompareSettingsEventArgs e)
        {
              NotifyPropertyChanged(nameof(CanUseBitChange));
        }

        private void UpdateBitMaskViewModels()
        {
            while (BitMaskEdgeItems.Count > _irqModule.PortWidthBits)
            {
                BitMaskEdgeItems.RemoveAt(BitMaskEdgeItems.Count - 1);
                BitEdgeIndicatorItems.RemoveAt(BitEdgeIndicatorItems.Count - 1);
            }

            for (int bitIdx = BitMaskEdgeItems.Count; bitIdx < _irqModule.PortWidthBits; bitIdx++)
            {
                var bmItem = new BitMaskEdgeItemViewModel(bitIdx);
                bmItem.PropertyChanged += BmItem_PropertyChanged;
                BitMaskEdgeItems.Add(bmItem);

                var indItem = new BitEdgeIndicatorItemViewModel(bitIdx);
                BitEdgeIndicatorItems.Add(indItem);
            }
        }

        private void BmItem_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(BitMaskEdgeItemViewModel.BitRisingEdge))
            {
                UpdateRisingEdgeValue();
            }
            else if (e.PropertyName == nameof(BitMaskEdgeItemViewModel.BitFallingEdge))
            {
                UpdateFallingEdgeValue();
            }
        }

        private void UpdateRisingEdgeValue()
        {
            uint edgesValue = 0;
            uint bit = 1;
            foreach (BitMaskEdgeItemViewModel bitMaskEdgeItem in BitMaskEdgeItems)
            {
                if (bitMaskEdgeItem.BitRisingEdge)
                {
                    edgesValue |= bit;
                }
                bit <<= 1;
            }
            _irqModule.PortBitChangeRisingEdge = edgesValue;
            UpdateRisingEdgesText(edgesValue);
        }

        private void UpdateRisingEdgesText(uint edgesValue)
        {
            _valueRisingEdgesText = AdqCommon.FormatHexValue(edgesValue);
            _errorInfo.Remove(nameof(ValueRisingEdgesText));
            NotifyPropertyChanged(nameof(ValueRisingEdgesText));
            NotifyPropertyChanged(nameof(IsInputValid));
        }

        private void UpdateFallingEdgeValue()
        {
            uint edgesValue = 0;
            uint bit = 1;
            foreach (BitMaskEdgeItemViewModel bitMaskEdgeItem in BitMaskEdgeItems)
            {
                if (bitMaskEdgeItem.BitFallingEdge)
                {
                    edgesValue |= bit;
                }
                bit <<= 1;
            }
            _irqModule.PortBitChangeFallingEdge = edgesValue;
            UpdateFallingEdgesText(edgesValue);
        }

        private void UpdateFallingEdgesText(uint edgesValue)
        {
            _valueFallingEdgesText = AdqCommon.FormatHexValue(edgesValue);
            _errorInfo.Remove(nameof(ValueFallingEdgesText));
            NotifyPropertyChanged(nameof(ValueFallingEdgesText));
            NotifyPropertyChanged(nameof(IsInputValid));
        }

        public ObservableCollection<BitMaskEdgeItemViewModel> BitMaskEdgeItems { get; } 
            = new ObservableCollection<BitMaskEdgeItemViewModel>();
        public ObservableCollection<BitEdgeIndicatorItemViewModel> BitEdgeIndicatorItems { get; } 
            = new ObservableCollection<BitEdgeIndicatorItemViewModel>();

        public void ClearStatusIndicators()
        {
            UpdateStatusRisingEdges(0);
            UpdateStatusFallingEdges(0);
        }

        private string _valueRisingEdgesText = "0";
        public string ValueRisingEdgesText
        {
            get { return _valueRisingEdgesText; }
            set
            {
                if (_valueRisingEdgesText != value)
                {
                    _valueRisingEdgesText = value;

                    string msg = InputValidation.ValidateHexValue(_valueRisingEdgesText, out uint edgesValue);

                    if (!string.IsNullOrEmpty(msg))
                    {
                        _errorInfo.AddOrUpdate(nameof(ValueRisingEdgesText), msg);
                    }
                    else
                    {
                        msg = _irqModule.ValidatePortMaxValue(edgesValue);
                        if (!string.IsNullOrEmpty(msg))
                        {
                            _errorInfo.AddOrUpdate(nameof(ValueRisingEdgesText), msg);
                        }
                        else
                        {
                            _errorInfo.Remove(nameof(ValueRisingEdgesText));
                            _irqModule.PortBitChangeRisingEdge = edgesValue;
                            _valueRisingEdgesText = AdqCommon.FormatHexValue(edgesValue);
                            UpdateRisingBitMaskEdges(edgesValue);
                        }
                    }

                    NotifyPropertyChanged(nameof(ValueRisingEdgesText));
                    NotifyPropertyChanged(nameof(IsInputValid));
                }
            }
        }

        private string _valueFallingEdgesText = "0";
        public string ValueFallingEdgesText
        {
            get { return _valueFallingEdgesText; }
            set
            {
                if (_valueFallingEdgesText != value)
                {
                    _valueFallingEdgesText = value;

                    string msg = InputValidation.ValidateHexValue(value, out uint edgesValue);

                    if (!string.IsNullOrEmpty(msg))
                    {
                        _errorInfo.AddOrUpdate(nameof(ValueFallingEdgesText), msg);
                    }
                    else
                    {
                        msg = _irqModule.ValidatePortMaxValue(edgesValue);
                        if (!string.IsNullOrEmpty(msg))
                        {
                            _errorInfo.AddOrUpdate(nameof(ValueFallingEdgesText), msg);
                        }
                        else
                        {
                            _errorInfo.Remove(nameof(ValueFallingEdgesText));
                            _irqModule.PortBitChangeFallingEdge = edgesValue;
                            _valueFallingEdgesText = AdqCommon.FormatHexValue(edgesValue);
                            UpdateFallingBitMaskEdges(edgesValue);
                        }
                    }

                    NotifyPropertyChanged(nameof(ValueFallingEdgesText));
                    NotifyPropertyChanged(nameof(IsInputValid));
                }
            }
        }

        private void BitChangeStatusChanged(object sender, BitChangeStatusChangedEventArgs e)
        {
            UpdateStatusRisingEdges(e.StatusRisingEdge);
            UpdateStatusFallingEdges(e.StatusFallingEdge);
        }

        private void UpdateStatusRisingEdges(uint edgesValue)
        {
            if (StatusBitChangeRisingEdges != edgesValue)
            {
                StatusBitChangeRisingEdges = edgesValue;
                NotifyPropertyChanged(nameof(StatusBitChangeRisingEdges));
                UpdateRisingBitIndicatorsEdges(edgesValue);
            }
        }

        private void UpdateStatusFallingEdges(uint edgesValue)
        {
            if (StatusBitChangeFallingEdges != edgesValue)
            {
                StatusBitChangeFallingEdges = edgesValue;
                NotifyPropertyChanged(nameof(StatusBitChangeFallingEdges));
                UpdateFallingBitIndicatorsEdges(edgesValue);
            }
        }

        public uint StatusBitChangeRisingEdges { get; private set; }
        public uint StatusBitChangeFallingEdges { get; private set; }

        private void UpdateRisingBitMaskEdges(uint edgesValue)
        {
            int bit = 1;
            foreach (BitMaskEdgeItemViewModel bitMaskEdgeItem in BitMaskEdgeItems)
            {
                bitMaskEdgeItem.BitRisingEdge = (bit & edgesValue) != 0;
                bit <<= 1;
            }
        }

        private void UpdateFallingBitMaskEdges(uint edgesValue)
        {
            int bit = 1;
            foreach (BitMaskEdgeItemViewModel bitMaskEdgeItem in BitMaskEdgeItems)
            {
                bitMaskEdgeItem.BitFallingEdge = (bit & edgesValue) != 0;
                bit <<= 1;
            }
        }

        private void UpdateRisingBitIndicatorsEdges(uint edgesValue)
        {
            int bit = 1;
            foreach (BitEdgeIndicatorItemViewModel bitIndicatorItem in BitEdgeIndicatorItems)
            {
                bitIndicatorItem.RisingEdge = (bit & edgesValue) != 0;
                bit <<= 1;
            }
        }

        private void UpdateFallingBitIndicatorsEdges(uint edgesValue)
        {
            int bit = 1;
            foreach (BitEdgeIndicatorItemViewModel bitIndicatorItem in BitEdgeIndicatorItems)
            {
                bitIndicatorItem.FallingEdge = (bit & edgesValue) != 0;
                bit <<= 1;
            }
        }

        void BitChangeViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IrqModule.IsModuleIdle))
            {
                UpdateModuleIdleStatus();
            }
        }

        private void UpdateModuleIdleStatus()
        {
            NotifyPropertyChanged(nameof(CanUseBitChange));
        }
    }
}
