﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the Interrupt project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using Interrupt.Models;
using Interrupt.Utility;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows;

namespace Interrupt.ViewModels
{
    class MainViewModel : ErrorInfoViewModelBase
    {
        private readonly ModuleLocator _moduleLocator;

        public RelayCommand ResetModuleCommand { get; private set; }
        public RelayCommand ClearStatusIndicatorsCommand { get; private set; }
        public RelayCommand CloseCommand { get; private set; }
        public RelayCommand EnableDisableIrqCommand { get; private set; }

        public MainViewModel()
        {
            InitCommands();
            _moduleLocator = new ModuleLocator(MainViewModel_PropertyChanged);
            _moduleLocator.InitModuleLocator();
            _irqWaitTimeoutText = AdqCommon.DefaultIrqWaitTimeoutMilliseconds.ToString();
            InitViewModels();
        }

        private void InitCommands()
        {
            ResetModuleCommand = new RelayCommand(ResetModuleExecute, CanResetModuleExecute);
            ClearStatusIndicatorsCommand = new RelayCommand(ClearStatusIndicatorsExecute);
            CloseCommand = new RelayCommand(CloseApplicationExecute);
            EnableDisableIrqCommand = new RelayCommand(EnableDisableIrqExecute, CanEnableDisableIrqExecute);
        }

        private bool CanEnableDisableIrqExecute(object obj)
        {
            return _selectedModule != null && IsInputValid();
        }

        private void EnableDisableIrqExecute(object obj)
        {
            if (_selectedModule.IsModuleIdle)
            {
                ClearStatusIndicators();
                _selectedModule.EnableInterrupt();
            }
            else
            {
                _selectedModule.DisableInterrupt();
            }
        }

        private void ClearStatusIndicators()
        {
            BitChangeViewModel.ClearStatusIndicators();
            PortStatusViewModel.ClearStatusIndicators();
            BitPatternCompareViewModel.ClearStatusIndicators();
            ExternalIrqViewModel.ClearStatusIndicators();
            StatusIrqCount = 0;
            StatusTriggerValue = 0;
        }


        string _enableButtonContent = "Enable IRQ";
        public string EnableButtonContent
        {
            get { return _enableButtonContent; }
            set
            {
                _enableButtonContent = value;
                NotifyPropertyChanged(nameof(EnableButtonContent));
            }
        }

        private void UpdateModuleIdleStatus()
        {
            EnableButtonContent = (_selectedModule.IsModuleIdle) ? "Enable IRQ" : "Disable IRQ";
            NotifyPropertyChanged(nameof(IsModuleAvailable));
            NotifyPropertyChanged(nameof(IsModuleSelectable));
        }

        private bool CanResetModuleExecute(object obj)
        {
            return _selectedModule != null;
        }

        private void ResetModuleExecute(object obj)
        {
            bool success = _selectedModule.ResetModule();
            if (success)
            {
                AppendInfoText(string.Format("Reset board {0}, module {1}", _selectedModule.BoardIndex, _selectedModule.ModuleIndex));
            }
        }

        public List<IrqModule> IrqModules
        {
            get { return _moduleLocator.IrqModules; }
        }

        private IrqModule _selectedModule;
        public IrqModule SelectedModule
        {
            get { return _selectedModule; }
            set
            {
                if (_selectedModule == value) return;
                if (_selectedModule != null)
                {
                    _selectedModule.PropertyChanged -= MainViewModel_PropertyChanged;
                    _selectedModule.CommonIrqStatusChanged -= CommonIrqStatusChanged;

                }
                _selectedModule = value;
                if (_selectedModule != null)
                {
                    _selectedModule.PropertyChanged += MainViewModel_PropertyChanged;
                    _selectedModule.CommonIrqStatusChanged += CommonIrqStatusChanged;
                    _resetIrqCount = _selectedModule.ResetIRQCount;
                    _irqWaitTimeoutText = _selectedModule.TimeoutMilliseconds.ToString();
                    _useCallbackFunction = _selectedModule.UseIrqCallbackFunction;
                    _useIrqWaitFunction = !_useCallbackFunction;
                }
                else
                {
                    _irqWaitTimeoutText = AdqCommon.DefaultIrqWaitTimeoutMilliseconds.ToString();
                }
                StatusIrqCount = 0;
                StatusTriggerValue = 0;

                NotifyPropertyChanged(nameof(SelectedModule));
                NotifyPropertyChanged(nameof(IrqDescription));
                NotifyPropertyChanged(nameof(ResetIrqCount));
                NotifyPropertyChanged(nameof(IsModuleAvailable));
                NotifyPropertyChanged(nameof(IrqWaitTimeoutText));
                NotifyPropertyChanged(nameof(UseCallbackFunction));
                NotifyPropertyChanged(nameof(UseIrqWaitFunction));
                UpdateViewModels();
                ResetModuleCommand.RaiseCanExecuteChanged();
                EnableDisableIrqCommand.RaiseCanExecuteChanged();
                SelectFirstSupportedIrqTabItem();
            }
        }

        private void UpdateViewModels()
        {
            BitChangeViewModel.UpdateIrqModule(_selectedModule);
            BitPatternCompareViewModel.UpdateIrqModule(_selectedModule);
            PortStatusViewModel.UpdateIrqModule(_selectedModule);
            ExternalIrqViewModel.UpdateIrqModule(_selectedModule);
        }

        private void SelectFirstSupportedIrqTabItem()
        {
            if (_selectedModule != null)
            {
                if (_selectedModule.SupportsBitChange)
                {
                    SelectedTabItemIndex = 0;
                }
                else if (_selectedModule.SupportsBitPatternCompare)
                {
                    SelectedTabItemIndex = 1;
                }
                else if (_selectedModule.SupportedPortStatusErrors != 0 || _selectedModule.SupportedFaultRecovery != 0)
                {
                    SelectedTabItemIndex = 2;
                }
                else
                {
                    SelectedTabItemIndex = 3;
                }
            }
            else
            {
                SelectedTabItemIndex = 0;
            }
        }

        private int _selectedTabItemIndex;
        public int SelectedTabItemIndex
        {
            get { return _selectedTabItemIndex; }
            set
            {
                if (_selectedTabItemIndex != value)
                {
                    _selectedTabItemIndex = value;
                    NotifyPropertyChanged(nameof(SelectedTabItemIndex));
                }
            }
        }

        public bool IsModuleSelectable
        {
            get
            {
                return (_moduleLocator.IrqModules.Count > 0 && _selectedModule == null) || IsModuleAvailable;
            }
        }

        public bool IsModuleAvailable
        {
            get { return _selectedModule != null && _selectedModule.IsModuleIdle; }
        }

        public string IrqDescription
        {
            get
            {
                return (_selectedModule != null) ? _selectedModule.IrqDescription : string.Empty;
            }
        }

        private bool _useCallbackFunction = true;
        public bool UseCallbackFunction
        {
            get { return _useCallbackFunction; }
            set
            {
                if (_useCallbackFunction != value)
                {
                    _useCallbackFunction = value;
                    UseIrqWaitFunction = !value;
                    if (_selectedModule != null)
                    {
                        _selectedModule.UseIrqCallbackFunction = value;
                    }
                    NotifyPropertyChanged(nameof(UseCallbackFunction));
                }
            }
        }

        private bool _useIrqWaitFunction;
        public bool UseIrqWaitFunction
        {
            get { return !_useCallbackFunction; }
            set
            {
                if (_useIrqWaitFunction != value)
                {
                    _useIrqWaitFunction = value;
                    UseCallbackFunction = !value;
                    if (_selectedModule != null)
                    {
                        _selectedModule.UseIrqCallbackFunction = !value;
                    }
                    NotifyPropertyChanged(nameof(UseIrqWaitFunction));
                }
            }
        }

        private bool _resetIrqCount;
        public bool ResetIrqCount
        {
            get { return _resetIrqCount; }
            set
            {
                if (_resetIrqCount != value)
                {
                    _resetIrqCount = value;
                    if (_selectedModule != null)
                    {
                        _selectedModule.ResetIRQCount = value;
                    }
                    NotifyPropertyChanged(nameof(ResetIrqCount));
                }
            }
        }

        private string _irqWaitTimeoutText;
        public string IrqWaitTimeoutText
        {
            get { return _irqWaitTimeoutText; }
            set
            {
                if (_selectedModule == null) return;
                if (_irqWaitTimeoutText != value)
                {
                    _irqWaitTimeoutText = value;
                    string msg = ValidateTimeoutValue(_irqWaitTimeoutText, out uint timeoutValue);
                    if (!string.IsNullOrEmpty(msg))
                    {
                        _errorInfo.AddOrUpdate(nameof(IrqWaitTimeoutText), msg);
                    }
                    else
                    {
                        _selectedModule.TimeoutMilliseconds = timeoutValue;
                        _errorInfo.Remove(nameof(IrqWaitTimeoutText));
                    }

                    NotifyPropertyChanged(nameof(IrqWaitTimeoutText));
                    EnableDisableIrqCommand.RaiseCanExecuteChanged();
                }
            }
        }

        private string ValidateTimeoutValue(string textValue, out uint timeout)
        {
            timeout = 0;
            string msg = string.Empty;
            if (string.IsNullOrEmpty(textValue))
            {
                msg = "The value cannot be empty";
            }
            else
            {
                if (!uint.TryParse(textValue, out timeout))
                {
                    msg = "The entered text \"" + textValue + "\" is not a valid number";
                }
            }
            return msg;
        }

        private void InitViewModels()
        {
            BitChangeViewModel = new BitChangeViewModel();
            BitChangeViewModel.PropertyChanged += MainViewModel_PropertyChanged;
            BitPatternCompareViewModel = new BitPatternCompareViewModel();
            BitPatternCompareViewModel.PropertyChanged += MainViewModel_PropertyChanged;
            PortStatusViewModel = new PortStatusViewModel();
            PortStatusViewModel.PropertyChanged += MainViewModel_PropertyChanged;
            ExternalIrqViewModel = new ExternalIrqViewModel();
            ExternalIrqViewModel.PropertyChanged += MainViewModel_PropertyChanged;
        }

        public BitChangeViewModel BitChangeViewModel { get; private set; }
        public BitPatternCompareViewModel BitPatternCompareViewModel { get; private set; }
        public PortStatusViewModel PortStatusViewModel { get; private set; }
        public ExternalIrqViewModel ExternalIrqViewModel { get; private set; }

        private void CommonIrqStatusChanged(object sender, CommonIrqStatusChangedEventArgs e)
        {
            StatusIrqCount = e.IrqCount;
            StatusTriggerValue = e.TriggerValue;
            AppendInfoText(e.InfoText);
        }

        private uint _statusIrqCount;
        public uint StatusIrqCount
        {
            get { return _statusIrqCount; }
            set
            {
                if (_statusIrqCount != value)
                {
                    _statusIrqCount = value;
                    NotifyPropertyChanged(nameof(StatusIrqCount));
                }
            }
        }

        private uint _statusTriggerValue;
        public uint StatusTriggerValue
        {
            get { return _statusTriggerValue; }
            set
            {
                if (_statusTriggerValue != value)
                {
                    _statusTriggerValue = value;
                    NotifyPropertyChanged(nameof(StatusTriggerValue));
                }
            }
        }

        string _infoText = string.Empty;
        public string InfoText
        {
            get { return _infoText; }
            set
            {
                _infoText = value;
                NotifyPropertyChanged(nameof(InfoText));
            }
        }

        private void AppendInfoText(string text)
        {
            if (!string.IsNullOrEmpty(_infoText))
            {
                _infoText += Environment.NewLine;
            }
            _infoText += text;
            NotifyPropertyChanged(nameof(InfoText));
        }

        private void ClearStatusIndicatorsExecute(object obj)
        {
            ClearStatusIndicators();
        }

        private void CloseApplicationExecute(object obj)
        {
            if (_selectedModule != null && !_selectedModule.IsModuleIdle)
            {
                _selectedModule.DisableInterrupt();
            }

            Application.Current.MainWindow.Close();
        }

        void MainViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(ModuleLocator.InfoText))
            {
                AppendInfoText(_moduleLocator.InfoText);
            }
            else if (e.PropertyName == nameof(_selectedModule.ModuleInfoText))
            {
                AppendInfoText(_selectedModule.ModuleInfoText);
            }
            else if (e.PropertyName == nameof(_selectedModule.ModuleErrorText))
            {
                AppendInfoText(_selectedModule.ModuleErrorText);
            }
            else if (e.PropertyName == nameof(_selectedModule.IsModuleIdle))
            {
                UpdateModuleIdleStatus();
            }
            else if (e.PropertyName == nameof(IsInputValid))
            {
                EnableDisableIrqCommand.RaiseCanExecuteChanged();
            }
        }

        public override bool IsInputValid()
        {
            if (!BitChangeViewModel.IsInputValid())
            {
                return false;
            }

            if (!BitPatternCompareViewModel.IsInputValid())
            {
                return false;
            }

            if (!PortStatusViewModel.IsInputValid())
            {
                return false;
            }

            if (!ExternalIrqViewModel.IsInputValid())
            {
                return false;
            }

            return _errorInfo.Count == 0;
        }
    }
}
