﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the ReadBuffer project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using ReadBuffer.Utility;
using System.Collections.Generic;

namespace ReadBuffer.ALLDAQ
{
    public class AnalogInputModule : InputModule
    {
        private List<AnalogRange> _ranges = new List<AnalogRange>();

        public AnalogInputModule(string boardName, uint boardIndex, uint moduleIndex, 
            ref AdqDriver.ADQGeneralModuleInfo moduleInfo, ref AdqDriver.ADQAIModuleInfo adqAiModuleInfo)
            : base(boardName, boardIndex, moduleIndex, ref moduleInfo)
        {
            AddRanges(ref adqAiModuleInfo);
            SelectedRange = _ranges[0];
            NumberOfChannels = adqAiModuleInfo.NumberOfChannels;
            BuildModuleName();
        }

        private void AddRanges(ref AdqDriver.ADQAIModuleInfo adqAiModuleInfo)
        {
            for (uint i = 0; i < adqAiModuleInfo.NumberOfRanges; i++)
            {
                AdqDriver.ADQAIRange aiRange = adqAiModuleInfo.Ranges[i];
                var physicalRange = new AnalogRange(
                    aiRange.PhysicalUnit,
                    aiRange.DigitalMin,
                    aiRange.DigitalMax,
                    aiRange.PhysicalMin,
                    aiRange.PhysicalMax);
                _ranges.Add(physicalRange);
            }
        }

        public uint NumberOfChannels { get; }

        public int NumberOfRanges
        {
            get
            {
                return _ranges.Count;
            }
        }

        private void BuildModuleName()
        {
            _moduleName = string.Format("{0}, Board {1}, Module {2}, Type {3}, {4} {5}, {6} clock, max. {7}",
                _modelName,
                _boardIndex, 
                _moduleIndex,
                AdqCommon.GetModuleTypeName(_moduleType), 
                NumberOfChannels,
                NumberOfChannels == 1 ? "Channel" : "Channels",
                AdqCommon.GetClockType(_clockType), 
                AdqCommon.FormatFrequency(MaxfrequencyHz));
        }

        public uint SelectedChannelIndex { get; private set; }

        public AnalogRange GetRange(int index)
        {
            return _ranges[index];
        }
        public AnalogRange SelectedRange { get; private set; }
        public uint SelectedRangeIndex { get; private set; }

        internal uint ConfigureModule(uint channelIndex, uint rangeIndex)
        {
            uint adqError = AdqDriver.adqAISetChannelAndRange(_boardIndex, _moduleIndex, channelIndex, rangeIndex);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                ErrorTextChanged("adqAISetChannelAndRange", adqError);
            }
            else
            {
                SelectedChannelIndex = channelIndex;
                SelectedRangeIndex = rangeIndex;
                SelectedRange = _ranges[(int)rangeIndex];
            }
            return adqError;
        }
    }
}
