﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the ReadBuffer project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using ReadBuffer.Utility;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace ReadBuffer.ALLDAQ
{
    public class ModuleLocator
    {
        private uint _adqOpenError = AdqCommon.NotOpen;
        private readonly ICollection<IInputModule> _inputModules;
        private readonly ICollection<uint> _moduleTypes;

        public event EventHandler<InfoTextChangedEventArgs> InfoTextChanged;

        public ModuleLocator(ICollection<IInputModule> inputModules, ICollection<uint> moduleTypes,
            EventHandler<InfoTextChangedEventArgs> infoTextChangedHandler)
        {
            _inputModules = inputModules;
            _moduleTypes = moduleTypes;
            InfoTextChanged += infoTextChangedHandler;
        }

        public void FindInputModules()
        {
            _adqOpenError = OpenAdqDriver();
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                uint numberOfBoards = GetNumberOfBoards();
                for (uint boardIdx = 0; boardIdx < numberOfBoards; boardIdx++)
                {
                    uint adqError = AdqDriver.adqGetNumberOfModules(boardIdx, out uint numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("adqGetNumberOfModules", adqError);
                        continue;
                    }

                    adqError = GetModules(boardIdx, numberOfModules);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        continue;
                    }
                }
            }
        }

        private uint GetModules(uint boardIndex, uint numberOfModules)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;
            for (uint moduleIdx = 0; moduleIdx < numberOfModules; moduleIdx++)
            {
                adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, out uint moduleType);
                if (adqError != AdqErrors.ADQ_ERROR_NONE)
                {
                    SetFunctionErrorInfo("adqGetModuleType", adqError);
                    continue;
                }

                if (_moduleTypes.Contains(moduleType))
                {
                    adqError = GetModuleInfo(boardIndex, moduleIdx, moduleType);
                    if (adqError != AdqErrors.ADQ_ERROR_NONE)
                    {
                        SetFunctionErrorInfo("GetModuleInfo", adqError);
                        continue;
                    }
                }
            }
            return adqError;
        }

        private uint GetModuleInfo(uint boardIndex, uint moduleIdx, uint moduleType)
        {
            uint adqError = AdqErrors.ADQ_ERROR_NONE;

            if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DI)
            {
                adqError = GetDIModuleInfo(boardIndex, moduleIdx);
            }
            else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_DIO)
            {
                adqError = GetDIOModuleInfo(boardIndex, moduleIdx);
            }
            else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_CNTR)
            {
                adqError = GetCounterModuleInfo(boardIndex, moduleIdx);
            }
            else if (moduleType == AdqConstants.ADQ_MODULE_TYPE_AI)
            {
                adqError = GetAIModuleInfo(boardIndex, moduleIdx);
            }
            return adqError;
        }

        private uint GetDIModuleInfo(uint boardIndex, uint moduleIndex)
        {
            uint adqError;
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDiModuleInfo = new AdqDriver.ADQDIModuleInfo(AdqDriver.ADQDIModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDiModuleInfo.Size);
            Marshal.StructureToPtr(adqDiModuleInfo, moduleInfo, false);

            adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDiModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDiModuleInfo = (AdqDriver.ADQDIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDiModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDiModuleInfo.Size, infoSize));

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));

                string boardName = GetBoardName(boardIndex);
                var adqModule = new DigitalInputModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDiModuleInfo.PortWidthBits);
                _inputModules.Add(adqModule);
            }
            return adqError;
        }

        private uint GetDIOModuleInfo(uint boardIndex, uint moduleIndex)
        {
            uint adqError;
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqDioModuleInfo = new AdqDriver.ADQDIOModuleInfo(AdqDriver.ADQDIOModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqDioModuleInfo.Size);
            Marshal.StructureToPtr(adqDioModuleInfo, moduleInfo, false);

            adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqDioModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqDioModuleInfo = (AdqDriver.ADQDIOModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQDIOModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqDioModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqDioModuleInfo.Size, infoSize));

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));

                string boardName = GetBoardName(boardIndex);
                var adqModule = new DigitalInputModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, adqDioModuleInfo.PortWidthBits);
                _inputModules.Add(adqModule);
            }
            return adqError;
        }

        private uint GetCounterModuleInfo(uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqCntrModuleInfo = new AdqDriver.ADQCNTRModuleInfo(AdqDriver.ADQCNTRModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqCntrModuleInfo.Size);
            Marshal.StructureToPtr(adqCntrModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqCntrModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqCntrModuleInfo = (AdqDriver.ADQCNTRModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQCNTRModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqCntrModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqCntrModuleInfo.Size, infoSize));

            string boardName = GetBoardName(boardIndex);
            if (adqCntrModuleInfo.CounterType != AdqConstants.ADQ_CNTR_TYPE_8254)
            {
                UpdateInfoText(string.Format("Board {0}, {1}, Module {2} counter type not supported by this program.",
                                  boardIndex, boardName, moduleIndex));
                return AdqErrors.ADQ_ERROR_UNSUPPORTED;
            }

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));

                var adqModule = new CounterModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, ref adqCntrModuleInfo);
                _inputModules.Add(adqModule);
            }
            return adqError;
        }

        private uint GetAIModuleInfo(uint boardIndex, uint moduleIndex)
        {
            var adqGeneralModuleInfo = new AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID);
            var adqAiModuleInfo = new AdqDriver.ADQAIModuleInfo(AdqDriver.ADQAIModuleInfo_ID);

            IntPtr moduleInfo = Marshal.AllocHGlobal((int)adqAiModuleInfo.Size);
            Marshal.StructureToPtr(adqAiModuleInfo, moduleInfo, false);

            uint adqError = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                ref adqGeneralModuleInfo,
                adqAiModuleInfo.Size,
                moduleInfo,
                out uint infoSize);
            adqAiModuleInfo = (AdqDriver.ADQAIModuleInfo)Marshal.PtrToStructure(moduleInfo, typeof(AdqDriver.ADQAIModuleInfo));
            Marshal.FreeHGlobal(moduleInfo);

            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetModuleInfo", adqError);
                return adqError;
            }

            Debug.Assert(adqAiModuleInfo.Size == infoSize,
                string.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqAiModuleInfo.Size, infoSize));

            string boardName = GetBoardName(boardIndex);
            if (adqAiModuleInfo.RangeInfoPerChannel)
            {
                var args = new InfoTextChangedEventArgs
                {
                    InfoText = string.Format("Board {0}, {1}, Module {2} has individual ranges per channel."
                                  + Environment.NewLine
                                  + "Please update to a newer version of this program.",
                                  boardIndex, boardName, moduleIndex)
                };
                RaiseInfoTextChanged(args);
                return AdqErrors.ADQ_ERROR_UNSUPPORTED;
            }

            if (IsClockControlledModule(adqGeneralModuleInfo.ClockType))
            {
                Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0, string.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex));

                var adqModule = new AnalogInputModule(boardName, boardIndex, moduleIndex, ref adqGeneralModuleInfo, ref adqAiModuleInfo);
                _inputModules.Add(adqModule);
            }
            return adqError;
        }

        private bool IsClockControlledModule(uint clockType)
        {
            return clockType == AdqConstants.ADQ_CLOCK_TYPE_HARDWARE || clockType == AdqConstants.ADQ_CLOCK_TYPE_SOFTWARE;
        }

        private string GetBoardName(uint boardIndex)
        {
            var boardInfo = new AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID);

            uint adqError = AdqDriver.adqGetBoardInfo(boardIndex, ref boardInfo);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetBoardInfo", adqError);
                return "";
            }
            return boardInfo.BoardName;
        }

        private uint OpenAdqDriver()
        {
            uint adqError = AdqDriver.adqOpen();
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqOpen", adqError);
            }
            return adqError;
        }

        public void CloseAdqDriver()
        {
            if (_adqOpenError == AdqErrors.ADQ_ERROR_NONE)
            {
                AdqDriver.adqClose();
                _adqOpenError = AdqCommon.NotOpen;
            }
        }

        private uint GetNumberOfBoards()
        {
            uint adqError = AdqDriver.adqGetNumberOfBoards(out uint numBoards);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetFunctionErrorInfo("adqGetNumberOfBoards", adqError);
                return 0;
            }
            return numBoards;
        }

        private void SetFunctionErrorInfo(string functionName, uint adqError)
        {
            UpdateInfoText(string.Format("API-function {0} returned Error number: 0x{1:X}", functionName, adqError));
        }

        private void UpdateInfoText(string text)
        {
            var args = new InfoTextChangedEventArgs
            {
                InfoText = text
            };
            RaiseInfoTextChanged(args);
        }

        protected virtual void RaiseInfoTextChanged(InfoTextChangedEventArgs e)
        {
            InfoTextChanged?.Invoke(this, e);
        }
    }
}
