﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// https://www.alldaq.com
//
// Purpose:
//       This file is part of the ReadBuffer project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using System;
using System.Collections.Generic;
using System.Globalization;

namespace ReadBuffer.Utility
{
    public static class AdqCommon
    {
        private const double _oneMicrosecond = 0.000001;
        private const double _oneMillisecond = 0.001;
        private const double _oneSecond = 1.0;
        private const double _oneMinute = 60;
        private const double _oneHour = 3600;

        public static uint NotOpen = AdqErrors.ADQ_ERROR_TYPE_ERROR | AdqErrors.ADQ_ERROR_NOT_OPEN;
        public static readonly uint DefaultNumberOfValues = 50;
        public static readonly uint[] Counter8254Modes = {AdqConstants.ADQ_CNTR_8254_MODE_0,
                                                          AdqConstants.ADQ_CNTR_8254_MODE_1,
                                                          AdqConstants.ADQ_CNTR_8254_MODE_2,
                                                          AdqConstants.ADQ_CNTR_8254_MODE_3,
                                                          AdqConstants.ADQ_CNTR_8254_MODE_4,
                                                          AdqConstants.ADQ_CNTR_8254_MODE_5};
        public static string GetModuleTypeName(uint moduleType)
        {
            return GetDictionaryValue(_moduleTypeNames, moduleType);
        }

        public static string GetPhysicalUnitName(uint physicalUnit)
        {
            return GetDictionaryValue(_physicalUnitNames, physicalUnit);
        }

        public static string GetPhysicalUnitSymbol(uint physicalUnit)
        {
            return GetDictionaryValue(_physicalUnitSymbols, physicalUnit);
        }

        public static string GetClockType(uint clockType)
        {
            return GetDictionaryValue(_clockTypeNames, clockType);
        }

        public static string GetCounterClockSourceName(uint counterClockSource)
        {
            return GetDictionaryValue(_counterClockSourceNames, counterClockSource);
        }

        private static string GetDictionaryValue(IDictionary<uint, string> dictionary, uint key)
        {
            bool found = dictionary.TryGetValue(key, out string stringValue);
            if (!found)
            {
                stringValue = string.Format("0x{0:X8}", key);
            }
            return stringValue;
        }

        private static readonly Dictionary<uint, string> _moduleTypeNames = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_MODULE_TYPE_DIO, "DIO"},
            {AdqConstants.ADQ_MODULE_TYPE_DO, "DO"},
            {AdqConstants.ADQ_MODULE_TYPE_DI, "DI"},
            {AdqConstants.ADQ_MODULE_TYPE_CNTR, "Counter"},
            {AdqConstants.ADQ_MODULE_TYPE_EXTINT, "External Interrupt"},
            {AdqConstants.ADQ_MODULE_TYPE_AI, "AI"},
            {AdqConstants.ADQ_MODULE_TYPE_AO, "AO"},
            {AdqConstants.ADQ_MODULE_TYPE_I2C, "I2C"},
            {AdqConstants.ADQ_MODULE_TYPE_PWM, "PWM"},
            {AdqConstants.ADQ_MODULE_TYPE_FIN, "FIN"}
        };

        private static readonly Dictionary<uint, string> _physicalUnitNames = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_PHYSICAL_UNIT_VOLT, "Volt"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_AMPERE, "Ampere"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_OHM, "Ohm"}
        };

        private static readonly Dictionary<uint, string> _physicalUnitSymbols = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_PHYSICAL_UNIT_VOLT, "V"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_AMPERE, "A"},
            {AdqConstants.ADQ_PHYSICAL_UNIT_OHM, "Ω"}
        };

        private static readonly Dictionary<uint, string> _clockTypeNames = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_CLOCK_TYPE_HARDWARE, "Hardware"},
            {AdqConstants.ADQ_CLOCK_TYPE_SOFTWARE, "Software"},
            {AdqConstants.ADQ_CLOCK_TYPE_NONE, "None"}
        };

        private static readonly Dictionary<uint, string> _counterClockSourceNames = new Dictionary<uint, string>
        {
            {AdqConstants.ADQ_CNTR_CLOCK_SOURCE_DEFAULT, "Default"},
            {AdqConstants.ADQ_CNTR_CLOCK_SOURCE_EXTERNAL, "External"},
            {AdqConstants.ADQ_CNTR_CLOCK_SOURCE_INTERNAL_10_MHZ, "Internal 10 MHz"},
            {AdqConstants.ADQ_CNTR_CLOCK_SOURCE_INTERNAL_66_MHZ, "Internal 66 MHz"}
        };


        public static string GetModelName(string boardName)
        {
            int posADQ = boardName.IndexOf("ADQ");
            if (posADQ < 0) posADQ = 0;

            int posEnd = boardName.IndexOf(" ", posADQ);
            if (posEnd < 0)
            {
                posEnd = boardName.Length;
            }

            return boardName.Substring(posADQ, posEnd - posADQ);
        }

        public static string FormatPhysicalValue(double physicalValue)
        {
            return FormatPhysicalValue("{0:0.##}", physicalValue);
        }

        public static string FormatPhysicalValue(string formatString, double physicalValue)
        {
            string physicalTxt;
            double tmpVal = Math.Abs(physicalValue);
            if (tmpVal == 0)
            {
                physicalTxt = "0";
            }
            else if (tmpVal < 0.000001)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000000000.0) + "n";
            }
            else if (tmpVal < 0.001)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000000.0) + "µ";
            }
            else if (tmpVal < 1.0)
            {
                physicalTxt = string.Format(formatString, physicalValue * 1000.0) + "m";
            }
            else
            {
                physicalTxt = string.Format(formatString, physicalValue);
            }
            return physicalTxt;
        }

        public static string FormatFrequency(double frequency)
        {
            string s;
            if (frequency >= 1000000)
            {
                s = string.Format("{0:0.######}MHz", frequency / 1000000);
            }
            else if (frequency >= 1000)
            {
                s = string.Format("{0:0.######}kHz", frequency / 1000);
            }
            else
            {
                s = string.Format("{0:0.########}Hz", frequency);
            }
            return s;
        }

        public static string FormatTimeToWait(double seconds)
        {
            string s = string.Empty;
            try
            {
                var ts = TimeSpan.FromSeconds(seconds);
                s = string.Concat(ts.Days >= 1 ? string.Concat(ts.Days, " d ") : "",
                                  ts.Hours >= 1 ? string.Concat(ts.Hours, " h ") : "",
                                  ts.Minutes >= 1 ? string.Concat(ts.Minutes, " min ") : "",
                                  ts.Seconds >= 1 ? string.Concat(ts.Seconds, " sec ") : "",
                                  ts.Milliseconds >= 1 ? string.Concat(ts.Milliseconds, " ms") : "");

                if (s == string.Empty && seconds > 0)
                {
                    s = "< 1ms";
                }
            }
            catch (OverflowException)
            {
                ;
            }
            return s;
        }

        public static string FormatSampleTime(double seconds)
        {
            if (seconds == 0.0) return "0";

            string s = "";

            if (seconds < _oneMicrosecond)
            {
                s = string.Format("{0:0##}ns", seconds * 1000000000);
            }
            else if (seconds < _oneMillisecond)
            {
                s = string.Format(CultureInfo.InvariantCulture, "{0:0.#######}µs", seconds * 1000000);
            }
            else if (seconds < _oneMinute)
            {
                s = string.Format(CultureInfo.InvariantCulture, "{0:0.#######}s", seconds);
            }
            else if (seconds < _oneHour)
            {
                int minutes = (int)(seconds / _oneMinute);
                double sec = seconds % _oneMinute;
                s = string.Format(CultureInfo.InvariantCulture, "{0}:{1:0.0000000}m", minutes, sec);
            }
            else
            {
                int hours = (int)(seconds / _oneHour);
                int minutes = (int)(seconds % _oneHour);
                double sec = seconds % _oneMinute;
                s = string.Format(CultureInfo.InvariantCulture, "{0}:{1}:{2:0.0000000}h", hours, minutes, sec);
            }
            return s;
        }
    }
}
