﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the WriteBufferAO project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.InteropServices;
using WriteBufferAO.Utility;

namespace WriteBufferAO.Models
{
    public class AoModule : INotifyPropertyChanged
    {
        private readonly uint _defaultRepeatBufferCount = 2;
        private WaveForm _waveForm;
        private uint[] _values;
        public event PropertyChangedEventHandler PropertyChanged;

        public AoModule(uint boardIndex, uint moduleIndex, ref AdqDriver.ADQGeneralModuleInfo generalInfo, ref AdqDriver.ADQAOModuleInfo aoModuleInfo)
        {
            BoardIndex = boardIndex;
            ModuleIndex = moduleIndex;
            MaxFrequency = generalInfo.ClockBaseFrequencyHz / generalInfo.ClockMinimumDivider;
            DefaultOutputFrequency = MaxFrequency / 100;
            _repeatBufferCount = _defaultRepeatBufferCount;
            AoRanges = new List<AoRange>();
            ModuleName = string.Format("Module {0}, max. {1}", moduleIndex, AdqCommon.FormatFrequency(MaxFrequency));
            AddRanges(ref aoModuleInfo);
            DesiredOutputFrequency = DefaultOutputFrequency;
        }

        private void AddRanges(ref AdqDriver.ADQAOModuleInfo aoModuleInfo)
        {
            for (uint i = 0; i < aoModuleInfo.NumberOfRanges; i++)
            {
                AdqDriver.ADQAORange adqRange = aoModuleInfo.Ranges[i];
                var aoRange = new AoRange(i,
                    adqRange.DigitalMin,
                    adqRange.DigitalMax,
                    adqRange.PhysicalUnit,
                    adqRange.PhysicalMin,
                    adqRange.PhysicalMax);
                AoRanges.Add(aoRange);
            }

            if (aoModuleInfo.NumberOfRanges >= 1)
            {
                SelectedAoRange = AoRanges[0];
            }
        }

        public AoRange SelectedAoRange { get; set; }
        public List<AoRange> AoRanges { get; }
        public uint BoardIndex { get; }
        public uint ModuleIndex { get; }
        public string ModuleName { get; }
        public double MaxFrequency { get; }
        public double DefaultOutputFrequency { get; }

        private double _desiredOutputFrequency;
        public double DesiredOutputFrequency
        {
            get { return _desiredOutputFrequency; }
            set
            {
                if (_desiredOutputFrequency != value)
                {
                    _desiredOutputFrequency = value;
                    if (value > 0 && value <= MaxFrequency)
                    {
                        SetActualOutputFrequency(value);
                    }
                }
            }
        }

        private void SetActualOutputFrequency(double value)
        {
            double desiredFrequency = value;
            uint adqError = AdqDriver.adqSetFrequency(BoardIndex, ModuleIndex, desiredFrequency, out double actualFrequency);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                ActualOutputFrequency = actualFrequency;
                RaisePropertyChanged(nameof(ActualOutputFrequency));
            }
            else
            {
                SetErrorText("adqSetFrequency", adqError);
            }
        }

        public double ActualOutputFrequency { get; private set; }

        uint _repeatBufferCount;
        public uint RepeatBufferCount
        {
            get { return _repeatBufferCount; }
            set
            {
                if (_repeatBufferCount != value)
                {
                    _repeatBufferCount = value;
                    RaisePropertyChanged(nameof(RepeatBufferCount));
                }
            }
        }

        public uint AdqError { get; set; }

        public uint NumberOfValuesWritten { get; set; }

        public void SetWaveForm(WaveForm waveForm)
        {
            _waveForm = waveForm;
            _values = PhysicalToDigital(waveForm.Values);
        }

        public void SetArbitraryWaveForm(List<double> physicalValues)
        {
            _values = PhysicalToDigital(physicalValues.ToArray());
        }

        public void WriteBuffer()
        {
            NumberOfValuesWritten = 0;
            uint numberOfValues = (uint)_values.Length;
            GCHandle gch = GCHandle.Alloc(_values, GCHandleType.Pinned);
            AdqError = AdqDriver.adqWriteBuffer(BoardIndex, ModuleIndex,
                numberOfValues, _values, _repeatBufferCount, out uint numValuesWritten);
            NumberOfValuesWritten = numValuesWritten;
            gch.Free();
        }

        private uint[] PhysicalToDigital(double[] physicalValues)
        {
            uint valueCountAboveMax = 0;
            uint valueCountBelowMin = 0;

            uint[] values = _values;
            if (values == null || values.Length != physicalValues.Length)
            {
                values = new uint[physicalValues.Length];
            }

            for (int i = 0; i < values.Length; i++)
            {
                double value = physicalValues[i];
                if (value > SelectedAoRange.PhysicalMax)
                {
                    value = SelectedAoRange.PhysicalMax;
                    valueCountAboveMax++;
                }
                else if (value < SelectedAoRange.PhysicalMin)
                {
                    value = SelectedAoRange.PhysicalMin;
                    valueCountBelowMin++;
                }

                values[i] = SelectedAoRange.PhysicalToDigital(value);
            }

            if (valueCountAboveMax > 0 || valueCountBelowMin > 0)
            {
                SetInfoText(string.Format("The signal is clipped: Values above range max: {0}, values below range min: {1}", 
                    valueCountAboveMax, valueCountBelowMin));
            }

            return values;
        }

        public bool GetModuleStatus()
        {
            uint adqError = AdqDriver.adqWriteContinuousStatus(BoardIndex, ModuleIndex, false, out uint numAccept, out bool isModuleRunning);
            if (adqError != AdqErrors.ADQ_ERROR_NONE)
            {
                SetErrorText("adqWriteContinuousStatus", adqError);
            }
            return isModuleRunning;
        }

        public void Stop()
        {
            uint adqError = AdqDriver.adqStop(BoardIndex, ModuleIndex);
            if (adqError == AdqErrors.ADQ_ERROR_NONE)
            {
                SetInfoText("stopped");
            }
            else
            {
                SetErrorText("adqStop", adqError);
            }
        }

        public string ValidateRepeatBufferCount(uint count)
        {
            string msg = string.Empty;
            if (count == 0)
            {
                msg = "The number must be greater than 0";
            }
            return msg;
        }

        public string ValidateDesiredOutputFrequency(double frequency)
        {
            string msg = string.Empty;
            if (frequency <= 0)
            {
                msg = "The output frequency must be greater than 0";
            }
            else if (frequency > MaxFrequency)
            {
                msg = "The output frequency cannot be greater than the maximum output frequency";
            }
            return msg;
        }

        private void SetErrorText(string functionName, uint adqError)
        {
            ModuleErrorText = string.Format("Board {0}, Module {1}: Function {2} returned error number: 0x{3:X}",
                BoardIndex,
                ModuleIndex,
                functionName, 
                adqError);
        }

        string _errorText = string.Empty;
        public string ModuleErrorText
        {
            get { return _errorText; }
            set
            {
                _errorText = value;
                if (value != string.Empty)
                {
                    RaisePropertyChanged(nameof(ModuleErrorText));
                }
            }
        }

        private void SetInfoText(string info)
        {
            ModuleInfoText = string.Format("Board {0} Module {1}: {2}",
                BoardIndex,
                ModuleIndex,
                info);
        }

        string _infoText = string.Empty;
        public string ModuleInfoText
        {
            get { return _infoText; }
            set
            {
                _infoText = value;
                if (value != string.Empty)
                {
                    RaisePropertyChanged(nameof(ModuleInfoText));
                }
            }
        }

        void RaisePropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}
