﻿//-----------------------------------------------------------------------------
// ALLDAQ, a division of ALLNET GmbH
// http://www.alldaq.com
//
// Purpose:
//       This file is part of the WriteBufferAO project of the ALLDAQ-SDK.  
//       It shows the usage of the API functions to access
//       ALLDAQ Data Acquisition Boards
// 
//-----------------------------------------------------------------------------

using ALLDAQ.Boards;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;
using WriteBufferAO.Messaging;
using WriteBufferAO.Models;
using WriteBufferAO.Utility;

namespace WriteBufferAO.ViewModels
{
    public class AoModuleViewModel : INotifyPropertyChanged, IDataErrorInfo, IDisposable
    {
        private readonly AoModule _aoModule;
        private readonly Dictionary<string, string> _errorInfo = new Dictionary<string, string>();
        private readonly double _defaultSignalLength = 1.0;
        private readonly string _fmtStr = "{0:0.###}";
        private readonly WaveForm _waveForm;
        private DispatcherTimer _dispatcherTimer;
        private BackgroundWorker _worker;
        private readonly IDialogService _dialogService;
        private const int _tabItemIndexArbitrarySignal = 2;
        public RelayCommand StartStopCommand { get; private set; }

        public event PropertyChangedEventHandler PropertyChanged;

        public AoModuleViewModel(AoModule aoModule, Brush backgroundColor, IDialogService dialogService)
        {
            _aoModule = aoModule;
            BackgroundColor = backgroundColor;
            _dialogService = dialogService;
            _aoModule.PropertyChanged += AoModuleViewModel_PropertyChanged;

            _desiredOutputFrequency = string.Format(_fmtStr, _aoModule.DefaultOutputFrequency);

            _signalLength = _defaultSignalLength;
            _signalLengthText = string.Format(_fmtStr, _signalLength);
            _repeatBufferCount = _aoModule.RepeatBufferCount.ToString();
            UpdateNumberOfValues();
            InitSignalViewModels();
            InitCommands();
            InitDispatcherTimer();
            _waveForm = new WaveForm();

            Messenger.Default.Send(new InfoMessage(string.Format("Selected Board {0}, {1}",
                _aoModule.BoardIndex, _aoModule.ModuleName)), AdqCommon.MessengerContextInfo);
        }

        private void InitCommands()
        {
            StartStopCommand = new RelayCommand(StartStopWriteBufferExecute, CanStartStopWriteBufferExecute);
        }

        private void InitDispatcherTimer()
        {
            _dispatcherTimer = new DispatcherTimer();
            _dispatcherTimer.Tick += new EventHandler(DispatcherTimer_Tick);
            _dispatcherTimer.Interval = new TimeSpan(0, 0, 0, 0, 500);
        }

        private bool CanStartStopWriteBufferExecute(object obj)
        {
            return IsInputValid;
        }

        private void StartStopWriteBufferExecute(object obj)
        {
            if (_isModuleIdle)
            {
                GenerateWaveForm();
            }
            else
            {
                _aoModule.Stop();
            }
        }

        bool _isModuleIdle = true;
        public bool IsModuleIdle
        {
            get { return _isModuleIdle; }
            set
            {
                if (_isModuleIdle != value)
                {
                    _isModuleIdle = value;

                    ButtonContent = _isModuleIdle ? "Start" : "Stop";
                    RaisePropertyChanged(nameof(IsModuleIdle));
                }
            }
        }

        string _buttonContent = "Start";
        public string ButtonContent
        {
            get { return _buttonContent; }
            set
            {
                _buttonContent = value;
                RaisePropertyChanged(nameof(ButtonContent));
            }
        }

        private void GenerateWaveForm()
        {
            if (_selectedTabItemIndex == _tabItemIndexArbitrarySignal)
            {
                _aoModule.SetArbitraryWaveForm(ArbitrarySignalViewModel.ArbitraryValues);
            }
            else
            {
                _waveForm.InitWaveForm(_aoModule.ActualOutputFrequency, _signalLength, NumberOfValues);
                foreach (var signalViewModel in SignalViewModels)
                {
                    signalViewModel.ApplySignalGenerator(_waveForm);
                }
                _aoModule.SetWaveForm(_waveForm);
            }

            StartBackgroundWorker();
        }

        private int _selectedTabItemIndex;
        public int SelectedTabItemIndex
        {
            get { return _selectedTabItemIndex; }
            set
            {
                if (_selectedTabItemIndex != value)
                {
                    _selectedTabItemIndex = value;
                    RaisePropertyChanged(nameof(SelectedTabItemIndex));
                }
            }
        }

        public void StopModule()
        {
            if (!_isModuleIdle)
            {
                _aoModule.Stop();
            }
        }

        private void StartBackgroundWorker()
        {
            IsModuleIdle = false;
            Messenger.Default.Send(new ModuleStatusMessage(this, false), AdqCommon.MessengerContextStatus);
            _aoModule.AdqError = AdqErrors.ADQ_ERROR_NONE;
            AppendInfoStartWriting();
            _worker = new BackgroundWorker
            {
                WorkerReportsProgress = false,
                WorkerSupportsCancellation = false
            };
            _worker.DoWork += Worker_DoWork;
            _worker.RunWorkerCompleted += Worker_RunWorkerCompleted;
            CommandManager.InvalidateRequerySuggested();
            _worker.RunWorkerAsync(_aoModule);
        }

        private void Worker_DoWork(object sender, DoWorkEventArgs e)
        {
            AoModule aoModule = e.Argument as AoModule;
            aoModule.WriteBuffer();
            e.Result = aoModule;
        }

        private void Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            AppendInfoValuesWritten();
            _dispatcherTimer.Start();
        }

        private void DispatcherTimer_Tick(object sender, EventArgs e)
        {
            CheckModuleStatus();
        }

        private void CheckModuleStatus()
        {
            bool isModuleRunning = _aoModule.GetModuleStatus();

            if (!isModuleRunning)
            {
                _dispatcherTimer.Stop();
                IsModuleIdle = true;
                AppendInfoOutputEnded();
                Messenger.Default.Send(new ModuleStatusMessage(this, true), AdqCommon.MessengerContextStatus);
                CommandManager.InvalidateRequerySuggested();
            }
        }

        private void AppendInfoStartWriting()
        {
            string s = string.Format("Board {0} Module {1}, output started, frequency: {2}, values: {3}, repeat: {4}",
                                _aoModule.BoardIndex,
                                _aoModule.ModuleIndex,
                                ActualOutputFrequency,
                                NumberOfValues,
                                _repeatBufferCount);
            Messenger.Default.Send(new InfoMessage(s), AdqCommon.MessengerContextInfo);
        }

        private void AppendInfoValuesWritten()
        {
            string s = string.Format("Board {0} Module {1}, values written to internal data buffer",
                                _aoModule.BoardIndex,
                                _aoModule.ModuleIndex);
            Messenger.Default.Send(new InfoMessage(s), AdqCommon.MessengerContextInfo);
        }

        private void AppendInfoOutputEnded()
        {
            string s = string.Format("Board {0} Module {1}, output ended, number of values written: {2}",
                                _aoModule.BoardIndex,
                                _aoModule.ModuleIndex,
                                _aoModule.NumberOfValuesWritten);
            Messenger.Default.Send(new InfoMessage(s), AdqCommon.MessengerContextInfo);
        }

        public List<AoRange> AoRanges
        {
            get { return _aoModule.AoRanges; }
        }

        public AoRange SelectedAoRange
        {
            get { return _aoModule.SelectedAoRange; }
            set
            {
                if (_aoModule.SelectedAoRange != value)
                {
                    _aoModule.SelectedAoRange = value;
                    RaisePropertyChanged(nameof(SelectedAoRange));
                }
            }
        }

        public string ModuleName
        {
            get { return _aoModule.ModuleName; }
        }

        public Brush BackgroundColor { get; }

        private string _desiredOutputFrequency;
        public string DesiredOutputFrequency
        {
            get { return _desiredOutputFrequency; }
            set
            {
                if (_desiredOutputFrequency != value)
                {
                    _desiredOutputFrequency = value;
                    string msg = InputValidation.ValidateValueIsPositive(_desiredOutputFrequency, out double frequency);
                    if (string.IsNullOrEmpty(msg))
                    {
                        msg = _aoModule.ValidateDesiredOutputFrequency(frequency);
                    }

                    UpdateErrorInfo(nameof(DesiredOutputFrequency), msg);
                    if (string.IsNullOrEmpty(msg))
                    {
                        _aoModule.DesiredOutputFrequency = frequency;
                    }
                    RaisePropertyChanged(nameof(DesiredOutputFrequency));
                    StartStopCommand.RaiseCanExecuteChanged();
                }
            }
        }

        public string ActualOutputFrequency
        {
            get
            {
                return AdqCommon.FormatFrequency(_aoModule.ActualOutputFrequency);
            }
        }

        private double _signalLength;
        public double SignalLength
        {
            get { return _signalLength; }
            set
            {
                if (_signalLength != value)
                {
                    _signalLength = value;
                    if (_selectedTabItemIndex != _tabItemIndexArbitrarySignal)
                    {
                        UpdateNumberOfValues();
                    }
                }
            }
        }

        private string _signalLengthText;
        public string SignalLengthText
        {
            get { return _signalLengthText; }
            set
            {
                if (_signalLengthText != value)
                {
                    _signalLengthText = value;
                    string msg = InputValidation.ValidateValueIsPositive(_signalLengthText, out double signalLen);
                    UpdateErrorInfo(nameof(SignalLengthText), msg);

                    if (string.IsNullOrEmpty(msg))
                    {
                        SignalLength = signalLen;
                    }

                    RaisePropertyChanged(nameof(SignalLengthText));
                    StartStopCommand.RaiseCanExecuteChanged();
                }
            }
        }

        private void UpdateSignalSetttings()
        {
            if (_selectedTabItemIndex == _tabItemIndexArbitrarySignal)
            {
                double signalLength = ArbitrarySignalViewModel.NumValuesReadFromFile / _aoModule.ActualOutputFrequency;

                SignalLengthText = string.Format(_fmtStr, signalLength);
                UpdateNumberOfValues();
            }
        }

        private void UpdateNumberOfValues()
        {
            NumberOfValues = (uint)Math.Round(_signalLength * _aoModule.ActualOutputFrequency);
            UpdateNumberOfValuesInfo();
            RaisePropertyChanged(nameof(NumberOfValues));
        }

        private void UpdateNumberOfValuesInfo()
        {
            NumberOfValuesInfo = string.Format(CultureInfo.CurrentCulture, "Total values: {0:N0}, duration {1}",
                NumberOfValues * _aoModule.RepeatBufferCount,
                AdqCommon.FormatTimeSpan(_signalLength * _aoModule.RepeatBufferCount));
            RaisePropertyChanged(nameof(NumberOfValuesInfo));
        }

        public uint NumberOfValues { get; private set; }
        public string NumberOfValuesInfo { get; private set; }

        private string _repeatBufferCount;
        public string RepeatBufferCount
        {
            get { return _repeatBufferCount; }
            set
            {
                if (_repeatBufferCount != value)
                {
                    _repeatBufferCount = value;

                    string msg = ValidateRepeatBufferCount(_repeatBufferCount, out uint count);
                    UpdateErrorInfo(nameof(RepeatBufferCount), msg);

                    if (string.IsNullOrEmpty(msg))
                    {
                        _aoModule.RepeatBufferCount = count;
                        UpdateNumberOfValuesInfo();
                    }

                    RaisePropertyChanged(nameof(RepeatBufferCount));
                    StartStopCommand.RaiseCanExecuteChanged();
                }
            }
        }

        public bool IsInputValid
        {
            get
            {
                return (_errorInfo.Count == 0) && GetSignalsInputValid();
            }
        }

        private bool GetSignalsInputValid()
        {
            bool isValid = true;

            if (_selectedTabItemIndex == _tabItemIndexArbitrarySignal)
            {
                isValid = ArbitrarySignalViewModel.IsGeneratorInputValid;
            }
            else
            {
                foreach (SignalViewModel sg in SignalViewModels)
                {
                    if (!sg.IsGeneratorInputValid)
                    {
                        isValid = false;
                        break;
                    }
                }
            }
            return isValid;
        }

        private void UpdateErrorInfo(string propertyName, string msg)
        {
            if (!string.IsNullOrEmpty(msg))
            {
                _errorInfo.AddOrUpdate(propertyName, msg);
            }
            else
            {
                _errorInfo.Remove(propertyName);
            }
        }

        public string Error => "AoModuleViewModel.Error";

        public string this[string propertyName]
        {
            get
            {
                return _errorInfo.ContainsKey(propertyName) ? _errorInfo[propertyName] : null;
            }
        }

        private string ValidateRepeatBufferCount(string textValue, out uint count)
        {
            count = 0;
            string msg;

            if (string.IsNullOrEmpty(textValue))
            {
                msg = "The value cannot be empty";
            }
            else
            {
                if (!uint.TryParse(textValue, out count))
                {
                    msg = "The entered text '" + textValue + "' is not a valid number";
                }
                else
                {
                    msg = _aoModule.ValidateRepeatBufferCount(count);
                }
            }
            return msg;
        }

        private void InitSignalViewModels()
        {
            SignalViewModels = new List<SignalViewModel>();

            var signalVmSettings = new SignalViewModelSettings
            {
                SignalName = "Main signal",
                IsEnabled = true
            };

            var signalSettings = new SignalSettings
            {
                PhaseShift = 0.0,
                SignalBeginFrequency = 10.0,
                SignalEndFrequency = 50.0,
                SignalBeginPhysicalMin = Math.Round(_aoModule.SelectedAoRange.PhysicalMin / 2.0),
                SignalBeginPhysicalMax = 0.0,
                SignalEndPhysicalMin = 0.0,
                SignalEndPhysicalMax = Math.Round(_aoModule.SelectedAoRange.PhysicalMax / 2.0),
                RangePhysicalMin = _aoModule.SelectedAoRange.PhysicalMin,
                RangePhysicalMax = _aoModule.SelectedAoRange.PhysicalMax
            };

            var signalViewModel = new SignalViewModel(signalVmSettings, signalSettings);
            signalViewModel.PropertyChanged += AoModuleViewModel_PropertyChanged;
            SignalViewModels.Add(signalViewModel);

            signalVmSettings = new SignalViewModelSettings
            {
                SignalName = "Superimposed signal",
                IsEnabled = false
            };

            signalSettings = new SignalSettings
            {
                PhaseShift = 0.0,
                SignalBeginFrequency = 50.0,
                SignalEndFrequency = 100.0,
                SignalBeginPhysicalMin = Math.Round(_aoModule.SelectedAoRange.PhysicalMin / 5.0),
                SignalBeginPhysicalMax = 0.0,
                SignalEndPhysicalMin = 0.0,
                SignalEndPhysicalMax = Math.Round(_aoModule.SelectedAoRange.PhysicalMax / 5.0),
                RangePhysicalMin = _aoModule.SelectedAoRange.PhysicalMin,
                RangePhysicalMax = _aoModule.SelectedAoRange.PhysicalMax
            };

            signalViewModel = new SignalViewModel(signalVmSettings, signalSettings);
            signalViewModel.PropertyChanged += AoModuleViewModel_PropertyChanged;
            SignalViewModels.Add(signalViewModel);

            ArbitrarySignalViewModel = new ArbitrarySignalViewModel(_dialogService);
            ArbitrarySignalViewModel.PropertyChanged += AoModuleViewModel_PropertyChanged;
        }
        
        public List<SignalViewModel> SignalViewModels { get; private set; }
        public ArbitrarySignalViewModel ArbitrarySignalViewModel { get; private set; }

        void RaisePropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        void AoModuleViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(ActualOutputFrequency))
            {
                RaisePropertyChanged(e.PropertyName);
                if (_selectedTabItemIndex == _tabItemIndexArbitrarySignal)
                {
                    UpdateSignalSetttings();
                }
                else
                {
                    UpdateNumberOfValues();
                }
            }
            else if (e.PropertyName == nameof(RepeatBufferCount))
            {
                RaisePropertyChanged(e.PropertyName);
            }
            else if (e.PropertyName == nameof(SignalViewModel.IsGeneratorInputValid))
            {
                RaisePropertyChanged(nameof(IsInputValid));
                StartStopCommand.RaiseCanExecuteChanged();
            }
            else if (e.PropertyName == nameof(ArbitrarySignalViewModel.NumValuesReadFromFile))
            {
                UpdateSignalSetttings();
            }
            else if (e.PropertyName == nameof(AoModule.ModuleErrorText))
            {
                Messenger.Default.Send(new InfoMessage(_aoModule.ModuleErrorText), AdqCommon.MessengerContextInfo);
            }
            else if (e.PropertyName == nameof(AoModule.ModuleInfoText))
            {
                Messenger.Default.Send(new InfoMessage(_aoModule.ModuleInfoText), AdqCommon.MessengerContextInfo);
            }
        }

        public void Dispose()
        {
            _aoModule.PropertyChanged -= AoModuleViewModel_PropertyChanged;
        }
    }
}