﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the AIReadBuffer project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.Runtime.InteropServices
Imports AIReadBuffer.ALLDAQ.Boards

Public Class AnalogInputModule
    Private _channelList As AdqDriver.ADQAIChannel()

    Public Property AdqError As UInteger
    Public Event InfoTextChanged As EventHandler(Of ModuleInfoTextChangedEventArgs)

    Public Sub New(boardName As String, boardIndex As UInteger, moduleIndex As UInteger,
                   ByRef generalInfo As AdqDriver.ADQGeneralModuleInfo, ByRef aiModuleInfo As AdqDriver.ADQAIModuleInfo)
        ModelName = AdqCommon.GetModelName(boardName)
        Me.BoardIndex = boardIndex
        Me.ModuleIndex = moduleIndex
        NumberOfChannels = aiModuleInfo.NumberOfChannels
        NumberOfRanges = aiModuleInfo.NumberOfRanges
        MaxFrequencyHz = generalInfo.ClockBaseFrequencyHz / generalInfo.ClockMinimumDivider
        DesiredSampleFrequency = MaxFrequencyHz
        BuildModuleName()
        NumberOfFrames = AdqCommon.DefaultNumberOfFrames
        AddRanges(aiModuleInfo)
        DefaultRange = Ranges(0)
    End Sub

    Private Sub AddRanges(ByRef adqAiModuleInfo As AdqDriver.ADQAIModuleInfo)
        Dim rangeIdx As UInteger = 0
        While rangeIdx < adqAiModuleInfo.NumberOfRanges
            Dim range As AdqDriver.ADQAIRange = adqAiModuleInfo.Ranges(CInt(rangeIdx))

            Dim physRange = New AnalogRange(
                rangeIdx,
                range.PhysicalUnit,
                range.DigitalMin,
                range.DigitalMax,
                range.PhysicalMin,
                range.PhysicalMax)
            Ranges.Add(physRange)
            rangeIdx += 1UI
        End While
    End Sub

    Public ReadOnly Property ModelName As String
    Public ReadOnly Property BoardIndex As UInteger
    Public ReadOnly Property ModuleIndex As UInteger

    Private _moduleName As String
    Public ReadOnly Property ModuleName As String
        Get
            Return _moduleName
        End Get
    End Property

    Public ReadOnly Property NumberOfChannels As UInteger
    Public ReadOnly Property NumberOfRanges As UInteger
    Public ReadOnly Property ChannelSettings As List(Of ChannelSetting) = New List(Of ChannelSetting)
    Public ReadOnly Property MaxFrequencyHz As Double

    Private Sub BuildModuleName()
        _moduleName = String.Format("{0}, Board {1}, Module {2}, {3} {4}, {5} {6}, max. {7}",
            ModelName,
            BoardIndex,
            ModuleIndex,
            NumberOfChannels,
            IIf(NumberOfChannels = 1, "Channel", "Channels"),
            NumberOfRanges,
            IIf(NumberOfRanges = 1, "Range", "Ranges"),
            AdqCommon.FormatFrequency(MaxFrequencyHz))
    End Sub

    Public Sub AddChannelSetting(channelIndex As UInteger, range As AnalogRange)
        ChannelSettings.Add(New ChannelSetting(channelIndex, range))
    End Sub

    Public Sub ConfigureModule()
        If NumberOfFrames = 0 Then
            Throw New ApplicationException("Please enter a valid number of frames.")
        End If

        If FrameLength = 0 Then
            Throw New ApplicationException("Please select channels in the channel list.")
        End If

        If DesiredSampleFrequency <= 0 Then
            Throw New ApplicationException("Please enter a valid sample frequency.")
        End If

        SetChannelListAndFrequency()
        _sampleTime = 1.0 / _actualSampleFrequency
        AdqError = AdqErrors.ADQ_ERROR_NONE
        Dim numberOfTotalValues As UInteger = NumberOfFrames * FrameLength
        If _frames Is Nothing OrElse _frames.Length <> numberOfTotalValues Then
            _frames = New UInteger(CInt(numberOfTotalValues - 1)) {}
            _physicalFrames = New Double(CInt(numberOfTotalValues - 1)) {}
        End If

        _numberOfFramesRead = 0
    End Sub

    Public Sub UpdateChannelList()
        If _channelList Is Nothing Then
            'In Visual Basic, the size of an array is declared with the upper bound of the array
            _channelList = New AdqDriver.ADQAIChannel(CInt(NumberOfChannels - 1)) {}
        End If

        For i = 0 To _channelList.Length - 1
            _channelList(i).acquireChannel = False
            _channelList(i).rangeIndex = 0
        Next

        _minResolution = Double.MaxValue
        For Each chSetting In ChannelSettings
            _channelList(CInt(chSetting.ChannelIndex)).acquireChannel = True
            _channelList(CInt(chSetting.ChannelIndex)).rangeIndex = chSetting.Range.RangeIndex
            If _minResolution > chSetting.Range.PhysicalResolution Then
                _minResolution = chSetting.Range.PhysicalResolution
            End If
        Next
        FrameLength = CUInt(ChannelSettings.Count)
    End Sub

    Private Sub SetChannelListAndFrequency()
        Dim adqError As UInteger = AdqDriver.adqAISetChannelListAndFrequency(
            BoardIndex, ModuleIndex, _channelList, NumberOfChannels, DesiredSampleFrequency, _actualSampleFrequency)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqAISetChannelListAndFrequency", adqError)
        End If
    End Sub

    Public Sub ReadFramesFromModule()
        _totalNumberOfValuesRead = 0
        Dim gch As GCHandle = GCHandle.Alloc(_frames, GCHandleType.Pinned)
        AdqError = AdqDriver.adqAIReadBuffer(BoardIndex,
                                             ModuleIndex,
                                             FrameLength,
                                             NumberOfFrames,
                                             _frames,
                                             _numberOfFramesRead)
        _totalNumberOfValuesRead = FrameLength * _numberOfFramesRead
        gch.Free()
        CalculatePhysicalFrames()
    End Sub

    Private Sub CalculatePhysicalFrames()
        Dim chanSetting As ChannelSetting
        Dim Range As AnalogRange
        Dim idx As Integer = 0
        While idx < FrameLength
            chanSetting = ChannelSettings(idx)
            Range = chanSetting.Range
            Dim valIdx As Integer = idx
            While valIdx < _totalNumberOfValuesRead
                _physicalFrames(valIdx) = Range.DigitalToPhysical(_frames(valIdx))
                valIdx += CInt(FrameLength)
            End While
            idx += 1
        End While
    End Sub

    Public Function StopAcquisition() As UInteger
        Dim adqError As UInteger = AdqDriver.adqStop(BoardIndex, ModuleIndex)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqStop", adqError)
        End If
        Return adqError
    End Function

    Public Function Reset() As UInteger
        Dim adqError As UInteger = AdqDriver.adqResetModule(BoardIndex, ModuleIndex)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqResetModule", adqError)
        End If
        Return adqError
    End Function

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        InfoText = String.Format("Function {0} returned Error number: 0x{1:X}", functionName, adqError)
    End Sub

    Protected Sub OnModuleInfoTextChanged(e As ModuleInfoTextChangedEventArgs)
        RaiseEvent InfoTextChanged(Me, e)
    End Sub

    Private _infoText As String = String.Empty
    Public Property InfoText As String
        Get
            Return _infoText
        End Get
        Set
            If _infoText <> Value Then
                _infoText = Value
                Dim args As ModuleInfoTextChangedEventArgs = New ModuleInfoTextChangedEventArgs With {
                    .InfoText = _infoText
                }
                OnModuleInfoTextChanged(args)
            End If
        End Set
    End Property

    Public Property NumberOfFrames As UInteger

    Public Property FrameLength As UInteger

    Private _totalNumberOfValuesRead As UInteger
    ReadOnly Property TotalNumberOfValuesRead As UInteger
        Get
            Return _totalNumberOfValuesRead
        End Get
    End Property

    Private _numberOfFramesRead As UInteger
    ReadOnly Property NumberOfFramesRead As UInteger
        Get
            Return _numberOfFramesRead
        End Get
    End Property

    Private _frames As UInteger()
    ReadOnly Property Frames As UInteger()
        Get
            Return _frames
        End Get
    End Property

    Private _physicalFrames As Double()
    ReadOnly Property PhysicalFrames As Double()
        Get
            Return _physicalFrames
        End Get
    End Property

    Private _minResolution As Double
    ReadOnly Property MinResolution As Double
        Get
            Return _minResolution
        End Get
    End Property

    Private _actualSampleFrequency As Double
    ReadOnly Property ActualSampleFrequency As Double
        Get
            Return _actualSampleFrequency
        End Get
    End Property

    Private _sampleTime As Double
    ReadOnly Property SampleTime As Double
        Get
            Return _sampleTime
        End Get
    End Property

    Public Property DesiredSampleFrequency As Double

    Public Sub ClearFrameBuffers()
        _frames = Nothing
        _physicalFrames = Nothing
    End Sub

    Public Function GetSecondsToWait() As Double
        If _actualSampleFrequency <= 0 Then
            Return 0.0
        End If

        Return NumberOfFrames / _actualSampleFrequency
    End Function

    ReadOnly Property DefaultRange As AnalogRange

    ReadOnly Property Ranges As List(Of AnalogRange) = New List(Of AnalogRange)
End Class
