﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the AIReadBuffer project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports AIReadBuffer.ALLDAQ.Boards
Imports System.Text
Imports System.Windows.Forms.DataVisualization.Charting
Imports System.ComponentModel
Imports System.IO

Public Class FormMain
    Private ReadOnly _gridLineColor As Color = Color.DarkSlateGray
    Private ReadOnly _labelColor As Color = Color.LightSkyBlue
    Private Const _rangeSubItemIndex As Integer = 1
    Private Const _idxChartTypeLine As Integer = 0
    Private Const _idxChartTypeLineAndMarkers As Integer = 1
    Private Const _idxXAxisLabelsTime As Integer = 0
    Private Const _fmtStringPhysValues As String = "{0:0.000000}"

    Private _selectedUnitSymbol As String
    Private _moduleLocator As ModuleLocator
    Private ReadOnly _aiModules As List(Of AnalogInputModule) = New List(Of AnalogInputModule)
    Private _selectedAiModule As AnalogInputModule
    Private _selectedModuleItemIndex As Integer = -1
    Private _isFormClosing As Boolean
    Private _moduleIsBusy As Boolean
    Private _hitinfoChannelList As ListViewHitTestInfo
    Private WithEvents CoBoxRanges As ComboBox = New ComboBox()

    Private Sub FormMain_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        AdqCommon.SetDoubleBuffered(lvChannelList)
        InitRangeComboBox()
        coBoxChartType.SelectedIndex = _idxChartTypeLine
        coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime
        EnableUI(False)
        InitChart()
        Try
            _moduleLocator = New ModuleLocator(_aiModules, AddressOf OnModuleInfoTextChanged)
            _moduleLocator.FindAdqModules()

            FillModuleCombobox()
            SelectFirstComboboxItem(coBoxModules)
            textBoxInfo.AppendText(String.Format("Number of analog input modules found: {0}" & Environment.NewLine, _aiModules.Count))
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The device driver is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub InitRangeComboBox()
        CoBoxRanges.DropDownStyle = ComboBoxStyle.DropDownList
        CoBoxRanges.Parent = lvChannelList
        CoBoxRanges.Visible = False
        AddHandler CoBoxRanges.DropDownClosed, AddressOf CoBoxRanges_DropDownClosed
        AddHandler CoBoxRanges.Leave, AddressOf CoBoxRanges_Leave
        AddHandler CoBoxRanges.KeyPress, AddressOf CoBoxRanges_KeyPress
    End Sub

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()
        For Each aiModule As AnalogInputModule In _aiModules
            coBoxModules.Items.Add(aiModule.ModuleName)
        Next
    End Sub

    Private Sub OnModuleInfoTextChanged(sender As Object, e As ModuleInfoTextChangedEventArgs)
        textBoxInfo.AppendText(e.InfoText & Environment.NewLine)
    End Sub

    Private Sub DisplayErrorText(functionName As String, adqError As UInteger)
        textBoxInfo.AppendText(
            String.Format("Function {0} returned Error number: &H{1:X}" & Environment.NewLine,
                functionName,
                adqError))
    End Sub

    Private Sub SelectFirstComboboxItem(coBox As ComboBox)
        If coBox.Items.Count > 0 Then
            coBox.SelectedIndex = 0
        End If
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(sender As Object, e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If coBoxModules.SelectedIndex = _selectedModuleItemIndex Then Return

        _selectedModuleItemIndex = coBoxModules.SelectedIndex
        ClearDisplayedValues()

        EnableUI(coBoxModules.SelectedIndex >= 0)
        btnStartStop.Enabled = coBoxModules.SelectedIndex >= 0

        If _selectedAiModule IsNot Nothing Then
            RemoveHandler _selectedAiModule.InfoTextChanged, AddressOf OnModuleInfoTextChanged
            _selectedAiModule.ClearFrameBuffers()
        End If

        If coBoxModules.SelectedIndex < 0 Then
            _selectedAiModule = Nothing
            textBoxSampleFrequency.Text = String.Empty
            _selectedUnitSymbol = String.Empty
        Else
            _selectedAiModule = _aiModules(coBoxModules.SelectedIndex)
            AddHandler _selectedAiModule.InfoTextChanged, AddressOf OnModuleInfoTextChanged
            coBoxNumOfFrames.Text = _selectedAiModule.NumberOfFrames.ToString()
            textBoxSampleFrequency.Text = String.Format("{0}", _selectedAiModule.DesiredSampleFrequency)
            _selectedUnitSymbol = _selectedAiModule.Ranges(0).UnitSymbol
        End If

        textBoxActualFrequency.Text = String.Empty
        FillChannelsAndRanges(_selectedAiModule)
        UpdateNumberOfValues()
    End Sub

    Private Sub FillChannelsAndRanges(aiModule As AnalogInputModule)
        FillChannels(aiModule)
        FillRanges(aiModule)
        UpdateRangesToolstripMenuItems(aiModule)
    End Sub

    Private Sub FillChannels(aiModule As AnalogInputModule)
        lvChannelList.Items.Clear()
        If aiModule Is Nothing Then Return

        Dim chIdx As Integer = 0
        While chIdx < aiModule.NumberOfChannels
            Dim lvItem As ListViewItem = lvChannelList.Items.Add("Channel " + chIdx.ToString())
            lvItem.SubItems.Add(aiModule.DefaultRange.RangeText)
            lvItem.Tag = aiModule.DefaultRange
            lvItem.BackColor = Color.Black
            lvItem.ForeColor = AdqCommon.ChannelColors(chIdx Mod AdqCommon.ChannelColors.Length)
            chIdx += 1
        End While
        CheckFirstChannels(2)
    End Sub

    Private Sub CheckFirstChannels(count As Integer)
        Dim idx As Integer = 0
        While idx < lvChannelList.Items.Count
            lvChannelList.Items(idx).Checked = idx < count
            idx += 1
        End While
    End Sub

    Private Sub FillRanges(aiModule As AnalogInputModule)
        CoBoxRanges.Items.Clear()
        If aiModule Is Nothing Then Return

        For Each range In aiModule.Ranges
            CoBoxRanges.Items.Add(New RangeComboBoxItem(range))
        Next
        SelectFirstComboboxItem(CoBoxRanges)
    End Sub

    Private Sub UpdateRangesToolstripMenuItems(aiModule As AnalogInputModule)
        tsMnuItemSetRange.DropDownItems.Clear()

        If aiModule IsNot Nothing AndAlso aiModule.NumberOfRanges > 1 Then
            For Each range In aiModule.Ranges
                Dim item = New ToolStripMenuItem(range.RangeText, Nothing, AddressOf TsMnuItemSetRange_Click) With {
                    .Tag = range
                }
                tsMnuItemSetRange.DropDownItems.Add(item)
            Next
        Else
            Dim item = New ToolStripMenuItem("No selection available") With {
                .Enabled = False
            }
            tsMnuItemSetRange.DropDownItems.Add(item)
        End If
    End Sub

    Private Sub TsMnuItemSetRange_Click(sender As Object, e As EventArgs)
        If (TypeOf sender Is ToolStripMenuItem) Then
            Dim range As AnalogRange = CType(CType(sender, ToolStripMenuItem).Tag, AnalogRange)
            SetChannelsToRange(range)
        End If
    End Sub

    Private Sub SetChannelsToRange(range As AnalogRange)
        For Each lvItem As ListViewItem In lvChannelList.Items
            lvItem.Tag = range
            lvItem.SubItems(_rangeSubItemIndex).Text = range.RangeText
        Next
    End Sub

    Private Sub LvChannelList_MouseDown(sender As Object, e As MouseEventArgs) Handles lvChannelList.MouseDown
        _hitinfoChannelList = lvChannelList.HitTest(e.X, e.Y)
    End Sub

    Private Sub LvChannelList_MouseUp(sender As Object, e As MouseEventArgs) Handles lvChannelList.MouseUp
        If e.Button <> MouseButtons.Left OrElse _moduleIsBusy OrElse _selectedAiModule Is Nothing OrElse _selectedAiModule.Ranges.Count <= 1 OrElse
            _hitinfoChannelList Is Nothing OrElse _hitinfoChannelList.SubItem Is Nothing Then Return

        Dim hitInfo As ListViewHitTestInfo = lvChannelList.HitTest(e.X, e.Y)
        If _hitinfoChannelList.Item Is hitInfo.Item AndAlso _hitinfoChannelList.Item.SubItems(_rangeSubItemIndex) Is hitInfo.SubItem Then
            Dim range As AnalogRange = CType(_hitinfoChannelList.Item.Tag, AnalogRange)
            CoBoxRanges.Bounds = hitInfo.SubItem.Bounds
            CoBoxRanges.SelectedIndex = CInt(range.RangeIndex)
            CoBoxRanges.Visible = True
            CoBoxRanges.BringToFront()
            CoBoxRanges.Focus()
        End If
    End Sub

    Private Sub CoBoxRanges_KeyPress(sender As Object, e As KeyPressEventArgs) Handles CoBoxRanges.KeyPress
        If e.KeyChar = ChrW(Keys.Enter) Then
            UpdateListViewItemChannelList()
        ElseIf e.KeyChar = ChrW(Keys.Escape) Then
            Dim range As AnalogRange = CType(_hitinfoChannelList.Item.Tag, AnalogRange)
            CoBoxRanges.SelectedIndex = CInt(range.RangeIndex)
            CoBoxRanges.Visible = False
        End If
    End Sub

    Private Sub CoBoxRanges_DropDownClosed(sender As Object, e As EventArgs) Handles CoBoxRanges.DropDownClosed
        UpdateListViewItemChannelList()
    End Sub

    Private Sub CoBoxRanges_Leave(sender As Object, e As EventArgs) Handles CoBoxRanges.Leave
        UpdateListViewItemChannelList()
    End Sub

    Private Sub UpdateListViewItemChannelList()
        If _hitinfoChannelList Is Nothing OrElse _hitinfoChannelList.SubItem Is Nothing OrElse CoBoxRanges.SelectedItem Is Nothing Then Return

        Dim rangeItem As RangeComboBoxItem = CType(CoBoxRanges.SelectedItem, RangeComboBoxItem)
        _hitinfoChannelList.Item.Tag = rangeItem.Range
        _hitinfoChannelList.Item.SubItems(_rangeSubItemIndex).Text = rangeItem.ToString()

        CoBoxRanges.Visible = False
        lvChannelList.Focus()
    End Sub

    Private Sub ContextMnuStripChanList_Opening(sender As Object, e As CancelEventArgs) Handles contxtMnuStripChanList.Opening
        If CoBoxRanges.Visible Then
            CoBoxRanges.Visible = False
        End If
    End Sub

    Private Sub EnableUI(enabled As Boolean)
        coBoxModules.Enabled = enabled
        textBoxSampleFrequency.Enabled = enabled
        coBoxNumOfFrames.Enabled = enabled
        coBoxChartType.Enabled = enabled
        coBoxXAxisLabels.Enabled = enabled
        btnSetConfig.Enabled = enabled
        btnReset.Enabled = enabled
        chBoxDisplayTextValues.Enabled = enabled
        lvChannelList.Enabled = enabled
        tsMnuItemSelectAll.Enabled = enabled
        tsMnuItemSelectNone.Enabled = enabled
        tsMnuItemInvertSelection.Enabled = enabled
        tsMenuItemSaveGraphic.Enabled = enabled
    End Sub

    Private Sub CoBoxNumOfFrames_TextChanged(sender As Object, e As EventArgs) Handles coBoxNumOfFrames.TextChanged
        If _selectedAiModule Is Nothing Then Return

        _selectedAiModule.NumberOfFrames = ParseNumberOfFrames(coBoxNumOfFrames.Text)
        UpdateNumberOfValues()
    End Sub

    Private Sub UpdateNumberOfValues()
        If _selectedAiModule Is Nothing Then
            textBoxNumberOfValues.Text = String.Empty
            Return
        End If

        textBoxNumberOfValues.Text = (_selectedAiModule.NumberOfFrames * _selectedAiModule.FrameLength).ToString()
    End Sub

    Private Sub TextBoxSampleFrequency_TextChanged(sender As Object, e As EventArgs) Handles textBoxSampleFrequency.TextChanged
        If _selectedAiModule Is Nothing Then Return

        _selectedAiModule.DesiredSampleFrequency = ParseFrequency(textBoxSampleFrequency.Text)
    End Sub

    Private Function ParseFrequency(sampleFrequency As String) As Double
        Dim actualFrequency As Double
        Dim isValid As Boolean = Double.TryParse(sampleFrequency, actualFrequency)
        If Not isValid OrElse actualFrequency <= 0 Then
            Return 0.0
        End If

        Return actualFrequency
    End Function

    Private Function ParseNumberOfFrames(numberOfFrames As String) As UInteger
        Dim numValues As UInteger
        Dim isValid As Boolean = UInteger.TryParse(numberOfFrames, numValues)
        If Not isValid Then
            Return 0
        End If

        Return numValues
    End Function

    Private Sub LvChannelList_ItemChecked(sender As Object, e As ItemCheckedEventArgs) Handles lvChannelList.ItemChecked
        _selectedAiModule.FrameLength = CUInt(lvChannelList.CheckedItems.Count)
        textBoxNumberOfSelectedChannels.Text = _selectedAiModule.FrameLength.ToString()
        UpdateNumberOfValues()
    End Sub

    Private Sub BtnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        Dim adqError = _selectedAiModule.Reset()
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            textBoxInfo.AppendText(
                String.Format("Reset Board {0}, Module {1}" & Environment.NewLine,
                              _selectedAiModule.BoardIndex, _selectedAiModule.ModuleIndex))
        End If
    End Sub

    Private Sub BtnSetConfig_Click(sender As Object, e As EventArgs) Handles btnSetConfig.Click
        Try
            SetAiModuleConfiguration()
        Catch ex As Exception
            textBoxInfo.AppendText(ex.Message + Environment.NewLine)
        End Try
    End Sub

    Private Sub BtnStartStop_Click(sender As Object, e As EventArgs) Handles btnStartStop.Click
        If _selectedAiModule Is Nothing Then Return

        If _moduleIsBusy Then
            _selectedAiModule.StopAcquisition()
            btnStartStop.Text = "Start"
            Return
        End If

        ClearDisplayedValues()
        EnableUI(False)
        btnStartStop.Text = "Stop"
        Try
            SetAiModuleConfiguration()
            _moduleIsBusy = True
            AddTextToCombobox(coBoxNumOfFrames, coBoxNumOfFrames.Text)
            BackgroundWorker1.RunWorkerAsync(_selectedAiModule)
        Catch ex As Exception
            _moduleIsBusy = False
            EnableUI(True)
            btnStartStop.Text = "Start"
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub SetAiModuleConfiguration()
        UpdateChannelSettings()
        _selectedAiModule.ConfigureModule()
        UpdateUI()
    End Sub

    Private Sub UpdateChannelSettings()
        _selectedAiModule.ChannelSettings.Clear()
        For Each lvItem As ListViewItem In lvChannelList.CheckedItems
            Dim range As AnalogRange = CType(lvItem.Tag, AnalogRange)
            _selectedAiModule.AddChannelSetting(CUInt(lvItem.Index), range)
        Next
        _selectedAiModule.UpdateChannelList()
    End Sub

    Private Sub UpdateUI()
        textBoxActualFrequency.Text = AdqCommon.FormatFrequency(_selectedAiModule.ActualSampleFrequency)

        Dim seconds As Double = _selectedAiModule.GetSecondsToWait()
        lblDuration.Text = AdqCommon.FormatTimeToWait(seconds)
    End Sub

    Private Sub DisplayValues(aiModule As AnalogInputModule)
        DisplayAnalogValuesAsText(aiModule)
        DisplayAnalogValuesAsChart(aiModule)
    End Sub

    Private Sub DisplayAnalogValuesAsText(aiModule As AnalogInputModule)
        If Not chBoxDisplayTextValues.Checked Then Return

        UseWaitCursor = True
        Try
            Dim sb = New StringBuilder()
            sb.AppendLine(String.Format("Frames read: {0}", aiModule.NumberOfFramesRead))
            Dim frameLength As UInteger = aiModule.FrameLength
            Dim currentFrameOffset As UInteger = 0
            Dim time As Double = 0
            Dim frameIdx As UInteger = 0
            While frameIdx < aiModule.NumberOfFramesRead
                sb.Append(String.Format("{0:000}; {1:0.0000000sec}", frameIdx, time))
                Dim idx As Integer = 0
                While idx < frameLength
                    Dim chanSetting As ChannelSetting = aiModule.ChannelSettings(idx)

                    Dim digitalValue As UInteger = aiModule.Frames(CInt(currentFrameOffset + idx))
                    Dim physicalValue As Double = aiModule.PhysicalFrames(CInt(currentFrameOffset + idx))
                    sb.Append(String.Format("; Ch{0}: &H{1:X8}, {1} -> {2:+0.00000000;-0.00000000;0}{3}",
                        chanSetting.ChannelIndex,
                        digitalValue,
                        physicalValue,
                        chanSetting.Range.UnitSymbol))
                    idx += 1
                End While
                sb.AppendLine()
                time += aiModule.SampleTime
                currentFrameOffset += frameLength
                frameIdx += 1UI
            End While
            textBoxInfo.AppendText(sb.ToString())
        Finally
            UseWaitCursor = False
        End Try
    End Sub

    Private Sub DisplayAnalogValuesAsChart(aiModule As AnalogInputModule)
        Dim frameLength As Integer = CInt(aiModule.FrameLength)

        Chart1.SuspendLayout()
        UseWaitCursor = True
        Try
            UpdateChart(aiModule)
            InitSeriesSubMenus()
            Dim xValueDelta As Double = GetXValueDelta(aiModule)
            Dim points0 As DataPointCollection = Nothing
            Dim frameIdx As Integer = 0
            While frameIdx < frameLength
                Dim chSeries As Series = Chart1.Series(frameIdx)
                Dim valIdx As Integer = frameIdx
                Dim xValue As Double = 0.0
                Dim pointIdx As Integer = 0
                Dim tsLabel As String
                While (valIdx < aiModule.TotalNumberOfValuesRead)
                    Dim physicalValue As Double = aiModule.PhysicalFrames(valIdx)
                    Dim dp = New DataPoint(pointIdx, physicalValue)
                    If (points0 Is Nothing) Then
                        tsLabel = GetXAsixLabel(xValue)
                        xValue += xValueDelta
                    Else
                        tsLabel = points0(pointIdx).AxisLabel
                    End If
                    dp.AxisLabel = tsLabel
                    chSeries.Points.Add(dp)
                    valIdx += frameLength
                    pointIdx += 1
                End While
                If (frameIdx = 0) Then
                    points0 = chSeries.Points
                End If
                frameIdx += 1
            End While
            Chart1.ChartAreas(0).RecalculateAxesScale()
        Finally
            UseWaitCursor = False
            Chart1.ResumeLayout()
        End Try
    End Sub

    Private Function GetXAsixLabel(xValue As Double) As String
        If coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime Then
            Return AdqCommon.FormatSampleTime(xValue)
        End If
        Return CInt(xValue).ToString()
    End Function

    Private Function GetXValueDelta(aiModule As AnalogInputModule) As Double
        Return If(coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime, aiModule.SampleTime, 1.0)
    End Function

    Private Sub InitSeriesSubMenus()
        tsMnuItemShowSeries.DropDownItems.Clear()

        If Chart1.Series.Count = 0 Then
            Dim item = New ToolStripMenuItem("No series present") With {
                .Enabled = False
            }
            tsMnuItemShowSeries.DropDownItems.Add(item)
            Return
        End If

        For Each chSeries As Series In Chart1.Series

            Dim item = New ToolStripMenuItem(chSeries.LegendText, Nothing, AddressOf TsMnuItemShowSeries_Click) With {
                .Checked = True,
                .CheckOnClick = True,
                .Tag = chSeries
            }
            tsMnuItemShowSeries.DropDownItems.Add(item)
        Next

        tsMnuItemShowSeries.DropDownItems.Add(New ToolStripSeparator())
        Dim itemShowAll = New ToolStripMenuItem("Show all Series", Nothing, AddressOf TsMnuItemShowAllSeries_Click)
        tsMnuItemShowSeries.DropDownItems.Add(itemShowAll)
        Dim itemHideAll = New ToolStripMenuItem("Hide all Series", Nothing, AddressOf TsMnuItemHideAllSeries_Click)
        tsMnuItemShowSeries.DropDownItems.Add(itemHideAll)
    End Sub

    Private Sub TsMnuItemShowSeries_Click(sender As Object, e As EventArgs)
        If (TypeOf sender Is ToolStripMenuItem) Then
            Dim chSeries As Series = CType(CType(sender, ToolStripMenuItem).Tag, Series)
            chSeries.Enabled = CType(sender, ToolStripMenuItem).Checked
        End If
    End Sub

    Private Sub TsMnuItemShowAllSeries_Click(sender As Object, e As EventArgs)
        ShowAllSeries(True)
    End Sub

    Private Sub TsMnuItemHideAllSeries_Click(sender As Object, e As EventArgs)
        ShowAllSeries(False)
    End Sub

    Private Sub ShowAllSeries(showSeries As Boolean)
        For Each item As ToolStripItem In tsMnuItemShowSeries.DropDownItems
            If TypeOf item Is ToolStripMenuItem AndAlso TypeOf item.Tag Is Series Then
                Dim chSeries As Series = CType(item.Tag, Series)
                CType(item, ToolStripMenuItem).Checked = showSeries
                chSeries.Enabled = showSeries
            End If
        Next
    End Sub

    Private Sub InitChart()
        Chart1.BackColor = Color.Black
        Dim ca As ChartArea = Chart1.ChartAreas(0)
        ca.BackColor = Color.Black

        ca.AxisX.MajorGrid.LineColor = _gridLineColor
        ca.AxisX.MajorTickMark.LineColor = _gridLineColor
        ca.AxisX.MinorGrid.LineColor = _gridLineColor
        ca.AxisX.MinorTickMark.LineColor = _gridLineColor
        ca.AxisX.LineColor = _gridLineColor
        ca.AxisX.LineWidth = 1
        ca.AxisX.LabelStyle.ForeColor = _labelColor
        ca.AxisX.TitleForeColor = _labelColor
        ca.AxisX.ScrollBar.ButtonColor = Color.DarkGray
        ca.AxisX.ScrollBar.BackColor = Color.Gray
        ca.CursorX.IsUserSelectionEnabled = True

        ca.AxisY.MajorGrid.LineColor = _gridLineColor
        ca.AxisY.MajorTickMark.LineColor = _gridLineColor
        ca.AxisY.MinorGrid.LineColor = _gridLineColor
        ca.AxisY.MinorTickMark.LineColor = _gridLineColor
        ca.AxisY.LineColor = _gridLineColor
        ca.AxisY.LineWidth = 1
        ca.AxisY.LabelStyle.ForeColor = _labelColor
        ca.AxisY.LabelStyle.Format = _fmtStringPhysValues
        ca.AxisY.TitleForeColor = _labelColor
        ca.AxisY.ScrollBar.ButtonColor = Color.DarkGray
        ca.AxisY.ScrollBar.BackColor = Color.Gray
        ca.CursorY.IsUserSelectionEnabled = True
    End Sub

    Private Sub UpdateChart(aiModule As AnalogInputModule)
        Dim ca As ChartArea = Chart1.ChartAreas(0)
        ca.AxisX.Title = coBoxXAxisLabels.Text
        ca.AxisY.Title = aiModule.ChannelSettings(0).Range.UnitName

        Dim scrollSize As Double = aiModule.MinResolution / 4.0
        ca.AxisY.ScaleView.SmallScrollMinSize = scrollSize
        ca.AxisY.ScaleView.SmallScrollSize = scrollSize
        ca.CursorY.Interval = aiModule.MinResolution
        ca.AxisY.Maximum = Double.NaN
        ca.AxisY.Minimum = Double.NaN

        For Each chanSetting In aiModule.ChannelSettings
            InitSeries(chanSetting.ChannelIndex, Chart1, coBoxChartType.SelectedIndex)
        Next
    End Sub

    Private Sub InitSeries(channelIndex As UInteger, chart As Chart, chartTypeIndex As Integer)
        RemoveHandler chart.GetToolTipText, AddressOf Chart1_GetToolTipText

        Dim chSeries As Series = New Series("Channel " & channelIndex.ToString())
        chSeries.LegendText = chSeries.Name
        If chartTypeIndex = _idxChartTypeLine Then
            chSeries.ChartType = SeriesChartType.Line
        ElseIf chartTypeIndex = _idxChartTypeLineAndMarkers Then
            chSeries.ChartType = SeriesChartType.Line
            chSeries.MarkerStyle = AdqCommon.MarkerStyles(CInt(channelIndex Mod AdqCommon.MarkerStyles.Length))
        Else
            chSeries.ChartType = SeriesChartType.Point
        End If

        AddHandler chart.GetToolTipText, AddressOf Chart1_GetToolTipText
        chSeries.Color = AdqCommon.ChannelColors(CInt(channelIndex Mod AdqCommon.ChannelColors.Length))
        chSeries.XValueType = ChartValueType.UInt32
        chSeries.YValueType = ChartValueType.Double
        chSeries.IsXValueIndexed = False
        chart.Series.Add(chSeries)
    End Sub

    Private Sub Chart1_GetToolTipText(sender As Object, e As ToolTipEventArgs)
        If e.HitTestResult.ChartElementType = ChartElementType.DataPoint Then
            Dim series = e.HitTestResult.Series
            If series IsNot Nothing Then
                Dim dp As DataPoint = series.Points(e.HitTestResult.PointIndex)
                e.Text = FormatTooltipText(dp)
            End If
        ElseIf e.HitTestResult.ChartElementType = ChartElementType.Axis OrElse
            e.HitTestResult.ChartElementType = ChartElementType.TickMarks Then
            If (TypeOf sender Is Chart) Then
                Dim result As HitTestResult = CType(sender, Chart).HitTest(e.X, e.Y, ChartElementType.DataPoint)
                If result.ChartElementType = ChartElementType.DataPoint Then
                    Dim dp As DataPoint = result.Series.Points(result.PointIndex)
                    e.Text = FormatTooltipText(dp)
                End If
            End If
        End If
    End Sub

    Private Function FormatTooltipText(dp As DataPoint) As String
        Return String.Format("X={0}, Y={1}",
                        dp.XValue,
                        AdqCommon.FormatPhysicalValue(_fmtStringPhysValues, dp.YValues(0)) & _selectedUnitSymbol)
    End Function

    Private Sub BackgroundWorker1_DoWork(sender As Object, e As DoWorkEventArgs) Handles BackgroundWorker1.DoWork
        Dim aiModule As AnalogInputModule = CType(e.Argument, AnalogInputModule)
        aiModule.ReadFramesFromModule()
        e.Result = aiModule
    End Sub

    Private Sub BackgroundWorker1_RunWorkerCompleted(sender As Object, e As RunWorkerCompletedEventArgs) Handles BackgroundWorker1.RunWorkerCompleted
        If _isFormClosing Then Return

        Dim aiModule As AnalogInputModule = CType(e.Result, AnalogInputModule)
        _moduleIsBusy = False
        If aiModule.AdqError = AdqErrors.ADQ_ERROR_NONE Then
            DisplayValues(aiModule)
        Else
            DisplayErrorText("adqAIReadBuffer", aiModule.AdqError)
        End If
        btnStartStop.Text = "Start"
        EnableUI(True)
    End Sub

    Private Sub FormMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles MyBase.FormClosing
        If _moduleIsBusy Then
            Dim result As DialogResult = MessageBox.Show("The acquisition is still running." _
                                     & Environment.NewLine _
                                     & "Do you want to close the program?",
                                     "Module busy",
                                     MessageBoxButtons.YesNo,
                                     MessageBoxIcon.Question)
            If result = DialogResult.Yes Then
                _isFormClosing = True
                _selectedAiModule.StopAcquisition()
                e.Cancel = False
            Else
                e.Cancel = True
            End If
        End If
    End Sub

    Private Sub BtnClose_Click(sender As Object, e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub FormMain_FormClosed(sender As Object, e As FormClosedEventArgs) Handles MyBase.FormClosed
        CloseAdqDriver()
    End Sub

    Private Sub CloseAdqDriver()
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub BtnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        ClearDisplayedValues()
    End Sub

    Private Sub ClearDisplayedValues()
        Chart1.Series.Clear()
        InitSeriesSubMenus()
        textBoxInfo.Clear()
    End Sub

    Private Sub TextBoxInfo_KeyDown(sender As Object, e As KeyEventArgs) Handles textBoxInfo.KeyDown
        If e.Modifiers = Keys.Control AndAlso e.KeyCode = Keys.A Then
            textBoxInfo.SelectionStart = 0
            textBoxInfo.SelectionLength = textBoxInfo.Text.Length
        End If
    End Sub

    Private Sub TsMnuItemSelectAll_Click(sender As Object, e As EventArgs) Handles tsMnuItemSelectAll.Click
        CheckListViewItems(True)
    End Sub

    Private Sub TsMnuItemInvertSelection_Click(sender As Object, e As EventArgs) Handles tsMnuItemInvertSelection.Click
        For i As Integer = 0 To lvChannelList.Items.Count - 1
            lvChannelList.Items(i).Checked = Not lvChannelList.Items(i).Checked
        Next
    End Sub

    Private Sub TsMnuItemSelectNone_Click(sender As Object, e As EventArgs) Handles tsMnuItemSelectNone.Click
        CheckListViewItems(False)
    End Sub

    Private Sub CheckListViewItems(check As Boolean)
        For i As Integer = 0 To lvChannelList.Items.Count - 1
            lvChannelList.Items(i).Checked = check
        Next
    End Sub

    Private Function GetImageFileName(aiModule As AnalogInputModule, fileExt As String) As String
        Dim s As String = String.Empty
        If aiModule IsNot Nothing Then
            s = String.Format("{0}, Board {1}, Module {2}, ",
                    aiModule.ModelName,
                    aiModule.BoardIndex,
                    aiModule.ModuleIndex)
        End If
        Return s & Date.Now.ToString("yyyy-MM-dd HH.mm.ss") & fileExt
    End Function

    Private Sub TsMenuItemSaveGraphic_Click(sender As Object, e As EventArgs) Handles tsMenuItemSaveGraphic.Click
        SaveFileDialog1.FileName = GetImageFileName(_selectedAiModule, ".png")
        SaveFileDialog1.DefaultExt = "png"
        SaveFileDialog1.ShowDialog()
    End Sub

    Private Function GetImageFormat(filterIndex As Integer) As ChartImageFormat
        Dim imgFormat As ChartImageFormat

        Select Case filterIndex
            Case 1
                imgFormat = ChartImageFormat.Png
            Case 2
                imgFormat = ChartImageFormat.Gif
            Case 3
                imgFormat = ChartImageFormat.Bmp
            Case 4
                imgFormat = ChartImageFormat.Tiff
            Case 5
                imgFormat = ChartImageFormat.Jpeg
            Case Else
                imgFormat = ChartImageFormat.Png
        End Select
        Return imgFormat
    End Function

    Private Sub SaveFileDialog1_FileOk(sender As Object, e As CancelEventArgs) Handles SaveFileDialog1.FileOk
        Dim imgFormat As ChartImageFormat = GetImageFormat(SaveFileDialog1.FilterIndex)
        Chart1.SaveImage(SaveFileDialog1.FileName, imgFormat)
    End Sub

    Private Sub TsMnuItemCopyToClipboard_Click(sender As Object, e As EventArgs) Handles tsMnuItemCopyToClipboard.Click
        Using ms As MemoryStream = New MemoryStream()
            Chart1.SaveImage(ms, ChartImageFormat.Png)
            Dim Bitmap = New Bitmap(ms)
            Clipboard.SetImage(Bitmap)
        End Using
    End Sub

    Private Sub AddTextToCombobox(coBox As ComboBox, text As String)
        If Not coBox.Items.Contains(text) Then
            coBox.Items.Insert(0, text)
        End If
    End Sub

    Private Sub BtnInfo_Click(sender As Object, e As EventArgs) Handles btnInfo.Click
        AboutBox.ShowDialog()
    End Sub
End Class



