﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the AIReadContinuous project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.IO
Imports System.Windows.Forms.DataVisualization.Charting
Imports AIReadContinuous.ALLDAQ.Boards

Public Class FormMain
    Private ReadOnly _gridLineColor As Color = Color.DarkSlateGray
    Private ReadOnly _labelColor As Color = Color.LightSkyBlue
    Private Const _rangeSubItemIndex As Integer = 1
    Private Const _idxChartTypeLine As Integer = 0
    Private Const _idxChartTypeLineAndMarkers As Integer = 1
    Private Const _idxXAxisLabelsTime As Integer = 0
    Private Const _bit0Mask As Integer = 1
    Private Const _bit1Mask As Integer = 2
    Private Const _fmtStringPhysValues As String = "{0:0.000000}"
    Private Const _txtInfinite As String = "Infinite"
    Private Const _txtStartingAcquisition As String = "Starting acquisition..."
    Private Const _txtWaitingForStartTrigger As String = "Waiting for start trigger..."
    Private Const _txtStartTriggerEventOccured As String = "Start trigger event occurred."
    Private Const _txtAcquisitionStarted As String = "Acquisition started."
    Private Const _txtErrorOccurred As String = "Error occurred."
    Private Const _txtReady As String = "Ready"

    Private _selectedUnitSymbol As String
    Private _moduleLocator As ModuleLocator
    Private ReadOnly _aiModules As List(Of AnalogInputModule) = New List(Of AnalogInputModule)
    Private _selectedAiModule As AnalogInputModule
    Private _selectedModuleItemIndex As Integer = -1

    Private _hitinfoChannelList As ListViewHitTestInfo
    Private _isAppClosing As Boolean
    Private WithEvents CoBoxRanges As ComboBox = New ComboBox()
    Private _xAxisLabels As String()
    Private _imgInitialDirectory As String

    Private Delegate Sub DisplayChartValuesDelegate()
    Private _updateChartDelegate As DisplayChartValuesDelegate

    Private Delegate Sub UpdatePropertyDelegate(changedProperty As AdqCommon.PropertyType)
    Private _updatePropertyDelegate As UpdatePropertyDelegate


    Private Sub FormMain_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        AdqCommon.SetDoubleBuffered(lvChannelList)
        InitRangeComboBox()
        coBoxChartType.SelectedIndex = _idxChartTypeLine
        coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime
        DisableComboBoxMouseWheel()
        EnableUI(False)
        InitChart()
        _updateChartDelegate = New DisplayChartValuesDelegate(AddressOf UpdateChartGraphic)
        _updatePropertyDelegate = New UpdatePropertyDelegate(AddressOf UpdateProperty)
        Try
            _moduleLocator = New ModuleLocator(_aiModules, AddressOf OnModuleInfoTextChanged)
            _moduleLocator.FindAdqModules()

            FillModuleCombobox()
            SelectFirstComboboxItem(coBoxModules)
            textBoxInfo.AppendText(String.Format("Number of analog input modules found: {0}" & Environment.NewLine, _aiModules.Count))
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The device driver is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
            tabControlMain.SelectedTab = tpInfo
        End Try
    End Sub

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()
        For Each aiModule As AnalogInputModule In _aiModules
            coBoxModules.Items.Add(aiModule.ModuleName)
        Next
    End Sub

    Private Sub DisableComboBoxMouseWheel()
        AddHandler coBoxModules.MouseWheel, AddressOf ComboBox_MouseWheel
    End Sub

    Private Sub ComboBox_MouseWheel(sender As Object, e As MouseEventArgs)
        Dim hme As HandledMouseEventArgs = CType(e, HandledMouseEventArgs)
        hme.Handled = True
    End Sub

    Private Sub InitChart()
        Chart1.BackColor = Color.Black
        Dim ca As ChartArea = Chart1.ChartAreas(0)
        ca.BackColor = Color.Black
    End Sub

    Private Sub UpdateChart(aiModule As AnalogInputModule)
        RemoveHandler Chart1.GetToolTipText, AddressOf Chart_GetToolTipText
        Chart1.Series.Clear()
        Chart1.Legends.Clear()

        Chart1.ChartAreas.Clear()
        Dim ca = New ChartArea()
        Chart1.ChartAreas.Add(ca)
        InitChartTitles(ca, aiModule.ModuleName, aiModule.ChannelSettings(0).Range.UnitName)
        InitChartArea(ca, aiModule.MinResolution)
        InitLegend(Chart1, ca, chkBoxShowLegend.Checked)
        For Each chanSetting As ChannelSetting In aiModule.ChannelSettings
            Dim cvt As ChartValueType = GetChartXValueType(coBoxXAxisLabels.SelectedIndex)
            InitSeries(chanSetting.ChannelIndex, Chart1, coBoxChartType.SelectedIndex, cvt)
        Next
    End Sub

    Private Sub InitChartTitles(chArea As ChartArea, chartTitle As String, yAxisTitle As String)
        Chart1.Titles.SuspendUpdates()
        Chart1.Titles.Clear()

        Dim Title = New Title()
        Chart1.Titles.Add(Title)
        InitChartTitle(Title, chArea, chartTitle)

        Title = New Title()
        Chart1.Titles.Add(Title)
        InitYAxisTitle(Title, chArea, yAxisTitle)

        Title = New Title()
        Chart1.Titles.Add(Title)
        InitXAxisTitle(Title, chArea, coBoxXAxisLabels.Text)

        Chart1.Titles.ResumeUpdates()
    End Sub

    Private Sub InitChartTitle(chTitle As Title, chArea As ChartArea, titleText As String)
        chTitle.DockedToChartArea = chArea.Name
        chTitle.Text = titleText
        chTitle.Alignment = ContentAlignment.TopCenter
        chTitle.Docking = Docking.Top
        chTitle.IsDockedInsideChartArea = False
        chTitle.ForeColor = _labelColor
        chTitle.Position.Auto = False
        chTitle.Position.X = 3
        chTitle.Position.Y = 0
        chTitle.Position.Width = 94
        chTitle.Position.Height = 5
    End Sub

    Private Sub InitYAxisTitle(chTitle As Title, chArea As ChartArea, titleText As String)
        chTitle.DockedToChartArea = chArea.Name
        chTitle.Text = titleText
        chTitle.Position.Auto = False
        chTitle.Position.X = 0.5F
        chTitle.Position.Y = 0
        chTitle.Position.Width = 5
        chTitle.Position.Height = 95
        chTitle.Alignment = ContentAlignment.TopCenter
        chTitle.Docking = Docking.Left
        chTitle.IsDockedInsideChartArea = False
        chTitle.TextOrientation = TextOrientation.Rotated270
        chTitle.ForeColor = _labelColor
    End Sub

    Private Sub InitXAxisTitle(chTitle As Title, chArea As ChartArea, titleText As String)
        chTitle.DockedToChartArea = chArea.Name
        chTitle.Text = titleText
        chTitle.Position.Auto = False
        chTitle.Position.X = 3
        chTitle.Position.Y = 95
        chTitle.Position.Width = 94
        chTitle.Position.Height = 5
        chTitle.Alignment = ContentAlignment.BottomCenter
        chTitle.Docking = Docking.Bottom
        chTitle.IsDockedInsideChartArea = False
        chTitle.ForeColor = _labelColor
    End Sub

    Private Sub InitChartArea(chArea As ChartArea, resolutionYAxis As Double)
        chArea.BackColor = Color.Black

        chArea.AxisX.MajorGrid.LineColor = _gridLineColor
        chArea.AxisX.MajorTickMark.LineColor = _gridLineColor
        chArea.AxisX.MinorGrid.LineColor = _gridLineColor
        chArea.AxisX.MinorTickMark.LineColor = _gridLineColor
        chArea.AxisX.LineColor = _gridLineColor
        chArea.AxisX.LineWidth = 1
        chArea.AxisX.LabelStyle.ForeColor = _labelColor
        chArea.AxisX.ScrollBar.ButtonColor = Color.DarkGray
        chArea.AxisX.ScrollBar.BackColor = Color.Gray
        chArea.CursorX.IsUserSelectionEnabled = True

        chArea.AxisY.MajorGrid.LineColor = _gridLineColor
        chArea.AxisY.MajorTickMark.LineColor = _gridLineColor
        chArea.AxisY.MinorGrid.LineColor = _gridLineColor
        chArea.AxisY.MinorTickMark.LineColor = _gridLineColor
        chArea.AxisY.LineColor = _gridLineColor
        chArea.AxisY.LineWidth = 1
        chArea.AxisY.IsLabelAutoFit = False
        chArea.AxisY.LabelStyle.ForeColor = _labelColor
        chArea.AxisY.LabelStyle.Format = _fmtStringPhysValues
        chArea.AxisY.ScrollBar.ButtonColor = Color.DarkGray
        chArea.AxisY.ScrollBar.BackColor = Color.Gray

        Dim scrollSize As Double = resolutionYAxis / 4.0
        chArea.AxisY.ScaleView.SmallScrollMinSize = scrollSize
        chArea.AxisY.ScaleView.SmallScrollSize = scrollSize
        chArea.CursorY.Interval = scrollSize
        chArea.CursorY.IsUserSelectionEnabled = True
    End Sub

    Private Sub InitSeries(channelIndex As UInteger, ch As Chart, chartTypeIndex As Integer, chXValueType As ChartValueType)
        Dim chSeries = New Series With
            {
                .Name = "Channel " & channelIndex.ToString()
            }

        If chartTypeIndex = _idxChartTypeLine Then
            chSeries.ChartType = SeriesChartType.Line
        ElseIf chartTypeIndex = _idxChartTypeLineAndMarkers Then
            chSeries.ChartType = SeriesChartType.Line
            chSeries.MarkerStyle = AdqCommon.MarkerStyles(CInt(channelIndex Mod AdqCommon.MarkerStyles.Length))
        Else
            chSeries.ChartType = SeriesChartType.Point
        End If
        chSeries.Color = AdqCommon.ChannelColors(CInt(channelIndex Mod AdqCommon.ChannelColors.Length))
        chSeries.YValueType = ChartValueType.Double
        chSeries.XValueType = chXValueType
        chSeries.IsXValueIndexed = False
        ch.Series.Add(chSeries)
    End Sub

    Private Sub InitLegend(ch As Chart, chArea As ChartArea, isEnabled As Boolean)
        Dim chLegend = New Legend("Legend1") With
            {
                .ForeColor = Color.Gray,
                .BackColor = Color.Transparent,
                .DockedToChartArea = chArea.Name,
                .Enabled = isEnabled
            }
        ch.Legends.Add(chLegend)
    End Sub

    Private Function GetChartXValueType(selectedIndex As Integer) As ChartValueType
        Return If(selectedIndex = _idxXAxisLabelsTime, ChartValueType.String, ChartValueType.UInt32)
    End Function

    Private Sub OnModuleInfoTextChanged(sender As Object, e As InfoTextChangedEventArgs)
        If _isAppClosing Then Return

        AddInfoTextLine(e.InfoText)
    End Sub

    Private Sub AddInfoTextLine(text As String)
        textBoxInfo.AppendText(String.Format("[{0}] {1}" & Environment.NewLine,
                Date.Now.ToString("HH:mm:ss.fffffff"), text))
    End Sub

    Private Sub OnModulePropertyChanged(sender As Object, e As ModulePropertyChangedEventArgs)
        If _isAppClosing Then Return

        If e.ChangedProperty = AdqCommon.PropertyType.DisplayBuffer Then
            BeginInvoke(_updateChartDelegate)
        Else
            BeginInvoke(_updatePropertyDelegate, New Object() {e.ChangedProperty})
        End If
    End Sub

    Private Sub UpdateProperty(changedProperty As AdqCommon.PropertyType)
        If changedProperty = AdqCommon.PropertyType.UpdateNumberOfFrames Then
            UpdateNumberOfFrames()
        ElseIf changedProperty = AdqCommon.PropertyType.ActualSampleFrequency Then
            UpdateActualSampleFrequency()
        ElseIf changedProperty = AdqCommon.PropertyType.DisplayBufferLength Then
            UpdateNumberOfFramesToDisplay()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionStarting Then
            UpdateAcquisitionStarting()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionStarted Then
            UpdateAcquisitionStarted()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionEnded Then
            AcquisitionEnded()
        ElseIf changedProperty = AdqCommon.PropertyType.WaitingForStartTrigger Then
            UpdateWaitingForStartTrigger()
        ElseIf changedProperty = AdqCommon.PropertyType.StartTriggerEventOccured Then
            UpdateStartTriggerEventOccured()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionErrorOccurred Then
            UpdateAcquisitionErrorOccured()
        End If
    End Sub

    Private Sub UpdateNumberOfFrames()
        textBoxFramesAcquired.Text = _selectedAiModule.TotalFramesAcquired.ToString("N0")
        textBoxFramesDisplayed.Text = _selectedAiModule.TotalFramesDisplayed.ToString("N0")
        textBoxFramesSkipped.Text = _selectedAiModule.TotalFramesSkipped.ToString("N0")
        progressUpdateTimer.Start()
    End Sub

    Private Sub UpdateActualSampleFrequency()
        textBoxActualFrequency.Text = AdqCommon.FormatFrequency(_selectedAiModule.ActualSampleFrequency)
    End Sub

    Private Sub UpdateNumberOfFramesToDisplay()
        textBoxFramesToDisplay.Text = _selectedAiModule.DisplayBufferNumberOfFrames.ToString()
    End Sub

    Private Sub UpdateAcquisitionStarting()
        AddInfoTextLine(_txtStartingAcquisition)
        lblAcquisitionStatus.BackColor = Color.LimeGreen
        lblAcquisitionStatus.Text = _txtStartingAcquisition
    End Sub

    Private Sub UpdateAcquisitionStarted()
        AddInfoTextLine(_txtAcquisitionStarted)
        lblAcquisitionStatus.BackColor = Color.Lime
        lblAcquisitionStatus.Text = _txtAcquisitionStarted
    End Sub

    Private Sub UpdateWaitingForStartTrigger()
        AddInfoTextLine(_txtWaitingForStartTrigger)
        lblAcquisitionStatus.BackColor = Color.Gold
        lblAcquisitionStatus.Text = _txtWaitingForStartTrigger
    End Sub

    Private Sub UpdateStartTriggerEventOccured()
        AddInfoTextLine(_txtStartTriggerEventOccured & " " & _txtAcquisitionStarted)
        lblAcquisitionStatus.BackColor = Color.LightGreen
        lblAcquisitionStatus.Text = _txtStartTriggerEventOccured & Environment.NewLine & _txtAcquisitionStarted
    End Sub

    Private Sub UpdateAcquisitionErrorOccured()
        lblAcquisitionStatus.BackColor = Color.OrangeRed
        lblAcquisitionStatus.Text = _txtErrorOccurred
    End Sub

    Private Sub ProgressUpdateTimer_Tick(sender As Object, e As EventArgs) Handles progressUpdateTimer.Tick
        If Not _selectedAiModule.RequestUpdateNumberOfFrames Then
            progressUpdateTimer.Stop()
            _selectedAiModule.RequestUpdateNumberOfFrames = True
        End If
    End Sub

    Private Sub GenerateAxisLabels(numberOfFrames As UInteger)
        If _xAxisLabels Is Nothing OrElse _xAxisLabels.Length <> numberOfFrames Then
            _xAxisLabels = New String(CInt(numberOfFrames - 1)) {}
        End If

        Dim xValue As Double = 0.0
        Dim xValueDelta As Double = GetXValueDelta()
        If coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime Then
            For i As Integer = 0 To CInt(numberOfFrames - 1)
                _xAxisLabels(i) = AdqCommon.FormatSampleTime(xValue)
                xValue += xValueDelta
            Next
        Else
            For i As Integer = 0 To CInt(numberOfFrames - 1)
                _xAxisLabels(i) = xValue.ToString()
                xValue += xValueDelta
            Next
        End If
    End Sub

    Private Function GetXValueDelta() As Double
        Return If(coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime, _selectedAiModule.SampleTime, 1.0)
    End Function

    Private Sub UpdateChartGraphic()
        Chart1.SuspendLayout()
        For i = 0 To Chart1.Series.Count - 1
            RemovePoints(Chart1.Series(i))
        Next
        Chart1.ResumeLayout()
        DisplayChartValues()
    End Sub

    Private Sub RemovePoints(chSeries As Series)
        While chSeries.Points.Count > 0
            chSeries.Points.RemoveAt(chSeries.Points.Count - 1)
        End While
        chSeries.Points.Clear()
    End Sub

    Private Sub DisplayChartValues()
        Dim frameLength As Integer = CInt(_selectedAiModule.FrameLength)
        Dim bufferLength As UInteger = _selectedAiModule.DisplayBufferLength
        Dim displayFramesBuffer As Double() = _selectedAiModule.DisplayBuffer

        Chart1.SuspendLayout()
        Dim frameIdx As Integer = 0
        While frameIdx < frameLength
            Dim chSeries As Series = Chart1.Series(frameIdx)
            Dim valIdx As Integer = frameIdx
            Dim pointIdx As Integer = 0
            While valIdx < bufferLength
                Dim physicalValue As Double = displayFramesBuffer(valIdx)
                Dim dp = New DataPoint(pointIdx, physicalValue) With
                    {
                        .AxisLabel = _xAxisLabels(pointIdx)
                    }

                chSeries.Points.Add(dp)
                valIdx += frameLength
                pointIdx += 1
            End While

            frameIdx += 1
        End While
        Chart1.ChartAreas(0).RecalculateAxesScale()
        Chart1.ResumeLayout()
        graphicUpdateTimer.Start()
    End Sub

    Private Sub GraphicUpdateTimer_Tick(sender As Object, e As EventArgs) Handles graphicUpdateTimer.Tick
        graphicUpdateTimer.Stop()
        _selectedAiModule.RequestFramesToDisplay = True
    End Sub

    Private Sub SelectFirstComboboxItem(coBox As ComboBox)
        If coBox.Items.Count > 0 Then
            coBox.SelectedIndex = 0
        End If
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(sender As Object, e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If coBoxModules.SelectedIndex = _selectedModuleItemIndex Then Return

        _selectedModuleItemIndex = coBoxModules.SelectedIndex
        ClearDisplayedValues()
        Chart1.Titles.Clear()

        EnableUI(coBoxModules.SelectedIndex >= 0)
        btnStartStop.Enabled = coBoxModules.SelectedIndex >= 0

        If _selectedAiModule IsNot Nothing Then
            RemoveHandler _selectedAiModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
            RemoveHandler _selectedAiModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
            _selectedAiModule.ClearModuleResources()
        End If

        If coBoxModules.SelectedIndex < 0 Then
            _selectedAiModule = Nothing
            textBoxSampleFrequency.Text = String.Empty
            textBoxDisplaySeconds.Text = String.Empty
            _selectedUnitSymbol = String.Empty

        Else
            _selectedAiModule = _aiModules(coBoxModules.SelectedIndex)
            AddHandler _selectedAiModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
            AddHandler _selectedAiModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
            SetStopAfterFramesText(_selectedAiModule.StopAfterFrames)
            textBoxSampleFrequency.Text = String.Format("{0:0.###}", _selectedAiModule.DesiredSampleFrequency)
            textBoxDisplaySeconds.Text = String.Format("{0:0.###}", _selectedAiModule.DisplaySeconds)
            _selectedUnitSymbol = _selectedAiModule.DefaultRange.UnitSymbol
        End If

        textBoxActualFrequency.Text = String.Empty
        textBoxFramesToDisplay.Text = String.Empty

        FillChannelsAndRanges(_selectedAiModule)
        EnableHardwareTriggerSettings(_selectedAiModule)
    End Sub

    Private Sub FillChannelsAndRanges(aiModule As AnalogInputModule)
        FillRanges(aiModule)
        FillChannels(aiModule)
        UpdateRangesToolstripMenuItems(aiModule)
    End Sub

    Private Sub FillChannels(aiModule As AnalogInputModule)
        lvChannelList.Items.Clear()
        If aiModule Is Nothing Then Return

        Dim chIdx As Integer = 0
        While chIdx < aiModule.NumberOfChannels
            Dim lvItem As ListViewItem = lvChannelList.Items.Add("Channel " & chIdx.ToString())
            lvItem.SubItems.Add(aiModule.DefaultRange.RangeText)
            lvItem.Tag = aiModule.DefaultRange
            lvItem.BackColor = Color.Black
            lvItem.ForeColor = AdqCommon.ChannelColors(chIdx Mod AdqCommon.ChannelColors.Length)
            chIdx += 1
        End While

        CheckFirstChannels(2)
    End Sub

    Private Sub CheckFirstChannels(count As Integer)
        Dim idx As Integer = 0
        While idx < lvChannelList.Items.Count
            lvChannelList.Items(idx).Checked = idx < count
            idx += 1
        End While
    End Sub

    Private Sub FillRanges(aiModule As AnalogInputModule)
        CoBoxRanges.Items.Clear()
        If aiModule Is Nothing Then Return

        For Each range As AnalogRange In aiModule.Ranges
            CoBoxRanges.Items.Add(New RangeComboBoxItem(range))
        Next
        SelectFirstComboboxItem(CoBoxRanges)
    End Sub

    Private Sub UpdateRangesToolstripMenuItems(aiModule As AnalogInputModule)
        tsMnuItemSetRange.DropDownItems.Clear()

        If aiModule IsNot Nothing AndAlso aiModule.NumberOfRanges > 1 Then
            For Each range As AnalogRange In aiModule.Ranges
                Dim item = New ToolStripMenuItem(range.RangeText, Nothing, AddressOf TsMnuItemSetRange_Click) With
                    {
                        .Tag = range
                    }
                tsMnuItemSetRange.DropDownItems.Add(item)
            Next
        Else
            Dim item = New ToolStripMenuItem("No selection available") With
                {
                    .Enabled = False
                }
            tsMnuItemSetRange.DropDownItems.Add(item)
        End If
    End Sub

    Private Sub TsMnuItemSetRange_Click(sender As Object, e As EventArgs)
        If (TypeOf sender Is ToolStripMenuItem) Then
            Dim range As AnalogRange = CType(CType(sender, ToolStripMenuItem).Tag, AnalogRange)
            SetChannelsToRange(range)
        End If
    End Sub

    Private Sub SetChannelsToRange(range As AnalogRange)
        For Each lvItem As ListViewItem In lvChannelList.Items
            lvItem.Tag = range
            lvItem.SubItems(_rangeSubItemIndex).Text = range.RangeText
        Next
    End Sub

    Private Sub LvChannelList_MouseDown(sender As Object, e As MouseEventArgs) Handles lvChannelList.MouseDown
        _hitinfoChannelList = lvChannelList.HitTest(e.X, e.Y)
    End Sub

    Private Sub LvChannelList_MouseUp(sender As Object, e As MouseEventArgs) Handles lvChannelList.MouseUp
        If e.Button <> MouseButtons.Left OrElse _selectedAiModule.ModuleIsBusy OrElse _selectedAiModule Is Nothing OrElse
            _selectedAiModule.Ranges.Count <= 1 OrElse _hitinfoChannelList Is Nothing OrElse
            _hitinfoChannelList.SubItem Is Nothing Then
            Return
        End If

        Dim hitInfo As ListViewHitTestInfo = lvChannelList.HitTest(e.X, e.Y)
        If _hitinfoChannelList.Item Is hitInfo.Item AndAlso _hitinfoChannelList.Item.SubItems(_rangeSubItemIndex) Is hitInfo.SubItem Then
            Dim range As AnalogRange = CType(_hitinfoChannelList.Item.Tag, AnalogRange)
            CoBoxRanges.Bounds = hitInfo.SubItem.Bounds
            CoBoxRanges.SelectedIndex = CInt(range.RangeIndex)
            CoBoxRanges.Visible = True
            CoBoxRanges.BringToFront()
            CoBoxRanges.Focus()
        End If
    End Sub

    Private Sub InitRangeComboBox()
        CoBoxRanges.DropDownStyle = ComboBoxStyle.DropDownList
        CoBoxRanges.Parent = lvChannelList
        CoBoxRanges.Visible = False
        AddHandler CoBoxRanges.DropDownClosed, AddressOf CoBoxRanges_DropDownClosed
        AddHandler CoBoxRanges.Leave, AddressOf CoBoxRanges_Leave
        AddHandler CoBoxRanges.KeyPress, AddressOf CoBoxRanges_KeyPress
    End Sub

    Private Sub CoBoxRanges_KeyPress(sender As Object, e As KeyPressEventArgs) Handles CoBoxRanges.KeyPress
        If e.KeyChar = ChrW(Keys.Enter) Then
            UpdateListViewItemChannelList()
        ElseIf e.KeyChar = ChrW(Keys.Escape) Then
            Dim range As AnalogRange = CType(_hitinfoChannelList.Item.Tag, AnalogRange)
            CoBoxRanges.SelectedIndex = CInt(range.RangeIndex)
            CoBoxRanges.Visible = False
        End If
    End Sub

    Private Sub CoBoxRanges_DropDownClosed(sender As Object, e As EventArgs) Handles CoBoxRanges.DropDownClosed
        UpdateListViewItemChannelList()
    End Sub

    Private Sub CoBoxRanges_Leave(sender As Object, e As EventArgs) Handles CoBoxRanges.Leave
        UpdateListViewItemChannelList()
    End Sub

    Private Sub UpdateListViewItemChannelList()
        If _hitinfoChannelList Is Nothing OrElse _hitinfoChannelList.SubItem Is Nothing OrElse CoBoxRanges.SelectedItem Is Nothing Then Return

        Dim rangeItem As RangeComboBoxItem = CType(CoBoxRanges.SelectedItem, RangeComboBoxItem)
        _hitinfoChannelList.Item.Tag = rangeItem.Range
        _hitinfoChannelList.Item.SubItems(_rangeSubItemIndex).Text = rangeItem.ToString()

        CoBoxRanges.Visible = False
        lvChannelList.Focus()
    End Sub

    Private Sub ContextMnuChanList_Opening(sender As Object, e As System.ComponentModel.CancelEventArgs) Handles contextMnuChanList.Opening
        If CoBoxRanges.Visible Then
            CoBoxRanges.Visible = False
        End If
    End Sub

    Private Sub EnableUI(enabled As Boolean)
        coBoxModules.Enabled = enabled
        textBoxSampleFrequency.Enabled = enabled
        textBoxDisplaySeconds.Enabled = enabled
        coBoxStopAfterFrames.Enabled = enabled
        coBoxChartType.Enabled = enabled
        coBoxXAxisLabels.Enabled = enabled
        btnReset.Enabled = enabled
        tsMnuItemSelectAll.Enabled = enabled
        tsMnuItemSelectNone.Enabled = enabled
        tsMnuItemInvertSelection.Enabled = enabled
        btnClear.Enabled = enabled
    End Sub

    Private Sub TextBoxSampleFrequency_TextChanged(sender As Object, e As EventArgs) Handles textBoxSampleFrequency.TextChanged
        If _selectedAiModule Is Nothing Then Return
        _selectedAiModule.DesiredSampleFrequency = ParseDoubleValue(textBoxSampleFrequency.Text)
    End Sub

    Private Sub TextBoxDisplaySeconds_TextChanged(sender As Object, e As EventArgs) Handles textBoxDisplaySeconds.TextChanged
        If _selectedAiModule Is Nothing Then Return
        _selectedAiModule.DisplaySeconds = ParseDoubleValue(textBoxDisplaySeconds.Text)
    End Sub

    Private Function ParseDoubleValue(valueText As String) As Double
        Dim value As Double
        Dim isValid As Boolean = Double.TryParse(valueText, value)
        If Not isValid OrElse value <= 0 Then Return 0.0

        Return value
    End Function

    Private Sub SetStopAfterFramesText(stopAfterFrames As ULong)
        If stopAfterFrames = AdqConstants.ADQ_NUMBER_UNKNOWN Then
            coBoxStopAfterFrames.Text = _txtInfinite
        Else
            coBoxStopAfterFrames.Text = stopAfterFrames.ToString()
        End If
    End Sub

    Private Sub CoBoxStopAfterFrames_Leave(sender As Object, e As EventArgs) Handles coBoxStopAfterFrames.Leave
        If _selectedAiModule Is Nothing Then Return
        _selectedAiModule.StopAfterFrames = ParseStopAfterFrames(coBoxStopAfterFrames.Text)
    End Sub

    Private Function ParseStopAfterFrames(stopAfterFramesText As String) As ULong
        If stopAfterFramesText = _txtInfinite Then
            Return AdqConstants.ADQ_NUMBER_UNKNOWN
        End If

        Dim numFrames As ULong
        Dim isValid As Boolean = ULong.TryParse(stopAfterFramesText, numFrames)
        If Not isValid Then
            coBoxStopAfterFrames.Text = _txtInfinite
            Return AdqConstants.ADQ_NUMBER_UNKNOWN
        End If
        Return numFrames
    End Function

    Private Sub ChkBoxShowLegend_CheckedChanged(sender As Object, e As EventArgs) Handles chkBoxShowLegend.CheckedChanged
        If Chart1.Legends.Count > 0 Then
            Chart1.Legends(0).Enabled = chkBoxShowLegend.Checked
        End If
    End Sub

    Private Sub LvChannelList_ItemCheck(sender As Object, e As ItemCheckEventArgs) Handles lvChannelList.ItemCheck
        If _selectedAiModule.ModuleIsBusy Then
            e.NewValue = e.CurrentValue
        End If
    End Sub

    Private Sub LvChannelList_ItemChecked(sender As Object, e As ItemCheckedEventArgs) Handles lvChannelList.ItemChecked
        _selectedAiModule.FrameLength = CUInt(lvChannelList.CheckedItems.Count)
        textBoxNumberOfSelectedChannels.Text = _selectedAiModule.FrameLength.ToString()
    End Sub

    Private Sub EnableHardwareTriggerSettings(aiModule As AnalogInputModule)
        textBoxTriggerDescription.Text = GetTriggerDescription(aiModule)
        Dim triggerAvailable As Boolean = aiModule IsNot Nothing AndAlso aiModule.HardwareTrigger IsNot Nothing

        chkBoxTriggerStartRisingEdge0.Enabled = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartSource And _bit0Mask) <> 0
        chkBoxTriggerStartRisingEdge1.Enabled = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartSource And _bit1Mask) <> 0

        chkBoxTriggerStartFallingEdge0.Enabled = chkBoxTriggerStartRisingEdge0.Enabled
        chkBoxTriggerStartFallingEdge1.Enabled = chkBoxTriggerStartRisingEdge1.Enabled

        chkBoxTriggerStopRisingEdge0.Enabled = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopSource And _bit0Mask) <> 0
        chkBoxTriggerStopRisingEdge1.Enabled = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopSource And _bit1Mask) <> 0

        chkBoxTriggerStopFallingEdge0.Enabled = chkBoxTriggerStopRisingEdge0.Enabled
        chkBoxTriggerStopFallingEdge1.Enabled = chkBoxTriggerStopRisingEdge1.Enabled

        chkBoxTriggerStartRisingEdge0.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartRisingEdges And _bit0Mask) <> 0
        chkBoxTriggerStartRisingEdge1.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartRisingEdges And _bit1Mask) <> 0

        chkBoxTriggerStartFallingEdge0.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartFallingEdges And _bit0Mask) <> 0
        chkBoxTriggerStartFallingEdge1.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStartFallingEdges And _bit1Mask) <> 0

        chkBoxTriggerStopRisingEdge0.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopRisingEdges And _bit0Mask) <> 0
        chkBoxTriggerStopRisingEdge1.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopRisingEdges And _bit1Mask) <> 0

        chkBoxTriggerStopFallingEdge0.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopFallingEdges And _bit0Mask) <> 0
        chkBoxTriggerStopFallingEdge1.Checked = triggerAvailable AndAlso (aiModule.HardwareTrigger.TriggerStopFallingEdges And _bit1Mask) <> 0

        chkBoxTriggerTimeout.Enabled = triggerAvailable AndAlso CheckStartTriggerEnabledCondition()
        chkBoxTriggerTimeout.Checked = triggerAvailable AndAlso aiModule.HardwareTrigger.UseTriggerTimeout

        textBoxTriggerTimeout.Enabled = chkBoxTriggerTimeout.Checked
        textBoxTriggerTimeout.Text = GetTriggerTimeoutText(aiModule, triggerAvailable)

        chkBoxChannelListPerTrigger.Enabled = triggerAvailable AndAlso CheckChanListPerTriggerCondition()
        chkBoxChannelListPerTrigger.Checked = triggerAvailable AndAlso aiModule.HardwareTrigger.ChannelListPerTrigger

        chkBoxAutomaticRearm.Enabled = triggerAvailable AndAlso CheckRearmConditions()
        chkBoxAutomaticRearm.Checked = triggerAvailable AndAlso aiModule.HardwareTrigger.TriggerAutomaticRearm
    End Sub

    Private Function GetTriggerDescription(aiModule As AnalogInputModule) As String
        If aiModule Is Nothing Then
            Return String.Empty
        End If
        Return aiModule.TriggerDescription
    End Function

    Private Function GetTriggerTimeoutText(aiModule As AnalogInputModule, triggerAvailable As Boolean) As String
        If triggerAvailable Then
            Return aiModule.HardwareTrigger.TriggerTimeoutMilliseconds.ToString()
        End If
        Return String.Empty
    End Function

    Private Sub ChkBoxTriggerTimeout_CheckedChanged(sender As Object, e As EventArgs) Handles chkBoxTriggerTimeout.CheckedChanged
        textBoxTriggerTimeout.Enabled = chkBoxTriggerTimeout.Checked
    End Sub

    Private Sub ChkBoxTriggerStartEdge_CheckedChanged(sender As Object, e As EventArgs) Handles _
        chkBoxTriggerStartRisingEdge1.CheckedChanged, chkBoxTriggerStartRisingEdge0.CheckedChanged,
        chkBoxTriggerStartFallingEdge1.CheckedChanged, chkBoxTriggerStartFallingEdge0.CheckedChanged

        chkBoxChannelListPerTrigger.Enabled = CheckChanListPerTriggerCondition()
        chkBoxAutomaticRearm.Enabled = CheckRearmConditions()
        chkBoxTriggerTimeout.Enabled = CheckStartTriggerEnabledCondition()
    End Sub

    Private Sub ChkBoxTriggerStopEdge_CheckedChanged(sender As Object, e As EventArgs) Handles _
        chkBoxTriggerStopRisingEdge1.CheckedChanged, chkBoxTriggerStopRisingEdge0.CheckedChanged,
        chkBoxTriggerStopFallingEdge1.CheckedChanged, chkBoxTriggerStopFallingEdge0.CheckedChanged

        chkBoxAutomaticRearm.Enabled = CheckRearmConditions()
    End Sub

    Private Function CheckChanListPerTriggerCondition() As Boolean
        Return _selectedAiModule.HardwareTrigger.SupportsChannelListPerTrigger _
                AndAlso (chkBoxTriggerStartRisingEdge0.Checked _
                OrElse chkBoxTriggerStartRisingEdge1.Checked _
                OrElse chkBoxTriggerStartFallingEdge0.Checked _
                OrElse chkBoxTriggerStartFallingEdge1.Checked)
    End Function

    Private Function CheckRearmConditions() As Boolean
        Return CheckStartTriggerEnabledCondition() _
                AndAlso (chkBoxTriggerStopRisingEdge0.Checked _
                OrElse chkBoxTriggerStopRisingEdge1.Checked _
                OrElse chkBoxTriggerStopFallingEdge0.Checked _
                OrElse chkBoxTriggerStopFallingEdge1.Checked)
    End Function

    Private Function CheckStartTriggerEnabledCondition() As Boolean
        Return chkBoxTriggerStartRisingEdge0.Checked _
                OrElse chkBoxTriggerStartRisingEdge1.Checked _
                OrElse chkBoxTriggerStartFallingEdge0.Checked _
                OrElse chkBoxTriggerStartFallingEdge1.Checked
    End Function

    Private Sub BtnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        _selectedAiModule.ResetModule()
    End Sub

    Private Sub BtnStartStop_Click(sender As Object, e As EventArgs) Handles btnStartStop.Click
        If _selectedAiModule Is Nothing Then Return

        If _selectedAiModule.ModuleIsBusy Then
            _selectedAiModule.CancelAcquisition()
            Return
        End If

        ClearDisplayedValues()
        EnableUI(False)
        btnStartStop.Text = "Stop"
        Try
            UpdateChannelSettings()
            UpdateModuleTriggerSettings()
            _selectedAiModule.InitAcquisition()
            GenerateAxisLabels(_selectedAiModule.DisplayBufferNumberOfFrames)
            UpdateChart(_selectedAiModule)
            _selectedAiModule.RequestFramesToDisplay = True
            _selectedAiModule.RequestUpdateNumberOfFrames = True
            _selectedAiModule.StartAcquisition()
        Catch ex As Exception
            tabControlMain.SelectedTab = tpInfo
            ResetUiStatus()
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub ResetUiStatus()
        EnableUI(True)
        btnStartStop.Text = "Start"
        lblAcquisitionStatus.BackColor = Color.White
        lblAcquisitionStatus.Text = _txtReady
    End Sub

    Private Sub AcquisitionEnded()
        progressUpdateTimer.Stop()
        If _isAppClosing Then Return

        If Not _selectedAiModule.ModuleIsBusy Then
            ResetUiStatus()
            EnableChartTooltipFunction()
        End If
    End Sub

    Private Sub EnableChartTooltipFunction()
        If Chart1.Series.Count > 0 Then
            AddHandler Chart1.GetToolTipText, AddressOf Chart_GetToolTipText
        End If
    End Sub

    Private Sub Chart_GetToolTipText(sender As Object, e As ToolTipEventArgs)
        If e.HitTestResult.ChartElementType = ChartElementType.DataPoint Then
            Dim chSeries = e.HitTestResult.Series
            If chSeries IsNot Nothing Then
                Dim dp As DataPoint = chSeries.Points(e.HitTestResult.PointIndex)
                e.Text = FormatTooltipText(dp)
            End If
        ElseIf e.HitTestResult.ChartElementType = ChartElementType.Axis OrElse
            e.HitTestResult.ChartElementType = ChartElementType.TickMarks Then
            If (TypeOf sender Is Chart) Then
                Dim result As HitTestResult = CType(sender, Chart).HitTest(e.X, e.Y, ChartElementType.DataPoint)
                If result.ChartElementType = ChartElementType.DataPoint Then
                    Dim dp As DataPoint = result.Series.Points(result.PointIndex)
                    e.Text = FormatTooltipText(dp)
                End If
            End If
        End If
    End Sub

    Private Function FormatTooltipText(dp As DataPoint) As String
        Return String.Format("X={0}, Y={1}",
                        dp.XValue,
                        AdqCommon.FormatPhysicalValue(_fmtStringPhysValues, dp.YValues(0)) & _selectedUnitSymbol)
    End Function

    Private Sub UpdateChannelSettings()
        _selectedAiModule.ChannelSettings.Clear()
        For Each item As ListViewItem In lvChannelList.CheckedItems
            Dim range As AnalogRange = CType(item.Tag, AnalogRange)
            _selectedAiModule.AddChannelSetting(CUInt(item.Index), range)
        Next
        _selectedAiModule.UpdateChannelList()
    End Sub

    Private Sub UpdateModuleTriggerSettings()
        If _selectedAiModule.HardwareTrigger Is Nothing Then Return

        Dim timeout As UInteger = 0
        Dim useTimeout As Boolean = chkBoxTriggerTimeout.Enabled AndAlso chkBoxTriggerTimeout.Checked

        If useTimeout Then
            Try
                timeout = UInteger.Parse(textBoxTriggerTimeout.Text)
            Catch ex As Exception
                Throw New ApplicationException("Error trigger Timeout value: " _
                        & textBoxTriggerTimeout.Text + Environment.NewLine + ex.Message)
            End Try
        End If
        _selectedAiModule.HardwareTrigger.TriggerTimeoutMilliseconds = timeout
        _selectedAiModule.HardwareTrigger.UseTriggerTimeout = useTimeout

        _selectedAiModule.HardwareTrigger.ChannelListPerTrigger = chkBoxChannelListPerTrigger.Enabled AndAlso chkBoxChannelListPerTrigger.Checked
        _selectedAiModule.HardwareTrigger.TriggerAutomaticRearm = chkBoxAutomaticRearm.Enabled AndAlso chkBoxAutomaticRearm.Checked

        Dim startTriggerRisingEdge As UInteger = 0
        Dim startTriggerFallingEdge As UInteger = 0
        If _selectedAiModule.HardwareTrigger.TriggerStartSource <> 0 Then
            If (_selectedAiModule.HardwareTrigger.TriggerStartSource And _bit0Mask) <> 0 Then
                If chkBoxTriggerStartRisingEdge0.Checked Then
                    startTriggerRisingEdge = CUInt(startTriggerRisingEdge Or _bit0Mask)
                End If

                If chkBoxTriggerStartFallingEdge0.Checked Then
                    startTriggerFallingEdge = CUInt(startTriggerFallingEdge Or _bit0Mask)
                End If
            End If

            If (_selectedAiModule.HardwareTrigger.TriggerStartSource And _bit1Mask) <> 0 Then
                If chkBoxTriggerStartRisingEdge1.Checked Then
                    startTriggerRisingEdge = CUInt(startTriggerRisingEdge Or _bit1Mask)
                End If

                If chkBoxTriggerStartFallingEdge1.Checked Then
                    startTriggerFallingEdge = CUInt(startTriggerFallingEdge Or _bit1Mask)
                End If
            End If
        End If

        _selectedAiModule.HardwareTrigger.TriggerStartRisingEdges = startTriggerRisingEdge
        _selectedAiModule.HardwareTrigger.TriggerStartFallingEdges = startTriggerFallingEdge

        Dim stopTriggerRisingEdge As UInteger = 0
        Dim stopTriggerFallingEdge As UInteger = 0
        If _selectedAiModule.HardwareTrigger.TriggerStopSource <> 0 Then
            If (_selectedAiModule.HardwareTrigger.TriggerStopSource And _bit0Mask) <> 0 Then
                If chkBoxTriggerStopRisingEdge0.Checked Then
                    stopTriggerRisingEdge = CUInt(stopTriggerRisingEdge Or _bit0Mask)
                End If

                If chkBoxTriggerStopFallingEdge0.Checked Then
                    stopTriggerFallingEdge = CUInt(stopTriggerFallingEdge Or _bit0Mask)
                End If
            End If

            If (_selectedAiModule.HardwareTrigger.TriggerStopSource And _bit1Mask) <> 0 Then

                If chkBoxTriggerStopRisingEdge1.Checked Then
                    stopTriggerRisingEdge = CUInt(stopTriggerRisingEdge Or _bit1Mask)
                End If

                If chkBoxTriggerStopFallingEdge1.Checked Then
                    stopTriggerFallingEdge = CUInt(stopTriggerFallingEdge Or _bit1Mask)
                End If
            End If
        End If

        _selectedAiModule.HardwareTrigger.TriggerStopRisingEdges = stopTriggerRisingEdge
        _selectedAiModule.HardwareTrigger.TriggerStopFallingEdges = stopTriggerFallingEdge
    End Sub

    Private Sub FormMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles MyBase.FormClosing
        If _selectedAiModule IsNot Nothing AndAlso _selectedAiModule.ModuleIsBusy Then
            Dim result As DialogResult = MessageBox.Show("The acquisition is still running." _
                                         & Environment.NewLine _
                                         & "Do you want to close the program?",
                                         "Module busy",
                                         MessageBoxButtons.YesNo,
                                         MessageBoxIcon.Warning)
            If result = DialogResult.Yes Then
                _isAppClosing = True
                _selectedAiModule.RequestFramesToDisplay = False
                graphicUpdateTimer.Stop()
                progressUpdateTimer.Stop()
                RemoveHandler _selectedAiModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
                RemoveHandler _selectedAiModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
                _selectedAiModule.StopAcquisition()
                e.Cancel = False
            Else
                e.Cancel = True
            End If
            Return
        End If
        _isAppClosing = True
    End Sub

    Private Sub BtnClose_Click(sender As Object, e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub FormMain_FormClosed(sender As Object, e As FormClosedEventArgs) Handles MyBase.FormClosed
        CloseAdqDriver()
    End Sub

    Private Sub CloseAdqDriver()
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub ClearDisplayedValues()
        Chart1.Series.Clear()
        textBoxInfo.Clear()
        textBoxFramesAcquired.Text = String.Empty
        textBoxFramesSkipped.Text = String.Empty
        textBoxFramesDisplayed.Text = String.Empty
    End Sub

    Private Sub BtnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        ClearDisplayedValues()
    End Sub

    Private Sub TextBoxInfo_KeyDown(sender As Object, e As KeyEventArgs) Handles textBoxInfo.KeyDown
        If e.Modifiers = Keys.Control AndAlso e.KeyCode = Keys.A Then
            textBoxInfo.SelectionStart = 0
            textBoxInfo.SelectionLength = textBoxInfo.Text.Length
        End If
    End Sub

    Private Sub TsMnuItemSelectAll_Click(sender As Object, e As EventArgs) Handles tsMnuItemSelectAll.Click
        CheckListViewItems(True)
    End Sub

    Private Sub TsMnuItemInvertSelection_Click(sender As Object, e As EventArgs) Handles tsMnuItemInvertSelection.Click
        For i As Integer = 0 To lvChannelList.Items.Count - 1
            lvChannelList.Items(i).Checked = Not lvChannelList.Items(i).Checked
        Next
    End Sub

    Private Sub TsMnuItemSelectNone_Click(sender As Object, e As EventArgs) Handles tsMnuItemSelectNone.Click
        CheckListViewItems(False)
    End Sub

    Private Sub CheckListViewItems(check As Boolean)
        For i As Integer = 0 To lvChannelList.Items.Count - 1
            lvChannelList.Items(i).Checked = check
        Next
    End Sub

    Private Sub BtnInfo_Click(sender As Object, e As EventArgs) Handles btnInfo.Click
        AboutBox.ShowDialog()
    End Sub

    Private Function GetImageFileName(aiModule As AnalogInputModule, fileExt As String) As String
        Dim s As String = String.Empty
        If aiModule IsNot Nothing Then
            s = String.Format("{0}, Board {1}, Module {2}, ",
                    aiModule.ModelName,
                    aiModule.BoardIndex,
                    aiModule.ModuleIndex)
        End If
        Return s & Date.Now.ToString("yyyy-MM-dd HH.mm.ss") & fileExt
    End Function

    Private Function GetImageFormat(filterIndex As Integer) As ChartImageFormat
        Dim imgFormat As ChartImageFormat

        Select Case filterIndex
            Case 1
                imgFormat = ChartImageFormat.Png
            Case 2
                imgFormat = ChartImageFormat.Gif
            Case 3
                imgFormat = ChartImageFormat.Bmp
            Case 4
                imgFormat = ChartImageFormat.Tiff
            Case 5
                imgFormat = ChartImageFormat.Jpeg
            Case Else
                imgFormat = ChartImageFormat.Png
        End Select
        Return imgFormat
    End Function

    Private Sub TsMenuItemSaveGraphic_Click(sender As Object, e As EventArgs) Handles tsMenuItemSaveGraphic.Click
        saveFileDialog1.FileName = GetImageFileName(_selectedAiModule, ".png")
        saveFileDialog1.InitialDirectory = _imgInitialDirectory
        saveFileDialog1.Filter = "PNG files (*.png)|*.png|Gif files (*.gif)|*.gif|BMP files (*.bmp)|*.bmp|Tiff files (*.tif)|*.tif|Jpeg files (*.jpg)|*.jpg"
        Dim result As DialogResult = saveFileDialog1.ShowDialog()

        If result = DialogResult.OK Then
            _imgInitialDirectory = Path.GetDirectoryName(saveFileDialog1.FileName)
            Dim imgFormat As ChartImageFormat = GetImageFormat(saveFileDialog1.FilterIndex)
            Chart1.SaveImage(saveFileDialog1.FileName, imgFormat)
        End If
    End Sub

    Private Sub TsMnuItemCopyToClipboard_Click(sender As Object, e As EventArgs) Handles tsMnuItemCopyToClipboard.Click
        Using ms As MemoryStream = New MemoryStream()
            Chart1.SaveImage(ms, ChartImageFormat.Png)
            Dim Bitmap = New Bitmap(ms)
            Clipboard.SetImage(Bitmap)
        End Using
    End Sub
End Class
