﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the AOSingle project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports AOSingle.ALLDAQ.Boards

Public Class ModuleLocator
    Implements IDisposable, INotifyPropertyChanged

    Private _disposed As Boolean = False
    Private _adqOpenError As UInteger = AdqCommon.NotOpen

    Private Event PropertyChanged As PropertyChangedEventHandler _
        Implements INotifyPropertyChanged.PropertyChanged

    Public Sub New(handler As PropertyChangedEventHandler)
        AddHandler PropertyChanged, handler
    End Sub

    Public Sub InitModuleLocator()
        Try
            _adqOpenError = OpenAdqDriver()
            If _adqOpenError <> AdqErrors.ADQ_ERROR_NONE Then Return

            FindAdqModules()
            If AoModules.Count = 0 Then
                InfoText = "No FIN modules found."
            End If
        Catch ex As DllNotFoundException
            InfoText = "The device driver is not properly installed." & Environment.NewLine _
                & ex.Message
        End Try
    End Sub

    Public Sub FindAdqModules()
        Dim numberOfBoards As UInteger = GetNumberOfBoards()
        Dim boardIdx As UInteger = 0

        While boardIdx < numberOfBoards
            Dim numberOfModules As UInteger
            Dim adqError As UInteger = AdqDriver.adqGetNumberOfModules(boardIdx, numberOfModules)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetNumberOfModules", adqError)
                boardIdx += 1UI
                Continue While
            End If

            GetModuleTypes(boardIdx, numberOfModules)
            boardIdx += 1UI
        End While
    End Sub

    Private Function OpenAdqDriver() As UInteger
        Dim adqError As UInteger = AdqDriver.adqOpen()
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqOpen", adqError)
        End If
        Return adqError
    End Function

    Private Function GetNumberOfBoards() As UInteger
        Dim numBoards As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetNumberOfBoards(numBoards)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetNumberOfBoards", adqError)
            Return 0
        End If
        Return numBoards
    End Function

    Private Function GetModuleTypes(boardIndex As UInteger, numberOfModules As UInteger) As UInteger
        Dim adqError As UInteger = AdqErrors.ADQ_ERROR_NONE
        Dim moduleIdx As UInteger = 0

        While moduleIdx < numberOfModules
            Dim moduleType As UInteger
            adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, moduleType)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetModuleType", adqError)
                moduleIdx += 1UI
                Continue While
            End If

            If moduleType = AdqConstants.ADQ_MODULE_TYPE_AO Then
                GetAOModuleInfo(boardIndex, moduleIdx)
            End If
            moduleIdx += 1UI
        End While
        Return adqError
    End Function

    Private Function GetAOModuleInfo(boardIndex As UInteger, moduleIndex As UInteger) As UInteger
        Dim adqError As UInteger
        Dim adqGeneralModuleInfo = New AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID)
        Dim adqAoModuleInfo = New AdqDriver.ADQAOModuleInfo(AdqDriver.ADQAOModuleInfo_ID)

        Dim ModuleInfo As IntPtr = Marshal.AllocHGlobal(CInt(adqAoModuleInfo.Size))
        Marshal.StructureToPtr(adqAoModuleInfo, ModuleInfo, False)

        Dim infoSize As UInteger
        adqError = AdqDriver.adqGetModuleInfo(boardIndex, moduleIndex, adqGeneralModuleInfo, adqAoModuleInfo.Size, ModuleInfo, infoSize)
        adqAoModuleInfo = CType(Marshal.PtrToStructure(ModuleInfo, GetType(AdqDriver.ADQAOModuleInfo)), AdqDriver.ADQAOModuleInfo)
        Marshal.FreeHGlobal(ModuleInfo)

        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetModuleType", adqError)
            Return adqError
        End If

        Debug.Assert(adqAoModuleInfo.Size = infoSize,
                         String.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                            boardIndex, moduleIndex, adqAoModuleInfo.Size, infoSize))

        Dim boardName As String = GetBoardName(boardIndex)
        Dim aoModule = New AoModule(boardName, boardIndex, moduleIndex, adqAoModuleInfo)
        AoModules.Add(aoModule)

        Return adqError
    End Function

    Private Function GetBoardName(boardIndex As UInteger) As String
        Dim s As String = String.Empty
        Dim boardInfo = New AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID)

        Dim adqError As UInteger = AdqDriver.adqGetBoardInfo(boardIndex, boardInfo)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            s = boardInfo.BoardName
        End If
        Return s
    End Function

    Public ReadOnly Property AoModules As List(Of AoModule) = New List(Of AoModule)

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        InfoText = String.Format("API-function {0} returned Error number: &H{1:X}", functionName, adqError)
    End Sub

    Private Sub CloseAdqDriver()
        If _adqOpenError = AdqErrors.ADQ_ERROR_NONE Then
            AdqDriver.adqClose()
            _adqOpenError = AdqCommon.NotOpen
        End If
    End Sub

    Dim _infoText As String = String.Empty
    Public Property InfoText As String
        Get
            Return _infoText
        End Get
        Set(value As String)
            _infoText = value
            NotifyPropertyChanged(NameOf(InfoText))
        End Set
    End Property

    Public Sub NotifyPropertyChanged(ByVal propertyName As String)
        RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub

    Protected Overloads Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
    End Sub

    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not _disposed Then
            _disposed = True
            CloseAdqDriver()
        End If
    End Sub
End Class
