﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file is part of the CommonSettings project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports CommonSettings.ALLDAQ.Boards

Public Class Board
    Public Event InfoTextChanged As EventHandler(Of InfoTextChangedEventArgs)

    Public Sub New(boardName As String, boardIndex As UInteger, ByRef boardInfo As AdqDriver.ADQBoardInfo)
        Me.BoardIndex = boardIndex
        BuildBoardName(boardName)
        FillTTLVoltageLevels(boardInfo.TTLVoltageLevel)
    End Sub

    Public ReadOnly Property BoardIndex As UInteger

    Private _boardName As String
    Public ReadOnly Property BoardName As String
        Get
            Return _boardName
        End Get
    End Property

    Public ReadOnly Property VoltageLevels As List(Of KeyValuePair(Of UInteger, String)) = New List(Of KeyValuePair(Of UInteger, String))

    Private Sub BuildBoardName(boardName As String)
        _boardName = String.Format("{0}, Board {1}",
            AdqCommon.GetModelName(boardName),
            BoardIndex)
    End Sub

    Private Sub FillTTLVoltageLevels(ttlVoltageLevel As UInteger)
        AddVoltageLevel(ttlVoltageLevel, AdqConstants.ADQ_TTL_VOLTAGE_LEVEL_3_3_VOLTS)
        AddVoltageLevel(ttlVoltageLevel, AdqConstants.ADQ_TTL_VOLTAGE_LEVEL_5_0_VOLTS)
    End Sub

    Private Sub AddVoltageLevel(ttlVoltageLevel As UInteger, voltageConstant As UInteger)
        If (ttlVoltageLevel And voltageConstant) = voltageConstant Then
            VoltageLevels.Add(New KeyValuePair(Of UInteger, String)(voltageConstant, AdqCommon.GetVoltageText(voltageConstant)))
        End If
    End Sub

    Public Function GetTTLVoltage(ByRef voltage As UInteger) As Boolean
        Dim adqError As UInteger = AdqDriver.adqGetBoardTTLVoltage(BoardIndex, voltage)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetBoardTTLVoltage", adqError)
        End If
        Return adqError = AdqErrors.ADQ_ERROR_NONE
    End Function

    Public Function SetTTLVoltage(voltage As UInteger) As Boolean
        Dim adqError As UInteger = AdqDriver.adqSetBoardTTLVoltage(BoardIndex, voltage)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqSetBoardTTLVoltage", adqError)
        End If
        Return adqError = AdqErrors.ADQ_ERROR_NONE
    End Function

    Public Function Reset() As Boolean
        Dim adqError As UInteger = AdqDriver.adqResetBoard(BoardIndex)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqResetBoard", adqError)
        End If
        Return adqError = AdqErrors.ADQ_ERROR_NONE
    End Function

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        RaiseOnInfoTextChanged(String.Format("Function {0} returned Error number: &H{1:X}",
                functionName,
                adqError))
    End Sub

    Protected Sub RaiseOnInfoTextChanged(infoText As String)
        If InfoTextChangedEvent IsNot Nothing Then
            Dim args = New InfoTextChangedEventArgs(infoText)
            RaiseEvent InfoTextChanged(Me, args)
        End If
    End Sub
End Class
