﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'      This file contains a programming example for Intel counter 82C54 ICs
'      on ALLDAQ Data Acquisition Boards
'-----------------------------------------------------------------------------

Imports System.Runtime.InteropServices
Imports Counter8254.ALLDAQ.Boards

Public Class ModuleLocator
    Private Const _maxCounters As Integer = 3
    Private _adqOpenError As UInteger = AdqCommon.NotOpen
    Private ReadOnly _cntr8254Modules As ICollection(Of Counter8254Modules)
    Private ReadOnly _moduleType As UInteger
    Public Event OnInfoTextChanged As EventHandler(Of InfoTextChangedEventArgs)

    Public Sub New(cntr8254Modules As ICollection(Of Counter8254Modules), moduleType As UInteger,
                   ByRef infoTextChangedHandler As EventHandler(Of InfoTextChangedEventArgs))
        _cntr8254Modules = cntr8254Modules
        _moduleType = moduleType
        AddHandler OnInfoTextChanged, infoTextChangedHandler
    End Sub

    Private Function OpenAdqDriver() As UInteger
        Dim adqError As UInteger = AdqDriver.adqOpen()
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqOpen", adqError)
        End If
        Return adqError
    End Function

    Public Sub CloseAdqDriver()
        If _adqOpenError = AdqErrors.ADQ_ERROR_NONE Then
            AdqDriver.adqClose()
            _adqOpenError = AdqCommon.NotOpen
        End If
    End Sub

    Private Function GetNumberOfBoards() As UInteger
        Dim numBoards As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetNumberOfBoards(numBoards)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetNumberOfBoards", adqError)
            Return 0
        End If
        Return numBoards
    End Function

    Public Sub FindAdqModules()
        _adqOpenError = OpenAdqDriver()
        If _adqOpenError = AdqErrors.ADQ_ERROR_NONE Then
            Dim numberOfBoards As UInteger = GetNumberOfBoards()
            Dim boardIdx As UInteger = 0
            While boardIdx < numberOfBoards
                Dim numberOfModules As UInteger
                Dim adqError = AdqDriver.adqGetNumberOfModules(boardIdx, numberOfModules)
                If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                    SetFunctionErrorInfo("adqGetNumberOfModules", adqError)
                    boardIdx += 1UI
                    Continue While
                End If

                GetModuleTypes(boardIdx, numberOfModules)
                boardIdx += 1UI
            End While
        End If
    End Sub

    Private Function GetModuleTypes(boardIndex As UInteger, numberOfModules As UInteger) As UInteger
        Dim adqError As UInteger = AdqErrors.ADQ_ERROR_NONE
        Dim counterIndex As Integer = 0
        Dim counters As Counter8254Modules = Nothing
        Dim boardName As String = String.Empty
        Dim moduleIdx As UInteger = 0
        While moduleIdx < numberOfModules
            Dim moduleType As UInteger
            adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, moduleType)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetModuleType", adqError)
                Exit While
            End If

            If (moduleType = _moduleType) AndAlso (moduleType = AdqConstants.ADQ_MODULE_TYPE_CNTR) Then
                If String.IsNullOrEmpty(boardName) Then
                    boardName = GetBoardName(boardIndex)
                End If

                Dim cntr = GetCounterModule(boardIndex, moduleIdx)
                If cntr Is Nothing Then
                    Exit While
                End If

                If counterIndex = 0 Then
                    counters = New Counter8254Modules(boardIndex, boardName)
                End If
                counters.AddCounter(cntr)
                counterIndex += 1
            End If

            If IsLastCounterInGroup(moduleIdx, numberOfModules, moduleType, counterIndex) Then
                _cntr8254Modules.Add(counters)
                counterIndex = 0
            End If

            moduleIdx += 1UI
        End While
        Return adqError
    End Function

    Private Function IsLastCounterInGroup(moduleIndex As UInteger, numberOfModules As UInteger, moduleType As UInteger, counterIndex As Integer) As Boolean
        If counterIndex = _maxCounters Then Return True

        If counterIndex > 0 Then
            If (moduleType <> _moduleType) OrElse (moduleIndex = (numberOfModules - 1UI)) Then
                Return True
            End If
        End If
        Return False
    End Function


    Private Function GetCounterModule(boardIndex As UInteger, moduleIndex As UInteger) As CounterModule
        Dim adqGeneralModuleInfo As AdqDriver.ADQGeneralModuleInfo = New AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID)
        Dim adqCounterModuleInfo As AdqDriver.ADQCNTRModuleInfo = New AdqDriver.ADQCNTRModuleInfo(AdqDriver.ADQCNTRModuleInfo_ID)

        Dim moduleInfo As IntPtr = Marshal.AllocHGlobal(CInt(adqCounterModuleInfo.Size))
        Marshal.StructureToPtr(adqCounterModuleInfo, moduleInfo, False)

        Dim infoSize As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetModuleInfo(
            boardIndex,
            moduleIndex,
            adqGeneralModuleInfo,
            adqCounterModuleInfo.Size,
            moduleInfo,
            infoSize)
        adqCounterModuleInfo = CType(Marshal.PtrToStructure(moduleInfo, GetType(AdqDriver.ADQCNTRModuleInfo)), AdqDriver.ADQCNTRModuleInfo)
        Marshal.FreeHGlobal(moduleInfo)

        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetModuleInfo", adqError)
            Return Nothing
        End If

        Debug.Assert(adqCounterModuleInfo.Size = infoSize,
                         String.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                            boardIndex, moduleIndex, adqCounterModuleInfo.Size, infoSize))

        If adqCounterModuleInfo.CounterType <> AdqConstants.ADQ_CNTR_TYPE_8254 Then
            UpdateInfoText(String.Format("Board {0}, Module {1}, counter type is not supported.", boardIndex, moduleIndex))
            Return Nothing
        End If

        Debug.Assert(adqCounterModuleInfo.WidthBits = 16,
                String.Format("Board {0}, Module {1}, API-function adqGetModuleInfo counter 8254 width is {2}, but should be 16.",
                boardIndex, moduleIndex, adqCounterModuleInfo.WidthBits))

        Return New CounterModule(boardIndex, moduleIndex, adqCounterModuleInfo.WidthBits, adqCounterModuleInfo.ClockSources)
    End Function

    Private Function GetBoardName(boardIndex As UInteger) As String
        Dim s As String = String.Empty
        Dim boardInfo = New AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID)

        Dim adqError As UInteger = AdqDriver.adqGetBoardInfo(boardIndex, boardInfo)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            s = boardInfo.BoardName
        End If
        Return s
    End Function

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        UpdateInfoText(String.Format("API-function {0} returned Error number: &H{1:X}", functionName, adqError))
    End Sub

    Private Sub UpdateInfoText(text As String)
        Dim args = New InfoTextChangedEventArgs With {
            .InfoText = text
        }
        RaiseOnModuleInfoTextChanged(args)
    End Sub

    Protected Sub RaiseOnModuleInfoTextChanged(e As InfoTextChangedEventArgs)
        RaiseEvent OnInfoTextChanged(Me, e)
    End Sub
End Class
