﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the DIReadContinuous project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.IO
Imports System.Windows.Forms.DataVisualization.Charting
Imports System.Runtime.InteropServices
Imports DIReadContinuous.DIReadContinuous.ALLDAQ
Imports DIReadContinuous.ALLDAQ.Boards

Partial Public Class FormMain
    Inherits Form

    Private Const _idxChartTypeLine As Integer = 0
    Private Const _idxChartTypeLineAndMarkers As Integer = 1
    Private Const _idxChartTypePoint As Integer = 2
    Private Const _idxXAxisLabelsTime As Integer = 0

    Private _moduleLocator As ModuleLocator
    Private ReadOnly _modules As List(Of DigitalInputModule) = New List(Of DigitalInputModule)()
    Private _selectedModule As DigitalInputModule
    Private _isAppClosing As Boolean
    Private _selectedModuleItemIndex As Integer = -1
    Private _imgInitialDirectory As String = String.Empty
    Private Const _txtInfinite As String = "Infinite"
    Private Const _txtStartingAcquisition As String = "Starting acquisition..."
    Private Const _txtAcquisitionStarted As String = "Acquisition started."
    Private Const _txtErrorOccurred As String = "Error occurred."
    Private Const _txtReady As String = "Ready"
    Private Const _txtInvalidInput As String = "Invalid input"
    Private Delegate Sub UpdatePropertyDelegate(ByVal changedProperty As AdqCommon.PropertyType)
    Private _updatePropertyDelegate As UpdatePropertyDelegate
    Private Delegate Sub DisplayChartValuesDelegate()
    Private _updateChartDelegate As DisplayChartValuesDelegate

    Private _isSampleFrequencyValid As Boolean
    Private _isNumberOfValuesValid As Boolean

    Public Sub New()
        InitializeComponent()
    End Sub

    Private Sub FormMain_Load(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
        EnableUI(False)
        _updatePropertyDelegate = New UpdatePropertyDelegate(AddressOf UpdateProperty)
        _updateChartDelegate = New DisplayChartValuesDelegate(AddressOf UpdateChartGraphic)
        InitChart()

        Try
            _moduleLocator = New ModuleLocator(_modules, AddressOf OnModuleInfoTextChanged)
            _moduleLocator.FindModules()
            FillModuleCombobox()
            SelectFirstComboboxItem(coBoxModules)
            SelectFirstComboboxItem(coBoxChartXAxisLabels)
            SelectFirstComboboxItem(coBoxChartSeriesType)
            textBoxInfo.AppendText(String.Format("Number of clock controlled digital input modules found: {0}" & Environment.NewLine, _modules.Count))
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The device driver is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
            tabControlMain.SelectedTab = tabPageInfo
        End Try
    End Sub

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()

        For Each diMmodule As DigitalInputModule In _modules
            coBoxModules.Items.Add(diMmodule.ModuleName)
        Next
    End Sub

    Private Sub SelectFirstComboboxItem(ByVal coBox As ComboBox)
        If coBox.Items.Count > 0 Then
            coBox.SelectedIndex = 0
        End If
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(ByVal sender As Object, ByVal e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If coBoxModules.SelectedIndex = _selectedModuleItemIndex Then Return
        _selectedModuleItemIndex = coBoxModules.SelectedIndex
        ClearDisplayedValues()
        chart.Titles.Clear()
        EnableUI(coBoxModules.SelectedIndex >= 0)
        btnStart.Enabled = coBoxModules.SelectedIndex >= 0

        If _selectedModule IsNot Nothing Then
            RemoveHandler _selectedModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
            RemoveHandler _selectedModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
            _selectedModule.ClearModuleResources()
        End If

        If coBoxModules.SelectedIndex < 0 Then
            _selectedModule = Nothing
            textBoxSampleFrequency.Text = String.Empty
            textBoxValuesToDisplay.Text = String.Empty
        Else
            _selectedModule = _modules(coBoxModules.SelectedIndex)
            AddHandler _selectedModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
            AddHandler _selectedModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
            SetStopAfterValuesText(_selectedModule.StopAfterValues)
            textBoxSampleFrequency.Text = String.Format("{0:0.###}", _selectedModule.DesiredFrequency)
            textBoxValuesToDisplay.Text = String.Format("{0:0.###}", _selectedModule.DisplayNumberOfValues)
        End If

        textBoxActualSampleFrequency.Text = String.Empty
        textBoxValuesDisplayed.Text = String.Empty
        FillBits(_selectedModule)
    End Sub

    Private Sub ListView1_ItemCheck(ByVal sender As Object, ByVal e As ItemCheckEventArgs) Handles listView1.ItemCheck
        If _selectedModule.ModuleIsBusy Then
            e.NewValue = e.CurrentValue
            Return
        End If

        Dim sCount As Integer = listView1.CheckedItems.Count

        If e.NewValue = CheckState.Checked Then
            sCount += 1
        End If

        If e.NewValue = CheckState.Unchecked Then
            sCount -= 1
        End If

        textBoxSelectedBits.Text = String.Format("{0}", sCount)
        _selectedModule.BitsSelected = sCount
    End Sub

    Private Sub FillBits(ByVal diModule As DigitalInputModule)
        listView1.Items.Clear()

        If diModule Is Nothing Then
            Dim lvItem As ListViewItem = New ListViewItem With {
                .Checked = False,
                .Text = "No selection available"
            }
            listView1.Enabled = False
            listView1.Items.Add(lvItem)
            Return
        End If

        listView1.Enabled = True

        For bitIdx As Integer = 0 To CInt(diModule.PortWidthBits - 1UI)
            Dim lvItem As ListViewItem = New ListViewItem With {
                .Checked = False,
                .Text = String.Format("Bit {0}", bitIdx)
            }
            lvItem.ForeColor = AdqCommon.BitColors(bitIdx Mod AdqCommon.BitColors.Length)
            listView1.Items.Add(lvItem)
        Next

        Dim idx As Integer = 0

        While idx < listView1.Items.Count AndAlso idx < 2
            listView1.Items(idx).Checked = True
            idx += 1
        End While
    End Sub

    Private Sub UpdateBits()
        If _selectedModule IsNot Nothing Then
            textBoxSelectedBits.Text = String.Format("{0}", listView1.CheckedItems.Count)
            _selectedModule.BitsSelected = listView1.CheckedItems.Count
        End If
    End Sub

    Private Sub OnModuleInfoTextChanged(ByVal sender As Object, ByVal e As InfoTextChangedEventArgs)
        If _isAppClosing Then Return
        AddInfoTextLine(e.InfoText)
    End Sub

    Private Sub AddInfoTextLine(ByVal text As String)
        textBoxInfo.AppendText(String.Format("[{0}] {1}" & Environment.NewLine, Date.Now.ToString("HH:mm:ss.fff"), text))
    End Sub

    Private Sub OnModulePropertyChanged(ByVal sender As Object, ByVal e As ModulePropertyChangedEventArgs)
        If _isAppClosing Then Return

        If e.ChangedProperty = AdqCommon.PropertyType.DisplayBuffer Then
            BeginInvoke(_updateChartDelegate)
        Else
            BeginInvoke(_updatePropertyDelegate, New Object() {e.ChangedProperty})
        End If
    End Sub

    Private Sub UpdateProperty(ByVal changedProperty As AdqCommon.PropertyType)
        If changedProperty = AdqCommon.PropertyType.UpdateNumberOfValues Then
            UpdateNumberOfValues()
        ElseIf changedProperty = AdqCommon.PropertyType.ActualSampleFrequency Then
            UpdateActualSampleFrequency()
        ElseIf changedProperty = AdqCommon.PropertyType.UpdateValuesToDisplay Then
            UpdateValuesToDisplay()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionStarting Then
            UpdateAcquisitionStarting()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionStarted Then
            UpdateAcquisitionStarted()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionEnded Then
            AcquisitionEnded()
        ElseIf changedProperty = AdqCommon.PropertyType.AcquisitionErrorOccurred Then
            UpdateAcquisitionErrorOccured()
        End If
    End Sub

    Private Sub UpdateActualSampleFrequency()
        textBoxActualSampleFrequency.Text = AdqCommon.FormatFrequency(_selectedModule.ActualSampleFrequency)
        textBoxMinDisplayWidth.Text = AdqCommon.FormatSampleTime(1 / _selectedModule.ActualSampleFrequency * _selectedModule.DisplayNumberOfValues)
    End Sub

    Private Sub UpdateNumberOfValues()
        textBoxValuesAcquired.Text = _selectedModule.TotalValuesAcquired.ToString("N0")
        textBoxValuesDisplayed.Text = _selectedModule.TotalValuesDisplayed.ToString("N0")
        textBoxValuesSkipped.Text = _selectedModule.TotalValuesSkipped.ToString("N0")
        progressUpdateTimer.Start()
    End Sub

    Private Sub UpdateValuesToDisplay()
        textBoxValuesToDisplay.Text = _selectedModule.DisplayNumberOfValues.ToString()
    End Sub

    Private Sub UpdateAcquisitionStarting()
        AddInfoTextLine(_txtStartingAcquisition)
        lblAcquisitionStatus.BackColor = Color.LimeGreen
        lblAcquisitionStatus.Text = _txtStartingAcquisition
    End Sub

    Private Sub UpdateAcquisitionStarted()
        AddInfoTextLine(_txtAcquisitionStarted)
        lblAcquisitionStatus.BackColor = Color.Lime
        lblAcquisitionStatus.Text = _txtAcquisitionStarted
    End Sub

    Private Sub UpdateAcquisitionErrorOccured()
        lblAcquisitionStatus.BackColor = Color.OrangeRed
        lblAcquisitionStatus.Text = _txtErrorOccurred
    End Sub

    Private Sub AcquisitionEnded()
        progressUpdateTimer.[Stop]()
        If _isAppClosing Then Return

        If Not _selectedModule.ModuleIsBusy Then
            ResetUiStatus()
        End If
    End Sub

    Private Sub TextBoxSampleFrequency_TextChanged(ByVal sender As Object, ByVal e As EventArgs) Handles textBoxSampleFrequency.TextChanged
        If _selectedModule Is Nothing Then Return

        Dim sampleFrequency As Double
        _isSampleFrequencyValid = Double.TryParse(textBoxSampleFrequency.Text, sampleFrequency)
        If (Not _isSampleFrequencyValid) OrElse (sampleFrequency <= 0.0) OrElse (sampleFrequency > _selectedModule.MaxFrequency) Then
            _isSampleFrequencyValid = False
            _selectedModule.DesiredFrequency = 0.0
            Return
        End If
        _selectedModule.DesiredFrequency = sampleFrequency
        UpdateMinDisplayWidth()
    End Sub

    Private Sub TextBoxValuesToDisplay_TextChanged(ByVal sender As Object, ByVal e As EventArgs) Handles textBoxValuesToDisplay.TextChanged
        If _selectedModule Is Nothing Then Return

        Dim numberOfValuesToDisplay As UInteger
        _isNumberOfValuesValid = UInteger.TryParse(textBoxValuesToDisplay.Text, numberOfValuesToDisplay)
        If (Not _isNumberOfValuesValid) OrElse (numberOfValuesToDisplay = 0) Then
            _isNumberOfValuesValid = False
            _selectedModule.DisplayNumberOfValues = 0
            Return
        End If
        _selectedModule.DisplayNumberOfValues = numberOfValuesToDisplay
        UpdateMinDisplayWidth()
    End Sub

    Private Sub UpdateMinDisplayWidth()
        If (Not _isSampleFrequencyValid) OrElse (Not _isNumberOfValuesValid) Then
            textBoxMinDisplayWidth.Text = _txtInvalidInput
            Return
        End If

        Dim sampleTime As Double = 1.0 / _selectedModule.DesiredFrequency * _selectedModule.DisplayNumberOfValues
        textBoxMinDisplayWidth.Text = AdqCommon.FormatSampleTime(sampleTime)
    End Sub

    Private Sub CoBoxStopAfterValues_Leave(ByVal sender As Object, ByVal e As EventArgs) Handles coBoxStopAfterValues.Leave
        If _selectedModule Is Nothing Then Return
        _selectedModule.StopAfterValues = ParseStopAfterValues(coBoxStopAfterValues.Text)
    End Sub

    Private Sub SetStopAfterValuesText(ByVal stopAfterFrames As ULong)
        If stopAfterFrames = AdqConstants.ADQ_NUMBER_UNKNOWN Then
            coBoxStopAfterValues.Text = _txtInfinite
        Else
            coBoxStopAfterValues.Text = stopAfterFrames.ToString()
        End If
    End Sub

    Private Sub EnableUI(ByVal enabled As Boolean)
        coBoxModules.Enabled = enabled
        textBoxSampleFrequency.Enabled = enabled
        textBoxValuesToDisplay.Enabled = enabled
        coBoxStopAfterValues.Enabled = enabled
        coBoxChartSeriesType.Enabled = enabled
        coBoxChartXAxisLabels.Enabled = enabled
        btnReset.Enabled = enabled
        btnClear.Enabled = enabled
    End Sub

    Private Sub ResetUiStatus()
        EnableUI(True)
        btnStart.Text = "Start"
        lblAcquisitionStatus.BackColor = Color.White
        lblAcquisitionStatus.Text = _txtReady
    End Sub

    Private Sub BtnReset_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnReset.Click
        If _selectedModule IsNot Nothing Then
            _selectedModule.Reset()
        End If
    End Sub

    Private Sub BtnClose_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub BtnClear_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnClear.Click
        ClearDisplayedValues()
    End Sub

    Private Sub ClearDisplayedValues()
        chart.Series.Clear()
        textBoxInfo.Clear()
        textBoxValuesAcquired.Text = String.Empty
        textBoxValuesSkipped.Text = String.Empty
        textBoxValuesDisplayed.Text = String.Empty
    End Sub

    Private Sub FormMain_FormClosed(ByVal sender As Object, ByVal e As FormClosedEventArgs) Handles MyBase.FormClosed
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub FormMain_FormClosing(ByVal sender As Object, ByVal e As FormClosingEventArgs) Handles MyBase.FormClosing
        If _selectedModule IsNot Nothing AndAlso _selectedModule.ModuleIsBusy Then
            Dim result As DialogResult = MessageBox.Show("The acquisition is still running." & Environment.NewLine _
                & "Do you want to close the program?", "Module busy", MessageBoxButtons.YesNo, MessageBoxIcon.Warning)

            If result = DialogResult.Yes Then
                _isAppClosing = True
                _selectedModule.RequestValuesToDisplay = False
                progressUpdateTimer.[Stop]()
                graphicUpdateTimer.[Stop]()
                RemoveHandler _selectedModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
                RemoveHandler _selectedModule.ModulePropertyChanged, AddressOf OnModulePropertyChanged
                _selectedModule.StopAcquisition()
                e.Cancel = False
            Else
                e.Cancel = True
            End If
            Return
        End If
        _isAppClosing = True
    End Sub

    Private Sub BtnStartStop_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnStart.Click
        If _selectedModule Is Nothing Then Return

        If _selectedModule.ModuleIsBusy Then
            _selectedModule.CancelAcquisition()
            UpdateBits()
            Return
        End If

        ClearDisplayedValues()
        EnableUI(False)
        btnStart.Text = "Stop"

        Try
            _selectedModule.InitAcquisition()
            UpdateChart()
            _selectedModule.RequestValuesToDisplay = True
            _selectedModule.RequestUpdateNumberOfValues = True
            _selectedModule.StartAcquisition()
        Catch ex As Exception
            ResetUiStatus()
            AddInfoTextLine(ex.Message)
            tabControlMain.SelectedTab = tabPageInfo
        End Try
    End Sub

    Private Sub BtnInfo_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnInfo.Click
        Dim aboutBox = New AboutBox()
        aboutBox.ShowDialog()
    End Sub

    Private Sub ProgressUpdateTimer_Tick(ByVal sender As Object, ByVal e As EventArgs) Handles progressUpdateTimer.Tick
        If Not _selectedModule.RequestUpdateNumberOfValues Then
            progressUpdateTimer.[Stop]()
            _selectedModule.RequestUpdateNumberOfValues = True
        End If
    End Sub

    Private Sub GraphicUpdateTimer_Tick(ByVal sender As Object, ByVal e As EventArgs) Handles graphicUpdateTimer.Tick
        graphicUpdateTimer.[Stop]()
        _selectedModule.RequestValuesToDisplay = True
    End Sub

    Private Function ParseStopAfterValues(ByVal stopAfterValuesText As String) As ULong
        If stopAfterValuesText = _txtInfinite Then
            Return AdqConstants.ADQ_NUMBER_UNKNOWN
        End If

        Dim numFrames As ULong = Nothing
        Dim isValid As Boolean = ULong.TryParse(stopAfterValuesText, numFrames)

        If Not isValid OrElse numFrames <= 0 Then
            coBoxStopAfterValues.Text = _txtInfinite
            Return AdqConstants.ADQ_NUMBER_UNKNOWN
        End If

        Return numFrames
    End Function

    Private Sub InitChart()
        chart.BackColor = AdqCommon.DefaultBackgroundColor
        chart.Update()
    End Sub

    Private Sub UpdateChart()
        chart.Legends.Clear()
        chart.Series.Clear()
        chart.ChartAreas.Clear()
        InitChartXAxisTitle()
        Dim position As UInteger = 0

        For Each lvItem As ListViewItem In listView1.Items

            If lvItem.Checked Then
                AddGraph(lvItem, listView1.CheckedItems.Count, position)
                position += 1UI
            End If
        Next
    End Sub

    Private Sub AddGraph(ByVal lvi As ListViewItem, ByVal numberOfChartAreas As Integer, ByVal position As UInteger)
        Dim legend As Legend = New Legend With {
            .Name = "legend" & lvi.Index.ToString(),
            .ForeColor = lvi.ForeColor,
            .IsDockedInsideChartArea = True,
            .Docking = Docking.Top,
            .BackColor = Color.Transparent
        }
        chart.Legends.Add(legend)

        Dim chartarea As ChartArea = New ChartArea With {
            .Name = "chartarea" & lvi.Index.ToString(),
            .BackColor = AdqCommon.DefaultBackgroundColor
        }

        chartarea.AxisY.Maximum = 1.05
        chartarea.AxisY.Interval = 1
        chartarea.CursorX.IsUserSelectionEnabled = True
        chartarea.CursorY.IsUserSelectionEnabled = True
        chartarea.AxisX.MajorGrid.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisX.MajorTickMark.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisX.MinorGrid.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisX.MinorTickMark.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisX.LineColor = Color.Transparent
        chartarea.AxisX.LineWidth = 1
        chartarea.AxisX.LabelStyle.ForeColor = AdqCommon.DefaultLabelColor
        chartarea.AxisX.TitleForeColor = AdqCommon.DefaultLabelColor
        chartarea.AxisX.ScrollBar.ButtonColor = AdqCommon.DefaultScrollBarColor
        chartarea.AxisX.ScrollBar.BackColor = AdqCommon.DefaultScrollBarColor
        chartarea.AxisY.MajorGrid.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisY.MajorTickMark.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisY.MinorGrid.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisY.MinorTickMark.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisY.LineColor = AdqCommon.DefaultAxisColor
        chartarea.AxisY.LineWidth = 1
        chartarea.AxisY.LabelStyle.ForeColor = AdqCommon.DefaultLabelColor
        chartarea.AxisY.TitleForeColor = AdqCommon.DefaultLabelColor
        chartarea.AxisY.ScrollBar.ButtonColor = AdqCommon.DefaultScrollBarColor
        chartarea.AxisY.ScrollBar.BackColor = AdqCommon.DefaultScrollBarColor
        If radioButtonVertical.Checked Then
            chartarea.Position.Auto = False
            chartarea.Position.X = 3
            chartarea.Position.Y = 100.0F / numberOfChartAreas * position
            chartarea.Position.Height = 100.0F / numberOfChartAreas
            chartarea.Position.Width = 95
        ElseIf radioButtonAuto.Checked Then
            chartarea.Position.Auto = True
        End If
        chart.ChartAreas.Add(chartarea)

        Dim chartSeries As Series = New Series With {
            .Color = AdqCommon.BitColors(lvi.Index Mod AdqCommon.BitColors.Length),
            .Name = "series" & lvi.Index.ToString(),
            .Tag = lvi.Index,
            .YValueType = ChartValueType.Int32,
            .XValueType = ChartValueType.String,
            .IsXValueIndexed = False,
            .Legend = legend.Name,
            .LegendText = lvi.Text
        }
        legend.DockedToChartArea = chartarea.Name

        Select Case coBoxChartSeriesType.SelectedIndex
            Case _idxChartTypeLine
                chartSeries.ChartType = SeriesChartType.StepLine
            Case _idxChartTypeLineAndMarkers
                chartSeries.ChartType = SeriesChartType.StepLine
                chartSeries.MarkerStyle = MarkerStyle.Circle
            Case _idxChartTypePoint
                chartSeries.ChartType = SeriesChartType.Point
                chartSeries.MarkerStyle = MarkerStyle.Circle
            Case Else
                chartSeries.ChartType = SeriesChartType.StepLine
                chartSeries.MarkerStyle = MarkerStyle.Circle
        End Select

        For i As Integer = 0 To CInt(_selectedModule.DisplayBufferNumberOfValues - 1)
            chartSeries.Points.AddY(0)
            chartSeries.Points(chartSeries.Points.Count - 1).AxisLabel = GetXAsixLabel(0)
        Next

        chartSeries.ChartArea = chartarea.Name
        chart.Series.Add(chartSeries)
    End Sub

    Private Sub InitChartXAxisTitle()
        chart.Titles.Clear()
        Dim title As Title = New Title(coBoxChartXAxisLabels.Text) With {
            .Alignment = ContentAlignment.BottomCenter,
            .Docking = Docking.Bottom,
            .IsDockedInsideChartArea = False,
            .ForeColor = AdqCommon.DefaultLabelColor
        }
        title.Position.Auto = False
        title.Position.X = 3
        title.Position.Y = 95
        title.Position.Width = 94
        title.Position.Height = 5
        chart.Titles.Add(title)
    End Sub

    Private Function GetXAsixLabel(ByVal xValue As Double) As String
        If coBoxChartXAxisLabels.SelectedIndex = _idxXAxisLabelsTime Then
            Return AdqCommon.FormatSampleTime(xValue)
        End If

        Return CInt(xValue).ToString()
    End Function

    Private Function GetXValueDelta() As Double
        Return If(coBoxChartXAxisLabels.SelectedIndex = _idxXAxisLabelsTime, _selectedModule.SampleTime, 1.0)
    End Function

    Private Function GetXValue(ByVal numberOfValuesReturned As UInteger) As Double
        Return If(coBoxChartXAxisLabels.SelectedIndex = _idxXAxisLabelsTime,
            (_selectedModule.TotalValuesAcquired - numberOfValuesReturned) * _selectedModule.SampleTime,
            _selectedModule.TotalValuesAcquired - numberOfValuesReturned)
    End Function

    Private Sub UpdateChartGraphic()
        DisplayChartValues()
        graphicUpdateTimer.Start()
    End Sub

    Private Sub DisplayChartValues()
        Dim displayValuesBuffer As UInteger() = _selectedModule.NewValuesDisplayBuffer
        Dim numberOfValuesReturned As UInteger = _selectedModule.NumberOfValidValues
        Dim xValueBufferStart As Double = GetXValue(numberOfValuesReturned)
        Dim xValueDelta As Double = GetXValueDelta()

        chart.SuspendLayout()
        For Each chartSeries As Series In chart.Series
            Dim xValue As Double = xValueBufferStart
            For i As Integer = 0 To CInt(numberOfValuesReturned - 1)
                chartSeries.Points.RemoveAt(0)
                chartSeries.Points.AddY(GetBitValue(CInt(chartSeries.Tag), displayValuesBuffer(i)))
                chartSeries.Points(chartSeries.Points.Count - 1).AxisLabel = GetXAsixLabel(xValue)
                xValue += xValueDelta
            Next
        Next
        chart.ResumeLayout()
    End Sub

    Private Function GetBitValue(ByVal bitNumber As Integer, ByVal value As UInteger) As Integer
        Return CInt(value >> bitNumber) And 1
    End Function

    Private Function GetImageFileName(ByVal diModule As DigitalInputModule, ByVal fileExt As String) As String
        Dim s As String = String.Empty

        If diModule IsNot Nothing Then
            s = String.Format("{0}, Board {1}, Module {2}, ", diModule.ModelName, diModule.BoardIndex, diModule.ModuleIndex)
        End If

        Return s & Date.Now.ToString("yyyy-MM-dd HH.mm.ss") & fileExt
    End Function

    Private Function GetImageFormat(ByVal filterIndex As Integer) As ChartImageFormat
        Dim imgFormat As ChartImageFormat

        Select Case filterIndex
            Case 1
                imgFormat = ChartImageFormat.Png
            Case 2
                imgFormat = ChartImageFormat.Gif
            Case 3
                imgFormat = ChartImageFormat.Bmp
            Case 4
                imgFormat = ChartImageFormat.Tiff
            Case 5
                imgFormat = ChartImageFormat.Jpeg
            Case Else
                imgFormat = ChartImageFormat.Png
        End Select

        Return imgFormat
    End Function

    Private Sub TsMenuItemSaveGraphic_Click(ByVal sender As Object, ByVal e As EventArgs) Handles tsMenuItemSaveGraphic.Click
        saveFileDialog1.FileName = GetImageFileName(_selectedModule, ".png")
        saveFileDialog1.InitialDirectory = _imgInitialDirectory
        saveFileDialog1.Filter = "PNG files (*.png)|*.png|Gif files (*.gif)|*.gif|BMP files (*.bmp)|*.bmp|Tiff files (*.tif)|*.tif|Jpeg files (*.jpg)|*.jpg"
        Dim result As DialogResult = saveFileDialog1.ShowDialog()

        If result = DialogResult.OK Then
            _imgInitialDirectory = Path.GetDirectoryName(saveFileDialog1.FileName)
            Dim imgFormat As ChartImageFormat = GetImageFormat(saveFileDialog1.FilterIndex)
            chart.SaveImage(saveFileDialog1.FileName, imgFormat)
        End If
    End Sub

    Private Sub TsMenuItemCopyToClipboard_Click(ByVal sender As Object, ByVal e As EventArgs) Handles tsMenuItemCopyToClipboard.Click
        Using ms As MemoryStream = New MemoryStream()
            chart.SaveImage(ms, ChartImageFormat.Png)
            Dim bitmap = New Bitmap(ms)
            Clipboard.SetImage(bitmap)
        End Using
    End Sub

    Private Sub ToolStripMenuItemSelectAll_Click(ByVal sender As Object, ByVal e As EventArgs) Handles toolStripMenuItemSelectAll.Click
        For i As Integer = 0 To listView1.Items.Count - 1
            listView1.Items(i).Checked = True
        Next
    End Sub

    Private Sub ToolStripMenuItemInvertSelection_Click(ByVal sender As Object, ByVal e As EventArgs) Handles toolStripMenuItemInvertSelection.Click
        For i As Integer = 0 To listView1.Items.Count - 1
            listView1.Items(i).Checked = Not listView1.Items(i).Checked
        Next
    End Sub

    Private Sub ToolStripMenuItemSelectNone_Click(ByVal sender As Object, ByVal e As EventArgs) Handles toolStripMenuItemSelectNone.Click
        For i As Integer = 0 To listView1.Items.Count - 1
            listView1.Items(i).Checked = False
        Next
    End Sub
End Class
