﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file contains a programming example for the digital input modules
'       on ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports DISingle.ALLDAQ.Boards

Public Class FormMain
    Private _moduleLocator As ModuleLocator
    Private ReadOnly _diModules As List(Of DigitalModule) = New List(Of DigitalModule)
    Private _selectedModule As DigitalModule

    Private Sub FormMain_Shown(sender As Object, e As EventArgs) Handles MyBase.Shown
        Try
            EnableUI(False)

            Dim moduleTypes As List(Of UInteger) = InitModuleTypesToFind()
            _moduleLocator = New ModuleLocator(_diModules, moduleTypes, AddressOf OnInfoTextChanged)
            _moduleLocator.FindAdqModules()

            FillModuleCombobox()
            If coBoxModules.Items.Count > 0 Then
                coBoxModules.SelectedIndex = 0
            End If
        Catch ex As DllNotFoundException
            AddInfoTextLine("The ALLDAQ device driver is not properly installed." & Environment.NewLine & ex.Message)
        End Try
    End Sub

    Private Function InitModuleTypesToFind() As List(Of UInteger)
        Dim moduleTypes As List(Of UInteger) = New List(Of UInteger) From {
            AdqConstants.ADQ_MODULE_TYPE_DI,
            AdqConstants.ADQ_MODULE_TYPE_DIO
        }
        Return moduleTypes
    End Function

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()
        For Each diModule As DigitalModule In _diModules
            coBoxModules.Items.Add(diModule.ModuleName)
        Next
    End Sub

    Private Sub EnableUI(enable As Boolean)
        btnReadValue.Enabled = enable
        btnReset.Enabled = enable
    End Sub

    Private Sub BtnReadValue_Click(sender As Object, e As EventArgs) Handles btnReadValue.Click
        Dim value As UInteger = 0
        Dim success As Boolean = _selectedModule.ReadSingle(value)
        If success Then
            DisplayValue(value)
        End If
    End Sub

    Private Sub DisplayValue(value As UInteger)
        AddInfoTextLine(String.Format("Read from Board {0}, Module {1}",
                                      _selectedModule.BoardIndex, _selectedModule.ModuleIndex) _
            & String.Format(", Value hex: &H{0:X2}", value) _
            & ", bin: " & AdqCommon.IntToBin(value, _selectedModule.PortWidthBits) _
            & ", dec: " & value.ToString())
    End Sub

    Private Sub BtnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        Dim success As Boolean = _selectedModule.ResetModule()
        If success Then
            AddInfoTextLine(String.Format("Reset board {0} module {1}",
            _selectedModule.BoardIndex,
            _selectedModule.ModuleIndex))
        End If
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(sender As Object, e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If _selectedModule IsNot Nothing Then
            RemoveHandler _selectedModule.OnModuleInfoTextChanged, AddressOf OnInfoTextChanged
        End If

        EnableUI(coBoxModules.SelectedIndex >= 0)

        If coBoxModules.SelectedIndex >= 0 Then
            _selectedModule = _diModules(coBoxModules.SelectedIndex)
            AddHandler _selectedModule.OnModuleInfoTextChanged, AddressOf OnInfoTextChanged
        Else
            _selectedModule = Nothing
        End If

        ShowModuleDescription(_selectedModule)
    End Sub

    Private Sub ShowModuleDescription(diModule As DigitalModule)
        If diModule IsNot Nothing Then
            textBoxDescription.Text = diModule.Description
        Else
            textBoxDescription.Text = ""
        End If
    End Sub

    Private Sub OnInfoTextChanged(sender As Object, e As InfoTextChangedEventArgs)
        AddInfoTextLine(e.InfoText)
    End Sub

    Private Sub AddInfoTextLine(text As String)
        textBoxInfo.AppendText(String.Format("[{0}] {1}" & Environment.NewLine,
                Date.Now.ToString("HH:mm:ss.fff"), text))
    End Sub

    Private Sub BtnClose_Click(sender As Object, e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub FormMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles MyBase.FormClosing
        CloseAdqDriver()
    End Sub

    Private Sub CloseAdqDriver()
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub BtnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        textBoxInfo.Clear()
    End Sub

    Private Sub BtnInfo_Click(sender As Object, e As EventArgs) Handles btnInfo.Click
        AboutBox.ShowDialog()
    End Sub
End Class


