﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the DOSingle project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.Globalization
Imports DOSingle.ALLDAQ.Boards

Public Class FormMain
    Private _moduleLocator As ModuleLocator
    Private ReadOnly _doModules As List(Of DigitalModule) = New List(Of DigitalModule)
    Private _selectedModule As DigitalModule

    Private Sub FormMain_Shown(sender As Object, e As EventArgs) Handles MyBase.Shown
        Try
            EnableUI(False)

            Dim moduleTypes As List(Of UInteger) = InitModuleTypesToFind()
            _moduleLocator = New ModuleLocator(_doModules, moduleTypes, AddressOf InfoTextChanged)
            _moduleLocator.FindAdqModules()

            FillModuleCombobox()
            If coBoxModules.Items.Count > 0 Then
                coBoxModules.SelectedIndex = 0
            End If
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The ALLDAQ device driver is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Function InitModuleTypesToFind() As List(Of UInteger)
        Dim moduleTypes As List(Of UInteger) = New List(Of UInteger) From {
            AdqConstants.ADQ_MODULE_TYPE_DO,
            AdqConstants.ADQ_MODULE_TYPE_DIO
        }
        Return moduleTypes
    End Function

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()
        For Each diModule As DigitalModule In _doModules
            coBoxModules.Items.Add(diModule.ModuleName)
        Next
    End Sub

    Private Sub CloseAdqDriver()
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub EnableUI(enabled As Boolean)
        btnWriteValue1.Enabled = enabled
        btnWriteValue2.Enabled = enabled
        btnWriteValue3.Enabled = enabled
        btnReset.Enabled = enabled
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(sender As Object, e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If _selectedModule IsNot Nothing Then
            RemoveHandler _selectedModule.OnModuleInfoTextChanged, AddressOf InfoTextChanged
        End If

        EnableUI(coBoxModules.SelectedIndex >= 0)
        If coBoxModules.SelectedIndex >= 0 Then
            _selectedModule = _doModules(coBoxModules.SelectedIndex)
            AddHandler _selectedModule.OnModuleInfoTextChanged, AddressOf InfoTextChanged
            textBoxValue1.Text = String.Format("{0:X}", _selectedModule.OutputValue1)
            textBoxValue2.Text = String.Format("{0:X}", _selectedModule.OutputValue2)
            textBoxValue3.Text = String.Format("{0:X}", _selectedModule.OutputValue3)
        Else
            _selectedModule = Nothing
        End If

        ShowModuleDescription()
    End Sub

    Private Sub ShowModuleDescription()
        If _selectedModule IsNot Nothing Then
            textBoxDescription.Text = _selectedModule.Description
        Else
            textBoxDescription.Text = ""
        End If
    End Sub

    Private Sub BtnWriteValue1_Click(sender As Object, e As EventArgs) Handles btnWriteValue1.Click
        WriteEnteredValue(textBoxValue1, valueNumber:=1)
    End Sub

    Private Sub BtnWriteValue2_Click(sender As Object, e As EventArgs) Handles btnWriteValue2.Click
        WriteEnteredValue(textBoxValue2, valueNumber:=2)
    End Sub

    Private Sub BtnWriteValue3_Click(sender As Object, e As EventArgs) Handles btnWriteValue3.Click
        WriteEnteredValue(textBoxValue3, valueNumber:=3)
    End Sub

    Private Sub WriteEnteredValue(txtBox As TextBox, valueNumber As Integer)
        Try
            Dim valueToWrite As UInteger = ConvertStringValueToUint(txtBox.Text)
            Dim succeeded As Boolean = _selectedModule.WriteSingle(valueToWrite)
            If succeeded Then
                DisplayValue(valueToWrite)
                UpdateModuleOutputValue(valueToWrite, valueNumber)
            End If
        Catch ex As Exception
            MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            txtBox.Focus()
        End Try
    End Sub

    Private Sub UpdateModuleOutputValue(valueToWrite As UInteger, valueNumber As Integer)
        Select Case valueNumber
            Case 1
                _selectedModule.OutputValue1 = valueToWrite
            Case 2
                _selectedModule.OutputValue2 = valueToWrite
            Case 3
                _selectedModule.OutputValue3 = valueToWrite
        End Select
    End Sub

    Private Function ConvertStringValueToUint(valueStr As String) As UInteger
        Dim x As UInteger = UInteger.Parse(valueStr, NumberStyles.AllowHexSpecifier)
        If x > _selectedModule.MaxPortValue Then
            Throw New ArgumentOutOfRangeException(String.Format("Value out of range. Maximum allowed value: &H{0:X}", _selectedModule.MaxPortValue))
        End If
        Return x
    End Function

    Private Sub DisplayValue(ByVal value As UInteger)
        AddInfoTextLine(String.Format("Wrote to Board {0}, Module {1}", _selectedModule.BoardIndex, _selectedModule.ModuleIndex) _
            & String.Format(", Value hex: &H{0:X2}", value) _
            & ", bin: " & AdqCommon.IntToBin(value, _selectedModule.PortWidthBits) _
            & ", dec: " & value.ToString())
    End Sub

    Private Sub BtnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        Dim success As Boolean = _selectedModule.ResetModule()
        If success Then
            AddInfoTextLine(String.Format("Reset board {0} module {1}" & Environment.NewLine,
            _selectedModule.BoardIndex,
            _selectedModule.ModuleIndex))
        End If
    End Sub

    Private Sub InfoTextChanged(sender As Object, e As InfoTextChangedEventArgs)
        AddInfoTextLine(e.InfoText)
    End Sub

    Private Sub AddInfoTextLine(text As String)
        textBoxInfo.AppendText(String.Format("[{0}] {1}" & Environment.NewLine,
                Date.Now.ToString("HH:mm:ss.fff"), text))
    End Sub

    Private Sub BtnClose_Click(sender As Object, e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub FormMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles MyBase.FormClosing
        CloseAdqDriver()
    End Sub

    Private Sub BtnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        textBoxInfo.Clear()
    End Sub

    Private Sub BtnInfo_Click(sender As Object, e As EventArgs) Handles btnInfo.Click
        AboutBox.ShowDialog()
    End Sub
End Class

