﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the FINRead project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.ComponentModel
Imports FINRead.ALLDAQ.Boards

Public Class ModuleLocator
    Implements IDisposable, INotifyPropertyChanged

    Private _disposed As Boolean
    Private _adqOpenError As UInteger = AdqCommon.NotOpen
    Private Event PropertyChanged As PropertyChangedEventHandler _
        Implements INotifyPropertyChanged.PropertyChanged

    Public Sub New(handler As PropertyChangedEventHandler)
        AddHandler PropertyChanged, handler
    End Sub

    Public Sub InitModuleLocator()
        Try
            _adqOpenError = OpenAdqDriver()
            If _adqOpenError <> AdqErrors.ADQ_ERROR_NONE Then Return

            FindFinModules()
            If FinModules.Count = 0 Then
                InfoText = "No FIN modules found."
            End If
        Catch ex As DllNotFoundException
            InfoText = "The device driver is not properly installed." & Environment.NewLine _
                & ex.Message
        End Try
    End Sub

    Private Function OpenAdqDriver() As UInteger
        Dim adqError As UInteger = AdqDriver.adqOpen()
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqOpen", adqError)
        End If
        Return adqError
    End Function

    Private Function GetNumberOfBoards() As UInteger
        Dim numBoards As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetNumberOfBoards(numBoards)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetNumberOfBoards", adqError)
            Return 0
        End If
        Return numBoards
    End Function

    Private Sub FindFinModules()
        Dim numberOfBoards As UInteger = GetNumberOfBoards()
        Dim boardIdx As UInteger = 0

        While boardIdx < numberOfBoards
            Dim numberOfModules As UInteger
            Dim adqError As UInteger = AdqDriver.adqGetNumberOfModules(boardIdx, numberOfModules)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetNumberOfModules", adqError)
                boardIdx += 1UI
                Continue While
            End If

            GetFinModuleTypes(boardIdx, numberOfModules)
            boardIdx += 1UI
        End While
    End Sub

    Private Function GetFinModuleTypes(boardIndex As UInteger, numberOfModules As UInteger) As UInteger
        Dim adqError As UInteger = AdqErrors.ADQ_ERROR_NONE
        Dim moduleIdx As UInteger = 0
        While moduleIdx < numberOfModules
            Dim currentModuleType As UInteger
            adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, currentModuleType)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetModuleType", adqError)
                moduleIdx += 1UI
                Continue While
            End If

            If currentModuleType = AdqConstants.ADQ_MODULE_TYPE_FIN Then
                Dim boardName As String = GetBoardName(boardIndex)
                FinModules.Add(New FinModule(boardName, boardIndex, moduleIdx))
            End If

            moduleIdx += 1UI
        End While
        Return adqError
    End Function

    Public ReadOnly Property FinModules As List(Of FinModule) = New List(Of FinModule)

    Private Function GetBoardName(boardIndex As UInteger) As String
        Dim s As String = String.Empty
        Dim boardInfo = New AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID)

        Dim adqError As UInteger = AdqDriver.adqGetBoardInfo(boardIndex, boardInfo)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            s = boardInfo.BoardName
        End If
        Return s
    End Function

    Private Sub CloseAdqDriver()
        If _adqOpenError = AdqErrors.ADQ_ERROR_NONE Then
            AdqDriver.adqClose()
            _adqOpenError = AdqCommon.NotOpen
        End If
    End Sub

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        InfoText = AdqCommon.GetFunctionErrorText(functionName, adqError)
    End Sub

    Dim _infoText As String = String.Empty
    Public Property InfoText As String
        Get
            Return _infoText
        End Get
        Set(value As String)
            _infoText = value
            NotifyPropertyChanged(NameOf(InfoText))
        End Set
    End Property

    Public Sub NotifyPropertyChanged(ByVal propertyName As String)
        RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub

    Protected Overloads Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
    End Sub

    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not _disposed Then
            _disposed = True
            CloseAdqDriver()
        End If
    End Sub
End Class
