﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the I2CReadWrite project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports I2CReadWrite.ALLDAQ.Boards

Public Class ModuleLocator
    Implements IDisposable, INotifyPropertyChanged

    Private _disposed As Boolean
    Private _adqOpenError As UInteger = AdqCommon.NotOpen

    Private Event PropertyChanged As PropertyChangedEventHandler _
        Implements INotifyPropertyChanged.PropertyChanged

    Public Sub New(handler As PropertyChangedEventHandler)
        AddHandler PropertyChanged, handler
    End Sub

    Public Sub InitModuleLocator()
        Try
            _adqOpenError = OpenAdqDriver()
            If (_adqOpenError <> AdqErrors.ADQ_ERROR_NONE) Then Return

            FindI2CModules()
            If (I2cModules.Count = 0) Then
                InfoText = "No I2C modules found."
            End If
        Catch ex As DllNotFoundException
            InfoText = "The device driver is not properly installed." & Environment.NewLine _
                & ex.Message
        End Try
    End Sub

    Private Function OpenAdqDriver() As UInteger
        Dim adqError As UInteger = AdqDriver.adqOpen()
        If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then
            SetFunctionErrorInfo("adqOpen", adqError)
        End If
        Return adqError
    End Function

    Private Sub FindI2CModules()
        Dim numberOfBoards As UInteger = GetNumberOfBoards()
        Dim boardIdx As UInteger = 0

        While boardIdx < numberOfBoards
            Dim numberOfModules As UInteger
            Dim adqError As UInteger = AdqDriver.adqGetNumberOfModules(boardIdx, numberOfModules)
            If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then
                SetFunctionErrorInfo("adqGetNumberOfModules", adqError)
                boardIdx += 1UI
                Continue While
            End If

            GetModuleTypes(boardIdx, numberOfModules)
            boardIdx += 1UI
        End While
    End Sub

    Private Function GetModuleTypes(boardIndex As UInteger, numberOfModules As UInteger) As UInteger
        Dim adqError As UInteger = AdqErrors.ADQ_ERROR_NONE

        Dim moduleIdx As UInteger = 0
        While moduleIdx < numberOfModules
            Dim currentModuleType As UInteger
            adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, currentModuleType)
            If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then
                SetFunctionErrorInfo("adqGetModuleType", adqError)
                moduleIdx += 1UI
                Continue While
            End If

            If (currentModuleType = AdqConstants.ADQ_MODULE_TYPE_I2C) Then
                GetI2CModuleInfo(boardIndex, moduleIdx)
            End If

            moduleIdx += 1UI
        End While
        Return adqError
    End Function

    Private Function GetI2CModuleInfo(boardIndex As UInteger, moduleIndex As UInteger) As UInteger
        Dim adqGeneralModuleInfo = New AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID)
        Dim adqI2cModuleInfo = New AdqDriver.ADQI2CModuleInfo(AdqDriver.ADQI2CModuleInfo_ID)

        Dim moduleInfo As IntPtr = Marshal.AllocHGlobal(CInt(adqI2cModuleInfo.Size))
        Marshal.StructureToPtr(adqI2cModuleInfo, moduleInfo, False)


        Dim infoSize As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetModuleInfo(
                boardIndex,
                moduleIndex,
                adqGeneralModuleInfo,
                adqI2cModuleInfo.Size,
                moduleInfo,
                 infoSize)
        adqI2cModuleInfo = CType(Marshal.PtrToStructure(moduleInfo, GetType(AdqDriver.ADQI2CModuleInfo)), AdqDriver.ADQI2CModuleInfo)
        Marshal.FreeHGlobal(moduleInfo)

        If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then
            SetFunctionErrorInfo("adqGetModuleInfo", adqError)
            Return adqError
        End If

        Dim boardName As String = GetBoardName(boardIndex)
        Dim i2cModule = New I2cModule(boardName, boardIndex, moduleIndex, adqGeneralModuleInfo, adqI2cModuleInfo)
        I2cModules.Add(i2cModule)
        Return adqError
    End Function

    Public ReadOnly Property I2cModules As List(Of I2cModule) = New List(Of I2cModule)

    Private Function GetNumberOfBoards() As UInteger
        Dim numBoards As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetNumberOfBoards(numBoards)
        If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then
            SetFunctionErrorInfo("adqGetNumberOfBoards", adqError)
            Return 0
        End If
        Return numBoards
    End Function

    Private Function GetBoardName(boardIndex As UInteger) As String
        Dim s As String = String.Empty
        Dim boardInfo = New AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID)

        Dim adqError As UInteger = AdqDriver.adqGetBoardInfo(boardIndex, boardInfo)
        If (adqError = AdqErrors.ADQ_ERROR_NONE) Then
            s = boardInfo.BoardName
        End If
        Return s
    End Function

    Private Sub CloseAdqDriver()
        If (_adqOpenError = AdqErrors.ADQ_ERROR_NONE) Then
            AdqDriver.adqClose()
            _adqOpenError = AdqCommon.NotOpen
        End If
    End Sub

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        InfoText = String.Format("API-function {0} returned Error number: &H{1:X}", functionName, adqError)
    End Sub

    Dim _infoText As String = String.Empty
    Public Property InfoText As String
        Get
            Return _infoText
        End Get
        Set(value As String)
            _infoText = value
            NotifyPropertyChanged(NameOf(InfoText))
        End Set
    End Property

    Public Sub NotifyPropertyChanged(ByVal propertyName As String)
        RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub

    Protected Overloads Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
    End Sub

    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not _disposed Then
            _disposed = True
            CloseAdqDriver()
        End If
    End Sub
End Class
