﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the Interrupt project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.Collections.ObjectModel
Imports System.ComponentModel

Public Class PortStatusViewModel
    Inherits ErrorInfoViewModelBase

    Private _irqModule As IrqModule

    Public Sub UpdateIrqModule(irqModule As IrqModule)
        If _supportsPortStatusError OrElse _supportsFaultRecovery Then
            RemoveHandler _irqModule.PortStatusChanged, AddressOf PortStatusChanged
            RemoveHandler _irqModule.PropertyChanged, AddressOf PortStatusViewModel_PropertyChanged
        End If

        _irqModule = irqModule
        _supportsPortStatusError = irqModule IsNot Nothing AndAlso irqModule.SupportedPortStatusErrors <> 0
        _supportsFaultRecovery = irqModule IsNot Nothing AndAlso irqModule.SupportedFaultRecovery <> 0

        InitPortStatusItemViewModels()
        If _supportsPortStatusError OrElse _supportsFaultRecovery Then
            UpdateEnabledPortStatusErrorsText(irqModule.EnabledPortStatusErrors)
            UpdatePortStatusErrorBitItems(irqModule.EnabledPortStatusErrors)
            UpdateEnabledFaultRecoveryText(irqModule.EnabledFaultRecovery)
            UpdateFaultRecoveryBitItems(irqModule.EnabledFaultRecovery)
            AddHandler _irqModule.PortStatusChanged, AddressOf PortStatusChanged
            AddHandler _irqModule.PropertyChanged, AddressOf PortStatusViewModel_PropertyChanged
        Else
            EnabledPortStatusBits.Clear()
            PortStatusIndicatorBits.Clear()
            UpdateEnabledPortStatusErrorsText(0)
            UpdateEnabledFaultRecoveryText(0)
        End If
        ClearStatusIndicators()
        NotifyPropertyChanged(NameOf(SupportsIrqType))
        UpdateModuleIdleStatus()
    End Sub

    Private Sub PortStatusViewModel_PropertyChanged(sender As Object, e As PropertyChangedEventArgs)
        If e.PropertyName = NameOf(IrqModule.IsModuleIdle) Then
            UpdateModuleIdleStatus()
        End If
    End Sub

    Private Sub UpdateModuleIdleStatus()
        NotifyPropertyChanged(NameOf(CanEditIrqSettings))
        NotifyPropertyChanged(NameOf(CanEditPortStatusErrorSettings))
        NotifyPropertyChanged(NameOf(CanEditFaultRecoverySettings))
    End Sub

    Public ReadOnly Property SupportsIrqType As Boolean
        Get
            Return _supportsPortStatusError OrElse _supportsFaultRecovery
        End Get
    End Property

    Private _supportsPortStatusError As Boolean
    Public ReadOnly Property CanEditPortStatusErrorSettings As Boolean
        Get
            Return _supportsPortStatusError AndAlso _irqModule.IsModuleIdle
        End Get
    End Property

    Private _supportsFaultRecovery As Boolean
    Public ReadOnly Property CanEditFaultRecoverySettings As Boolean
        Get
            Return _supportsFaultRecovery AndAlso _irqModule.IsModuleIdle
        End Get
    End Property

    Public ReadOnly Property CanEditIrqSettings As Boolean
        Get
            Return SupportsIrqType AndAlso _irqModule.IsModuleIdle
        End Get
    End Property

    Private Sub InitPortStatusItemViewModels()
        EnabledPortStatusBits.Clear()
        PortStatusIndicatorBits.Clear()

        Dim portStatusBits As UInteger = _irqModule.SupportedPortStatusErrors Or _irqModule.SupportedFaultRecovery
        Dim bitNumber As Integer = 0
        While bitNumber < 32 AndAlso portStatusBits <> 0
            If (portStatusBits And 1) <> 0 Then
                Dim statusErrorSupported As Boolean = IsBitSet(_irqModule.SupportedPortStatusErrors, bitNumber)
                Dim faultRecoverySupported As Boolean = IsBitSet(_irqModule.SupportedFaultRecovery, bitNumber)

                Dim bitItem As PortStatusItemViewModel = New PortStatusItemViewModel(bitNumber, statusErrorSupported, faultRecoverySupported)
                AddHandler bitItem.PropertyChanged, AddressOf BitItem_PropertyChanged
                EnabledPortStatusBits.Add(bitItem)

                bitItem = New PortStatusItemViewModel(bitNumber, statusErrorSupported, faultRecoverySupported)
                PortStatusIndicatorBits.Add(bitItem)
            End If

            portStatusBits >>= 1
            bitNumber += 1
        End While
    End Sub

    Private Function IsBitSet(value As UInteger, bitNumber As Integer) As Boolean
        Return If(bitNumber < 32, ((value >> bitNumber) And 1) <> 0, False)
    End Function

    Private Sub BitItem_PropertyChanged(sender As Object, e As PropertyChangedEventArgs)
        If e.PropertyName = NameOf(PortStatusItemViewModel.BitStatusError) Then
            UpdatePortStatusErrorsText()
        ElseIf e.PropertyName = NameOf(PortStatusItemViewModel.BitFaultRecovery) Then
            UpdateFaultRecoveryText()
        End If
    End Sub

    Private Sub UpdatePortStatusErrorsText()
        Dim statusValue As UInteger = 0
        Dim bit As UInteger = 1
        For Each enabledPortStatusBit As PortStatusItemViewModel In EnabledPortStatusBits
            If enabledPortStatusBit.BitStatusError Then
                statusValue += enabledPortStatusBit.BinaryValue
            End If
            bit <<= 1
        Next
        _irqModule.EnabledPortStatusErrors = statusValue
        UpdateEnabledPortStatusErrorsText(statusValue)
    End Sub

    Private Sub UpdateEnabledPortStatusErrorsText(statusValue As UInteger)
        _enabledPortStatusErrorsText = AdqCommon.FormatHexValue(statusValue)
        _errorInfo.Remove(NameOf(EnabledPortStatusErrorsText))
        NotifyPropertyChanged(NameOf(EnabledPortStatusErrorsText))
        NotifyPropertyChanged(NameOf(IsInputValid))
    End Sub

    Private Sub UpdateFaultRecoveryText()
        Dim faultRecoveryValue As UInteger = 0
        Dim bit As UInteger = 1
        For Each enabledPortStatusBit As PortStatusItemViewModel In EnabledPortStatusBits
            If enabledPortStatusBit.BitFaultRecovery Then
                faultRecoveryValue += enabledPortStatusBit.BinaryValue
            End If
            bit <<= 1
        Next
        _irqModule.EnabledFaultRecovery = faultRecoveryValue
        UpdateEnabledFaultRecoveryText(faultRecoveryValue)
    End Sub

    Private Sub UpdateEnabledFaultRecoveryText(faultRecoveryValue As UInteger)
        _enabledFaultRecoveryText = AdqCommon.FormatHexValue(faultRecoveryValue)
        _errorInfo.Remove(NameOf(EnabledFaultRecoveryText))
        NotifyPropertyChanged(NameOf(EnabledFaultRecoveryText))
        NotifyPropertyChanged(NameOf(IsInputValid))
    End Sub

    Public ReadOnly Property EnabledPortStatusBits As ObservableCollection(Of PortStatusItemViewModel) =
        New ObservableCollection(Of PortStatusItemViewModel)
    Public ReadOnly Property PortStatusIndicatorBits As ObservableCollection(Of PortStatusItemViewModel) =
        New ObservableCollection(Of PortStatusItemViewModel)

    Public Sub ClearStatusIndicators()
        UpdatePortStatusErrorIndicators(0)
        UpdateFaultRecoveryIndicators(0)
    End Sub

    Private _enabledPortStatusErrorsText As String = "0"
    Public Property EnabledPortStatusErrorsText As String
        Get
            Return _enabledPortStatusErrorsText
        End Get
        Set
            If _enabledPortStatusErrorsText <> Value Then
                _enabledPortStatusErrorsText = Value

                Dim statusErrors As UInteger
                Dim msg As String = InputValidation.ValidateHexValue(_enabledPortStatusErrorsText, statusErrors)

                If Not String.IsNullOrEmpty(msg) Then
                    _errorInfo.AddOrUpdate(NameOf(EnabledPortStatusErrorsText), msg)
                Else
                    msg = _irqModule.ValidateEnabledPortStatusErrorValue(statusErrors)
                    If Not String.IsNullOrEmpty(msg) Then
                        _errorInfo.AddOrUpdate(NameOf(EnabledPortStatusErrorsText), msg)
                    Else
                        _errorInfo.Remove(NameOf(EnabledPortStatusErrorsText))
                        _irqModule.EnabledPortStatusErrors = statusErrors
                        _enabledPortStatusErrorsText = AdqCommon.FormatHexValue(statusErrors)
                        UpdatePortStatusErrorBitItems(statusErrors)
                    End If
                End If

                NotifyPropertyChanged(NameOf(EnabledPortStatusErrorsText))
                NotifyPropertyChanged(NameOf(IsInputValid))
            End If
        End Set
    End Property

    Private _enabledFaultRecoveryText As String = "0"
    Public Property EnabledFaultRecoveryText As String
        Get
            Return _enabledFaultRecoveryText
        End Get
        Set
            If _enabledFaultRecoveryText <> Value Then
                _enabledFaultRecoveryText = Value

                Dim faultRecovery As UInteger
                Dim msg As String = InputValidation.ValidateHexValue(_enabledFaultRecoveryText, faultRecovery)

                If Not String.IsNullOrEmpty(msg) Then
                    _errorInfo.AddOrUpdate(NameOf(EnabledFaultRecoveryText), msg)
                Else
                    msg = _irqModule.ValidateEnabledFaultRecoveryValue(faultRecovery)
                    If Not String.IsNullOrEmpty(msg) Then
                        _errorInfo.AddOrUpdate(NameOf(EnabledFaultRecoveryText), msg)
                    Else
                        _errorInfo.Remove(NameOf(EnabledFaultRecoveryText))
                        _irqModule.EnabledFaultRecovery = faultRecovery
                        _enabledFaultRecoveryText = AdqCommon.FormatHexValue(faultRecovery)
                        UpdateFaultRecoveryBitItems(faultRecovery)
                    End If
                End If

                NotifyPropertyChanged(NameOf(EnabledFaultRecoveryText))
                NotifyPropertyChanged(NameOf(IsInputValid))
            End If
        End Set
    End Property

    Private Sub UpdatePortStatusErrorBitItems(statusError As UInteger)
        For Each enabledPortStatusBit As PortStatusItemViewModel In EnabledPortStatusBits
            enabledPortStatusBit.BitStatusError = ((statusError >> enabledPortStatusBit.BitNumber) And 1) <> 0
        Next
    End Sub

    Private Sub UpdateFaultRecoveryBitItems(faultRecovery As UInteger)
        For Each enabledPortStatusBit As PortStatusItemViewModel In EnabledPortStatusBits
            enabledPortStatusBit.BitFaultRecovery = ((faultRecovery >> enabledPortStatusBit.BitNumber) And 1) <> 0
        Next
    End Sub

    Private Sub PortStatusChanged(sender As Object, e As PortStatusChangedEventArgs)
        UpdatePortStatusErrorIndicators(e.StatusPortErrorValue)
        UpdateFaultRecoveryIndicators(e.StatusFaultRecoveryValue)
    End Sub

    Private Sub UpdatePortStatusErrorIndicators(portStatusValue As UInteger)
        If _portStatusErrorIndicatorValue <> portStatusValue Then
            _portStatusErrorIndicatorValue = portStatusValue
            NotifyPropertyChanged(NameOf(PortStatusErrorIndicatorValue))
            UpdatePortStatusErrorIndicatorBits(portStatusValue)
        End If
    End Sub

    Private Sub UpdateFaultRecoveryIndicators(faultRecoveryValue As UInteger)
        If _faultRecoveryIndicatorValue <> faultRecoveryValue Then
            _faultRecoveryIndicatorValue = faultRecoveryValue
            NotifyPropertyChanged(NameOf(FaultRecoveryIndicatorValue))
            UpdateFaultRecoveryIndicatorBits(faultRecoveryValue)
        End If
    End Sub

    Private _portStatusErrorIndicatorValue As UInteger
    Public ReadOnly Property PortStatusErrorIndicatorValue As UInteger
        Get
            Return _portStatusErrorIndicatorValue
        End Get
    End Property

    Private _faultRecoveryIndicatorValue As UInteger
    Public ReadOnly Property FaultRecoveryIndicatorValue As UInteger
        Get
            Return _faultRecoveryIndicatorValue
        End Get
    End Property

    Private Sub UpdatePortStatusErrorIndicatorBits(portStatusValue As UInteger)
        For Each portStatusIndicatorBit As PortStatusItemViewModel In PortStatusIndicatorBits
            portStatusIndicatorBit.BitStatusError =
                ((portStatusValue >> portStatusIndicatorBit.BitNumber) And 1) <> 0
        Next
    End Sub

    Private Sub UpdateFaultRecoveryIndicatorBits(faultRecoveryValue As UInteger)
        For Each portStatusIndicatorBit As PortStatusItemViewModel In PortStatusIndicatorBits
            portStatusIndicatorBit.BitFaultRecovery =
                ((faultRecoveryValue >> portStatusIndicatorBit.BitNumber) And 1) <> 0
        Next
    End Sub
End Class
