﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the ReadBuffer project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports ReadBuffer.ALLDAQ.Boards
Imports System.Text
Imports System.ComponentModel
Imports System.Windows.Forms.DataVisualization.Charting
Imports System.IO

Public Class FormMain
    Private ReadOnly _gridLineColor As Color = Color.DarkSlateGray
    Private ReadOnly _labelColor As Color = Color.LightSkyBlue
    Private Const _fmtStringPhysValues As String = "{0:0.000000}"
    Private Const _fmtStringDigitalValues As String = "{0}"
    Private Const _idxChartTypeLine As Integer = 0
    Private Const _idxChartTypeStepLine As Integer = 1
    Private Const _idxChartTypeLineAndMarkers As Integer = 2
    Private Const _idxChartTypeStepLineAndMarkers As Integer = 3
    Private Const _idxXAxisLabelsTime As Integer = 0

    Private _moduleLocator As ModuleLocator
    Private ReadOnly _inputModules As List(Of IInputModule) = New List(Of IInputModule)
    Private _selectedAdqModule As IInputModule
    Private _isModuleBusy As Boolean
    Private _isFormClosing As Boolean
    Private _selectedUnitSymbol As String

    Private _oldChartTypeIndex As Integer = -1
    Private _oldNumberOfChartAreas As Integer = 0
    Private _oldChartAlignmentIsAuto As Boolean

    Private Sub FormMain_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        coBoxChartType.SelectedIndex = _idxChartTypeLine
        coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime
        EnableUI(False)
        btnStartStop.Enabled = False
        InitChartXAxisTitle()
        Try
            Dim moduleTypes As List(Of UInteger) = InitModuleTypesToFind()
            _moduleLocator = New ModuleLocator(_inputModules, moduleTypes, AddressOf OnModuleInfoTextChanged)
            _moduleLocator.FindAdqModules()

            FillModuleCombobox()
            SelectFirstComboboxItem(coBoxModules)
            textBoxInfo.AppendText(String.Format("Number of input modules found: {0}" & Environment.NewLine, _inputModules.Count))
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The device driver is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Function InitModuleTypesToFind() As List(Of UInteger)
        Dim moduleTypes = New List(Of UInteger) From {
            AdqConstants.ADQ_MODULE_TYPE_DI,
            AdqConstants.ADQ_MODULE_TYPE_DIO,
            AdqConstants.ADQ_MODULE_TYPE_AI,
            AdqConstants.ADQ_MODULE_TYPE_CNTR
        }
        Return moduleTypes
    End Function

    Private Sub FillModuleCombobox()
        coBoxModules.Items.Clear()
        For Each inpModule As IInputModule In _inputModules
            coBoxModules.Items.Add(inpModule.ModuleName)
        Next
    End Sub

    Private Sub OnModuleInfoTextChanged(sender As Object, e As InfoTextChangedEventArgs)
        textBoxInfo.AppendText(e.InfoText & Environment.NewLine)
    End Sub

    Private Sub DisplayErrorText(functionName As String, adqError As UInteger)
        textBoxInfo.AppendText(
            String.Format("Function {0} returned Error number: &H{1:X}" & Environment.NewLine,
                functionName,
                adqError))
    End Sub

    Private Sub SelectFirstComboboxItem(coBox As ComboBox)
        If coBox.Items.Count > 0 Then
            coBox.SelectedIndex = 0
        End If
    End Sub

    Private Sub CoBoxModules_SelectedIndexChanged(sender As Object, e As EventArgs) Handles coBoxModules.SelectedIndexChanged
        If _selectedAdqModule IsNot Nothing Then
            _selectedAdqModule.ClearValueBuffer()
            RemoveHandler _selectedAdqModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
        End If

        ClearDisplayedValues()
        EnableUI(coBoxModules.SelectedIndex >= 0)
        If coBoxModules.SelectedIndex < 0 Then
            _selectedAdqModule = Nothing
            btnStartStop.Enabled = False
            Return
        End If

        _selectedAdqModule = _inputModules(coBoxModules.SelectedIndex)
        AddHandler _selectedAdqModule.ModuleInfoTextChanged, AddressOf OnModuleInfoTextChanged
        textBoxSampleFrequency.Text = String.Format("{0}", _selectedAdqModule.DesiredSampleFrequency)
        coBoxNumOfValues.Text = _selectedAdqModule.NumberOfValues.ToString()
        UpdateActualFrequency(_selectedAdqModule)
        UpdateTimeToWait(_selectedAdqModule)

        If TypeOf _selectedAdqModule Is AnalogInputModule Then
            TabControl1.SelectedTab = tabPageAnalogInput
            ShowAnalogInputConfiguration(CType(_selectedAdqModule, AnalogInputModule))
        ElseIf TypeOf _selectedAdqModule Is CounterModule Then
            TabControl1.SelectedTab = tabPageCounter
            ShowCounterConfiguration(CType(_selectedAdqModule, CounterModule))
        Else
            TabControl1.SelectedTab = tabPageDigitalInput
        End If
        btnStartStop.Enabled = True
    End Sub

    Private Sub ShowCounterConfiguration(cntrModule As CounterModule)
        FillCounterClockSources(cntrModule)
        coBoxClockSource.SelectedIndex = CInt(cntrModule.SelectedClockSourceIndex)
        coBoxCntrModes.SelectedIndex = CInt(cntrModule.SelectedCounterModeIndex)
        coBoxCntrStartValue.Text = cntrModule.SelectedStartValue.ToString()
    End Sub

    Private Sub FillCounterClockSources(cntrModule As CounterModule)
        coBoxClockSource.Items.Clear()
        For Each item As ClockSourceItem In cntrModule.ClockSourceValues
            coBoxClockSource.Items.Add(item.ClockSourceName)
        Next
    End Sub

    Private Sub ShowAnalogInputConfiguration(aiModule As AnalogInputModule)
        FillChannels(aiModule)
        FillRanges(aiModule)
    End Sub

    Private Sub FillChannels(aiModule As AnalogInputModule)
        If coBoxChannels.Items.Count <> aiModule.NumberOfChannels Then
            coBoxChannels.Items.Clear()
            Dim chanIdx As UInteger = 0
            While chanIdx < aiModule.NumberOfChannels
                coBoxChannels.Items.Add(chanIdx.ToString())
                chanIdx += 1UI
            End While
        End If
        coBoxChannels.SelectedIndex = CInt(aiModule.SelectedChannelIndex)
    End Sub

    Private Sub FillRanges(aiModule As AnalogInputModule)
        coBoxRanges.Items.Clear()
        Dim rangeIdx As Integer = 0
        While rangeIdx < aiModule.NumberOfRanges
            Dim range = aiModule.GetRange(rangeIdx)
            coBoxRanges.Items.Add(range.RangeText)
            rangeIdx += 1
        End While
        coBoxRanges.SelectedIndex = CInt(aiModule.SelectedRangeIndex)
    End Sub

    Private Sub EnableUI(enabled As Boolean)
        coBoxModules.Enabled = enabled
        textBoxSampleFrequency.Enabled = enabled
        coBoxNumOfValues.Enabled = enabled
        btnReset.Enabled = enabled
        chkBoxDisplayTextValues.Enabled = enabled
        coBoxChannels.Enabled = enabled
        coBoxRanges.Enabled = enabled
        coBoxChartType.Enabled = enabled
        coBoxXAxisLabels.Enabled = enabled
        coBoxCntrModes.Enabled = enabled
        coBoxClockSource.Enabled = enabled
        coBoxCntrStartValue.Enabled = enabled
    End Sub

    Private Sub CoBoxNumOfValues_TextChanged(sender As Object, e As EventArgs) Handles coBoxNumOfValues.TextChanged
        If _selectedAdqModule Is Nothing Then Return

        _selectedAdqModule.NumberOfValues = ParseNumberOfValues(coBoxNumOfValues.Text)
        UpdateTimeToWait(_selectedAdqModule)
    End Sub

    Private Sub TxtBoxNominalFrequency_TextChanged(sender As Object, e As EventArgs) Handles textBoxSampleFrequency.TextChanged
        If _selectedAdqModule Is Nothing Then Return

        _selectedAdqModule.DesiredSampleFrequency = ParseFrequency(textBoxSampleFrequency.Text)
        UpdateActualFrequency(_selectedAdqModule)
        UpdateTimeToWait(_selectedAdqModule)
    End Sub

    Private Sub UpdateActualFrequency(adqModule As IInputModule)
        textBoxActualFrequency.Text = AdqCommon.FormatFrequency(adqModule.ActualSampleFrequency)
    End Sub

    Private Sub UpdateTimeToWait(adqModule As IInputModule)
        Dim seconds As Double = adqModule.GetSecondsToWait()
        lblDuration.Text = AdqCommon.FormatTimeToWait(seconds)
    End Sub

    Private Function ParseFrequency(nominalFrequency As String) As Double
        Dim frequency As Double
        Dim isValid As Boolean = Double.TryParse(nominalFrequency, frequency)
        If Not isValid OrElse frequency <= 0 Then
            Return 0.0
        End If

        Return frequency
    End Function

    Private Function ParseNumberOfValues(numberOfValues As String) As UInteger
        Dim numValues As UInteger
        Dim isValid As Boolean = UInteger.TryParse(numberOfValues, numValues)
        If Not isValid Then
            Return 0
        End If

        Return numValues
    End Function

    Private Sub BtnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        _selectedAdqModule.Reset()
    End Sub

    Private Sub BtnStart_Click(sender As Object, e As EventArgs) Handles btnStartStop.Click
        If _isModuleBusy Then
            textBoxInfo.AppendText("Acquisition stopped by user.")
            StopRunningAcquisition()
            btnStartStop.Text = "Start"
            Return
        End If

        UpdateChart(coBoxChartType.SelectedIndex, _selectedAdqModule)
        Try
            _selectedAdqModule.ValidateConfigurationValues()
            Dim adqError As UInteger = ConfigureModule(_selectedAdqModule)
            If (adqError <> AdqErrors.ADQ_ERROR_NONE) Then Return

            EnableUI(False)
            btnStartStop.Text = "Stop"
            _isModuleBusy = True
            _selectedAdqModule.InitValueBuffer()
            _selectedAdqModule.AdqError = AdqErrors.ADQ_ERROR_NONE
            BackgroundWorker.RunWorkerAsync(_selectedAdqModule)
        Catch Ex As Exception
            _isModuleBusy = False
            EnableUI(True)
            btnStartStop.Text = "Start"
            MessageBox.Show(Ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try
    End Sub

    Private Function ConfigureModule(adqModule As IInputModule) As UInteger
        Dim adqError As UInteger = adqModule.SetSampleFrequency()
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then Return adqError

        If (TypeOf adqModule Is AnalogInputModule) Then
            Return ConfigureAiModule(CType(adqModule, AnalogInputModule))
        ElseIf (TypeOf adqModule Is DigitalInputModule) Then
            Return ConfigureDiModule(CType(adqModule, DigitalInputModule))
        ElseIf (TypeOf adqModule Is CounterModule) Then
            Return ConfigureCounterModule(CType(adqModule, CounterModule))
        End If
        Return AdqErrors.ADQ_ERROR_UNSUPPORTED
    End Function

    Private Function ConfigureAiModule(aiModule As AnalogInputModule) As UInteger
        Return aiModule.ConfigureModule(CUInt(coBoxChannels.SelectedIndex), CUInt(coBoxRanges.SelectedIndex))
    End Function

    Private Function ConfigureDiModule(diModule As DigitalInputModule) As UInteger
        Return diModule.ConfigureModule()
    End Function

    Private Function ConfigureCounterModule(cntrModule As CounterModule) As UInteger
        Dim startValue As UInteger = ParseNumberOfValues(coBoxCntrStartValue.Text)
        Return cntrModule.ConfigureModule(coBoxClockSource.SelectedIndex, coBoxCntrModes.SelectedIndex, startValue)
    End Function

    Private Sub DisplayValues(adqModule As IInputModule)
        If TypeOf adqModule Is DigitalInputModule Then
            DisplayDigitalValuesAsText(adqModule)
            DisplayBitValuesAsChart(CType(adqModule, DigitalInputModule))
        ElseIf TypeOf adqModule Is CounterModule Then
            DisplayDigitalValuesAsText(adqModule)
            DisplayDigitalValuesAsChart(adqModule)
        ElseIf TypeOf adqModule Is AnalogInputModule Then
            DisplayAnalogValuesAsText(CType(adqModule, AnalogInputModule))
            DisplayAnalogValuesAsChart(CType(adqModule, AnalogInputModule))
        End If
    End Sub

    Private Sub DisplayDigitalValuesAsText(adqModule As IInputModule)
        If Not chkBoxDisplayTextValues.Checked Then Return

        UseWaitCursor = True
        Try
            Dim time As Double = 0.0
            Dim sb = New StringBuilder()
            Dim valIdx As Integer = 0
            While valIdx < adqModule.NumberOfValuesRead
                sb.AppendLine(String.Format("{0:000}, {1:0.0000000}sec Digital: {2} --> &H{2:X8}",
                                            valIdx, time, adqModule.Values(valIdx)))
                time += adqModule.SampleTime
                valIdx += 1
            End While
            textBoxInfo.AppendText(sb.ToString())
        Finally
            UseWaitCursor = False
        End Try
    End Sub

    Private Sub DisplayAnalogValuesAsText(adqModule As AnalogInputModule)
        If Not chkBoxDisplayTextValues.Checked Then Return

        UseWaitCursor = True
        Try
            Dim aiModule As AnalogInputModule = adqModule
            Dim time As Double = 0.0
            Dim sb = New StringBuilder()
            Dim valIdx As Integer = 0
            While valIdx < adqModule.NumberOfValuesRead
                Dim physicalValue As Double = aiModule.SelectedRange.DigitalToPhysical(adqModule.Values(valIdx))
                sb.AppendLine(String.Format("{0:000}, {1:0.0000000}sec Digital: {2} - Physical {3:0.00000000}{4}",
                                            valIdx,
                                            time,
                                            adqModule.Values(valIdx),
                                            physicalValue,
                                            aiModule.SelectedRange.UnitSymbol))
                time += adqModule.SampleTime
                valIdx += 1
            End While
            textBoxInfo.AppendText(sb.ToString())
        Finally
            UseWaitCursor = False
        End Try
    End Sub

    Private Sub InitChartXAxisTitle()
        Chart1.Titles.Clear()
        Dim chTitle As Title = New Title(coBoxXAxisLabels.Text)
        chTitle.Position.Auto = False
        chTitle.Position.X = 3
        chTitle.Position.Y = 95
        chTitle.Position.Width = 94
        chTitle.Position.Height = 5
        chTitle.Alignment = ContentAlignment.BottomCenter
        chTitle.Docking = Docking.Bottom
        chTitle.IsDockedInsideChartArea = False
        chTitle.ForeColor = _labelColor
        Chart1.Titles.Add(chTitle)
    End Sub

    Private Sub UpdateChart(chartTypeIndex As Integer, adqModule As IInputModule)
        If TypeOf adqModule Is DigitalInputModule Then
            Dim diModule As DigitalInputModule = CType(adqModule, DigitalInputModule)
            AdjustChartAreasAndSeries(chartTypeIndex, CInt(diModule.PortWidthBits))
        Else
            AdjustChartAreasAndSeries(chartTypeIndex, 1)
        End If
    End Sub

    Private Sub AdjustChartAreasAndSeries(chartTypeIndex As Integer, numberOfChartAreas As Integer)
        If _oldChartAlignmentIsAuto <> rbAlignAuto.Checked Then
            _oldChartAlignmentIsAuto = rbAlignAuto.Checked
            _oldNumberOfChartAreas = 0
        End If

        If _oldNumberOfChartAreas <> numberOfChartAreas Then
            _oldChartTypeIndex = -1
            _oldNumberOfChartAreas = numberOfChartAreas
            InitChartAreas(numberOfChartAreas)
        End If

        If _oldChartTypeIndex <> chartTypeIndex Then
            _oldChartTypeIndex = chartTypeIndex
            textBoxInfo.Clear()
            InitSeries(chartTypeIndex, numberOfChartAreas)
        Else
            ClearDisplayedValues()
        End If
    End Sub

    Private Sub InitChartAreas(numberOfChartAreas As Integer)
        Chart1.BackColor = Color.Black
        Chart1.ChartAreas.Clear()

        If numberOfChartAreas > 1 Then
            Dim areaHeight As Single = (100.0F / numberOfChartAreas)

            Dim areaPositionY As Single = 0
            For i As Integer = 0 To numberOfChartAreas - 1
                Dim chartArea = New ChartArea("ChartArea" + i.ToString())
                InitChartArea(chartArea)

                chartArea.Position.Auto = rbAlignAuto.Checked
                If rbAlignVertical.Checked Then
                    chartArea.Position.X = 0
                    chartArea.Position.Width = 99
                    chartArea.Position.Y = areaPositionY
                    chartArea.Position.Height = areaHeight
                    areaPositionY += areaHeight
                End If
                chartArea.AxisY.Minimum = -0.05
                chartArea.AxisY.Maximum = 1.05
                chartArea.AxisY.Interval = 1
                chartArea.AxisY.IntervalOffset = 0.05
                Chart1.ChartAreas.Add(chartArea)
            Next
        Else
            Dim chartArea = New ChartArea("ChartArea0")
            InitChartArea(chartArea)
            chartArea.Position.Auto = False
            chartArea.Position.X = 0
            chartArea.Position.Width = 99
            chartArea.Position.Y = 0.25
            chartArea.Position.Height = 99.5
            chartArea.AxisY.Maximum = Double.NaN
            chartArea.AxisY.Minimum = Double.NaN

            Chart1.ChartAreas.Add(chartArea)
        End If
    End Sub

    Private Sub InitChartArea(chartArea As ChartArea)
        chartArea.BackColor = Color.Black

        chartArea.AxisX.MajorGrid.LineColor = _gridLineColor
        chartArea.AxisX.MajorTickMark.LineColor = _gridLineColor
        chartArea.AxisX.MinorGrid.LineColor = _gridLineColor
        chartArea.AxisX.MinorTickMark.LineColor = _gridLineColor
        chartArea.AxisX.LineColor = _gridLineColor
        chartArea.AxisX.LineWidth = 1
        chartArea.AxisX.LabelStyle.ForeColor = _labelColor
        chartArea.AxisX.TitleForeColor = _labelColor
        chartArea.AxisX.ScrollBar.ButtonColor = Color.DarkGray
        chartArea.AxisX.ScrollBar.BackColor = Color.Gray
        chartArea.CursorX.IsUserSelectionEnabled = True

        chartArea.AxisY.MajorGrid.LineColor = _gridLineColor
        chartArea.AxisY.MajorTickMark.LineColor = _gridLineColor
        chartArea.AxisY.MinorGrid.LineColor = _gridLineColor
        chartArea.AxisY.MinorTickMark.LineColor = _gridLineColor
        chartArea.AxisY.LineColor = _gridLineColor
        chartArea.AxisY.LineWidth = 1
        chartArea.AxisY.LabelStyle.ForeColor = _labelColor
        chartArea.AxisY.TitleForeColor = _labelColor
        chartArea.AxisY.ScrollBar.ButtonColor = Color.DarkGray
        chartArea.AxisY.ScrollBar.ButtonColor = Color.DarkGray
        chartArea.AxisY.ScrollBar.BackColor = Color.Gray
        chartArea.CursorY.IsUserSelectionEnabled = True
    End Sub

    Private Sub InitSeries(chartTypeIndex As Integer, numberOfChartAreas As Integer)
        RemoveHandler Chart1.GetToolTipText, AddressOf Chart1_GetToolTipText
        Chart1.Series.Clear()
        Chart1.Legends.Clear()

        For i As Integer = 0 To numberOfChartAreas - 1
            Dim chartSeries = New Series("Series" + i.ToString())
            If (chartTypeIndex = _idxChartTypeLine) Then
                chartSeries.ChartType = SeriesChartType.Line
            ElseIf (chartTypeIndex = _idxChartTypeStepLine) Then
                chartSeries.ChartType = SeriesChartType.StepLine
            ElseIf (chartTypeIndex = _idxChartTypeLineAndMarkers) Then
                chartSeries.ChartType = SeriesChartType.Line
                chartSeries.MarkerStyle = MarkerStyle.Circle
            ElseIf (chartTypeIndex = _idxChartTypeStepLineAndMarkers) Then
                chartSeries.ChartType = SeriesChartType.StepLine
                chartSeries.MarkerStyle = MarkerStyle.Circle
            Else
                chartSeries.ChartType = SeriesChartType.Point
            End If

            chartSeries.XValueType = ChartValueType.UInt32
            chartSeries.IsXValueIndexed = False
            chartSeries.ChartArea = Chart1.ChartAreas(i).Name

            Dim Legend = New Legend("Legend" + i.ToString()) With {
                .ForeColor = Color.BlanchedAlmond,
                .BackColor = Color.Transparent,
                .DockedToChartArea = chartSeries.ChartArea
            }
            Chart1.Legends.Add(Legend)
            chartSeries.Legend = Legend.Name
            Chart1.Series.Add(chartSeries)
        Next
        AddHandler Chart1.GetToolTipText, AddressOf Chart1_GetToolTipText
    End Sub

    Private Sub Chart1_GetToolTipText(sender As Object, e As ToolTipEventArgs)
        If e.HitTestResult.ChartElementType = ChartElementType.DataPoint Then
            Dim series As Series = e.HitTestResult.Series
            If series IsNot Nothing Then
                Dim dp As DataPoint = series.Points(e.HitTestResult.PointIndex)
                e.Text = FormatTooltipText(dp)
            End If
        ElseIf e.HitTestResult.ChartElementType = ChartElementType.Axis OrElse
            e.HitTestResult.ChartElementType = ChartElementType.TickMarks Then
            If TypeOf sender Is Chart Then
                Dim result As HitTestResult = CType(sender, Chart).HitTest(e.X, e.Y, ChartElementType.DataPoint)
                If result.ChartElementType = ChartElementType.DataPoint Then
                    Dim dp As DataPoint = result.Series.Points(result.PointIndex)
                    e.Text = FormatTooltipText(dp)
                End If
            End If
        End If
    End Sub

    Private Function FormatTooltipText(dp As DataPoint) As String
        If TypeOf _selectedAdqModule Is AnalogInputModule Then
            Return String.Format("X={0}, Y={1}",
                        dp.XValue,
                        AdqCommon.FormatPhysicalValue(_fmtStringPhysValues, dp.YValues(0)) & _selectedUnitSymbol)
        End If
        Return String.Format("X={0}, Y={1}",
                            dp.XValue, dp.YValues(0))
    End Function

    Private Sub DisplayBitValuesAsChart(diModule As DigitalInputModule)
        UseWaitCursor = True
        Try
            Dim points0 As DataPointCollection = Nothing
            Chart1.Titles(0).Text = coBoxXAxisLabels.Text
            Dim bit As Integer = 0
            While bit < diModule.PortWidthBits
                Dim chSeries As Series = Chart1.Series(bit)
                Dim ca As ChartArea = Chart1.ChartAreas(bit)
                UpdateChartArea(ca, 0.25, String.Empty, _fmtStringDigitalValues)
                UpdateLegendText(chSeries, "Bit " + bit.ToString())
                chSeries.Points.SuspendUpdates()
                Dim xValue As Double = 0.0
                Dim xValueDelta As Double = GetXValueDelta()
                Dim valueIdx As Integer = 0
                While valueIdx < diModule.NumberOfValuesRead
                    Dim dp = New DataPoint(valueIdx, (diModule.Values(valueIdx) >> bit) And 1)
                    If points0 Is Nothing Then
                        dp.AxisLabel = GetXAsixLabel(xValue)
                        xValue += xValueDelta
                    Else
                        dp.AxisLabel = points0(valueIdx).AxisLabel
                    End If
                    chSeries.Points.Add(dp)
                    valueIdx += 1
                End While
                If points0 Is Nothing Then
                    points0 = chSeries.Points
                End If
                chSeries.Points.ResumeUpdates()
                bit += 1
            End While
        Finally
            UseWaitCursor = False
        End Try
    End Sub

    Private Sub DisplayDigitalValuesAsChart(adqModule As IInputModule)
        UseWaitCursor = True
        Try
            Dim chSeries As Series = Chart1.Series(0)
            Dim ca As ChartArea = Chart1.ChartAreas(0)
            Chart1.Titles(0).Text = coBoxXAxisLabels.Text
            UpdateChartArea(ca, 0.25, "Digital value", _fmtStringDigitalValues)
            UpdateLegendText(chSeries, "Module " + adqModule.ModuleIndex.ToString())
            Dim xValue As Double = 0.0
            Dim xValueDelta As Double = GetXValueDelta()
            chSeries.Points.SuspendUpdates()
            Dim valIdx As Integer = 0
            While valIdx < adqModule.NumberOfValuesRead
                Dim dp = New DataPoint(valIdx, adqModule.Values(valIdx)) With {
                    .AxisLabel = GetXAsixLabel(xValue)
                }
                chSeries.Points.Add(dp)
                xValue += xValueDelta
                valIdx += 1
            End While
            ca.RecalculateAxesScale()
            chSeries.Points.ResumeUpdates()
        Finally
            UseWaitCursor = False
        End Try
    End Sub

    Private Sub DisplayAnalogValuesAsChart(aiModule As AnalogInputModule)
        UseWaitCursor = True
        Try
            Dim chSeries As Series = Chart1.Series(0)
            Dim ca As ChartArea = Chart1.ChartAreas(0)
            Chart1.Titles(0).Text = coBoxXAxisLabels.Text
            Dim range = aiModule.SelectedRange
            _selectedUnitSymbol = range.UnitSymbol
            Dim scrollSize As Double = range.PhysicalResolution / 4.0
            UpdateChartArea(ca, scrollSize, range.UnitName, _fmtStringPhysValues)
            UpdateLegendText(chSeries, "Channel " + aiModule.SelectedChannelIndex.ToString())
            Dim xValue As Double = 0.0
            Dim xValueDelta As Double = GetXValueDelta()
            chSeries.Points.SuspendUpdates()
            Dim valIdx As Integer = 0
            While valIdx < aiModule.NumberOfValuesRead
                Dim dp = New DataPoint(valIdx, range.DigitalToPhysical(aiModule.Values(valIdx))) With {
                    .AxisLabel = GetXAsixLabel(xValue)
                }
                chSeries.Points.Add(dp)
                xValue += xValueDelta
                valIdx += 1
            End While
            ca.RecalculateAxesScale()
            chSeries.Points.ResumeUpdates()
        Finally
            UseWaitCursor = False
        End Try
    End Sub


    Private Sub UpdateChartArea(ca As ChartArea, yScrollSize As Double, yAxisTitle As String, yAxisLabelFormat As String)
        ca.CursorY.Interval = yScrollSize
        ca.AxisY.ScaleView.SmallScrollMinSize = yScrollSize
        ca.AxisY.ScaleView.SmallScrollSize = yScrollSize
        ca.AxisY.Title = yAxisTitle
        ca.AxisY.LabelStyle.Format = yAxisLabelFormat
    End Sub

    Private Sub UpdateLegendText(chSeries As Series, legendText As String)
        chSeries.LegendText = legendText
        chSeries.IsVisibleInLegend = True
    End Sub

    Private Function GetXAsixLabel(xValue As Double) As String
        If coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime Then
            Return AdqCommon.FormatSampleTime(xValue)
        End If
        Return (xValue.ToString())
    End Function

    Private Function GetXValueDelta() As Double
        Return If(coBoxXAxisLabels.SelectedIndex = _idxXAxisLabelsTime, _selectedAdqModule.SampleTime, 1.0)
    End Function

    Private Sub RemoveAllSeriesPoints()
        For i As Integer = 0 To Chart1.Series.Count - 1
            RemovePoints(Chart1.Series(i))
        Next
    End Sub

    Private Sub RemovePoints(chartSeries As Series)
        chartSeries.IsVisibleInLegend = False
        chartSeries.Points.SuspendUpdates()
        While chartSeries.Points.Count > 0
            chartSeries.Points.RemoveAt(chartSeries.Points.Count - 1)
        End While
        chartSeries.Points.ResumeUpdates()
        chartSeries.Points.Clear()
    End Sub

    Private Sub BackgroundWorker_DoWork(sender As Object,
                                       e As DoWorkEventArgs
                                       ) Handles BackgroundWorker.DoWork
        Dim adqModule As IInputModule = CType(e.Argument, IInputModule)
        adqModule.ReadValuesFromModule()
        e.Result = adqModule
    End Sub

    Private Sub BackgroundWorker_RunWorkerCompleted(sender As Object,
                                                   e As RunWorkerCompletedEventArgs
                                                   ) Handles BackgroundWorker.RunWorkerCompleted
        If _isFormClosing Then Return

        Dim adqModule As IInputModule = CType(e.Result, IInputModule)
        _isModuleBusy = False
        If adqModule.AdqError = AdqErrors.ADQ_ERROR_NONE Then
            DisplayValues(adqModule)
        Else
            DisplayErrorText("adqReadBuffer", adqModule.AdqError)
        End If
        EnableUI(True)
        btnStartStop.Text = "Start"
    End Sub

    Private Sub StopRunningAcquisition()
        Dim isRunning As Boolean
        _selectedAdqModule.GetModuleStatus(isRunning)
        If isRunning Then
            _selectedAdqModule.StopAcquisition()
        End If
    End Sub

    Private Sub BtnClose_Click(sender As Object, e As EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub FormMain_FormClosed(sender As Object, e As FormClosedEventArgs) Handles MyBase.FormClosed
        CloseAdqDriver()
    End Sub

    Private Sub CloseAdqDriver()
        If _moduleLocator IsNot Nothing Then
            _moduleLocator.CloseAdqDriver()
        End If
    End Sub

    Private Sub BtnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        ClearDisplayedValues()
    End Sub

    Private Sub ClearDisplayedValues()
        RemoveAllSeriesPoints()
        textBoxInfo.Clear()
    End Sub

    Private Sub TextBoxInfo_KeyDown(sender As Object, e As KeyEventArgs) Handles textBoxInfo.KeyDown
        If e.Modifiers = Keys.Control AndAlso e.KeyCode = Keys.A Then
            textBoxInfo.SelectionStart = 0
            textBoxInfo.SelectionLength = textBoxInfo.Text.Length
        End If
    End Sub

    Private Sub FormMain_FormClosing(sender As Object, e As FormClosingEventArgs) Handles MyBase.FormClosing
        If _isModuleBusy Then
            Dim result As DialogResult
            result = MessageBox.Show("The acquisition is still running." _
                                     & Environment.NewLine _
                                     & "Do you want to close the program?",
                                     "Module busy",
                                     MessageBoxButtons.YesNo, MessageBoxIcon.Question)
            If result = DialogResult.Yes Then
                _isFormClosing = True
                StopRunningAcquisition()
                e.Cancel = False
            Else
                e.Cancel = True
            End If
        End If
    End Sub

    Private Sub TsMenuItemSaveGraphic_Click(sender As Object, e As EventArgs) Handles tsMenuItemSaveGraphic.Click
        SaveFileDialog1.FileName = GetImageFileName(_selectedAdqModule, ".png")
        SaveFileDialog1.DefaultExt = "png"
        SaveFileDialog1.ShowDialog()
    End Sub

    Private Function GetImageFileName(adqModule As IInputModule, fileExt As String) As String
        Dim s As String = String.Empty
        If adqModule IsNot Nothing Then
            s = String.Format("{0}, Board {1}, Module {2}, ",
                    adqModule.ModelName,
                    adqModule.BoardIndex,
                    adqModule.ModuleIndex)
        End If
        Return s & Date.Now.ToString("yyyy-MM-dd HH.mm.ss") & fileExt
    End Function

    Private Function GetImageFormat(filterIndex As Integer) As ChartImageFormat
        Dim imgFormat As ChartImageFormat

        Select Case filterIndex
            Case 1
                imgFormat = ChartImageFormat.Png
            Case 2
                imgFormat = ChartImageFormat.Gif
            Case 3
                imgFormat = ChartImageFormat.Bmp
            Case 4
                imgFormat = ChartImageFormat.Tiff
            Case 5
                imgFormat = ChartImageFormat.Jpeg
            Case Else
                imgFormat = ChartImageFormat.Png
        End Select
        Return imgFormat
    End Function

    Private Sub SaveFileDialog1_FileOk(sender As Object, e As CancelEventArgs) Handles SaveFileDialog1.FileOk
        Dim imgFormat As ChartImageFormat = GetImageFormat(SaveFileDialog1.FilterIndex)
        Chart1.SaveImage(SaveFileDialog1.FileName, imgFormat)
    End Sub

    Private Sub TsMnuItemCopyToClipboard_Click(sender As Object, e As EventArgs) Handles tsMnuItemCopyToClipboard.Click
        Using ms As MemoryStream = New MemoryStream()
            Chart1.SaveImage(ms, ChartImageFormat.Png)
            Dim Bitmap = New Bitmap(ms)
            Clipboard.SetImage(Bitmap)
        End Using
    End Sub

    Private Sub BtnInfo_Click(sender As Object, e As EventArgs) Handles btnInfo.Click
        AboutBox.ShowDialog()
    End Sub
End Class

