﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the WriteBufferAO project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports WriteBufferAO.ALLDAQ.Boards

Public Class AoModule
    Implements INotifyPropertyChanged

    Private ReadOnly _defaultRepeatBufferCount As UInteger = 2
    Private _values As UInteger()

    Public Event PropertyChanged As PropertyChangedEventHandler _
        Implements INotifyPropertyChanged.PropertyChanged

    Public Sub New(boardIndex As UInteger, moduleIndex As UInteger,
                   ByRef generalInfo As AdqDriver.ADQGeneralModuleInfo, ByRef aoModuleInfo As AdqDriver.ADQAOModuleInfo)
        Me.BoardIndex = boardIndex
        Me.ModuleIndex = moduleIndex
        MaxFrequency = generalInfo.ClockBaseFrequencyHz / generalInfo.ClockMinimumDivider
        DefaultOutputFrequency = MaxFrequency / 100
        _repeatBufferCount = _defaultRepeatBufferCount
        AoRanges = New List(Of AoRange)
        ModuleName = String.Format("Module {0}, max. {1}", moduleIndex, AdqCommon.FormatFrequency(MaxFrequency))
        AddRanges(aoModuleInfo)
        DesiredOutputFrequency = DefaultOutputFrequency
    End Sub

    Private Sub AddRanges(ByRef aoModuleInfo As AdqDriver.ADQAOModuleInfo)
        Dim rangeIdx As Integer = 0
        While rangeIdx < aoModuleInfo.NumberOfRanges
            Dim adqRange As AdqDriver.ADQAORange = aoModuleInfo.Ranges(rangeIdx)
            Dim aoRange = New AoRange(adqRange.DigitalMin,
                    adqRange.DigitalMax,
                    adqRange.PhysicalUnit,
                    adqRange.PhysicalMin,
                    adqRange.PhysicalMax)
            AoRanges.Add(aoRange)
            rangeIdx += 1
        End While

        If aoModuleInfo.NumberOfRanges >= 1 Then
            SelectedAoRange = AoRanges(0)
        End If
    End Sub

    Public Property SelectedAoRange As AoRange
    Public ReadOnly Property AoRanges As List(Of AoRange)
    Public ReadOnly Property BoardIndex As UInteger
    Public ReadOnly Property ModuleIndex As UInteger
    Public ReadOnly Property ModuleName As String
    Public ReadOnly Property MaxFrequency As Double
    Public ReadOnly Property DefaultOutputFrequency As Double

    Private _desiredOutputFrequency As Double
    Public Property DesiredOutputFrequency As Double
        Get
            Return _desiredOutputFrequency
        End Get
        Set
            If _desiredOutputFrequency <> Value Then
                _desiredOutputFrequency = Value
                If Value > 0 AndAlso Value <= MaxFrequency Then
                    SetActualOutputFrequency(Value)
                End If
            End If
        End Set
    End Property

    Private Sub SetActualOutputFrequency(value As Double)
        Dim actualFrequency As Double
        Dim desiredFrequency As Double = value
        Dim adqError As UInteger = AdqDriver.adqSetFrequency(BoardIndex, ModuleIndex, desiredFrequency, actualFrequency)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            _actualOutputFrequency = actualFrequency
            RaisePropertyChanged(NameOf(ActualOutputFrequency))
        Else
            SetErrorText("adqSetFrequency", adqError)
        End If
    End Sub

    Private _actualOutputFrequency As Double
    Public ReadOnly Property ActualOutputFrequency As Double
        Get
            Return _actualOutputFrequency
        End Get
    End Property

    Private _repeatBufferCount As UInteger
    Public Property RepeatBufferCount As UInteger
        Get
            Return _repeatBufferCount
        End Get
        Set
            If _repeatBufferCount <> Value Then
                _repeatBufferCount = Value
                RaisePropertyChanged(NameOf(RepeatBufferCount))
            End If
        End Set
    End Property

    Public Property AdqError As UInteger
    Public Property NumberOfValuesWritten As UInteger

    Public Sub SetWaveForm(waveForm As WaveForm)
        _values = PhysicalToDigital(waveForm)
    End Sub

    Public Sub WriteBuffer()
        NumberOfValuesWritten = 0
        Dim numValuesWritten As UInteger
        Dim numberOfValues As UInteger = CUInt(_values.Length)
        Dim gch As GCHandle = GCHandle.Alloc(_values, GCHandleType.Pinned)
        AdqError = AdqDriver.adqWriteBuffer(BoardIndex, ModuleIndex,
                numberOfValues, _values, _repeatBufferCount, numValuesWritten)
        NumberOfValuesWritten = numValuesWritten
        gch.Free()
    End Sub

    Private Function PhysicalToDigital(waveForm As WaveForm) As UInteger()
        Dim values As UInteger() = _values
        If (values Is Nothing) OrElse (values.Length <> waveForm.Values.Length) Then
            values = New UInteger(waveForm.Values.Length - 1) {}
        End If

        For i = 0 To values.Length - 1
            values(i) = SelectedAoRange.PhysicalToDigital(waveForm.Values(i))
        Next
        Return values
    End Function

    Public Function GetModuleStatus() As Boolean
        Dim numAccept As UInteger
        Dim isModuleRunning As Boolean
        Dim adqError As UInteger = AdqDriver.adqWriteContinuousStatus(BoardIndex, ModuleIndex, False, numAccept, isModuleRunning)
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetErrorText("adqWriteContinuousStatus", adqError)
        End If
        Return isModuleRunning
    End Function

    Public Sub StopOutput()
        Dim adqError As UInteger = AdqDriver.adqStop(BoardIndex, ModuleIndex)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            SetInfoText("stopped")
        Else
            SetErrorText("adqStop", adqError)
        End If
    End Sub

    Public Function ValidateRepeatBufferCount(count As UInteger) As String
        Dim msg As String = String.Empty
        If count = 0 Then
            msg = "The number must be greater than 0"
        End If
        Return msg
    End Function

    Public Function ValidateDesiredOutputFrequency(frequency As Double) As String
        Dim msg As String = String.Empty
        If frequency <= 0 Then
            msg = "The output frequency must be greater than 0"
        ElseIf frequency > MaxFrequency Then
            msg = "The output frequency cannot be greater than the maximum output frequency"
        End If
        Return msg
    End Function

    Private Sub SetErrorText(functionName As String, adqError As UInteger)
        ModuleErrorText = String.Format("Board {0}, Module {1}: Function {2} returned error number: 0x{3:X}",
                BoardIndex,
                ModuleIndex,
                functionName,
                adqError)
    End Sub

    Private _errorText As String = String.Empty
    Public Property ModuleErrorText As String
        Get
            Return _errorText
        End Get
        Set
            _errorText = Value
            If Value <> String.Empty Then
                RaisePropertyChanged(NameOf(ModuleErrorText))
            End If
        End Set
    End Property
    Private Sub SetInfoText(info As String)
        ModuleInfoText = String.Format("Board {0} Module {1}: {2}",
                BoardIndex,
                ModuleIndex,
                info)
    End Sub

    Private _infoText As String = String.Empty
    Public Property ModuleInfoText As String
        Get
            Return _infoText
        End Get
        Set
            _infoText = Value
            If (Value <> String.Empty) Then
                RaisePropertyChanged(NameOf(ModuleInfoText))
            End If
        End Set
    End Property

    Public Sub RaisePropertyChanged(ByVal propertyName As String)
        RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub
End Class
