﻿'-----------------------------------------------------------------------------
' ALLDAQ, a division of ALLNET GmbH
' https://www.alldaq.com
'
' Purpose:
'       This file Is part of the WriteBufferAO project of the ALLDAQ-SDK.  
'       It shows the usage of the API functions to access
'       ALLDAQ Data Acquisition Boards
' 
'-----------------------------------------------------------------------------

Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports WriteBufferAO.ALLDAQ.Boards

Public Class BoardLocator
    Implements IDisposable, INotifyPropertyChanged

    Private _disposed As Boolean = False
    Private _adqOpenError As UInteger = AdqCommon.NotOpen

    Private Event PropertyChanged As PropertyChangedEventHandler _
        Implements INotifyPropertyChanged.PropertyChanged

    Public Sub New(handler As PropertyChangedEventHandler)
        AddHandler PropertyChanged, handler
    End Sub

    Public Sub InitBoardLocator()
        Try
            _adqOpenError = OpenAdqDriver()
            If _adqOpenError <> AdqErrors.ADQ_ERROR_NONE Then
                Return
            End If

            _adqAoBoards = New List(Of AdqAoBoard)
            FindAdqBoards(AdqConstants.ADQ_MODULE_TYPE_AO)
        Catch ex As DllNotFoundException
            InfoText = "The device driver is not properly installed." & Environment.NewLine _
                & ex.Message
        End Try
    End Sub

    Private Sub FindAdqBoards(moduleType As UInteger)
        Dim numberOfBoards As UInteger = GetNumberOfBoards()
        Dim boardIdx As UInteger = 0

        While boardIdx < numberOfBoards
            Dim numberOfModules As UInteger
            Dim adqError As UInteger = AdqDriver.adqGetNumberOfModules(boardIdx, numberOfModules)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetNumberOfModules", adqError)
                boardIdx += 1UI
                Continue While
            End If

            GetModules(boardIdx, numberOfModules, moduleType)
            boardIdx += 1UI
        End While
    End Sub

    Private Function GetNumberOfBoards() As UInteger
        Dim numBoards As UInteger
        Dim adqError As UInteger = AdqDriver.adqGetNumberOfBoards(numBoards)
        If _adqOpenError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqGetNumberOfBoards", adqError)
            Return 0
        End If
        Return numBoards
    End Function

    Private Function GetModules(boardIndex As UInteger, numberOfModules As UInteger, moduleType As UInteger) As UInteger
        Dim adqError As UInteger = AdqErrors.ADQ_ERROR_NONE
        Dim aoBoard As AdqAoBoard = Nothing
        Dim moduleIdx As UInteger = 0

        While moduleIdx < numberOfModules
            Dim currentModuleType As UInteger
            adqError = AdqDriver.adqGetModuleType(boardIndex, moduleIdx, currentModuleType)
            If adqError <> AdqErrors.ADQ_ERROR_NONE Then
                SetFunctionErrorInfo("adqGetModuleType", adqError)
                moduleIdx += 1UI
                Continue While
            End If

            If moduleType = currentModuleType Then
                If aoBoard Is Nothing Then
                    aoBoard = GetBoardInfo(boardIndex)
                    If aoBoard Is Nothing Then
                        Exit While
                    End If
                    _adqAoBoards.Add(aoBoard)
                End If

                GetModuleInfo(boardIndex, moduleIdx, aoBoard)
            End If
            moduleIdx += 1UI
        End While
        Return adqError
    End Function

    Private Function GetModuleInfo(boardIndex As UInteger, moduleIndex As UInteger, aoBoard As AdqAoBoard) As UInteger
        Dim adqError As UInteger
        Dim adqGeneralModuleInfo As AdqDriver.ADQGeneralModuleInfo = New AdqDriver.ADQGeneralModuleInfo(AdqDriver.ADQGeneralModuleInfo_ID)
        Dim adqAoModuleInfo As AdqDriver.ADQAOModuleInfo = New AdqDriver.ADQAOModuleInfo(AdqDriver.ADQAOModuleInfo_ID)

        Dim ModuleInfo As IntPtr = Marshal.AllocHGlobal(CInt(adqAoModuleInfo.Size))
        Marshal.StructureToPtr(adqAoModuleInfo, ModuleInfo, False)

        Dim infoSize As UInteger
        adqError = AdqDriver.adqGetModuleInfo(boardIndex, moduleIndex, adqGeneralModuleInfo, adqAoModuleInfo.Size, ModuleInfo, infoSize)
        adqAoModuleInfo = CType(Marshal.PtrToStructure(ModuleInfo, GetType(AdqDriver.ADQAOModuleInfo)), AdqDriver.ADQAOModuleInfo)
        Marshal.FreeHGlobal(ModuleInfo)

        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            Return adqError
        End If

        Debug.Assert(adqAoModuleInfo.Size = infoSize,
                String.Format("Board {0}, Module {1}: API-function adqGetModuleInfo info size does not match - expected {2}, returned {3}",
                boardIndex, moduleIndex, adqAoModuleInfo.Size, infoSize))

        If IsClockControlledModule(adqGeneralModuleInfo.ClockType) Then
            Debug.Assert(adqGeneralModuleInfo.ClockMinimumDivider > 0,
                         String.Format("Board {0}: The ClockMinimumDivider of the module {1} is zero", boardIndex, moduleIndex))

            Dim aoModule = New AoModule(boardIndex, moduleIndex, adqGeneralModuleInfo, adqAoModuleInfo)
            aoBoard.AddModule(aoModule)
        End If
        Return adqError
    End Function

    Private Function IsClockControlledModule(clockType As UInteger) As Boolean
        Return clockType = AdqConstants.ADQ_CLOCK_TYPE_HARDWARE OrElse clockType = AdqConstants.ADQ_CLOCK_TYPE_SOFTWARE
    End Function

    Private Function GetBoardInfo(boardIndex As UInteger) As AdqAoBoard
        Dim aoBoard As AdqAoBoard = Nothing
        Dim boardInfo = New AdqDriver.ADQBoardInfo(AdqDriver.ADQBoardInfo_ID)

        Dim adqError As UInteger = AdqDriver.adqGetBoardInfo(boardIndex, boardInfo)
        If adqError = AdqErrors.ADQ_ERROR_NONE Then
            aoBoard = New AdqAoBoard(boardIndex, boardInfo)
        Else
            SetFunctionErrorInfo("adqGetBoardInfo", adqError)
        End If
        Return aoBoard
    End Function

    Private Function OpenAdqDriver() As UInteger
        Dim adqError As UInteger = AdqDriver.adqOpen()
        If adqError <> AdqErrors.ADQ_ERROR_NONE Then
            SetFunctionErrorInfo("adqOpen", adqError)
        End If
        Return adqError
    End Function

    Private Sub CloseAdqDriver()
        If _adqOpenError = AdqErrors.ADQ_ERROR_NONE Then
            AdqDriver.adqClose()
            _adqOpenError = AdqCommon.NotOpen
        End If
    End Sub

    Private _adqAoBoards As List(Of AdqAoBoard)
    Public ReadOnly Property AdqAoBoards As List(Of AdqAoBoard)
        Get
            Return _adqAoBoards
        End Get
    End Property

    Private Sub SetFunctionErrorInfo(functionName As String, adqError As UInteger)
        InfoText = String.Format("[{0}] API-function {1} returned Error number: &H{2:X}",
                Date.Now.ToString(AdqCommon.TimeFmt),
                functionName,
                adqError)
    End Sub

    Dim _infoText As String = String.Empty
    Public Property InfoText As String
        Get
            Return _infoText
        End Get
        Set(value As String)
            _infoText = value
            NotifyPropertyChanged(NameOf(InfoText))
        End Set
    End Property

    Public Sub NotifyPropertyChanged(ByVal propertyName As String)
        RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub

    Protected Overloads Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
    End Sub

    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not _disposed Then
            _disposed = True
            CloseAdqDriver()
        End If
    End Sub
End Class
